(* ******************************************************************** *)
(*
 * Project         : HOL-CSP - A Shallow Embedding of CSP in  Isabelle/HOL
 * Version         : 2.0
 *
 * Author          : Burkhart Wolff
 *                   (Based on HOL-CSP 1.0 by Haykal Tej and Burkhart Wolff)
 *
 * This file       : A Combined CSP Theory
 *
 * Copyright (c) 2009 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: ROOT.ML 8591 2009-09-26 09:58:14Z brucker $ *)


header{* Toplevel Theory *}

theory     CSP
imports    Bot Skip Stop Mprefix Det Ndet Seq Hide Sync Legacy
begin 

subsection{* Refinement Proof Rules *}

subsection{* The "Laws" of CSP *}


axioms

  hide_mprefix_distr   : "\<lbrakk> (B \<inter> A) = {} \<rbrakk> \<Longrightarrow> 
                          ((Mprefix A P) \\ B) = (Mprefix A (% x. ((P x) \\ B)))"
  hide_prefix_distr1   : "a : B  \<Longrightarrow> ((a \<rightarrow> P) \\ B) = (P \\ B)"
  hide_prefix_distr2   : "a ~: B \<Longrightarrow> ((a \<rightarrow> P) \\ B) = (a \<rightarrow> (P \\ B))"
  hide_det             : "((a \<rightarrow> P) \<box> (b \<rightarrow> Q)) \\ {a} = 
                          ((P \\ {a}) \<sqinter> ((P \\ {a}) \<box> (b \<rightarrow> (Q \\{a}))))" 

(*  SEQ_SKIPF            : "F(P `;` SKIP) = F P"*)


lemma mprefix_singl: "(Mprefix {a} P) = (a \<rightarrow> (P a))"
  by (auto simp: write0_def Process_eq_spec F_Mprefix D_Mprefix)
 

lemma mono_mprefix_ref: "\<forall>x. P x \<sqsubseteq> Q x \<Longrightarrow> Mprefix A P \<sqsubseteq> Mprefix A Q"
 by(simp add: mono_Mprefix0)

lemma mono_prefix_ref: "P \<sqsubseteq> Q \<Longrightarrow> (a \<rightarrow> P) \<sqsubseteq> (a \<rightarrow> Q)"
  by (simp add: write0_def mono_mprefix_ref)


lemma mono_ndet_ref: " \<lbrakk>P \<sqsubseteq> P'; S \<sqsubseteq> S'\<rbrakk> \<Longrightarrow> (P \<sqinter> S) \<sqsubseteq> (P' \<sqinter> S')"
by(auto simp: le_approx_def F_ndet D_ndet Ra_def F_ndet min_elems_def T_ndet)

lemma mono_det_ref: 
  " \<lbrakk>P \<sqsubseteq> P'; S \<sqsubseteq> S'\<rbrakk> \<Longrightarrow> (P \<box> S) \<sqsubseteq> (P' \<box> S')"
  apply (simp add: le_approx_def)
  apply (simp add: F_det D_det)
  apply (rule conjI)
  apply (rule subsetI)
  apply (rule UnI1)
  apply (erule conjE)
  apply (drule conjunct1)
  apply (drule conjunct1)
  apply (simp add: subsetD)
  apply (rule conjI)
  apply (rule subsetI)
  apply (rule UnI2)
  apply (erule conjE)
  apply (drule conjunct1)
  apply (drule conjunct1)
  apply (simp add: subsetD)
  apply (rule conjI)
  apply (rule allI)
  apply (rule impI)
  apply (simp add: Ra_def)
  apply (simp add: F_det)
  apply (drule conjunct2)
  apply (drule conjunct2)
  apply (drule conjunct1)
  back
  apply (drule conjunct1)
  back  
  apply (rule equalityI)
  apply (rule subsetI)
  apply (simp)
  apply (rule disjI1)
  apply (rule conjI)
  prefer 3
  apply (rule subsetI)
  apply (simp)
  apply (rule disjI1)
  apply (rule conjI)
  prefer 5
  apply (simp add: min_elems_def T_det)  
  sorry

(* lemma mono_Ren_ref: "P << Q ==> (P`[[`f`]]`) << (Q`[[`f`]]`)"
  
 done
 byev[cut_facts_tac prem 1,etac conjE 1,rtac conjI 1,ALLGOALS(rtac subsetI),
 ALLGOALS(asm_full_simp_tac(set_ss addsimps [F_Ren,D_Ren,split_def])),
 REPEAT(SOMEGOAL(etac disjE) ORELSE SOMEGOAL(etac conjE) ORELSE SOMEGOAL(etac exE)),
 rtac disjI1 2,rtac disjI2 3,ALLGOALS(res_inst_tac [("x","t1")] exI),
 REPEAT(SOMEGOAL(res_inst_tac [("x","t2")] exI)),
 ALLGOALS(asm_full_simp_tac(set_ss addsimps [subsetD]))];
 qed"mono_Ren_ref";
*)

lemma mono_hide_set_refD: 
  "P \<sqsubseteq> Q \<Longrightarrow> D (Q \\ A) \<subseteq> D (P \\ A)"
  apply (simp add: le_approx_def)
  apply (simp add: D_hiding)  
  apply (simp add: Ra_def min_elems_def)
  apply (auto)    
  apply (rule_tac x="t" in exI)
  apply (rule_tac x="u" in exI)
  apply (rule conjI)
  apply (simp)
  apply (rule conjI)
  apply (simp)
  apply (rule conjI)
  apply (simp)
  apply (rule disjI2)
  apply (rule_tac x="f" in exI)
  apply (rule conjI)
  apply (simp)
  apply (rule conjI)
  apply (simp)  
  apply (rule allI)
  apply (simp add: T_F_spec[symmetric])
  sorry

lemma mono_hide_set_refF:  "P \<sqsubseteq> Q \<Longrightarrow> F (Q \\ A) \<subseteq> F (P \\ A)"
 apply (simp add: le_approx_def)
  apply (simp add: F_hiding)  
  apply (simp add: Ra_def min_elems_def)
  apply (auto)    
  apply (rule_tac x="t" in exI)
  apply (rule conjI)
  apply (simp)
  sorry

lemma mono_hide_set_ref: " P \<sqsubseteq> Q \<Longrightarrow> P \\ A \<sqsubseteq> Q \\ A"
  apply (subst le_approx_def)
  apply (simp add: Ra_def min_elems_def)
  apply (simp add: mono_hide_set_refD)
  apply (rule conjI)
  apply (rule allI)
  apply (rule impI)
  apply (rule equalityI)
  apply (auto)
  sorry

lemma mono_HSI2a: " \<lbrakk>P \<le> Q; s \<notin> D (P \\ A)\<rbrakk> \<Longrightarrow> Ra (P \\ A) s \<subseteq> Ra (Q \\ A) s"
  sorry

lemma mono_HSI2b: " \<lbrakk>P \<le> Q; s \<notin> D (P \\ A)\<rbrakk> \<Longrightarrow> Ra (Q \\ A) s \<subseteq> Ra (P \\ A) s"
  sorry

lemma mono_HSI2: " \<lbrakk>P \<le> Q; s \<notin> D (P \\ A)\<rbrakk> \<Longrightarrow> Ra (P \\ A) s = Ra (Q \\ A) s"
  sorry

lemma mono_HSI31: 
  " \<lbrakk>tr_hide_set t (ev ` A) = tr_hide_set s (ev ` A); s \<in> T Q;
     \<forall>s. tr_hide_set t (ev ` A) = tr_hide_set s (ev ` A) \<longrightarrow>
         (s, ev ` A) \<notin> F Q\<rbrakk>
    \<Longrightarrow> \<exists>a. a \<in> ev ` A \<and> s @ [a] \<in> T Q"
  sorry

lemma help1: "[a,b]=[a]@[b]"
  by(simp)

lemma help2: "[a]<[a,b]"
  by(simp add: less_list_def le_list_def)

lemma mono_HSI32: 
  "\<lbrakk>tickFree t; t \<in> T Q;
     \<forall>s. tr_hide_set t (ev ` A) = tr_hide_set s (ev ` A) \<longrightarrow>
         (s, ev ` A) \<notin> F Q\<rbrakk>
    \<Longrightarrow> \<exists>f. f IsChainOver t \<and> f Congruent t ModuloHide A \<and> (\<forall>i. f i \<in> T Q)"
  sorry

lemma mono_HSI33: 
  "\<exists>n s. length s = n \<and> s \<le> t \<and> tr_hide_set s A = tr_hide_set t A"
  by (auto)

lemma mono_HSI34: 
  "\<exists>s. tr_hide_set s A = tr_hide_set t A \<and>
        s \<le> t \<and> (\<forall>s1<s. tr_hide_set s1 A \<noteq> tr_hide_set t A)"
  sorry

lemma mono_HSI35: 
  "\<lbrakk>s < t; tr_hide_set s A \<noteq> tr_hide_set t A\<rbrakk>
    \<Longrightarrow> tr_hide_set s A < tr_hide_set t A"
  sorry

lemma mono_HSI36: 
  "\<forall>ta. (\<exists>t u. front_tickFree u \<and>
                tickFree t \<and>
                ta = tr_hide_set t (ev ` A) @ u \<and>
                (t \<in> D P \<or>
                 (\<exists>f. f IsChainOver t \<and>
                      f Congruent t ModuloHide A \<and> (\<forall>i. f i \<in> T P)))) \<longrightarrow>
         \<not> ta < tr_hide_set t (ev ` A) \<Longrightarrow>
    \<exists>t1. tr_hide_set t1 (ev ` A) = tr_hide_set t (ev ` A) \<and>
         t1 \<le> t \<and> (t1 \<notin> D P \<or> t1 \<in> min_elems (D P))"
  sorry


lemma mono_HSI3: 
  " P \<le> Q \<Longrightarrow> min_elems (D (P \\ A)) \<subseteq> T (Q \\ A)"
  sorry

lemma mono_HSI: " P \<le> Q \<Longrightarrow> P \\ A \<le> Q \\ A"
  sorry

lemma mono_HS_rec: "mono (\<lambda> P. (P \\ A))"
  sorry


lemma mono_PaI_refD: 
  "P \<sqsubseteq> Q \<Longrightarrow> D (Q \<lbrakk>A \<rbrakk> S) \<subseteq> D (P \<lbrakk>A \<rbrakk> S)"
  sorry

lemma mono_PaI_refF: " P \<sqsubseteq> Q \<Longrightarrow> F (Q \<lbrakk>A \<rbrakk> S) \<subseteq> F (P \<lbrakk>A \<rbrakk> S)"
  sorry

lemma mono_PaI_ref_L: "P \<sqsubseteq> Q \<Longrightarrow> (P \<lbrakk>A \<rbrakk> S) \<sqsubseteq> (Q \<lbrakk>A \<rbrakk> S)"
  sorry


lemma mono_PaI_ref_R: "P \<sqsubseteq> Q \<Longrightarrow> (S \<lbrakk>A \<rbrakk> P) \<sqsubseteq> (S \<lbrakk>A \<rbrakk> Q)"
  sorry


lemma mono_PaI_ref: "\<lbrakk>P \<sqsubseteq> P'; Q \<sqsubseteq> Q'\<rbrakk> \<Longrightarrow> (P \<lbrakk>A \<rbrakk> Q) \<sqsubseteq> (P' \<lbrakk>A \<rbrakk> Q')"
  sorry


lemma mono_Inter_ref: " \<lbrakk>P \<sqsubseteq> P'; Q \<sqsubseteq> Q'\<rbrakk> \<Longrightarrow> (P|||Q) \<sqsubseteq> (P'|||Q')"
  sorry


lemma mono_Par_ref: "\<lbrakk>P \<sqsubseteq> P'; Q \<sqsubseteq> Q'\<rbrakk> \<Longrightarrow> (P||Q) \<sqsubseteq> (P'||Qk')"
  sorry

lemma least_process: " \<bottom> \<le> (P::'a process)"
  by(auto intro: Process.le_approx_implies_le_ref)

lemma subset_F_Bot: "F Q \<le> F \<bottom>"
  by(auto intro: Process.le_approx_implies_le_ref is_processT2_S)


lemma subset_D_Bot: "D Q \<le> D \<bottom>"
  by (auto intro: least_process le_ref1 D_ftF)


lemma  eq_F_Bot: "F P = F \<bottom> = (F \<bottom> \<subseteq> F P)"  
  by (auto, subgoal_tac "F P \<subseteq> F \<bottom>", 
         auto simp: subset_F_Bot Legacy.is_processT2_S)


lemma eq_D_Bot: "D P = D \<bottom> = (D \<bottom> \<subseteq> D P)"
  by (auto, subgoal_tac "D P \<subseteq> D \<bottom>", 
         auto simp: subset_D_Bot Legacy.D_ftF)



lemma ftF_D_Bot: "front_tickFree t = (t \<in> D \<bottom>)"
  by (simp add: D_Bot Bot_is_UU[symmetric])

lemma ftF_T_Bot: "front_tickFree t = (t \<in> T \<bottom>)"
   by (simp add: T_Bot Bot_is_UU[symmetric])


lemma D_Bot_eq_T_Bot: "D \<bottom> = T \<bottom>"
  by (simp add: T_Bot D_Bot Bot_is_UU[symmetric])

lemma is_processT6_S3: "\<lbrakk>fst x = []; tick \<notin> snd x; [tick] \<in> T P\<rbrakk> \<Longrightarrow> x \<in> F P"
  by (subst surjective_pairing, simp add: is_processT6_S2)

lemma div_lemma: "([] \<in> D P) = (P = \<bottom>)"
    sorry

lemma det_commute :"(P \<box> Q) =(Q \<box> P)"
  apply (simp add: det_def)
  by(rule_tac f="Abs_Process" in arg_cong) auto
  

lemma det_bot [simp]: "(P \<box> \<bottom>) = \<bottom>"
  by (auto simp: Process_eq_spec eq_F_Bot eq_D_Bot,
      auto simp: Bot_is_UU[symmetric] subset_F_Bot Legacy.is_processT2_S
                 subset_D_Bot F_det F_Bot D_det D_Bot Legacy.D_ftF)


  
lemma det_bot'[simp]: "(\<bottom> \<box> P) = \<bottom>"
by(subst det_commute, simp)

lemma det_id[simp]: "(P \<box> P) = P"
  apply (simp add: Process_eq_spec, rule conjI)
  apply (auto)
  apply (simp_all only: D_det F_det Un_def Sigma_def)
  apply (simp_all only: mem_Collect_eq)  
  apply (simp add: sym)
  apply ((erule disjE | erule conjE)+,simp)+
  apply (subst is_processT8, simp_all)
  apply (erule conjE)+
  apply (subst is_processT6_S2, simp_all)
done
  
lemma det_assoc: "((M \<box> P) \<box> Q) = (M \<box> (P \<box> Q))"
  apply (simp add: Process_eq_spec, rule conjI)
  apply (auto)
  apply (simp_all add: D_det F_det Un_def Sigma_def)  
  apply ((erule disjE | erule conjE)+, simp_all)+
  apply (rule disjI2 | rule conjI)+
  apply (simp_all add: T_det)
  apply ((erule disjE | erule conjE)+, simp_all)+
done

lemma det_STOP[simp]: "(P \<box> STOP) = P"
  apply (simp add: Process_eq_spec, rule conjI)
  apply (auto)
  apply (simp_all add: D_det F_det D_STOP F_STOP T_STOP Un_def Sigma_def)  
  apply ((erule disjE | erule conjE)+, simp_all)+
  apply (subst is_processT8, simp_all)
  apply (erule conjE)+
  apply (subst is_processT6_S2, simp_all)
done

lemma det_STOP'[simp]: "(STOP \<box> P) = P"
by(simp add: det_commute)



lemma ndet_id[simp]: "(P \<sqinter> P) = P"
  by (simp_all add: F_ndet D_ndet Process_eq_spec)  
 
lemma ndet_commute: "(P \<sqinter> Q) = (Q \<sqinter> P)"
  by(auto simp: ndet_def Un_commute)

lemma ndet_bot[simp]: "(P \<sqinter> \<bottom>) = \<bottom>"
  apply (simp add: Process_eq_spec, rule conjI)
  apply (rule equalityI)
  prefer 3
  apply (rule equalityI)
  apply (auto simp: D_ndet F_ndet subset_F_Bot subset_D_Bot 
                    Un_upper1 Un_upper2 Legacy.is_processT2_S Legacy.D_ftF)  
done

lemma ndet_bot'[simp]: "(\<bottom> \<sqinter> P) = \<bottom>"
  by(subst ndet_commute, simp)

lemma non_det_assoc: "((M \<sqinter> P) \<sqinter> Q) = (M \<sqinter> (P \<sqinter> Q))"
  by (simp_all add: F_ndet D_ndet  Process_eq_spec Un_assoc)  


lemma det_distrib: "(M \<box> (P \<sqinter> Q)) = ((M \<box> P) \<sqinter> (M \<box> Q))"
  apply (simp_all add:  Process_eq_spec F_det D_det F_ndet D_ndet Un_def , rule conjI)
  apply (rule set_eqI)
  prefer 2
  apply (rule equalityI)
  apply (rule subsetI)
  apply (simp)
  apply ((erule disjE | erule conjE)+, simp_all)+
  apply (rule subsetI)
  apply (simp)
  apply ((erule disjE | erule conjE)+, simp_all)+
  apply (auto simp: T_ndet)
done

lemma non_det_distrib: "(M \<sqinter> (P \<box> Q)) = ((M \<sqinter> P) \<box> (M \<sqinter> Q))"
  apply (simp_all add:  Process_eq_spec F_det D_det F_ndet D_ndet Un_def , 
         rule conjI)
  prefer 2
  apply (rule equalityI)
  apply (rule subsetI)
  apply (simp)
  apply ((erule disjE | erule conjE)+, simp_all)+
  apply (rule subsetI)
  apply (simp)
  apply ((erule disjE | erule conjE)+, simp_all)+
  apply (rule equalityI)
  apply (rule subsetI)
  apply (simp)
  apply (auto simp:T_ndet)
  apply (simp_all add: is_processT8 excluded_middle is_processT6_S2)
done

lemma pref_non_det: "(a \<rightarrow> (P \<sqinter> Q)) = ((a \<rightarrow> P) \<sqinter> (a \<rightarrow> Q))"
  apply (simp add: Process_eq_spec, rule conjI)
  apply (auto)
  apply (simp_all add: write0_def D_ndet F_ndet F_Mprefix D_Mprefix Un_def)
  apply ((erule disjE | erule conjE)+, simp_all)+
done


lemma Mprefix_STOP: "(Mprefix {} P) = STOP"
  apply (simp_all add: Process_eq_spec , rule conjI)
  apply (auto)
  apply (simp_all add:F_Mprefix D_Mprefix D_STOP F_STOP)
done

lemma mprefix_Un_distrD:
  "D(Mprefix (A \<union> B) P) = D (Mprefix A P \<box> Mprefix B P)"
  apply (auto)
  apply (simp_all add: D_det D_Mprefix Un_def)
  apply (erule disjE | erule conjE)+
  apply (simp add: image_def)
  apply (auto)
done

lemma mprefix_Un_distrF: 
  "F(Mprefix (A \<union> B) P) = F((Mprefix A P) \<box> (Mprefix B P))"
  apply (auto)
  apply ((erule disjE | erule conjE)+ | 
         simp_all add: F_det F_Mprefix Un_def image_def)+
  apply (auto)
done

lemma mprefix_Un_distr: "(Mprefix (A \<union> B) P) = ((Mprefix A P) \<box> (Mprefix B P))"
  by (simp_all add: Process_eq_spec mprefix_Un_distrD mprefix_Un_distrF)


lemma mnon_det_non_det: "(Mprefix (A \<union> {a}) P) = ((Mprefix A P) \<box> (a \<rightarrow> (P a)))"
  by (subst mprefix_Un_distr, subst mprefix_singl, simp)


lemma pref_det_non_det: "((a \<rightarrow> P) \<box> (a \<rightarrow> Q)) = ((a \<rightarrow> P) \<sqinter> (a \<rightarrow> Q))"
  apply (simp_all add: Process_eq_spec , rule conjI)
  apply (auto)
  apply (simp_all add: write0_def F_ndet D_ndet F_det D_det 
                      F_Mprefix D_Mprefix Un_def)
  apply ((erule disjE | erule conjE)+, simp_all)+
done


lemma SEQ_SKIPD: "D(P `;` SKIP) = D P"
  apply (simp_all add: D_seq D_SKIP)
  apply (auto simp: is_processT7)
  apply (case_tac "tickFree x")
  apply (rule_tac x="x" in exI)  
  apply (simp)
  apply (subgoal_tac "\<exists> s. x = s @ [tick]")
  apply (erule exE)
  apply (rule_tac x="s" in exI)    
  apply (auto simp: is_processT9)
  apply (rule_tac s="[tick]" and P="P" in append_T_imp_tickFree)
  apply (rule D_T)
  prefer 3  
  apply (rule nonTickFree_n_frontTickFree)
  prefer 2
  apply (rule_tac P="P" in is_processT2_TR[rule_format])
  by (simp_all add: D_T)
  


lemma SEQ_SKIPF: "F(P `;` SKIP) = F P"
  apply (simp_all add: F_seq D_seq F_SKIP D_SKIP)
  apply (auto)
  apply (subgoal_tac "b \<subseteq> insert tick b")
  apply (rule_tac  Y="insert tick b" in is_processT4[rule_format])
  apply (simp_all)
  apply (subst insert_def)
  apply (auto)
  apply (simp add: T_F is_processT6_S1)
  defer
  apply (subst is_processT8_Pair)
  apply (simp_all add: is_processT7)
  apply (case_tac "tick \<in> b")
  apply (subgoal_tac "insert tick b \<subseteq> b")
  apply (rule_tac  Y="b" in is_processT4[rule_format])
  apply (simp_all)
  apply (subgoal_tac "\<exists> s. a = s@[tick]")
  apply (erule exE)
  apply (case_tac "tickFree a")
  apply (simp_all)
  apply (case_tac "(a,insert tick b) \<in> F P")
  apply (simp_all)
  apply (erule_tac x="a" in allE)
  apply (erule_tac x="[]" in allE, simp)
  prefer 3
  apply (subgoal_tac "\<exists> s. a = s@[tick]")
  apply (erule exE)
  apply (simp_all)
  apply (erule_tac x="s" in allE)
  apply (erule_tac x="[tick]" in allE, simp)
  sorry


lemma SEQ_SKIP: "(P `;` SKIP) = P"
  by (simp_all add: Process_eq_spec SEQ_SKIPD SEQ_SKIPF)


lemma SKIP_SEQD: "D(SKIP `;` P) = D P"
  by (simp_all add: D_seq D_SKIP T_SKIP)
  

lemma SKIP_SEQF: "F(SKIP `;` P) = F P"
  apply (simp_all add: F_seq F_SKIP D_SKIP T_SKIP)
  by (auto simp: is_processT8_Pair)


lemma SKIP_SEQ: "(SKIP `;` P) = P"
  by (simp_all add: Process_eq_spec SKIP_SEQD SKIP_SEQF)

lemma ev_Neq_tick1:"ev a = b ==> b ~= tick"
  by (auto)


lemma Nelem_image_ev:
  " \<lbrakk>\<forall> c. c \<in> B \<longrightarrow> c \<notin> C; hd x \<in> ev ` B\<rbrakk> \<Longrightarrow> hd x \<notin> ev ` C"
  by (auto)


lemma mprefix_seqD:
  "D((Mprefix A P) `;` Q) = D(Mprefix A (\<lambda> x. (P x) `;` Q))"
  apply (simp_all add: D_Mprefix D_seq T_Mprefix)
  apply (auto)
  apply (rule_tac x="tl t1" in exI)
  apply (rule_tac x="t2" in exI)
  apply (case_tac t1)
  apply (simp_all)
  apply (case_tac t1)
  apply (simp_all)
  apply (rule_tac x="a" in exI)
  apply (rule conjI)
  apply (case_tac t1)
  apply (simp_all)
  apply (rule disjI2)
  apply (rule_tac x="tl t1" in exI)
  apply (rule_tac x="t2" in exI)
  apply (simp)
  apply (case_tac t1)
  apply (simp_all)
  apply (rule_tac x="ev xa # t1" in exI)
  apply (rule_tac x="t2" in exI)
  apply (simp)
  apply (case_tac x)
  apply (simp_all)
  apply (rule_tac x="ev xa # t1" in exI)
  apply (rule_tac x="t2" in exI)
  apply (simp)
  apply (case_tac x)
  apply (simp_all)
  apply (auto)
  by (erule_tac x="ev xa # t1" in allE, erule_tac x="t2" in allE, simp)+
  

lemma mprefix_seqF1:
  " a \<notin> B \<Longrightarrow> (A \<union> {a}) \<inter> B = A \<inter> B"
  by (auto)


lemma mprefix_seqF: "F((Mprefix A P) `;` Q) = F(Mprefix A (\<lambda>x. (P x) `;` Q))"
  apply (simp_all add: split_def mprefix_seqD F_D_SEQ_spec F_Mprefix T_Mprefix D_Mprefix)
  apply (rule equalityI)
  apply (rule subsetI)
  apply (simp_all)
  apply (erule disjE | erule conjE)+
  apply (rule disjI1, simp)
  apply (erule disjE | erule conjE | erule exE)+
  apply (rule disjI2)
  apply (rule conjI, simp)+
  apply (rule_tac x="a" in exI, simp_all)
  apply (rule disjI1)
  apply (case_tac "fst x", simp_all)
  apply (erule disjE | erule conjE | erule exE)+
  apply (rule disjI2)
  apply (rule conjI)
  apply (simp_all add: ev_Neq_tick1)
  apply (rule impI, simp)
  apply (rule conjI)
  apply (subst hd_append, simp)
  apply (rule conjI)
  apply (rule impI, simp)
  apply (rule impI, simp)
  apply (rule_tac x="a" in exI)
  apply (rule conjI)
  apply (subst hd_append, simp)
  apply (rule conjI)
  apply (rule impI, simp)
  apply (rule impI, simp)
  apply (rule disjI2)
  apply (rule disjI1)
  apply (rule_tac x="tl t1" in exI)
  apply (rule_tac x="t2" in exI, simp)
  apply (subgoal_tac "t1 \<noteq> []")
  apply (rule conjI)
  apply (subst tl_append2)
  apply (simp_all)
  apply (case_tac t1, simp_all)
  apply (erule disjE | erule conjE | erule exE)+
  apply (rule_tac x="a" in exI)
  apply (rule conjI, simp)
  apply (rule disjI2)+
  apply (simp)
  apply (rule conjI)
  apply (rule subsetI, simp)
  apply (subst image_def, simp)
  apply (rule subsetI, simp)
  apply (erule disjE | erule conjE | erule exE)+
  apply (rule disjI1)
  apply (rule conjI)
  apply (rule_tac x="a" in exI, simp)
  apply (case_tac "fst x", simp_all add: ev_Neq_tick1)
  apply (erule disjE | erule conjE | erule exE)+
  apply (rule disjI2)
  apply (rule disjI1)
  apply (rule_tac x="(ev a)#t1" in exI, simp)
  apply (rule_tac x="t2" in exI, simp)
  apply (rule conjI)
  apply (case_tac "fst x", simp_all)
  by (auto)


lemma mprefix_seq:
    "((Mprefix A P) `;` Q) = (Mprefix A (\<lambda>x. (P x) `;` Q))"
    by (simp_all add: Process_eq_spec mprefix_seqF mprefix_seqD)


lemma pref_seq: "((a \<rightarrow> P) `;` Q) = (a \<rightarrow> (P `;` Q))"
  by (simp add: write0_def mprefix_seq)


lemma STOP_SEQ: "(STOP `;` P) = STOP"
  by (auto simp: Process_eq_spec F_SEQ_spec D_seq F_STOP D_STOP T_STOP Un_def)


lemma prefix_stop_seq: "((a \<rightarrow> STOP) `;` P) = (a \<rightarrow> STOP)"
  by (simp add: pref_seq STOP_SEQ)


lemma prefix_skip_seq: "((a \<rightarrow> SKIP) `;` P) = (a \<rightarrow> P)"
  by (simp add: pref_seq SKIP_SEQ)

lemma Bot_SEQ: "(\<bottom> `;` P) = \<bottom>"
  apply(simp add: Bot_is_UU[symmetric])
  apply (simp add: Process_eq_spec D_Bot F_Bot D_seq F_seq T_Bot)
  apply (auto)
  apply (subgoal_tac "front_tickFree t2")
  apply (rule front_tickFree_append, simp_all)
  apply (rule_tac a="tick" in front_tickFree_implies_tickFree, simp)
  apply (simp_all add: is_processT2 front_tickFree_append)
  apply (tactic "distinct_subgoals_tac")
  apply (subgoal_tac "\<exists> S. (t2, S) \<in> F P")
  apply (erule exE)
  apply (subgoal_tac "front_tickFree t2")
  apply (rule front_tickFree_append, simp_all)
  apply (rule_tac a="tick" in front_tickFree_implies_tickFree, simp)
  apply (simp add: is_processT2)
  apply (rule exI)
  apply (simp add: is_processT8)
  apply (erule_tac x="[]" in allE)+
  apply (simp)
  apply (rule_tac x="[]" in exI)
  apply (simp)
done

lemma SEQ_Ndet_distrRD: "D((P \<sqinter> Q) `;` S) = D((P `;` S) \<sqinter> (Q `;` S))"
  by (simp_all add: D_seq D_ndet T_ndet Un_def, auto)
  

lemma SEQ_Ndet_distrRF: "F((P \<sqinter> Q) `;` S) = F((P `;` S) \<sqinter> (Q `;` S))"
  by (simp_all add: F_seq D_ndet T_ndet F_ndet SEQ_Ndet_distrRD Un_def, auto)



lemma SEQ_Ndet_distrR: "((P \<sqinter> Q) `;` S) = ((P `;` S) \<sqinter> (Q `;` S))"
  by (simp add:  Process_eq_spec SEQ_Ndet_distrRD SEQ_Ndet_distrRF)


lemma SEQ_Ndet_distrLD: "D(P `;` (Q \<sqinter> S)) = D((P `;` Q) \<sqinter> (P `;` S))"
    by (simp_all add: D_seq D_ndet T_ndet Un_def, auto)


lemma SEQ_Ndet_distrLF: "F(P `;` (Q \<sqinter> S)) = F((P `;` Q) \<sqinter> (P `;` S))"
   by (simp_all add: F_seq D_ndet T_ndet F_ndet SEQ_Ndet_distrRD Un_def, auto)


lemma SEQ_Ndet_distrL: "(P `;` (Q \<sqinter> S)) = ((P `;` Q) \<sqinter> (P `;` S))"
   by (simp add:  Process_eq_spec SEQ_Ndet_distrLD SEQ_Ndet_distrLF)


lemma SEQ_Det_distrRD: "D((P \<box> Q) `;` S) = D((P `;` S) \<box> (Q `;` S))"
  by (simp_all add: D_seq D_det T_det Un_def, auto)


lemma SEQ_Det_distrRF: "F((P \<box> Q) `;` S) \<subseteq> F((P `;` S) \<box> (Q `;` S))"
  apply (simp_all add: F_SEQ_spec D_det T_det F_det split_def D_seq T_seq Un_def)
  apply (rule subsetI, simp)
  apply (erule conjE | erule disjE)+
  apply (rule disjI1)
  apply (rule conjI, simp_all)
  apply (erule conjE | erule disjE)+
  apply (rule disjI2)
  apply (rule disjI1)
  apply (rule conjI)
  apply (simp_all)
  apply (rule disjI2)
  apply (erule conjE | erule disjE)+
  apply (rule disjI1, simp)
  apply (rule disjI2)
  apply (rule disjI2)
  apply (rule disjI1, simp)
  apply (erule conjE | erule disjE | erule exE)+
  apply (rule disjI2)
  apply (case_tac "fst x = []")
  apply (rule disjI2)
  apply (rule disjI2)
  apply (simp_all)
  sorry

find_theorems front_tickFree
lemma par_Int_botD: "D(P \<lbrakk> A \<rbrakk> \<bottom>) = D \<bottom>"
  apply (auto simp: eq_D_Bot)
  apply (simp add: Bot_is_UU[symmetric])
  apply (simp add: D_sync D_Bot T_Bot)
  apply (auto simp: Process.front_tickFree_append)
(*
  apply (rule_tac x="[]" in exI)
  apply (rule_tac x="[]" in exI)
  apply (rule_tac x="[]" in exI)
  apply (rule_tac x="x" in exI)
  apply (rule conjI,simp)
  apply (rule conjI,simp)
  apply (rule conjI,simp)
  apply (rule conjI, simp only: si_empty1, simp)
  apply (rule disjI2)
  apply (auto)
  apply (rule_tac X="{}" in F_T)
  by (simp add: is_processT1)
*)
sorry

lemma par_Int_botF: "F(P \<lbrakk> A \<rbrakk> \<bottom>) = F \<bottom>"
  sorry


lemma par_Int_bot[simp]: "(P \<lbrakk> A \<rbrakk> \<bottom>) = \<bottom>"
   by (simp add:  Process_eq_spec par_Int_botF par_Int_botD)

lemma par_Int_bot1[simp]: "(\<bottom> \<lbrakk> A \<rbrakk> P) = \<bottom>"
  sorry


lemma par_Int_skip_D: "D(SKIP \<lbrakk> A \<rbrakk> SKIP) = D SKIP"
  by (simp add: D_SKIP D_sync)


lemma par_Int_skip_F: "F(SKIP \<lbrakk> A \<rbrakk> SKIP) = F SKIP"
  apply (simp add: F_SKIP D_SKIP T_SKIP F_sync Un_def Sigma_def)
  sorry


lemma par_Int_skip: "(SKIP \<lbrakk> A \<rbrakk> SKIP) = SKIP"
   by (simp add:  Process_eq_spec par_Int_skip_F par_Int_skip_D)

lemma sync_commute: "(P \<lbrakk> A \<rbrakk> Q) = (Q \<lbrakk> A \<rbrakk> P)"
   sorry

lemmas Par_Int_commute = sync_commute

lemma Inter_commute: "(P ||| Q) = (Q ||| P)"
  by(rule sync_commute)

lemma Inter_skip_D: "D(P ||| SKIP) = D P"
  apply (simp add: D_SKIP D_sync T_SKIP)
  sorry

lemma Inter_skip_F1:
  " \<lbrakk>(t, X) \<in> F P; fst x setinterleaves ((t, [tick]), {tick})\<rbrakk>
    \<Longrightarrow> x \<in> F P"
  sorry

lemma Inter_skip_F: "F(P ||| SKIP) = F P"
  sorry

lemma Inter_skip1: "(P ||| SKIP) = P"
  sorry

lemma Inter_skip2: "(SKIP ||| P) = P"
  sorry


lemma skip_Neq_stop: "SKIP \<noteq> STOP"
  by (auto simp: Process_eq_spec F_SKIP F_STOP D_SKIP D_STOP Un_def)


lemma stop_Neq_skip:  "STOP \<noteq> SKIP"
  by (subst not_sym, simp_all add: skip_Neq_stop)


lemma Inter_stop_seq_stop_D: "D(P ||| STOP) = D(P `;` STOP)"
  apply (simp add: D_sync D_seq D_STOP T_STOP)
  (* apply (auto) *)
  sorry


lemma setH1: 
  "(X \<union> Y) \<inter> A \<union> X \<inter> Y \<union> A = X \<inter> Y \<union> A"
  by (auto)


lemma Inter_stop_seq_stop_F: "F(P ||| STOP) = F(P `;` STOP)"
  sorry


(****P `|||` STOP ~= P ****)

lemma Inter_stop_seq_stop:
 "(P ||| STOP) = (P `;` STOP)"
  by (auto simp: Process_eq_spec Inter_stop_seq_stop_D Inter_stop_seq_stop_F)


lemma Inter_stop_seq_stop1:
  "(STOP ||| P) = (P `;` STOP)"
  sorry

(* SKIP ist kein neutrales Element fuer Par_Int *)
     (*a->Bot |]{a}[| SKIP = STOP *);


lemma par_int_ndet_distribD:
  "D(M \<lbrakk> A \<rbrakk> (P \<sqinter> Q)) = D((M \<lbrakk> A \<rbrakk> P) \<sqinter> (M \<lbrakk> A \<rbrakk> Q))"
  sorry

lemma par_int_ndet_distribF:
  "F(M \<lbrakk> A \<rbrakk> (P \<sqinter> Q)) = F((M \<lbrakk> A \<rbrakk> P) \<sqinter> (M \<lbrakk> A \<rbrakk> Q))"
  sorry


lemma par_int_ndet_distrib:
  "(M \<lbrakk> A \<rbrakk> (P \<sqinter> Q)) = ((M \<lbrakk> A \<rbrakk> P) \<sqinter> (M \<lbrakk> A \<rbrakk> Q))"
  sorry
      
lemma par_int_ndet_distrib1:
  "((P \<sqinter> Q) \<lbrakk> A \<rbrakk> M) = ((P \<lbrakk> A \<rbrakk> M) \<sqinter> (Q \<lbrakk> A \<rbrakk> M))"
  sorry


lemma par_comm: "(P || Q) =(Q || P)"
  sorry


lemma par_ndet_distrib1:
  "(M || (P \<sqinter> Q)) = ((M || P) \<sqinter> (M || Q))"
  by (simp add: par_int_ndet_distrib)


lemma par_ndet_distrib2:
  "((P \<sqinter> Q) || M) = ((P || M) \<sqinter> (Q || M))"
  by (simp add: par_int_ndet_distrib1)


lemma par_stopD: "P \<noteq> \<bottom> \<Longrightarrow> D(P || STOP) = D STOP"
  sorry

lemma par_stopF: "P \<noteq> \<bottom> \<Longrightarrow> F(P || STOP) = F STOP"
  sorry

lemma par_stop: "P \<noteq> \<bottom>  \<Longrightarrow> (P || STOP) = STOP"
  sorry


lemma par_assocD1: "D((P || Q) || S) \<subseteq> D(P || (Q || S))"
  sorry


lemma par_assocD2: "D(P || (Q || S)) \<subseteq> D((P || Q) || S)"
  sorry

lemma par_assocD: "D((P || Q) || S) = D(P || (Q || S))"
  sorry

lemma par_assocF1: "F((P || Q) || S) \<subseteq> F(P || (Q || S))"
  sorry

lemma par_assocF2: "F(P || (Q || S)) \<subseteq> F((P || Q) || S)"
  sorry

lemma par_assocF: "F((P || Q) || S) = F(P || (Q || S))"
  sorry

lemma par_assoc: "((P || Q) || S) = (P || (Q || S))"
  sorry


lemma hide_set_botD: "D(\<bottom> \\ A) = D \<bottom>"
  sorry

lemma hide_set_botF: "F(\<bottom> \\ A) = F \<bottom>"
  sorry

lemma hide_set_bot[simp]: "(\<bottom> \\ A) = \<bottom>"
  sorry

lemma hide_set_STOPD: "D(STOP \\ A) = D STOP"
  sorry

lemma hide_set_STOPF: "F(STOP \\ A) = F STOP"
  sorry

lemma hide_set_STOP: "(STOP \\ A) = STOP"
  sorry

lemma hide_set_SKIPD: "D(SKIP \\ A) = D SKIP"
  sorry

lemma hide_set_SKIPF: "F(SKIP \\ A) = F SKIP"
  sorry

lemma hide_set_SKIP: "(SKIP \\ A) = SKIP"
  sorry

lemma hide_set_emptyD:  "D(P \\ {}) = D P"
 sorry


lemma hide_set_emptyF: "F(P \\ {}) = F P"
  sorry


lemma hide_set_empty: "(P \\ {}) = P"
  sorry (*by (simp add: hide_set_STOP) *)

lemma "D(P \\ (A Un B)) <= D((P \\ A) \\ B)"
  sorry

lemma "D((P \\ A) \\ B) <= D(P \\ (A Un B))"
  sorry

lemma f_mono: 
  " \<forall>i. f i < f (Suc i) \<Longrightarrow> i < j \<longrightarrow> f i < f j"
  sorry


lemma f_0_less_f_i: 
  "\<lbrakk>f 0 = t; \<forall>i. f i < f (Suc i); 1 \<le> i\<rbrakk> \<Longrightarrow> t < f i"
  sorry


lemma
  " \<exists>f. f IsChainOver t \<and>
        f Congruent t ModuloHide A \<and> (\<forall>i. f i \<in> T P \<or> f i \<in> T Q) \<Longrightarrow>
    \<exists>f. f IsChainOver t \<and>
        f Congruent t ModuloHide A \<and> ((\<forall>i. f i \<in> T P) \<or> (\<forall>i. f i \<in> T Q))"
   sorry

lemma "D((P \<sqinter> Q) \\ A) = D((P \\ A) \<sqinter> (Q \\ A))"
  sorry

(*************************************************)
(* (P `|]` Q) `\\` A ~= (P `\\` A) `|]` (Q `\\` A).        *)
(*  For counterexample see Hoare's book p.114    *)
(*************************************************)

lemma le_trans:
  " \<And>i. \<lbrakk>i \<le> j; j \<le> k\<rbrakk> \<Longrightarrow> i \<le> (k::nat)"
  by (auto)

lemma length_f_i1: 
  "f 0 = t \<and> (\<forall>i. f i < f (Suc i)) \<Longrightarrow> \<forall>i. length t + i \<le> length (f i)" 
  sorry

lemma f_i_Neq_Nil1: 
  " f 0 = t \<and> (\<forall>i. f i < f (Suc i)) \<Longrightarrow> \<forall>i. i \<noteq> 0 \<longrightarrow> f i \<noteq> []"
  sorry

lemma tl_f_i_less_tl_f_Suc1: 
  " f 0 = t \<and> (\<forall>i. f i < f (Suc i)) \<Longrightarrow>
    \<forall>i. i \<noteq> 0 \<longrightarrow> tl (f i) < tl (f (Suc i))"
  sorry

lemma length_f_i2: 
  "f 0 = t \<and> (\<forall>i. f i < f (Suc i)) \<Longrightarrow> \<forall>i. length t + i < length (f (Suc i))"
  sorry

lemma tl_f_Suc_i_Neq_Nil: 
  "f 0 = t \<and> (\<forall>i. f i < f (Suc i)) \<Longrightarrow> \<forall>i. i \<noteq> 0 \<longrightarrow> tl (f (Suc i)) \<noteq> []"
  sorry

lemma tl_f_i_less_tl_f_Suc2: 
  "f 0 = t \<and> (\<forall>i. f i < f (Suc i)) \<Longrightarrow> \<forall>i. f (Suc i) \<noteq> []"
  sorry

lemma tl_f_Suc_i_less_tl_f_Suc2: 
  "f 0 = t \<and> (\<forall>i. f i < f (Suc i)) \<Longrightarrow>
    (if i = 0 then t else tl (f (Suc i))) < tl (f (Suc (Suc i)))"
  sorry

(*
lemma Ren_SKIP: "SKIP`[[`f`]]` = SKIP"
  sorry

lemma Ren_STOP: "STOP`[[`f`]]` = STOP"
  sorry

lemma Ren_Bot: "\<bottom>`[[`f`]]` = \<bottom>"
  sorry

lemma Ren_Ndet: "((P `|~|` Q)`[[`f`]]`) = (P`[[`f`]]` `|~|` (Q`[[`f`]]`))"
  sorry

lemma Ren_prefix: "((a`->`P)`[[`f`]]`) = ((f a) `->` P`[[`f`]]`)"
  sorry

lemma Ren_detD: "D((P `[]` Q)`[[`f`]]`) = D((P`[[`f`]]`) `[]` (Q`[[`f`]]`))"
  sorry

lemma Ren_detF1: "F((P `[]` Q)`[[`f`]]`) <= F((P`[[`f`]]`) `[]` (Q`[[`f`]]`))"
  sorry


lemma Ren_detF2: "F((P`[[`f`]]`) `[]` (Q`[[`f`]]`)) <= F((P `[]` Q)`[[`f`]]`)"
  sorry

lemma Ren_detF: "F((P `[]` Q)`[[`f`]]`) = F((P`[[`f`]]`) `[]` (Q`[[`f`]]`))"
  sorry

lemma Ren_Det: "(P `[]` Q)`[[`f`]]` = ((P`[[`f`]]`) `[]` (Q`[[`f`]]`))"
  sorry
*)

lemma det_left_commute: "(M \<box> P \<box> Q) = (P \<box> M \<box> Q)"
  by (simp add: det_commute)

lemma det_left_id: "(M \<box> M \<box> Q) = (M \<box> Q)"
  by (simp add: det_id)


lemma non_det_left_commute: "(M \<sqinter> P \<sqinter> Q) = (P \<sqinter> M \<sqinter> Q)"
  by (simp add: ndet_commute)


lemma non_det_left_id: "(M \<sqinter> M \<sqinter> Q) = (M \<sqinter> Q)"
  by (simp add: ndet_id)

lemma par_left_commute: "(M || P || Q) = (P || M || Q)"
  by (simp add: par_comm)

lemma mprefix_Par_Int_distr3D1: 
  " \<lbrakk>B \<inter> C = {}; A \<subseteq> C\<rbrakk>
    \<Longrightarrow> D (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)
       \<subseteq> D (\<box> x \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q x))"
   sorry

 
lemma  mprefix_Par_Int_distr3D2: 
  "\<lbrakk>B \<inter> C = {}; A \<subseteq> C\<rbrakk>
    \<Longrightarrow> D (\<box> x \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q x))
       \<subseteq> D (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)"
  sorry


lemma  mprefix_Par_Int_distr3D: 
  "\<lbrakk>B \<inter> C = {}; A \<subseteq> C\<rbrakk>
    \<Longrightarrow> D (\<box> x \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q x))
       \<subseteq> D (Mprefix A P \<lbrakk>C\<rbrakk> Mprefix B Q)"
  sorry

lemma mprefix_Par_Int_distr3F1: 
  "\<lbrakk>B \<inter> C = {}; A \<subseteq> C\<rbrakk>
    \<Longrightarrow> F (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)
       \<subseteq> F (\<box> x \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q x))"
  sorry

lemma mprefix_Par_Int_distr3F2: 
  "\<lbrakk>B \<inter> C = {}; A \<subseteq> C\<rbrakk>
    \<Longrightarrow> F (\<box> x \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q x))
       \<subseteq> F (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)"
  sorry

lemma mprefix_Par_Int_distr3F: 
  " \<lbrakk>B \<inter> C = {}; A \<subseteq> C\<rbrakk>
    \<Longrightarrow> F (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q) =
       F (\<box> x \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q x))"
  sorry

lemma mprefix_Par_Int_distr3: 
  "\<lbrakk>B \<inter> C = {}; A \<subseteq> C\<rbrakk>
    \<Longrightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q) = \<box> x \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q x)"
  sorry

lemma mprefix_Par_Int_distr2:
  "\<lbrakk>A \<inter> C = {}; B \<subseteq> C\<rbrakk>
    \<Longrightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q) = \<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q)"
  sorry


lemma mprefix_Par2D1a:
  "\<lbrakk>B \<inter> C = {}; A \<inter> C = {};
     \<exists>t u r v.
        front_tickFree v \<and>
        tickFree r \<and>
        x = r @ v \<and>
        r setinterleaves ((t, u), ev ` C \<union> {tick}) \<and>
        t \<in> D (Mprefix A P) \<and> u \<in> T (Mprefix B Q)\<rbrakk>
    \<Longrightarrow> x \<in> D (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
              \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))"
  sorry

   
lemma mprefix_Par2D1b: 
  " \<lbrakk>B \<inter> C = {}; A \<inter> C = {};
     \<exists>t u r v.
        front_tickFree v \<and>
        tickFree r \<and>
        x = r @ v \<and>
        r setinterleaves ((t, u), ev ` C \<union> {tick}) \<and>
        t \<in> D (Mprefix B Q) \<and> u \<in> T (Mprefix A P)\<rbrakk>
    \<Longrightarrow> x \<in> D (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
              \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))"
  sorry

lemma mprefix_Par2D1: 
  " \<lbrakk>B \<inter> C = {}; A \<inter> C = {}\<rbrakk>
    \<Longrightarrow> D (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)
       \<subseteq> D (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
            \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))"
  sorry


lemma mprefix_Par2D2a: "[| B Int C = {}; A Int C = {}; x:D( \<box> x \<in> A \<rightarrow> ((P x) \<lbrakk> C \<rbrakk> (Mprefix B Q))) |] ==> x : D ((Mprefix A P) \<lbrakk> C \<rbrakk> (Mprefix B Q))"
  sorry


lemma mprefix_Par2D2b:
  " \<lbrakk>B \<inter> C = {}; A \<inter> C = {}; x \<in> D (\<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))\<rbrakk>
    \<Longrightarrow> x \<in> D (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)"
  sorry


lemma mprefix_Par2D2: 
  "\<lbrakk>B \<inter> C = {}; A \<inter> C = {}\<rbrakk>
    \<Longrightarrow> D (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
          \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))
       \<subseteq> D (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)"
  sorry

lemma mprefix_Par2D: 
  " \<lbrakk>B \<inter> C = {}; A \<inter> C = {}\<rbrakk>
    \<Longrightarrow> D (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q) =
       D (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
          \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))"
     sorry


lemma AuxEv1: 
  "\<forall>c. c \<in> A \<longrightarrow> c \<notin> B \<Longrightarrow> \<forall>c. c \<in> ev ` B \<longrightarrow> c \<notin> ev ` A"
  by (auto)


lemma mprefix_Par2F1: 
  " \<lbrakk>B \<inter> C = {}; A \<inter> C = {}\<rbrakk>
    \<Longrightarrow> F (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)
       \<subseteq> F (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
            \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))"
  sorry


lemma mprefix_Par2F2a: 
  "\<lbrakk>B \<inter> C = {}; A \<inter> C = {}; fst x \<noteq> []; hd (fst x) \<in> ev ` A;
     ev a = hd (fst x); (t, X) \<in> F (P a); (u, Y) \<in> F (Mprefix B Q);
     tl (fst x) setinterleaves ((t, u), ev ` C \<union> {tick});
     snd x = (X \<union> Y) \<inter> (ev ` C \<union> {tick}) \<union> X \<inter> Y\<rbrakk>
    \<Longrightarrow> \<exists>t u Xa Ya.
          (t = [] \<and> Xa \<inter> ev ` A = {} \<or>
           t \<noteq> [] \<and>
           hd t \<in> ev ` A \<and> (\<exists>a. ev a = hd t \<and> (tl t, Xa) \<in> F (P a))) \<and>
          (u = [] \<and> Ya \<inter> ev ` B = {} \<or>
           u \<noteq> [] \<and>
           hd u \<in> ev ` B \<and> (\<exists>a. ev a = hd u \<and> (tl u, Ya) \<in> F (Q a))) \<and>
          fst x setinterleaves ((t, u), ev ` C \<union> {tick}) \<and>
          (X \<union> Y) \<inter> (ev ` C \<union> {tick}) \<union> X \<inter> Y =
          (Xa \<union> Ya) \<inter> (ev ` C \<union> {tick}) \<union> Xa \<inter> Ya"
  sorry


lemma mprefix_Par2F2b: 
  "(\<exists>t u Xa Ya.
        P t Xa \<and>
        Q u Ya \<and>
        fst x setinterleaves ((t, u), ev ` C \<union> {tick}) \<and>
        (X \<union> Y) \<inter> (ev ` C \<union> {tick}) \<union> X \<inter> Y =
        (Xa \<union> Ya) \<inter> (ev ` C \<union> {tick}) \<union> Xa \<inter> Ya) =
    (\<exists>u t Ya Xa.
        Q u Ya \<and>
        P t Xa \<and>
        fst x setinterleaves ((u, t), ev ` C \<union> {tick}) \<and>
        (Y \<union> X) \<inter> (ev ` C \<union> {tick}) \<union> Y \<inter> X =
        (Ya \<union> Xa) \<inter> (ev ` C \<union> {tick}) \<union> Ya \<inter> Xa)"
  sorry

lemma mprefix_Par2F2: 
  "\<lbrakk>B \<inter> C = {}; A \<inter> C = {}\<rbrakk>
    \<Longrightarrow> F (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
          \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))
       \<subseteq> F (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)"
  sorry


lemma mprefix_Par2F: 
  "\<lbrakk>B \<inter> C = {}; A \<inter> C = {}\<rbrakk>
    \<Longrightarrow> F (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q) =
       F (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
          \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))"
  sorry


lemma mprefix_Par_Int2: 
  "\<lbrakk>B \<inter> C = {}; A \<inter> C = {}\<rbrakk>
    \<Longrightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q) =
       (\<box> x \<in>  A \<rightarrow> (P x \<lbrakk>C \<rbrakk> Mprefix B Q) \<box>
        \<box> y \<in>  B \<rightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Q y))"
  sorry



(* ((Mprefix x:A@(P x))`[|`C`|]`(Mprefix x:B@(Q x))) ~=
((Mprefix x:A-((A Int C)-B)@(P x))`[|`C`|]`(Mprefix x:B-((B Int C)-A)@(Q x)))

Z.B

a->SKIP `[|`{a}`|]` b `->` a `->` SKIP = b `->` a `->` SKIP aber
STOP `[|`{a}`|]` b `->` a `->` SKIP = b `->` STOP *)


lemma mprefix_Par_Int_distr1D1a: 
  "\<lbrakk>A \<subseteq> C; B \<subseteq> C;
     \<exists>t u r v.
        front_tickFree v \<and>
        tickFree r \<and>
        x = r @ v \<and>
        r setinterleaves ((t, u), ev ` C \<union> {tick}) \<and>
        t \<in> D (Mprefix A P) \<and> u \<in> T (Mprefix B Q)\<rbrakk>
    \<Longrightarrow> x \<in> D (\<box> x \<in>  A \<inter> B \<inter> C \<rightarrow> (P x \<lbrakk>C \<rbrakk> Q x))"
  sorry

lemma mprefix_Par_Int_distr1D1: 
  " \<lbrakk>A \<subseteq> C; B \<subseteq> C\<rbrakk>
    \<Longrightarrow> D (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)
       \<subseteq> D (\<box> x \<in>  A \<inter> B \<inter> C \<rightarrow> (P x \<lbrakk>C \<rbrakk> Q x))"
  sorry

lemma mprefix_Par_Int_distr1D2: 
  "\<lbrakk>A \<subseteq> C; B \<subseteq> C\<rbrakk>
    \<Longrightarrow> D (\<box> x \<in>  A \<inter> B \<inter> C \<rightarrow> (P x \<lbrakk>C \<rbrakk> Q x))
       \<subseteq> D (Mprefix A P \<lbrakk>C\<rbrakk> Mprefix B Q)"
  sorry

lemma mprefix_Par_Int_distr1D: 
  "\<lbrakk>A \<subseteq> C; B \<subseteq> C\<rbrakk>
    \<Longrightarrow> D (Mprefix A P \<lbrakk>C\<rbrakk> Mprefix B Q) =
       D (\<box> x \<in>  A \<inter> B \<inter> C \<rightarrow> (P x \<lbrakk>C \<rbrakk> Q x))"
  sorry

lemma mprefix_Par_Int_distr1F1: 
  "\<lbrakk>A \<subseteq> C; B \<subseteq> C\<rbrakk>
    \<Longrightarrow> F (Mprefix A P \<lbrakk>C\<rbrakk> Mprefix B Q)
       \<subseteq> F (\<box> x \<in> A \<inter> B \<inter> C \<rightarrow> (P x \<lbrakk>C \<rbrakk> Q x))"
  sorry


lemma mprefix_Par_Int_distr1F2: 
  "\<lbrakk>A\<subseteq>C; B\<subseteq>C\<rbrakk>
    \<Longrightarrow> F(\<box> x \<in> A \<inter> B \<inter> C\<rightarrow> (P x) \<lbrakk>C\<rbrakk> (Q x)) 
        \<subseteq> F((\<box> x \<in> A \<rightarrow> P x)\<lbrakk> C \<rbrakk> ((\<box> x \<in> B \<rightarrow> Q x)))"
  sorry

lemma mprefix_Par_Int_distr1F: 
  "\<lbrakk>A \<subseteq> C; B \<subseteq> C\<rbrakk>
    \<Longrightarrow> F (\<box> x \<in>  A \<inter> B \<inter> C \<rightarrow> (P x \<lbrakk>C \<rbrakk> Q x))
       \<subseteq> F (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q)"
  sorry

lemma mprefix_Par_Int_distr1: 
  " \<lbrakk>A \<subseteq> C; B \<subseteq> C\<rbrakk>
    \<Longrightarrow> (Mprefix A P \<lbrakk>C \<rbrakk> Mprefix B Q) = \<box> x \<in>  A \<inter> B \<inter> C \<rightarrow> (P x \<lbrakk>C \<rbrakk> Q x)"
  sorry

lemma mprefix_Par_Int_skipD: 
  " D (Mprefix A P \<lbrakk>B \<rbrakk> SKIP) = D (\<box> x \<in>  A - B \<rightarrow> (P x \<lbrakk>B \<rbrakk> SKIP))"
  sorry

lemma mprefix_Par_Int_skipF1: 
  " \<lbrakk>X \<inter> A = {}; tick \<notin> Y\<rbrakk>
    \<Longrightarrow> ((X \<union> Y) \<inter> (B \<union> {tick}) \<union> X \<inter> Y) \<inter> (A - B) = {}"
  by (auto)

lemma mprefix_Par_Int_skipF2: 
  " X \<inter> (ev ` A - ev ` B) = {} \<Longrightarrow>
    X =
    (X - ev ` A \<union> (X - {tick})) \<inter> (ev ` B \<union> {tick}) \<union>
    (X - ev ` A) \<inter> (X - {tick})"
  by (auto)

lemma mprefix_Par_Int_skipF3: 
  " \<lbrakk>x setinterleaves ((t, [tick]), ev ` B \<union> {tick}); t \<noteq> []\<rbrakk>
    \<Longrightarrow> hd t \<notin> ev ` B"
  sorry

lemma mprefix_Par_Int_skipF4: 
  " \<lbrakk>x \<noteq> []; ev a = hd x;
     tl x setinterleaves ((t, [tick]), ev ` B \<union> {tick}); hd x \<notin> ev ` B\<rbrakk>
    \<Longrightarrow> x setinterleaves ((x, [tick]), ev ` B \<union> {tick})"
  sorry


lemma mprefix_Par_Int_skipF: 
  "F((Mprefix A P) \<lbrakk>B\<rbrakk> SKIP) = F(Mprefix (A - B)(%x.(P x) \<lbrakk> B \<rbrakk> SKIP))"
  sorry

lemma mprefix_Par_Int_skip: 
  " F (Mprefix A P \<lbrakk>B\<rbrakk> SKIP) = F (\<box> x \<in>  A - B \<rightarrow> (P x \<lbrakk>B \<rbrakk> SKIP))"
  sorry


lemma mprefix_Par_Int_skip1: 
  " F (Mprefix A P \<lbrakk>B\<rbrakk> SKIP) = F (\<box> x \<in>  A - B \<rightarrow> (P x \<lbrakk>B \<rbrakk> SKIP))"
  sorry


lemma par_Int_skip_stop: "(SKIP \<lbrakk>A\<rbrakk> STOP) = STOP"
  sorry

lemma par_Int_skip_stop1: "(STOP \<lbrakk>A\<rbrakk> SKIP) = STOP"
  sorry


lemma Inter_skip_stop: "(SKIP ||| STOP) = STOP"
  by (simp add: Inter_skip2)

lemma Inter_stop_skip: "(STOP ||| SKIP) = STOP"
  by (simp add: Inter_skip1)

lemma prefix_Par_Int_skip1: 
  "a \<notin> A \<Longrightarrow> (a \<rightarrow> P \<lbrakk>A\<rbrakk> SKIP) = (a \<rightarrow> (P \<lbrakk>A\<rbrakk> SKIP))"
  apply (simp add: write0_def)
  sorry

lemma prefix_Par_Int_skip1a: 
  "a \<notin> A \<Longrightarrow> (SKIP \<lbrakk>A\<rbrakk> a \<rightarrow> P) = (a \<rightarrow> (SKIP \<lbrakk>A\<rbrakk> P))"
  sorry

lemma prefix_Par_Int_skip2: 
  "a \<in> A \<Longrightarrow> (a \<rightarrow> P \<lbrakk>A\<rbrakk> SKIP) = STOP"
  sorry

lemma prefix_Par_Int_skip: 
  "(a \<rightarrow> P \<lbrakk>A\<rbrakk> SKIP) = (if a \<in> A then STOP else (a \<rightarrow> (P \<lbrakk>A \<rbrakk> SKIP)))"
  by (auto simp add: prefix_Par_Int_skip2 prefix_Par_Int_skip1)
  

lemma prefix_Par_Int_skip2a: 
  "a \<in> A \<Longrightarrow> (SKIP \<lbrakk>A\<rbrakk> a \<rightarrow> P) = STOP"
  sorry

lemma prefix_par_Int1: 
  " \<lbrakk>a \<in> A; b \<in> A; a \<noteq> b\<rbrakk> \<Longrightarrow> (a \<rightarrow> P \<lbrakk>A\<rbrakk> b \<rightarrow> Q) = STOP"
  sorry

lemma prefix_par_Int2: "a:A ==> ((a \<rightarrow> P) \<lbrakk>A\<rbrakk> (a \<rightarrow> Q)) = (a \<rightarrow> (P \<lbrakk> A \<rbrakk> Q))"
  sorry

lemma prefix_par: "((a \<rightarrow> P) || (a \<rightarrow> Q)) = (a \<rightarrow> (P || Q))"
  sorry

lemma prefix_Par_Int3: 
  " \<lbrakk>a \<in> C; b \<notin> C\<rbrakk> \<Longrightarrow> (a \<rightarrow> P \<lbrakk>C\<rbrakk> b \<rightarrow> Q) = (b \<rightarrow> (a \<rightarrow> P \<lbrakk>C\<rbrakk> Q))"
  sorry

lemma prefix_Par_Int4: 
  " \<lbrakk>a \<notin> C; b \<in> C\<rbrakk> \<Longrightarrow> (a \<rightarrow> P \<lbrakk>C\<rbrakk> b \<rightarrow> Q) = (a \<rightarrow> (P \<lbrakk>C\<rbrakk> b \<rightarrow> Q))"
  sorry

lemma prefix_Inter: 
  "((a \<rightarrow> P) ||| (b \<rightarrow> Q)) = ((a \<rightarrow> (P ||| (b \<rightarrow> Q))) \<box> (b \<rightarrow> ((a \<rightarrow> P) ||| Q)))"
  sorry

lemma IF_SEQ: "((if b then P else Q) `;` S) =(if b then P `;` S else Q `;` S)"
  by (simp add: if_distrib)

lemma  
  " a \<noteq> b \<Longrightarrow>
    (((b \<rightarrow> SKIP) \<lbrakk>{b}\<rbrakk> (a \<rightarrow> SKIP))|||b \<rightarrow> SKIP) \<noteq>
    ((b \<rightarrow> SKIP) \<lbrakk>{b}\<rbrakk> ((a \<rightarrow> SKIP)|||(b \<rightarrow> SKIP)))"
  sorry

lemma " \<lbrakk>directed X; finite A\<rbrakk> \<Longrightarrow> (\<Squnion>P\<in>X. P \\ A) = lub X \\ A"
  sorry

lemma cont_imp_cont_lub: "cont f \<Longrightarrow> contlub (f)"
  by (simp add:cont2contlub)



lemma dir_image2: " \<lbrakk>directed X; mono f\<rbrakk> \<Longrightarrow> directed (f ` X)"
  sorry



lemma mono_contlub_imp_cont: " \<lbrakk>mono f; contlub f\<rbrakk> \<Longrightarrow> cont f"
  sorry

lemma prefix_contlub: "cont f \<Longrightarrow> contlub (\<lambda>x. a \<rightarrow> f x)"
  sorry


lemma prefix_cont: "cont f \<Longrightarrow> cont (\<lambda>x. a \<rightarrow> f x)"
  sorry

lemma elemDIselemHD: "t \<in> D P \<Longrightarrow> tr_hide_set t (ev ` A) \<in> D (P \\ A)"
  sorry


lemma "D((P \\ {a}) \\ {b}) \<subseteq> D(P \\ ({a} \<union> {b}))"
  sorry


lemma "D(P \\ ({a} \<union> {b})) \<subseteq> D((P \\ {a}) \\ {b})"
  sorry

lemma hide_set_UnF1: 
  "D((P \\ {a}) \\ {b}) = D(P \\ ({a} \<union> {b})) \<Longrightarrow> 
   F((P \\ {a}) \\ {b}) \<subseteq> F(P \\ ({a} \<union> {b}))"
  sorry


lemma hide_set_UnF: 
  "D((P \\ {a}) \\ {b}) = D(P \\ ({a} \<union> {b})) \<Longrightarrow>
   F(P \\ ({a} \<union> {b})) \<subseteq> F((P \\ {a}) \\ {b})"
  sorry


lemma alphabet1: 
  "P \\ A = P \<Longrightarrow> \<forall>t. t \<in> T P - D P \<longrightarrow> tr_hide_set t (ev ` A) = t"
  sorry


lemma alphabet2: 
  "P \\ A = P \<Longrightarrow> \<forall>t. t \<in> min_elems (D P) \<longrightarrow> tr_hide_set t (ev ` A) = t"
  sorry

lemma setinterl_Nil_tick: 
  " \<lbrakk>u = [] \<or> u = [tick]; r setinterleaves ((t, u), ev ` A \<union> {tick})\<rbrakk>
    \<Longrightarrow> r = t \<and> (\<forall>x. x \<in> ev ` A \<longrightarrow> \<not> member t x)"
  sorry


lemma NelemAlphSpec: 
  " P \\ A = P \<Longrightarrow>
    \<forall>t. t \<in> T P - D P \<union> min_elems (D P) \<longrightarrow> tr_hide_set t (ev ` A) = t"
  sorry

lemma NelemAlphSpec1a: 
  " (P \<lbrakk>A \<rbrakk> SKIP) \<sqsubseteq> P \<Longrightarrow> \<forall>t. t \<in> T P - D P \<longrightarrow> tr_hide_set t (ev ` A) = t"
  sorry


lemma NelemAlphSpec1b: 
  " (P \<lbrakk>A \<rbrakk> SKIP) \<sqsubseteq> P \<Longrightarrow>
    \<forall>t. t \<in> min_elems (D P) \<longrightarrow> tr_hide_set t (ev ` A) = t"
  sorry


lemma NelemAlphSpec1: 
  "(P \<lbrakk>A \<rbrakk> SKIP) = P \<Longrightarrow>
    \<forall>t. t \<in> T P - D P \<union> min_elems (D P) \<longrightarrow> tr_hide_set t (ev ` A) = t"
  sorry

lemma adm_eq1: 
  "\<lbrakk>cont f; cont g\<rbrakk> \<Longrightarrow> adm (\<lambda>x. f x = g x)"
  by (simp)

lemma NelemAlphRec: 
  "\<lbrakk>cont u; \<And>x. (x \<lbrakk>A \<rbrakk> SKIP) = x \<Longrightarrow> (u x \<lbrakk>A \<rbrakk> SKIP) = u x\<rbrakk>
    \<Longrightarrow> (mu u \<lbrakk>A \<rbrakk> SKIP) = mu u"
  sorry



section {* Infra-structure for Communication Primitives *}
(* status: rudimentary, partially proven, rule-set definitvely 
          incomplete, therefore automation on this level low *)




lemma read_read_sync:
assumes contained: "(\<And>y. c y \<in> C)"
shows  "((c `?` x \<rightarrow> P x)\<lbrakk>C\<rbrakk>(c `?` x \<rightarrow> Q x)) = 
           (c `?` x \<rightarrow> ((P x)\<lbrakk>C\<rbrakk>(Q x)))"
proof -
have A: "range c \<subseteq> C"   by(insert contained, auto)
show ?thesis
  by(auto simp: read_def o_def Set.Int_absorb2 mprefix_Par_Int_distr1 A)
qed

lemmas  read_Par_Int_distr1 = read_read_sync(* legacy *)

lemma read_read_nonsync_left:
" \<lbrakk>\<And>y. c y \<notin> C; \<And>y. d y \<in> C\<rbrakk> \<Longrightarrow>
     ((c `?` x \<rightarrow> (P x)) \<lbrakk>C\<rbrakk> (d `?` x \<rightarrow>(Q x))) = 
      (c `?` x \<rightarrow> ((P x) \<lbrakk>C\<rbrakk> (d `?` x \<rightarrow>(Q x))))"
by(auto simp: read_def o_def intro!: mprefix_Par_Int_distr2)

lemmas read_Par_Int_distr2 = read_read_nonsync_left (* legacy *)

lemma read_read_nonsync_right:
" \<lbrakk>\<And>y. c y \<notin> C; \<And>y. d y \<in> C\<rbrakk> \<Longrightarrow>
     ((d `?` x \<rightarrow> (Q x)) \<lbrakk>C\<rbrakk> (c `?` x \<rightarrow>(P x))) = 
      (c `?` x \<rightarrow> ((d `?` x \<rightarrow>(Q x))  \<lbrakk>C\<rbrakk> (P x) ))"
apply(subst sync_commute)
apply(auto simp:read_read_nonsync_left sync_commute)
done

lemmas read_Par_Int_distr3 = read_read_nonsync_right 
               (* legacy modulo commute *)

lemma write_read_sync:
assumes contained:    "\<And>y. c y \<in> C"
assumes is_construct: "inj c"
shows   "((c `!` a \<rightarrow> P)\<lbrakk>C\<rbrakk>(c `?` x \<rightarrow> Q x)) = 
         (c `!` a \<rightarrow> (P\<lbrakk>C\<rbrakk>(Q a)))"
proof -
  have A : "range c \<subseteq> C"   by(insert contained, auto)
  have B : " {c a} \<inter> range c \<inter> C = {c a}" by(insert contained, auto)
show ?thesis
  apply(simp add: read_def write_def o_def )
  apply(subst mprefix_Par_Int_distr1)
  apply(auto simp: A B contained is_construct mprefix_Par_Int_distr1
                   Hilbert_Choice.inv_f_f mprefix_singl)
 done
qed

lemmas write_ParInt_read = write_read_sync (* legacy *)


lemma read_write_sync:
assumes contained:    "\<And>y. c y \<in> C"
assumes is_construct: "inj c"
shows   "((c `?` x \<rightarrow> P x)\<lbrakk>C\<rbrakk>(c `!` a \<rightarrow> Q)) = 
         (c `!` a \<rightarrow> ((P a) \<lbrakk>C\<rbrakk> Q))"
apply(subst sync_commute)
apply(auto simp: write_read_sync contained is_construct)
apply(simp only:  sync_commute)
done

lemmas read_ParInt_write = read_write_sync (* legacy *)


lemma write_read_nonsync_left:
"\<lbrakk>d a \<notin> C; \<And>y. c y \<in> C\<rbrakk> \<Longrightarrow>
 ((d `!` a \<rightarrow> P)\<lbrakk>C\<rbrakk>(c `?` x \<rightarrow> Q x)) = 
 (d `!` a \<rightarrow> (P\<lbrakk>C\<rbrakk>(c `?` x \<rightarrow> Q x)))"
apply(simp add: write_def read_def o_def)
apply(subst mprefix_Par_Int_distr2)
by auto

lemmas write_ParInt_read2 = write_read_nonsync_left (* legacy *)

lemma write0_read_nonsync_left :
"\<lbrakk>d \<in> C; \<And>y. c y \<notin> C\<rbrakk> \<Longrightarrow> 
 ((d \<rightarrow> P)\<lbrakk>C\<rbrakk>(c `?` x \<rightarrow> Q x)) = 
 (c `?` x \<rightarrow> ((d \<rightarrow> P)\<lbrakk>C\<rbrakk>Q x))"
apply(simp add: read_def o_def)
apply(subst mprefix_singl[symmetric])
apply(subst mprefix_singl[symmetric])
apply(subst mprefix_Par_Int_distr3)
apply auto
done

lemmas prefix_ParInt_read2 = write0_read_nonsync_left(* legacy *)


lemma read_write0_nonsync_left:
"\<lbrakk>d \<in> C; \<And>y. c y \<notin> C\<rbrakk> \<Longrightarrow>
 ((c `?` x \<rightarrow> Q x)\<lbrakk>C\<rbrakk>(d \<rightarrow> P)) = 
 (c `?` x \<rightarrow> (Q x\<lbrakk>C\<rbrakk>(d \<rightarrow> P)))"
by(subst sync_commute, auto simp: prefix_ParInt_read2 sync_commute)
  
lemma write0_write_nonsync_right:
"\<lbrakk>d a \<notin> C; c \<in> C\<rbrakk> \<Longrightarrow>
 ((c \<rightarrow> Q)\<lbrakk>C\<rbrakk>(d `!` a \<rightarrow> P)) = 
 (d `!` a \<rightarrow> ((c \<rightarrow> Q)\<lbrakk>C\<rbrakk>P))"
apply(simp add: write_def )
apply(subst mprefix_singl[symmetric])
apply(subst mprefix_singl[symmetric])
apply(auto simp: mprefix_Par_Int_distr3)
done

lemmas prefix_ParInt_write2 = write0_write_nonsync_right(* legacy *)

lemma write_write0_nonsync_left:
"\<lbrakk>d a \<notin> C; c \<in> C\<rbrakk> \<Longrightarrow>
 ((d `!` a \<rightarrow> P)\<lbrakk>C\<rbrakk>(c \<rightarrow> Q)) = 
 (d `!` a \<rightarrow> (P\<lbrakk>C\<rbrakk>(c \<rightarrow> Q)))"
apply(subst sync_commute)
apply(auto simp: prefix_ParInt_write2 sync_commute)
done

lemmas write_ParInt_prefix2 = write_write0_nonsync_left(* legacy *)

lemma write0_write0_sync :
  "c\<in>C \<Longrightarrow> ((c \<rightarrow> P)\<lbrakk>C\<rbrakk>(c \<rightarrow> Q)) = (c \<rightarrow> (P\<lbrakk>C\<rbrakk>Q))"
sorry


lemmas sync_rules = 
                read_read_sync read_read_nonsync_left read_read_nonsync_right
                write_read_sync read_write_sync write_read_nonsync_left
                write0_read_nonsync_left read_write0_nonsync_left
                write0_write_nonsync_right write_write0_nonsync_left
                write0_write0_sync

lemma no_hide_read_1:
"(\<And>y. c y \<notin> B) \<Longrightarrow>
 ((c `?` x \<rightarrow> (P x)) \\ B) = (c `?` x \<rightarrow> ((P x) \\ B))"
apply(simp add: read_def o_def)
apply(subst hide_mprefix_distr)
apply auto
done

lemmas hide_read_distr1 = no_hide_read_1 (* legacy *)


lemma no_hide_write:
"(\<And>y. c y \<notin> B) \<Longrightarrow> ((c `!` a \<rightarrow> P) \\ B) = (c `!` a \<rightarrow> (P \\ B))"
apply(simp add: write_def)
apply(subst hide_mprefix_distr)
apply auto
done

lemmas hide_write_distr1 = no_hide_write

lemma hide_write:
"(c a) \<in> B ==> ((c `!` a \<rightarrow> P) \\ B) = (P \\ B)"
apply(auto simp: write_def)
sorry


lemmas hide_write_distr2 = hide_write

lemma hide_write0:
"c \<in> B \<Longrightarrow> ((c \<rightarrow> P) \\ B) = (P \\ B)"
sorry

lemmas hide_rules = no_hide_read_1 no_hide_write hide_write hide_write0

lemma mono_read_ref:
" (\<And>x. P x \<sqsubseteq> Q x) \<Longrightarrow>  (c `?` x \<rightarrow> (P x)) \<sqsubseteq> (c `?` x \<rightarrow> (Q x))"
by(simp add: read_def mono_mprefix_ref)

lemma mono_write_ref:
"(P \<sqsubseteq> Q) \<Longrightarrow>  (c `!` x \<rightarrow> P) \<sqsubseteq> (c `!` x \<rightarrow> Q)"
by(simp add: write_def mono_mprefix_ref)

lemma mono_write0_ref:
"(P \<sqsubseteq> Q) \<Longrightarrow>  (c \<rightarrow> P) \<sqsubseteq> (c \<rightarrow> Q)"
by(simp add: write0_def mono_mprefix_ref)

lemmas mono_rules =   mono_read_ref mono_write_ref mono_write0_ref

 


(* ************************************************************************* *)
(* 									     *)
(* Setup for rewriting							     *)
(* 									     *)
(* ************************************************************************* *)

(*lemma mu_ind1: "[|cont f; cont g; cont h; !! x y z. x << ((y \<lbrakk> A \<rbrakk> z) \\ A) ==> 
    (f x) << (((g y) \<lbrakk> A \<rbrakk> (h z)) \\ A)|] ==> 
    (%x. f x) << (((%y. g y) \<lbrakk> A \<rbrakk> (%z. h z)) \\ A)"
  sorry

lemma mu_ind2: "[|cont f; cont g; cont h; !! x y z. x << G y z ==> 
    (f x) << G (g y) (h z)|] ==> 
    (/x. f x) << G (/y. g y) (/z. h z)"
  sorry
*)

lemmas Det_commute = det_commute (* Legacy *)
lemmas non_det_id = ndet_id
lemmas Ndet_commute = ndet_commute
lemmas non_det_bot = ndet_bot

section{* Operational Semantics *}

datatype '\<alpha> sevent = sevent "'\<alpha> event" | \<tau>

inductive op_sem :: "['\<alpha> process, '\<alpha> sevent, '\<alpha> process] => bool"
                        ("_ \<longrightarrow>\<^sub> _ _" [0,0,60] 60) 
where refl   : "P \<longrightarrow>\<^sub> \<tau> P"
     |skip   : "SKIP \<longrightarrow>\<^sub>  (sevent(tick)) Bot"
     |mpref  : "y \<in> A \<Longrightarrow> (\<box> x\<in>A \<rightarrow> P x ) \<longrightarrow>\<^sub> (sevent(ev y)) (P y)"
     |refine : "P \<sqsubseteq> Q \<Longrightarrow> Q \<longrightarrow>\<^sub> a Q' \<Longrightarrow> P \<longrightarrow>\<^sub> a Q'"

(* TODO: Simplify syntax, find trace charn, find sem charn, find correctness proof. *)
(* BUT - good idea to go for a semi-semantic inductive definition first ! *)

end

