(* ******************************************************************** *)
(*
 * Project         : HOL-CSP - A Shallow Embedding of CSP in  Isabelle/HOL
 * Version         : 2.0
 *
 * Author          : Burkhart Wolff, Abdou Feliachi
 *                   (Based on HOL-CSP 1.0 by Haykal Tej and Burkhart Wolff)
 *
 * This file       : A Combined CSP Theory
 *
 * Copyright (c) 2009 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: ROOT.ML 8591 2009-09-26 09:58:14Z brucker $ *)

header{* The Hiding Operator *}

theory  Hide
imports Process (* Legacy *)
begin

primrec  trace_hide    :: "['\<alpha> trace,('\<alpha> event) set] => '\<alpha> trace" where
        "trace_hide [] A = []"
|       "trace_hide (x # s) A = (if x \<in> A 
                                 then trace_hide s A 
                                 else x # (trace_hide s A))"


definition IsChainOver :: "[nat => '\<alpha> list, '\<alpha> list] => bool" 
                           (infixl "IsChainOver" 70) where
          "f IsChainOver t = (f 0 = t  \<and> (\<forall> i. f i < f (Suc i)))"

definition CongruentModuloHide :: "[nat => '\<alpha> trace,'\<alpha> trace , '\<alpha> set] => bool" 
          ("_ Congruent _ ModuloHide _" 70) where
           "f Congruent t ModuloHide A \<equiv> 
                        \<forall> i. trace_hide (f i) (ev ` A) = trace_hide t (ev`A)"
 

definition
  hiding  :: "['\<alpha> process ,'\<alpha> set] => '\<alpha> process"     ("_ \\ _" [73,72] 72)  where  
  "P \\ A \<equiv> Abs_Process({(s,X). \<exists> t. s = trace_hide t (ev`A) \<and> (t,X \<union> (ev`A)) \<in> F P} \<union>
                        {(s,X). \<exists> t u. front_tickFree u \<and> tickFree t \<and> 
                                       s = trace_hide t (ev`A) @ u \<and> 
                                       (t \<in> D P \<or> (\<exists> f. (f IsChainOver t) \<and> 
                                       (f Congruent t ModuloHide A) \<and> 
                                        (\<forall> i. f i \<in> T P)))},
                        {s.     \<exists> t u. front_tickFree u \<and> 
                                       tickFree t \<and> s = trace_hide t (ev`A) @ u \<and> 
                                       (t \<in> D P \<or> (\<exists> f. (f IsChainOver t) \<and> 
                                                        (f Congruent t ModuloHide A) \<and> 
                                                        (\<forall> i. f i \<in> T P)))})"



axioms
  F_hiding    : "F(P \\ A) = {(s,X). \<exists> t. s = trace_hide t (ev`A) \<and> (t,X \<union> (ev`A)) \<in> F P} \<union>
                           {(s,X). \<exists> t u. front_tickFree u \<and> tickFree t \<and> 
                                          s = trace_hide t (ev`A) @ u \<and> 
                                          (t \<in> D P \<or> (\<exists> f. (f IsChainOver t) \<and> 
                                            (f Congruent t ModuloHide A) \<and> 
                                            (\<forall> i. f i \<in> T P))) }"

  D_hiding    : "D(P \\ A) = {s.     \<exists> t u. front_tickFree u \<and> tickFree t \<and> 
                                          s = trace_hide t (ev`A) @ u \<and> 
                                          (t \<in> D P \<or> (\<exists> f. (f IsChainOver t) \<and> 
                                            (f Congruent t ModuloHide A) \<and> (\<forall> i. f i \<in> T P)))}"

  T_hiding    : "T(P \\ A) = {s.     \<exists> t. s = trace_hide t (ev`A) \<and> t \<in> T P}"

  hiding_cont [simp]: "\<lbrakk>cont f; finite A\<rbrakk> \<Longrightarrow> cont (\<lambda>x. f x \\ A)"


(* Legacy *)
lemmas tr_hide_set_def = trace_hide_def
lemmas Hide_set_def    = hiding_def
lemmas F_hide_set      = F_hiding
lemmas D_hide_set      = D_hiding
lemmas T_hide_set      = T_hiding
lemmas hide_set_cont   = hiding_cont
       
  
end


 


