(* ******************************************************************** *)
(*
 * Project         : HOL-CSP - A Shallow Embedding of CSP in  Isabelle/HOL
 * Version         : 2.0
 *
 * Author          : Burkhart Wolff
 *                   (Based on HOL-CSP 1.0 by Haykal Tej and Burkhart Wolff)
 *
 * This file       : multi-prefix operator definition
 *
 * Copyright (c) 2009 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: ROOT.ML 8591 2009-09-26 09:58:14Z brucker $ *)

header{* The Multi-Prefix Operator Definition *}

theory  Mprefix 
imports Process Legacy
begin 

section{* Backpatch Isabelle 2009-1*}


(* re-introduced from Isabelle/HOLCF 2009-1; clearly
   a derived concept, but still a useful shortcut *)

definition
  contlub :: "('a::cpo \<Rightarrow> 'b::cpo) \<Rightarrow> bool"  -- "first cont. def" where
  "contlub f = (\<forall>Y. chain Y \<longrightarrow> f (\<Squnion> i. Y i) = (\<Squnion> i. f (Y i)))"

lemma contlubE:
  "\<lbrakk>contlub f; chain Y\<rbrakk> \<Longrightarrow> f (\<Squnion> i. Y i) = (\<Squnion> i. f (Y i))"
by (simp add: contlub_def)


lemma monocontlub2cont: "\<lbrakk>monofun f; contlub f\<rbrakk> \<Longrightarrow> cont f"
apply (rule contI)
apply (rule thelubE)
apply (erule (1) ch2ch_monofun)
apply (erule (1) contlubE [symmetric])
done

lemma contlubI:
  "(\<And>Y. chain Y \<Longrightarrow> f (\<Squnion> i. Y i) = (\<Squnion> i. f (Y i))) \<Longrightarrow> contlub f"
by (simp add: contlub_def)


lemma cont2contlub: "cont f \<Longrightarrow> contlub f"
apply (rule contlubI)
apply (rule Porder.po_class.lub_eqI [symmetric])
apply (erule (1) contE)
done


section{* The core of it . . . *}

definition   Mprefix     :: "['a set,'a => 'a process] => 'a process" where
  "Mprefix A P \<equiv>  Abs_Process(  
                              {(tr,ref). tr = [] \<and> ref Int (ev ` A) = {}} \<union>

                              {(tr,ref). tr \<noteq> [] \<and> hd tr \<in> (ev ` A) \<and>
                                         (\<exists> a. ev a = (hd tr)\<and>(tl tr,ref)\<in>F(P a))},
                              {d. d \<noteq> [] \<and>  hd d  \<in> (ev ` A) \<and>
                                         (\<exists> a. ev a = hd d \<and> tl d \<in> D(P a))})"
 
syntax(HOL)
  "@mprefix" :: "[pttrn,'a set,'a process]=>'a process"	("(3[-]_ : _ -> _)"[0,0,64]64)

syntax(xsymbols)
  "@mprefix" :: "[pttrn,'a set,'a process]=>'a process"	("(3\<box> _ \<in>  _ \<rightarrow> _)"[0,0,64]64)

translations
  "\<box> x \<in> A \<rightarrow> P" == "CONST Mprefix A (% x . P)"

subsection{* Well-foundedness of Mprefix *}


lemma is_process_REP_Mp  :
"is_process ({(tr,ref). tr=[] \<and>  ref \<inter> (ev ` A) = {}} \<union>
             {(tr,ref). tr \<noteq> [] \<and> hd tr \<in> (ev ` A) \<and>
                        (\<exists> a. ev a = (hd tr) \<and> (tl tr,ref) \<in> F(P a))},
             {d. d \<noteq>  [] \<and>  hd d \<in> (ev ` A) \<and>
                        (\<exists> a. ev a = hd d  \<and> tl d \<in> D(P a))})"
 (is "is_process(?f, ?d)")
proof  (simp only:is_process_def FAILURES_def DIVERGENCES_def
                  Product_Type.fst_conv Product_Type.snd_conv,
        intro conjI allI impI)
  case goal1
  have 1: "([],{}) \<in> ?f" by simp
  show ?case by(simp add: 1)
next  
  case goal2  note asm2 = goal2 
  {
    fix    s:: "'a event list" fix X::"'a event set"
    assume H : "(s, X) \<in> ?f"
    have       "front_tickFree s"
      apply(insert H, auto simp:mem_iff  front_tickFree_def tickFree_def
                           dest!:list_nonMt_append)
      apply(case_tac "ta", auto simp: front_tickFree_charn
                           dest! : is_processT2[rule_format])
      apply(simp add: tickFree_def mem_iff)
      done 
  } note 2 = this
  show ?case  by(rule 2[OF asm2])
next
  case goal3  note asm3 = goal3
  {
    fix s t :: "'a event list" 
    assume H : "(s @ t, {}) \<in> ?f"
    have       "(s, {}) \<in> ?f"
    using H  by(auto elim: is_processT3[rule_format])
  } note 3 = this
  show ?case by(rule 3[OF asm3])
next
  case goal4  note asm4 = goal4
  { 
    fix    s:: "'a event list" fix X Y::"'a event set"
    assume H1: "(s, Y) \<in> ?f"
    assume H2: "X \<subseteq> Y"
    have       "(s, X) \<in> ?f"
      using H1 H2 by(auto intro: is_processT4[rule_format])
   } note 4 = this
  show ?case by(rule 4 [where Ya2=Y])(simp_all only: asm4)
next
  case goal5  note asm5 = goal5 
  { 
    fix s:: "'a event list" fix X Y::"'a event set"
    assume H1 : "(s, X) \<in> ?f"
    assume H2 : "\<forall> c. c\<in>Y \<longrightarrow> (s @ [c], {}) \<notin> ?f"
    have 5:     "(s, X \<union> Y) \<in> ?f "
      using H1 H2 by(auto intro!: is_processT1 is_processT5[rule_format]) 
   } note 5 = this 
   show ?case  by(rule 5,simp only: asm5,
                  rule asm5[THEN conjunct2])
next
  case goal6  note asm6 = goal6
  {
    fix s:: "'a event list" fix X::"'a event set"
    assume H : "(s @ [tick], {}) \<in> ?f"
    have  6:   "(s, X - {tick}) \<in> ?f"
      using H by(cases s, auto dest!: is_processT6[rule_format])
  } note 6 = this
  show ?case by(rule 6[OF asm6])
next
  case goal7  note asm7 = goal7
  { 
    fix s t:: "'a event list" fix X::"'a event set"
    assume H1 : "s \<in>  ?d"
    assume H2 : " tickFree s"
    assume H3 : "front_tickFree t"
    have 7:     "s @ t \<in> ?d"
      using H1 H2 H3 by(auto intro!: is_processT7_S, cases s, simp_all)
  } note 7 = this
  show ?case by(rule 7, insert asm7, auto)
next
  case goal8  note asm8 = goal8
  { 
    fix s:: "'a event list" fix X::"'a event set"
    assume H : "s \<in> ?d"
    have 8:    "(s, X) \<in> ?f"
      using H by(auto simp: is_processT8_S)
  } note 8 = this
  show ?case by(rule 8[OF asm8])
next
  case goal9  note asm9 = goal9
  { 
  fix s:: "'a event list" 
  assume H: "s @ [tick] \<in> ?d"
  have 9:   "s \<in> ?d"
    using H apply(auto)
    apply(cases s, simp_all)
    apply(cases s, auto intro: is_processT9[rule_format])
    done
  } note 9 = this
  show ?case by(rule 9, rule asm9)
  qed


lemma Rep_Abs_Mp : 
assumes H1 : "f = {(tr, ref). tr = [] \<and> ref \<inter> ev ` A = {}} \<union>
                  {(tr, ref). tr \<noteq> [] \<and> hd tr \<in> ev ` A 
                             \<and> (\<exists>a. ev a = hd tr \<and> (tl tr, ref) \<in> F (P a))}"
    and H2 : "d = {d. d \<noteq>  [] \<and>  hd d \<in> (ev ` A) \<and> 
                     (\<exists> a. ev a = hd d  \<and> tl d \<in> D(P a))}"
shows "Rep_Process (Abs_Process (f,d)) = (f,d)"
by(subst Abs_Process_inverse, 
   simp_all only: H1 H2 CollectI Process_def is_process_REP_Mp)


subsection {* Projections in Prefix *}

lemma F_Mprefix : 
"F(\<box> x \<in> A \<rightarrow> P x) = {(tr,ref). tr=[] \<and>  ref \<inter> (ev ` A) = {}} \<union>
                     {(tr,ref). tr \<noteq> [] \<and> hd tr \<in> (ev ` A) \<and> 
                               (\<exists> a. ev a = (hd tr) \<and> (tl tr,ref) \<in> F(P a))}"
by(simp add:Mprefix_def F_def Rep_Abs_Mp FAILURES_def)


lemma D_Mprefix:
"D(\<box> x \<in> A \<rightarrow> P x) = {d. d \<noteq>  [] \<and>  hd d \<in> (ev ` A) \<and> 
                        (\<exists> a. ev a = hd d  \<and> tl d \<in> D(P a))}"
by(simp add:Mprefix_def D_def Rep_Abs_Mp DIVERGENCES_def)


lemma T_Mprefix:
 "T(\<box> x \<in> A \<rightarrow> P x)={s. s=[] \<or> (\<exists> a. a\<in>A \<and> s\<noteq>[] \<and> hd s = ev a \<and> tl s \<in> T(P a))}"
by(auto simp: T_F_spec[symmetric] F_Mprefix)

subsection {* Basic Properties *}

lemma tick_T_Mprefix [simp]: "[tick] \<notin> T(\<box> x \<in> A \<rightarrow> P x)"
by(simp add:T_Mprefix)


lemma Nil_Nin_D_Mprefix [simp]: "[] \<notin> D(\<box> x \<in> A \<rightarrow> P x)"
by(simp add: D_Mprefix)

subsection{* Proof of Continuity Rule *}

lemma mono_Mprefix1:
"\<forall>a. P a \<sqsubseteq> Q a \<Longrightarrow> D (Mprefix A Q) \<subseteq> D (Mprefix A P)"
apply(auto simp: D_Mprefix)
apply(erule_tac x=xa in allE)
by(auto elim: proc_ord1 [THEN subsetD])

lemma mono_Mprefix2:
"\<forall>x. P x \<sqsubseteq> Q x \<Longrightarrow>
 \<forall>s. s \<notin> D (Mprefix A P) \<longrightarrow> Ra (Mprefix A P) s = Ra (Mprefix A Q) s"
apply(auto simp: Ra_def D_Mprefix F_Mprefix)
apply(erule_tac x = xa in allE, simp add: proc_ord2a)+
done


lemma mono_Mprefix3 :
"\<forall>x. P x \<sqsubseteq> Q x \<Longrightarrow> min_elems (D (Mprefix A P)) \<subseteq> T (Mprefix A Q)"
apply(auto simp: min_elems_def D_Mprefix T_Mprefix image_def)
apply(erule_tac x=xa in allE)
apply(auto simp:min_elems_def dest!: proc_ord3)
sorry
(*
by ev[cut_facts_tac prem 1,rtac subsetI 1,
     asm_full_simp_tac(set_ss addsimps
               [min_elems_def,D_Mprefix,T_Mprefix,image_def,Bex_def])1,
     REPEAT(SOMEGOAL(etac disjE) ORELSE SOMEGOAL(etac conjE) 
				ORELSE SOMEGOAL(etac exE))];
byev[res_inst_tac [("x","a")] exI 1,
     eres_inst_tac [("x","a")] allE 1,
     subgoals_tac ["tl x : min_elems (D(P a))"] 1,etac swap1 2,
     asm_simp_tac(set_ss addsimps [proc_ord3D])1,
     asm_full_simp_tac(set_ss addsimps [min_elems_def,Nimp_conj])2,etac exE 2];
byev[res_inst_tac [("x","hd x # t")] exI 2,
     rtac (ev_elem_image_ev_spec RS iffD1) 1,
     ALLGOALS(asm_full_simp_tac(list_ss addsimps [imageD,less_hd_tl_spec])),
     fast_tac HOL_cs 1];
val mono_Mprefix3 = result();
*)
lemma mono_Mprefix0:
"\<forall>x. P x \<sqsubseteq> Q x \<Longrightarrow> Mprefix A P \<sqsubseteq> Mprefix A Q"
apply(simp add: process_ord_def mono_Mprefix1 mono_Mprefix3)
apply(rule mono_Mprefix2)
apply(auto simp: process_ord_def)
done


lemma mono_Mprefix : "monofun(Mprefix A)"
by(auto simp: Fun_Cpo.below_fun_def monofun_def mono_Mprefix0)

lemma contlub_Mprefix : "contlub(Mprefix A)"
apply(auto simp: contlub_def)
sorry
(* Was:

val prem = goal Cont_Mprefix.thy 
" [| directed Y ; (! xa: Mprefix A `` Y. x : D xa) |] ==> \
\ (? a. ev a = hd x & (! xa. xa : (%f. f a) `` Y --> tl x : D xa)) ";
by(cut_facts_tac prem 1);
by(asm_full_simp_tac(set_ss addsimps [D_Mprefix,image_def,Ball_def,
                                     ex_imp_shift,Bex_def])1);
by(res_inst_tac [("x","(inv ev)(hd x)")] exI 1);
 br conjI 1;
by(strip_tac 2);
be conjE 2; 
by(eres_inst_tac [("x","Mprefix A xb")] allE 2);
by(eres_inst_tac [("x","xb")] allE 2);
by(rotate_tac 1 2);
by(asm_full_simp_tac(set_ss addsimps []) 2);
by(REPEAT (etac conjE 2));
 be exE 2;
 back();
 be conjE 2;
by(res_inst_tac [("t","(inv ev (hd x))")] subst 2);
 ba 3;
by(FOLD "hd x" [] 2);
by(simp_tac(set_ss addsimps event.simps@[inj_def,inv_f_f])2);
by(cut_facts_tac [(hd prem) RS directed1] 1);
by(asm_full_simp_tac(set_ss addsimps [cont_elem_eq_not_empty RS sym])1);
be exE 1;
by(eres_inst_tac [("x","Mprefix A xa")] allE 1);
by(eres_inst_tac [("x","xa")] allE 1);
by(asm_full_simp_tac(set_ss)1);
by(REPEAT(etac conjE 1));
be exE 1;
back();
be conjE 1;
by(simp_tac(set_ss addsimps [inv_def])1);
br selectI2 1;
by(ALLGOALS (fast_tac HOL_cs));
val xf = result();


val prem = goal Cont_Mprefix.thy 
"directed Y ==> \
\lub (Mprefix A `` Y) = Mprefix A (lub Y)";
br proc_ord_proc_eq_spec 1;   (- heureka ! -)
by(fast_tac (HOL_cs addSIs prem@[mono_imp_LubImageLeImageLub,mono_Mprefix]) 1);
by(cut_facts_tac [(hd prem) RS directed1] 1);
by(simp_tac(HOL_ss addsimps prem@
	[mono_Mprefix,image_dir_is_dir,lub_is_lim_proc,lub_fun_charn3,
	 dir_image2,mono_def,le_fun_def,D_Mprefix,D_LUB,INTER_def, 
 	 Ball_def,subset_spec,ex_imp_shift]) 1);
by(asm_full_simp_tac(set_ss addsimps prem@[xf, Ball_def]) 1);
by(asm_full_simp_tac(set_ss addsimps prem@
	[cont_elem_eq_not_empty RS sym,D_Mprefix,
	image_def,Ball_def,subset_spec,Bex_def,ex_imp_shift]) 1);
by(fast_tac HOL_cs 1);
val cont_Pref2 = result();
*)

lemma cont_revert2cont_pointwise: 
"\<And>x. cont (f x) \<Longrightarrow> cont (\<lambda>x y. f y x)"
sorry (*mysterious lemma from the lost part ... *)

lemma Mprefix_cont [simp]: 
"(\<And>x. cont (f x)) \<Longrightarrow> cont (\<lambda>y. \<box> z \<in>  A \<rightarrow> f z y)"
apply(rule_tac f = "\<lambda>z y. (f y z)" in Cont.cont_compose)
apply(rule monocontlub2cont)
apply(auto intro: mono_Mprefix contlub_Mprefix cont_revert2cont_pointwise)
done

(* legacy *)
(* global substitution:   mono :=>  monofun *)
lemmas proc_ord1D = proc_ord1 [THEN subsetD]
(* \<lbrakk>?P1 \<sqsubseteq> ?Q1; ?c \<in> D ?Q1\<rbrakk> \<Longrightarrow> ?c \<in> D ?P1 *)
lemmas proc_ord2b = proc_ord2a [THEN sym]
lemmas   le_fun_def =  Fun_Cpo.below_fun_def
lemmas   cont_compose1 = Cont.cont_compose
lemmas   mono_contlub_imp_cont = monocontlub2cont



subsection{* High-level Syntax *}

definition  read     :: "['a=>'b,'a set, 'a => 'b process] => 'b process"
where      "read c A P  \<equiv> Mprefix(c ` A) (P o (inv c))"
definition "write"    :: "['a=>'b, 'a, 'b process] => 'b process"
where      "write c a P \<equiv> Mprefix {c a} (\<lambda> x. P)"
definition   write0   :: "['a, 'a process] => 'a process"
where   "write0 a P \<equiv> Mprefix {a} (\<lambda> x. P)"


syntax
  "_read"   :: "[id, pttrn, 'a process] => 'a process"
                                        ("(3_`?`_ /\<rightarrow> _)" [0,0,28] 28)
  "_readX"  :: "[id, pttrn, bool,'a process] => 'a process"
                                        ("(3_`?`_`|`_ /\<rightarrow> _)" [0,0,28] 28)
  "_readS"  :: "[id, pttrn, 'b set,'a process] => 'a process"
                                        ("(3_`?`_`:`_ /\<rightarrow> _)" [0,0,28] 28)
                                        (* NOT CSP_M *)
  "_write"  :: "[id, 'b, 'a process] => 'a process"
                                        ("(3_`!`_ /\<rightarrow> _)" [0,0,28] 28)
  "_writeS" :: "['a, 'a process] => 'a process"
                                        ("(3_ /\<rightarrow> _)" [0,28] 28)


translations
  "_read c p P"     == "CONST read c CONST UNIV (\<lambda>p. P)"
  "_write c p P"    == "CONST write c p P"
  "_readX c p b P"  => "CONST read c {p. b} (\<lambda>p. P)"
  "_writeS a P"     == "CONST write0 a P"

lemma read_cont[simp]:
"(\<And>x. cont (f x)) \<Longrightarrow> cont (\<lambda>y. c`?`x \<rightarrow> f x y)"
by(simp add:read_def o_def inv_def)


lemma write_cont[simp]:
  "(\<And>x. cont (P::('b::cpo \<Rightarrow> 'a process)))
   \<Longrightarrow> cont(\<lambda>x. (c `!` a \<rightarrow> P x))"
by(simp add:write_def)

lemma write0_cont[simp]:
  "cont (P::('b::cpo \<Rightarrow> 'a process))
   \<Longrightarrow> cont(\<lambda>x. (a \<rightarrow> P x))"
by(simp add:write0_def)

end




