(* ******************************************************************** *)
(*
        Project         : HOL-CSP - A Shallow Embedding of CSP in 
                          Isabelle/HOL

        $Source: /home/uniform/rep/isa_lang/CSP/univ_csp/Rec.thy,v $
        Author          : Haykal Tej, Burkhart Wolff
                          (Last modification by $Author: bu $)

        Affiliation     : (1)University of Bremen,
                             Bremen Institute for Safe Systems (BISS)
                          (2)Albert-Ludwigs Universität Freiburg
        Level           : 3 (proofs of continuity)
        This file       : process orderings and cpo's

        $Date: 1999/11/15 18:52:46 $
        $Revision: 1.10 $
*)
(* ******************************************************************** *)

(* Recursive Operator. Based on File: PROCESS_LUB.ml in Camileri's theory *)

Rec = Bot + Fix +

instance
   process    :: (term)ord	(* declares <=,< *)
				(* i.e. recursion ordering *)

consts
   min_elems  :: "('s::ord) set => 's set"
   Ra         :: "['s process, 's trace] => ('s refusal set)" 
   				(* refusals after *) 

   "<<"       :: "['s process, 's process] => bool" 	    (infixl 50)
				(* process refinement ordering *)

   "<|'_<<"   :: "[('s process) set, 's process] => bool"   (infixl 55)
   "<<|'_<<"  :: "[('s process) set, 's process] => bool"   (infixl 55)


defs
   min_elems_def "min_elems (X) == {s. s : X & (!t. t : X --> ~(t < s))}"
   Ra_def        "Ra P s == {X . (s,X) : (F P)}"

   process_ord_def
		 "(P::'a process) <= Q == 
			(D Q <= D P &
			 (!s. (s~:(D P))-->(Ra P s = Ra Q s))&
	                 ((min_elems (D P)) <= (T Q)))"

   process_less_def
                 "(P::'a process) < Q == (P <= Q & P ~= Q)"

   process_ref_ord_def
	 	 "(P::'a process) << Q == (D Q <= D P & F Q <= F P)"

   is_ref_ub_def "Y <|_<< x == (!y:Y. y << x)"
   is_ref_lub_def"Y <<|_<< x == (Y <|_<< x & (!y. Y <|_<< y --> x << y))"

end

