(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: networks.thy 9413 2012-10-07 18:52:48Z lukasbru $ *)

theory 
  networks
imports
  FWTesting
begin

declare [[testgen_profiling]]

text{* Please increase the number of iterations even further to avoid any RSFs. *}

declare [[testgen_iterations=10]]
declare [[testgen_SMT=true]]
 

definition
   subnet     :: "adr\<^isub>i\<^isub>p net" where
  "subnet     = {{(a,b) . (a<12) }}"

definition
  "subnet_01" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_01  =  {{(a,b) . ( 0<= a) \<and> (a< 8) }}"
definition
  "subnet_02" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_02  =  {{(a,b) . ( 8<= a) \<and> (a<16) }}"
definition
  "subnet_03" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_03  =  {{(a,b) . (16<= a) \<and> (a<32) }}"
definition
  "subnet_04" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_04  =  {{(a,b) . (32<= a) \<and> (a<40) }}"
definition
  "subnet_05" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_05  =  {{(a,b) . (40<= a) \<and> (a<48) }}"
definition
  "subnet_06" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_06  =  {{(a,b) . (48<= a) \<and> (a<56) }}"
definition
  "subnet_07" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_07  =  {{(a,b) . (56<= a) \<and> (a<64) }}"
definition
  "subnet_08" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_08  =  {{(a,b) . (64<= a) \<and> (a<72) }}"
definition
  "subnet_09" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_09  =  {{(a,b) . (72<= a) \<and> (a<80) }}"
definition
  "subnet_10" ::"adr\<^isub>i\<^isub>p net" where
  "subnet_10  =  {{(a,b) . (80<= a) \<and> (a<88) }}"
definition
  "internet" ::"adr\<^isub>i\<^isub>p net" where
  "internet  =  {{(a,b) . (12>= a)}}"

definition
 not_in_same_net :: "(adr\<^isub>i\<^isub>p,DummyContent) packet \<Rightarrow> bool" where
 "not_in_same_net x  = (\<not>(src x \<sqsubset> internet \<and> dest x \<sqsubset> internet))"

definition 
  policy_00 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_00 = deny_all"

definition 
  policy_01 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_01 = policy_00 ++ allow_from_port_to (80::port) subnet_01 internet"

definition 
  policy_02 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_02 = policy_01 ++ allow_from_port_to (80::port) subnet_02 internet"

definition 
  policy_03 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_03 = policy_02 ++ allow_from_port_to (80::port) subnet_03 internet"

definition 
  policy_04 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where 
 "policy_04 = policy_03 ++ allow_from_port_to (80::port) subnet_04 internet"

definition 
  policy_05 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_05 = policy_04 ++ allow_from_port_to (80::port) subnet_05 internet"

definition 
  policy_06 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_06 = policy_05 ++ allow_from_port_to (80::port) subnet_06 internet"

definition 
  policy_07 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_07 = policy_06 ++ allow_from_port_to (80::port) subnet_07 internet"

definition 
  policy_08 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_08 = policy_07 ++ allow_from_port_to (80::port) subnet_08 internet"

definition 
  policy_09 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_09 = policy_08 ++ allow_from_port_to (80::port) subnet_09 internet"

definition 
  policy_10 :: "(adr\<^isub>i\<^isub>p, DummyContent) packet \<mapsto> unit" where
 "policy_10 = policy_09 ++ allow_from_port_to (80::port) subnet_10 internet"

lemmas UnfoldPolicy = policy_00_def policy_01_def policy_02_def policy_03_def policy_04_def 
                      policy_05_def policy_06_def policy_07_def policy_08_def policy_09_def 
                      policy_10_def

lemmas UnfoldNetworks = subnet_def internet_def subnet_01_def subnet_02_def subnet_03_def
                        subnet_04_def subnet_05_def subnet_06_def subnet_07_def subnet_08_def 
                        subnet_09_def subnet_10_def

lemmas UnfoldAll = UnfoldPolicy UnfoldNetworks PortCombinators
                   src_def dest_def adr\<^isub>i\<^isub>pLemmas in_subnet_def content_def
                   not_in_same_net_def 

declare[[testgen_SMT=true]]
declare[[testgen_iterations=3]]


test_spec "not_in_same_net x \<and> content x = data \<longrightarrow> FUT x = policy_00 x"
apply (prepare_fw_spec)
apply (simp add: UnfoldAll)
  apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_00"


gen_test_data "policy_00"
thm policy_00.test_data

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_01 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
  apply (gen_test_cases "FUT")
apply (simp_all add:  UnfoldAll)
store_test_thm "policy_01"


gen_test_data "policy_01"

thm policy_01.test_data

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_02 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_02"

gen_test_data "policy_02"

thm policy_02.test_data

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_03 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
  apply (gen_test_cases "FUT")  
apply (simp_all add: UnfoldAll)
store_test_thm "policy_03"

gen_test_data "policy_03"

  thm policy_03.test_data
  thm policy_03.pos



test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_04 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_04"


gen_test_data "policy_04"

write_clocks "time_stats_networks.tex"

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_05 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_05"

gen_test_data "policy_05"

write_clocks "time_stats_networks.tex"

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_06 x"
apply (prepare_fw_spec)
 apply (simp add: UnfoldAll)
apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_06"


gen_test_data "policy_06"

write_clocks "time_stats_networks.tex"

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_07 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_07"


gen_test_data "policy_07"

write_clocks "time_stats_networks.tex"

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_08 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_08"


gen_test_data "policy_08"

write_clocks "time_stats_networks.tex"

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_09 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_09"

write_clocks "time_stats_networks.tex"


gen_test_data "policy_09"

test_spec "not_in_same_net x \<longrightarrow> FUT x = policy_10 x"
apply (prepare_fw_spec)
  apply (simp add: UnfoldAll)
apply (gen_test_cases "FUT")
 apply (simp_all add: UnfoldAll)
store_test_thm "policy_10"



gen_test_data "policy_10"


write_clocks "time_stats_networks.tex"

declare [[testgen_profiling = false]]


end
