(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * FWTesting.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************.***************************************************************)
(* $Id: Scenario3.thy 9440 2012-10-11 21:06:26Z lukasbru $ *)

theory "Scenario3"
imports FWTesting
begin


declare [[testgen_profiling]]



definition
 Intranet :: "adr\<^isub>i\<^isub>p net" where 
"Intranet = {{(a,b). a = 1}}"

definition
 DMZ :: "adr\<^isub>i\<^isub>p net" where
"DMZ = {{(a,b). a = 2}}"

definition
 Internet:: "adr\<^isub>i\<^isub>p net" where
"Internet = {{(a,b). a > 2}}" 


definition
 Policy:: "(adr\<^isub>i\<^isub>p net,port) Combinators" where
"Policy = DenyAll \<oplus>  
          AllowPortFromTo Internet DMZ 56 \<oplus> 
          AllowPortFromTo Intranet Internet 3 \<oplus>
          DenyAllFromTo Intranet Internet \<oplus>
          AllowPortFromTo Internet DMZ 65\<oplus> 
          AllowPortFromTo DMZ Intranet 32 \<oplus> 
          AllowPortFromTo Intranet Internet 54 \<oplus>
          AllowPortFromTo Internet DMZ 56 \<oplus> 
          AllowPortFromTo Intranet Internet 45 \<oplus>
          AllowPortFromTo Intranet Internet 76 \<oplus>
          AllowPortFromTo Internet Intranet 21 \<oplus> 
          AllowPortFromTo DMZ Intranet 32" 
 
 
 

lemmas PolicyLemmas = Policy_def
                        DMZ_def Internet_def Intranet_def 

lemmas PolicyL =   Policy_def

definition 
 not_in_same_net :: "(adr\<^isub>i\<^isub>p,DummyContent) packet \<Rightarrow> bool" where
"not_in_same_net x =  (((src x \<sqsubset> DMZ) \<longrightarrow> ( \<not> (dest x \<sqsubset> DMZ))) \<and>
                      ((src x \<sqsubset>  Internet) \<longrightarrow> ( \<not> (dest x \<sqsubset> Internet))) \<and>
                      ((src x \<sqsubset>  Intranet) \<longrightarrow> ( \<not> (dest x \<sqsubset> Intranet))))"
                   
consts fixID :: id
consts fixContent :: DummyContent

definition "fixElements p = (id p = fixID \<and> content p = fixContent)"

lemmas fixDefs = fixElements_def NetworkCore.id_def NetworkCore.content_def



lemma sets_distinct1: "(n::int) \<noteq> m \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
apply auto
done

lemma sets_distinct2: "(m::int) \<noteq> n \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
apply auto
done



lemma sets_distinct3: "{((a::int),(b::int)). a = n} \<noteq> {(a,b). a > n}"
apply auto
done


lemma sets_distinct4: "{((a::int),(b::int)). a > n} \<noteq> {(a,b). a = n}"
apply auto
done


lemma aux: "\<lbrakk>a \<in> c; a \<notin> d; c = d\<rbrakk> \<Longrightarrow> False"
apply auto
done


lemma sets_distinct5: "(s::int) < g \<Longrightarrow> {(a::int, b::int). a = s} \<noteq> {(a::int, b::int).  g < a}"
apply (auto simp: sets_distinct3)
apply (subgoal_tac "(s,4) \<in> {(a::int,b::int). a = (s)}")
apply (subgoal_tac "(s,4) \<notin> {(a::int,b::int). g < a}")
apply (erule aux)
apply assumption+
apply simp
apply (thin_tac "?S")+
apply simp
done


lemma sets_distinct6: "(s::int) < g \<Longrightarrow> {(a::int, b::int).  g < a} \<noteq> {(a::int, b::int).  a = s}"
apply (rule not_sym)
apply (rule sets_distinct5)
by simp


lemma distinctNets: "DMZ \<noteq>  Intranet \<and> DMZ \<noteq>  Internet \<and>  Intranet \<noteq>  DMZ \<and> Intranet \<noteq>  Internet \<and>  Internet \<noteq>  DMZ \<and> Internet \<noteq>  Intranet"
apply (simp add: PolicyLemmas sets_distinct1 sets_distinct2 sets_distinct3 sets_distinct4 sets_distinct5 sets_distinct6)
done



lemma aux5: "\<lbrakk>x \<noteq> a; y\<noteq>b; (x \<noteq> y \<and> x \<noteq> b) \<or>  (a \<noteq> b \<and> a \<noteq> y)\<rbrakk> \<Longrightarrow> {x,a} \<noteq> {y,b}"
  apply auto
done


lemma aux2: "{a,b} = {b,a}"
  apply auto
done


lemma noMT: "\<forall> x \<in> set (policy2list Policy). dom (C x) \<noteq> {}"
apply (simp add: PolicyLemmas)
apply (simp add: PLemmas PolicyLemmas)
apply (rule_tac x = 14 in exI)
apply simp
done



lemma ANDex: "allNetsDistinct (policy2list Policy)"
apply (simp add: PolicyL allNetsDistinct_def distinctNets)
apply (auto simp: PLemmas PolicyLemmas netsDistinct_def sets_distinct5 sets_distinct6)
done


lemma count_the_rules: "(int (length(policy2list (list2FWpolicy(normalize Policy)))) = post) \<and> (int(length (policy2list Policy)) = pre) \<and>
 (int (length((normalize Policy))) = Partitions)"
apply (insert distinctNets noMT)
apply (simp add: normalize_def PolicyL bothNets_def aux5 aux2 Nets_List_def, thin_tac "?X",thin_tac "?S")
oops



lemma normedPolicy: "normalize Policy = X"
apply (insert distinctNets noMT)
apply (simp add: normalize_def PolicyL bothNets_def aux5 aux2 Nets_List_def, thin_tac "?X",thin_tac "?S")
oops

fun (sequential) numberOfRules where 
 "numberOfRules (a\<oplus>b) = numberOfRules a + numberOfRules b"
 |"numberOfRules a = (1::int)" 




fun numberOfRulesList where 
 "numberOfRulesList  (x#xs) = ((numberOfRules x)#(numberOfRulesList xs)) "
 |"numberOfRulesList [] = []" 


lemma "numberOfRulesList (normalize Policy) = X"
apply (insert distinctNets noMT)
apply (simp add: normalize_def PolicyL bothNets_def aux5 aux2 Nets_List_def, thin_tac "?X",thin_tac "?S")
oops


lemma all_in_list: "all_in_list (policy2list Policy) (Nets_List Policy)"
apply (simp add: PolicyL)
apply (unfold Nets_List_def)
apply (unfold bothNets_def)
apply (insert distinctNets)
apply simp
apply (rule disjI2)
apply (rule disjI1)
apply (metis insert_commute)
done



declare [[testgen_iterations=25]]



lemmas normalizeUnfold =   normalize_def PolicyL Nets_List_def bothNets_def aux aux2 bothNets_def sets_distinct1 sets_distinct2 sets_distinct3 sets_distinct4 sets_distinct5 sets_distinct6  aux5 aux2

lemmas casesSimp = fixDefs packet_Nets_def PLemmas PolicyLemmas



lemma noMT2: "\<forall> x \<in> set (policy2list Policy). dom (C x) \<noteq> {}"
apply (simp add: PLemmas  normalize_def bothNets_def
                 PolicyLemmas aux5 aux2 Nets_List_def ) 
by (metis zless_add1_eq)


test_spec "fixElements x \<longrightarrow> FUT x = C Policy x"
apply (prepare_fw_spec)
apply (rule_tac PL = "(map C (rev (normalize (Policy))))" 
   in distrPUTLG1)
apply (simp add: Policy_def domDA1)
apply (rule fun_cong) back
apply (rule NormalisationIntegerPortProof.p_eq2)
apply (rule NormalisationIntegerPortProof.norm_notMT)
apply (simp add: Policy_def)
apply (rule NormalisationIntegerPortProof.C_eq_normalize)
apply (simp add: Policy_def) 
apply (rule ANDex)
apply (rule all_in_list)
apply (insert distinctNets noMT)
apply (insert noMT2)
apply (simp add: normalize_def PolicyL bothNets_def aux5 aux2 Nets_List_def,
                                 thin_tac "?X",thin_tac "?S",thin_tac "?S")

apply (simp add: PUTListGD_def noMT2)
apply (simp add: gatherDomain_def domain_reasoning dom_2 domDA1)
apply (tactic  
 "REPEAT(CHANGED(TRYALL((rtac @{thm conjI}) THEN' (rtac @{thm impI}))))")
apply (simp_all add: PLemmas PolicyLemmas split_tupled_all domain_reasoning)
apply (gen_test_cases FUT simp:deny_all_def)
store_test_thm "PolicyN"

gen_test_data "PolicyN"


thm PolicyN.test_data


test_spec "fixElements x  \<and> not_in_same_net x   \<longrightarrow>  FUT x = (C Policy) x"
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)
apply( simp add: PolicyLemmas not_in_same_net_def)
apply (simp only: PLemmas)
apply (simp)
apply (simp add: fixDefs PLemmas)
  apply (gen_test_cases "FUT")
  store_test_thm "OriginalPolicy"

gen_test_data "OriginalPolicy"


write_clocks "time_stats_Scenario3"

end
