(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * FWTesting.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************.***************************************************************)
(* $Id: Normalized.thy 9427 2012-10-11 13:32:50Z lukasbru $ *)

theory 
  "Normalized"
imports
  "../../PolicyDefinitions"
begin



declare [[testgen_profiling]]


definition Policy :: "(adr\<^isub>i\<^isub>p net,port) Combinators" where
"Policy = 
DenyAll \<oplus>  AllowPortFromTo one three 392 \<oplus> AllowPortFromTo three two 874 \<oplus> DenyAllFromTo one two\<oplus>
  AllowPortFromTo three two 763 \<oplus> DenyAllFromTo two three\<oplus> AllowPortFromTo two one 567 \<oplus> 
AllowPortFromTo two three 953 \<oplus> AllowPortFromTo three one 296 \<oplus> DenyAllFromTo one three\<oplus> 
AllowPortFromTo one two 789 \<oplus> AllowPortFromTo one three 692 \<oplus> AllowPortFromTo two three 257
"

lemma "dupl (policy2list Policy) = X"
apply (insert nets_different)
apply (simp add: Policy_def,thin_tac "?X")
oops





lemmas policies =  Policy_def 

lemmas UnfoldNetworkAndPolicy = UnfoldPolicy PLemmas policies


lemmas normalizeUnfold =  normalize_def policies Nets_List_def aux aux2 bothNets_def

lemmas casesSimp = fixDefs packet_Nets_def PLemmas UnfoldNetworkAndPolicy


lemma noMT2: "\<forall> x \<in> set (policy2list Policy). dom (C x) \<noteq> {}"
apply (simp add: UnfoldNetworkAndPolicy)
done


lemma count_the_rules: "(int (length(policy2list (list2FWpolicy(normalize Policy)))) = post) \<and> 
  (int(length (policy2list Policy)) = pre) \<and>
 (int (length((normalize Policy))) = Partitions)"
apply (insert nets_different noMT2)
apply (simp add:  normalizeUnfold)
apply (thin_tac "?X")+
oops



lemma normalize: "normalize Policy = X"
apply (insert nets_different noMT2)
apply (simp add: normalizeUnfold, thin_tac "?X",thin_tac "?S")
oops


lemma concAssoc3: "C (a \<oplus> (b \<oplus> c) \<oplus> d) = 
                   C (a \<oplus> b \<oplus> c \<oplus> d)"
apply (simp add: C.simps)
done



lemma dist1: "map C (rev (normalize Policy)) \<noteq> []"
by (insert nets_different noMT2,simp_all add: normalizeUnfold)

lemma dist2: "disjDomGD (map C (rev (normalize Policy)))"
apply (insert nets_different noMT2,simp_all add: normalizeUnfold, thin_tac "?X",thin_tac "?A")
apply (simp add: disjDomGD_def)
apply (simp add: domSubset1 domSubset2 ConcAssoc2 ConcAssoc concAssoc3)
apply (simp add: PLemmas UnfoldNetworkAndPolicy)
apply auto
done





lemma aND: "allNetsDistinct (policy2list Policy)"
apply (simp add: Policy_def allNetsDistinct_def nets_different)
apply (simp add: netsDistinct_def)
apply (auto simp: allNetsDistinct_def PLemmas UnfoldPolicy)
done

test_spec "fixElements x \<longrightarrow> FUT x = C Policy x"
apply (prepare_fw_spec)
apply (rule_tac PL = "map C (rev (normalize Policy))" 
   in distrPUTLG1)
apply (simp add: Policy_def domDA1)
apply (rule fun_cong) back
apply (rule NormalisationIntegerPortProof.p_eq2)
apply (rule NormalisationIntegerPortProof.norm_notMT)
apply (simp add: Policy_def)
apply (rule NormalisationIntegerPortProof.C_eq_normalize)
apply (simp add: Policy_def)
apply (rule aND)
apply (simp_all add: nets_different noMT2 normalizeUnfold)
apply (simp add: PUTListGD_def)
apply (simp add: gatherDomain_def domain_reasoning dom_2 domDA1)
apply (tactic  
 "REPEAT(CHANGED(TRYALL((rtac @{thm conjI}) THEN' (rtac @{thm impI}))))")
apply (simp_all add: PLemmas UnfoldPolicy)
apply (gen_test_cases FUT)
store_test_thm "PolicyN"




test_spec "fixElements x \<longrightarrow>  FUT x = C Policy x"
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)
apply (simp add: fixDefs packet_Nets_def)
apply (simp add:  UnfoldNetworkAndPolicy)
apply (gen_test_cases "FUT")
apply simp_all
store_test_thm "full"




write_clocks "../clock6.tex"
declare [[testgen_profiling=false]]


end
