theory 
  Scenario3
imports 
  FWTesting
begin

subsubsection{* Network Model *}

text{* To define NAT rules, first the respective addresses need to be
defined (as sets of Integers).  *} 

definition Adr1_1 :: "int set"
where "Adr1_1 = {d. d > 1 \<and> d < 8}"

definition Adr1_2 :: "int set" where
 "Adr1_2 = {d. d > 7 \<and> d < 16}"

definition Adr1_3 :: "int set" where
 "Adr1_3 = {d. d > 15 \<and> d < 24}"


definition Adr2_1 :: "int set" where
 "Adr2_1 = {d. d = 1004}"

definition Adr2_2 :: "int set" where
 "Adr2_2 = {d. d = 1005}"

definition Adr2_3 :: "int set" where
 "Adr2_3 = {d. d > 1015 \<and> d < 1024}"


definition subnet1 :: "adr\<^isub>i\<^isub>p\<^isub>p net" where
 "subnet1 = {{(d,e). d > 1 \<and> d < 256}}"

definition subnet2 :: "adr\<^isub>i\<^isub>p\<^isub>p net" where
 "subnet2 = {{(d,e). d > 1001 \<and> d < 1256}}"


subsubsection{* Policy Model *} 

text{* The filtering part of the policy is as before. But we also have
to model the NAT rules as is shown below. E.g. in \texttt{NAT3} all
packets which have their source address within Adr1\_3 are transformed
to a packet with a source address within Adr2\_3.  *}

definition
 Filter :: "(adr\<^isub>i\<^isub>p\<^isub>p,DummyContent) packet \<mapsto> unit" where
"Filter =   
   allow_from_to_ports_prot tcp {21::port,80,443,25} subnet1 subnet2 \<Oplus>
   allow_from_to_ports_prot udp {68::port,67,138} subnet1 subnet2 \<Oplus>
   allow_from_to_ports_prot tcp {45::port,346,457} subnet2 subnet1 \<Oplus> 
   deny_all"


definition NAT1 :: "(adr\<^isub>i\<^isub>p\<^isub>p,DummyContent) packet \<mapsto> 
                   (((adr\<^isub>i\<^isub>p\<^isub>p,DummyContent) packet) set)" where
 "NAT1 = srcPat2pool_IntProtocol Adr1_1 Adr2_1"


definition NAT2 where
 "NAT2 = srcPat2pool_IntProtocol Adr1_2 Adr2_2"

definition NAT3 where
 "NAT3 = srcNat2pool_IntProtocolPort Adr1_3 Adr2_3"

text {* The individual Nat rules are combined using the usual override
operator, with a default allow everything mapped to itself rule in the
end, and can then be combined with the filtering policy using one of
the standard UPF operators. NATID is the default NAT rule used for all
packets which should not be transformed.  *}

definition TestPolicy where
 "TestPolicy = ((\<lambda> (x,y). x) o_f 
 (((NAT1 \<Oplus> NAT2 \<Oplus> NAT3 \<Oplus> (D\<^sub>f (\<lambda>x.{x}))) \<Otimes>\<^sub>2 Filter) o (\<lambda> x. (x,x))))"


subsubsection{* Definitions of the testing constraints. *} 
text{* We use the standard testing constraints: *}
definition
  "accross_subnets x  \<equiv> 
 ((src x \<sqsubset> subnet1 \<and> (dest x \<sqsubset> subnet2)) \<or>
  (src x \<sqsubset> subnet2 \<and> (dest x \<sqsubset> subnet1)))"



subsubsection{* Test Generation *}


lemmas NAT = NAT1_def NAT2_def NAT3_def

lemmas policyLemmas = TestPolicy_def 
                      NATLemmas NAT Filter_def
                      ProtocolPortCombinators.ProtocolCombinators 
                      adr\<^isub>i\<^isub>p\<^isub>pLemmas adr\<^isub>i\<^isub>p\<^isub>pTestConstraints
                      packet_defs

lemmas subnets = subnet1_def subnet2_def 

lemmas Adr1 = Adr1_1_def Adr1_2_def Adr1_3_def  
lemmas Adr2 = Adr2_1_def Adr2_2_def Adr2_3_def  
lemmas Adr = Adr1 Adr2


text{* Test Case generation is similar as before. However, for
efficiency reasons, we do not completely unfold the policy before test
case generation. The way it is done here is the standard way when
using NAT rules. *}

lemma foo[simp]:"\<lbrakk>n < x; (x::int) < m; m = Suc n\<rbrakk> \<Longrightarrow> False"
by auto

lemma foo1[simp]:"\<lbrakk>A \<and> (n < x) \<and> B \<and> ((x::int) < (Suc n)) \<and> D\<rbrakk> \<Longrightarrow> False"
by auto

lemma foo2[simp]:"\<lbrakk>A \<and> (n < x) \<and> B \<and> ((x::int) < m); m = n+1\<rbrakk> \<Longrightarrow> False"
by auto


lemma foo3[simp]:" m = n + 1 \<Longrightarrow> (A \<and> E \<and> F \<and> G \<and> (n < x) \<and> B \<and> H \<and> ((x::int) < m)) = False"
apply auto
done

lemma foo4[simp]:" m = n + 1 \<Longrightarrow> (A \<and> (n < x) \<and> H \<and> ((x::int) < m)) = False"
apply auto
done


lemma foo5[simp]:" m = n + 1 \<Longrightarrow> (A \<and> (n < x)  \<and> H \<and> ((x::int) < m) \<and> B) = False"
apply auto
done

test_spec 
"\<And>x. \<lbrakk> accross_subnets x;port_positive x; fix_values x\<rbrakk>  \<Longrightarrow> 
FUT x = TestPolicy x"
apply (prepare_fw_spec)
apply (simp add: subnets policyLemmas accross_subnets_def)
(*apply (simp split: option.splits decision.splits prod.splits)*)
apply (gen_test_cases FUT simp: subnets Adr datasimps)
store_test_thm "Scenario3"


declare[[testgen_SMT=true]]
declare[[testgen_iterations=54]]

gen_test_data "Scenario3"
text {* For this scenario, 131 test cases are generated.*}

export_fw_data "Scenario3.data" Scenario3
thm Scenario3.pos

end
