theory 
  Scenario5
imports 
  FWTesting
begin


subsubsection{* Network Model *}

definition net1 :: "int set" where "net1 = {x. x > 1 \<and> x <  255}"
definition net2 :: "int set" where "net2 = {x. x > 1001 \<and> x < 1255}"
definition net3 :: "int set" where "net3 = {x. x > 2001 \<and> x < 2255}"
definition net4 :: "int set" where "net4 = {x. x > 3001 \<and> x < 3255}"


definition src1 :: "int set" where "src1 = {5}"
definition src2 :: "int set" where "src2 = {2004}"
definition to1 :: "int set"where "to1 = {15}"
definition to2 :: "int set"where "to2 = {2014}"

lemma src_in_net1[simp]: "src1 \<subseteq> net1"
apply (simp add: net1_def src1_def)
done


lemma src_in_net2[simp]: "src2 \<subseteq> net3"
apply (simp add: net3_def src2_def)
done

definition
 subnet1 :: "adr\<^isub>i\<^isub>p\<^isub>p net" where
"subnet1 = {{(a,b,c). a \<in> net1}}"

definition
 subnet2 :: "adr\<^isub>i\<^isub>p\<^isub>p net" where
"subnet2 = {{(a,b,c). a \<in> net2}}"

definition
 subnet3 :: "adr\<^isub>i\<^isub>p\<^isub>p net" where
"subnet3 = {{(a,b,c). a \<in> net3}}"

definition
 subnet4 :: "adr\<^isub>i\<^isub>p\<^isub>p net" where
"subnet4 = {{(a,b,c). a \<in> net4}}"



subsubsection{* Policy Model *}                    
definition
 Filter :: "(adr\<^isub>i\<^isub>p\<^isub>p net, protocol \<times> port) Combinators" where
"Filter =    (DenyAll \<oplus>
              AllowPortFromTo subnet1 subnet2 (tcp, (80::port)) \<oplus>
              AllowPortFromTo subnet1 subnet3 (tcp, (8080::port)) \<oplus>
              AllowPortFromTo subnet2 subnet4 (udp, (546::port)) \<oplus>
              AllowPortFromTo subnet2 subnet1 (tcp, (86::port)) \<oplus>
              AllowPortFromTo subnet4 subnet1 (udp, (12::port)) \<oplus>
              AllowPortFromTo subnet3 subnet2 (tcp, (32::port)) \<oplus>
              AllowPortFromTo subnet1 subnet4 (udp, (65::port)) \<oplus>
              AllowPortFromTo subnet2 subnet4 (tcp, (861::port)))"

definition NAT1  where
 "NAT1 = srcNat2pool_IntProtocolPort src1 to1"

definition NAT2  where
 "NAT2 = srcNat2pool_IntProtocolPort src2 to2"

definition NAT where
 "NAT = (NAT1 \<Oplus> NAT2 \<Oplus> (\<lambda>p. Some (allow {p})))"

definition Policy where
 "Policy = ((\<lambda> (x,y). x) o_f ((NAT \<Otimes>\<^sub>2 (Cp Filter)) o  (\<lambda>x. (x,x))))"



subsubsection{* Definitions of the testing constraints. *} 
text{* We use the standard testing constraints: *}

definition
 accross_subnets :: "(adr\<^isub>i\<^isub>p\<^isub>p,'b) packet \<Rightarrow> bool" where
"accross_subnets x = (((src x \<sqsubset> subnet1) \<longrightarrow> (\<not> (dest x \<sqsubset> subnet1))) \<and>  
                    ((src x \<sqsubset> subnet2) \<longrightarrow> (\<not> (dest x \<sqsubset> subnet2))) \<and>  
                    ((src x \<sqsubset> subnet3) \<longrightarrow> (\<not> (dest x \<sqsubset> subnet3))) \<and>  
                    ((src x \<sqsubset> subnet4) \<longrightarrow> (\<not> (dest x \<sqsubset> subnet4))) \<and>
                (src x \<sqsubset> subnet1 \<or> src x \<sqsubset> subnet2 \<or> src x \<sqsubset> subnet3 \<or> src x \<sqsubset> subnet4) \<and>  
               (dest x \<sqsubset> subnet1 \<or> dest x \<sqsubset> subnet2 \<or> dest x \<sqsubset> subnet3 \<or> dest x \<sqsubset> subnet4))"
                    

subsubsection{* Test Generation *}

lemmas policyLemmas = ProtocolPortCombinators.ProtocolCombinators 
                      adr\<^isub>i\<^isub>p\<^isub>pTestConstraints
                      adr\<^isub>i\<^isub>p\<^isub>pLemmas
                      packet_defs Policy_def NAT_def NAT1_def NAT2_def
                      Cp.simps NATLemmas
                      accross_subnets_def subnet1_def subnet2_def
                      subnet3_def subnet4_def dom_def 
                       

                     
lemmas nets = net1_def net2_def net3_def net4_def src1_def src2_def to1_def to2_def

text{* Again, we need to prove a couple of lemmas, they are of the
same kind as before. *}

lemma ex1: "(\<exists>a. a \<in> net1)"
apply (auto simp: nets)
apply (rule_tac x = 14 in exI)
by auto

lemma ex2: "(\<exists>a. a \<in> net2)"
apply (auto simp: nets)
apply (rule_tac x= 1013 in exI)
by auto


lemma ex3: "(\<exists>a. a \<in> net3)"
apply (auto simp: nets)
apply (rule_tac x= 2013 in exI)
by auto

lemma ex4: "(\<exists>a. a \<in> net4)"
apply (auto simp: nets)
apply (rule_tac x= 3013 in exI)
by auto


lemma nets_different: "subnet1 \<noteq> subnet2 \<and> subnet1 \<noteq> subnet3 \<and> 
                       subnet1 \<noteq> subnet4 \<and> subnet2 \<noteq> subnet4 \<and> 
                       subnet2 \<noteq> subnet1 \<and> subnet2 \<noteq> subnet3 \<and>
                       subnet3 \<noteq> subnet1 \<and> subnet3 \<noteq> subnet2 \<and> 
                       subnet3 \<noteq> subnet4 \<and> subnet4 \<noteq> subnet1 \<and> 
                       subnet4 \<noteq> subnet2 \<and> subnet4 \<noteq> subnet3"
apply (simp add: policyLemmas )
apply (rule conjI, rule setDistinct1, simp add: nets setDistinct4, 
     smt CollectD ex2 net2_def ex1 ex3 ex4 net1_def net3_def net4_def)+
apply (rule setDistinct1, simp add: nets setDistinct4, 
     smt CollectD ex2 net2_def ex1 ex3 ex4 net1_def net3_def net4_def)+
done

lemma noMT: "\<forall> x \<in> set (policy2list Filter). dom (Cp x) \<noteq> {}" 
apply (simp add: Filter_def subnet1_def subnet2_def subnet3_def subnet4_def
 adr\<^isub>i\<^isub>p\<^isub>pLemmas)
apply (rule conjI)
apply (metis NormalisationIPPProofs.DAnMT)
apply (simp add: PLemmas  ex1 ex2 ex3 ex4)
done


lemma ANDex: "allNetsDistinct (policy2list Filter)"
apply (simp add: Filter_def allNetsDistinct_def nets_different)
apply (auto simp: policyLemmas netsDistinct_def nets)
done


lemma all_in_list: "all_in_list (policy2list Filter) (Nets_List Filter)"
apply (simp add: Filter_def)
apply (unfold Nets_List_def)
apply (unfold bothNets_def)
apply (insert nets_different,simp)
apply (metis insert_commute)
done



lemma C_Filter_compile: " 
Cp (list2FWpolicy (normalizePr Filter)) =  Cp Filter"
apply (unfold normalizePr_def)
apply (rule C_eq_compile)
apply (simp add: Filter_def)
apply (rule all_in_list)
apply (rule ANDex)
done

lemma FilterUniv: "dom (Cp Filter) = UNIV" 
apply (simp add: Filter_def)
apply (rule domDA1Cp)
done



lemma NATuniv: "dom NAT = UNIV"
apply (simp add: dom_def NAT_def)
done


lemma PolUniv: "dom Policy = UNIV"
apply (simp add: Policy_def)
apply (simp add: Filter_def)
apply (auto simp: NATuniv domDA1Cp dom_2)
done


lemma net_noteq12: "(x \<in> net1) \<Longrightarrow> (x \<notin> net2)" 
by (auto simp: nets)
lemma net_noteq13: "(x \<in> net1) \<Longrightarrow> (x \<notin> net3)" 
by (auto simp: nets)
lemma net_noteq14: "(x \<in> net1) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)
lemma net_noteq23: "(x \<in> net2) \<Longrightarrow> (x \<notin> net3)" 
by (auto simp: nets)
lemma net_noteq24: "(x \<in> net2) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)
lemma net_noteq34: "(x \<in> net3) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)

lemma net_noteq40: "(x \<in> src1) \<Longrightarrow> (x \<notin> net2)" 
by (auto simp: nets)
lemma net_noteq41: "(x \<in> src1) \<Longrightarrow> (x \<notin> net3)" 
by (auto simp: nets)
lemma net_noteq42: "(x \<in> src1) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)
lemma net_noteq43: "(x \<in> to1) \<Longrightarrow> (x \<notin> net2)" 
by (auto simp: nets)
lemma net_noteq44: "(x \<in> to1) \<Longrightarrow> (x \<notin> net3)" 
by (auto simp: nets)
lemma net_noteq45: "(x \<in> to1) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)


lemma net_noteq50: "(x \<in> src2) \<Longrightarrow> (x \<notin> net1)" 
by (auto simp: nets)
lemma net_noteq51: "(x \<in> src2) \<Longrightarrow> (x \<notin> net2)" 
by (auto simp: nets)
lemma net_noteq52: "(x \<in> src2) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)
lemma net_noteq53: "(x \<in> to2) \<Longrightarrow> (x \<notin> net1)" 
by (auto simp: nets)
lemma net_noteq54: "(x \<in> to2) \<Longrightarrow> (x \<notin> net2)" 
by (auto simp: nets)
lemma net_noteq55: "(x \<in> to2) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)

lemma foo[simp]:"\<lbrakk>n < x; (x::int) < m; m = Suc n\<rbrakk> \<Longrightarrow> False"
by auto

lemma foo1[simp]:"\<lbrakk>A \<and> (n < x) \<and> B \<and> ((x::int) < (Suc n)) \<and> D\<rbrakk> \<Longrightarrow> False"
by auto

lemma foo2[simp]:"\<lbrakk>A \<and> (n < x) \<and> B \<and> ((x::int) < m); m = n+1\<rbrakk> \<Longrightarrow> False"
by auto


lemma foo3[simp]:" m = n + 1 \<Longrightarrow> (A \<and> E \<and> F \<and> G \<and> (n < x) \<and> B \<and> H \<and> ((x::int) < m)) = False"
apply auto
done

lemma foo4[simp]:" m = n + 1 \<Longrightarrow> (A \<and> (n < x) \<and> H \<and> ((x::int) < m)) = False"
apply auto
done


lemma foo5[simp]:" m = n + 1 \<Longrightarrow> (A \<and> (n < x)  \<and> H \<and> ((x::int) < m) \<and> B) = False"
apply auto
done


lemmas net_noteq = net_noteq12 net_noteq13 net_noteq14
                   net_noteq23 net_noteq24 net_noteq34
                   net_noteq40 net_noteq41 net_noteq42 
                   net_noteq43 net_noteq44 net_noteq45 
                   net_noteq50 net_noteq51 net_noteq52 
                   net_noteq53 net_noteq54 net_noteq55 

test_spec 
 "port_positive x \<and> accross_subnets x \<and> 
  fix_values x \<longrightarrow> 
 FUT x = Policy x"
apply (prepare_fw_spec)
apply (rule_tac PL = "(NAT \<Otimes>\<^sub>L (map Cp (rev (normalizePr (Filter))))) 
                       (op \<Otimes>\<^sub>2) (\<lambda> (x,y). x) (\<lambda> x. (x,x))" 
   in distrPUTLG1)
apply (simp add: PolUniv)
apply (simp add: Policy_def)
apply (rule sym)
apply (subst normalizePrNAT)
apply (simp add: Filter_def)
apply (rule ANDex)
apply (rule all_in_list)
apply simp
apply (insert nets_different noMT)
apply (simp add:  normalizePr_def Filter_def bothNets_def aux aux2 
                  Nets_List_def)
apply (rotate_tac -1, thin_tac "?X",rotate_tac -1, thin_tac "?X")
apply (simp add: PUTListGD_def)
apply (simp add: gatherDomain_def)
apply (insert NATuniv)
apply (simp add: domain_reasoningPr dom_2 NATuniv domDA1)
apply (tactic  
 "REPEAT(CHANGED(TRYALL((rtac @{thm conjI}) THEN' (rtac @{thm impI}))))")    
apply (simp_all add: policyLemmas split: prod.splits option.splits)
apply (tactic  
 "REPEAT(CHANGED(TRYALL((rtac @{thm conjI}) THEN' (rtac @{thm impI}))))")    
apply (gen_test_cases FUT simp: net_noteq)
apply (simp_all add: policyLemmas nets src1_def src2_def to1_def to2_def datasimps singleton_iff)
store_test_thm "Scenario5"

(*
Test theorem (gen_test_cases) 'Scenario5': 686 test cases in 1778.840 seconds
*)

declare[[testgen_SMT=true]]
declare [[testgen_iterations=1]]


gen_test_data "Scenario5" 

text{* 686 test cases are being generated. *}
(*
Test theorem (gen_test_data) 'Scenario5': 686 test cases in 1033.461 seconds
*)

export_fw_data "Scenario5.data" Scenario5

thm Scenario5.test_data


end
