(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: ethz-datatype-02.thy 9047 2010-11-17 18:23:54Z lukasbru $ *)

theory 
  "ethz-datatype-02"
imports 
  "ethz-datatype-base"
begin


definition
 D1a :: "(networks, DummyContent) FWPolicy" where
"D1a = allow_from_to_ports {(7::port),(1::port)} eth_intern any"

definition
 D1b :: "(networks, DummyContent) FWPolicy" where
"D1b = allow_from_to_ports {(7::port),(1::port)} eth_private any"

definition
 D1c :: "(networks, DummyContent) FWPolicy" where
"D1c = allow_from_to_ports  {(7::port),(1::port)} vpn_vpz any"

definition
 D3 :: "(networks, DummyContent) FWPolicy" where
"D3 = allow_from_to_ports {(7::port),(1::port)} any any"


definition
 DefaultPolicy :: "(networks,DummyContent) FWPolicy" where
"DefaultPolicy = D3 ++ D1c ++ D1b ++ D1a"

lemmas PolicyLemmas = PolicyLemmas DefaultPolicy_def D1a_def D1b_def D1c_def D3_def


test_spec "not_in_same_net x \<longrightarrow> FUT x =  (deny_all ++ DefaultPolicy) x"
apply (prepare_fw_spec)
apply (simp add: PolicyLemmas PolicyCombinators PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def PolicyLemmas) 
store_test_thm "DP"

gen_test_data "DP"
thm DP.test_data



test_spec "not_in_same_net x \<longrightarrow> FUT x =  (deny_all ++ DefaultGlobalPolicy ++ DefaultPolicy ++ MandatoryGlobalPolicy) x"
apply (prepare_fw_spec)
apply (simp add: PolicyLemmas PolicyCombinators PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def PolicyLemmas) 
store_test_thm "all"



end
