(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: ethz-datatype-base.thy 9047 2010-11-17 18:23:54Z lukasbru $ *)

theory 
  "ethz-datatype-base"
imports 
  FWTesting
begin


declare [[testgen_profiling]]


(* The subnetworks *)

datatype nets =  FWLInk |
 Any |
 I26|
 I27|
 I32|
 I4|
 Sip|
 Time|
 Dhcp|
 I4_32|
 I10_32|
 Eth_intern|
 Eth_private|
 Vpn_vpz
  
 
instance nets :: adr ..

types networks = "nets \<times> int"

definition
 FWLink :: "networks net" where
"FWLink = {{(a,b). a = FWLInk}}"

definition
 any :: "networks net" where
"any = {{(a,b). a = Any}}"

definition
 i26:: "networks net" where
"i26 = {{(a,b). a = I26}}"

definition
 i27:: "networks net" where
"i27 = {{(a,b). a = I27 }}"

definition
 i32:: "networks net" where
"i32 = {{(a,b). a =  I32}}"

definition
 i4:: "networks net" where
"i4 = {{(a,b). a = I4 }}"

definition
 sip:: "networks net" where
"sip = {{(a,b). a = Sip }}"

definition
 time:: "networks net" where
"time = {{(a,b). a = Time}}"

definition
 dhcp:: "networks net" where
"dhcp = {{(a,b). a =  Dhcp}}"
 
definition
 i4_32:: "networks net" where
"i4_32 = {{(a,b). a =  I4_32}}"

definition
 i10_32:: "networks net" where
"i10_32 = {{(a,b). a = I10_32}}"

definition
 eth_intern:: "networks net" where
"eth_intern = {{(a,b). a = Eth_intern }}"
 
definition
 eth_private:: "networks net" where
"eth_private = {{(a,b). a = Eth_private}}"

definition
 vpn_vpz :: "networks net" where
"vpn_vpz = {{(a,b). a = Vpn_vpz}}"


defs (overloaded)
 src_port_def: "src_port (x::(networks,'b) packet) \<equiv> snd (src x)"
 dest_port_def: "dest_port (x::(networks,'b) packet) \<equiv> snd (dest x)"



definition
(* Mandatory: Global *)

 MG1 :: "(networks,DummyContent) FWPolicy" where
"MG1 = allow_from_to_port (1::port) i26 any "

definition
 MG2 :: "(networks,DummyContent) FWPolicy" where
"MG2 = allow_from_to_ports {(1::port),(2::port),(4::port)} i27 any"

definition
 MG3 :: "(networks,DummyContent) FWPolicy" where
"MG3 = allow_from_port_to (1::port) any FWLink"

definition
 MG4 :: "(networks,DummyContent) FWPolicy" where
"MG4 = allow_from_to_port (3::port) FWLink FWLink
"
definition
 MG5 :: "(networks,DummyContent) FWPolicy" where
"MG5 = allow_from_to_port (5::port) i32 any"

definition
 MG6 :: "(networks,DummyContent) FWPolicy" where
"MG6 = allow_from_to_port (6::port) i32 i4 "
 
definition
 MG7 :: "(networks,DummyContent) FWPolicy" where
"MG7 = allow_from_to_ports {(6::port),(7::port)} FWLink i4"

definition
 MG8 :: "(networks,DummyContent) FWPolicy" where
"MG8 = allow_from_to_ports {(6::port),(7::port)} FWLink i4"

definition
 MG9 :: "(networks,DummyContent) FWPolicy" where
"MG9 = allow_from_to_port (5::port) any i32 "

definition
 MG10 :: "(networks,DummyContent) FWPolicy" where
"MG10 = allow_from_to_port (5::port) i32 any"

definition
 MG11 :: "(networks,DummyContent) FWPolicy" where
"MG11 = allow_from_to_port (1::port) FWLink any"

definition
 MG12 :: "(networks,DummyContent) FWPolicy" where
"MG12 = allow_from_to_port (3::port) FWLink FWLink"

definition
 MG13 :: "(networks,DummyContent) FWPolicy" where
"MG13 = allow_from_to_ports {(4::port), (2::port)} sip any "

(* Default Global *)

definition
 DG1 :: "(networks,DummyContent) FWPolicy" where
"DG1 = allow_from_to_port (8::port) time any "

definition
 DG2:: "(networks,DummyContent) FWPolicy" where
"DG2 = allow_from_to_port (1::port) dhcp any "

definition
 DG3:: "(networks,DummyContent) FWPolicy" where
"DG3 = deny_from_to_port (7::port) any any "

 


definition 
 MandatoryGlobalPolicy :: "(networks,DummyContent) FWPolicy" where
"MandatoryGlobalPolicy = MG13 ++ MG12 ++  MG11 ++  MG10 ++  MG9 ++  MG8 ++  MG7 ++  MG6 ++ MG5 ++ MG4 ++ MG3 ++ MG2 ++ MG1"   
      
definition
 MandatoryGlobalInsidePolicy :: "(networks,DummyContent) FWPolicy" where
"MandatoryGlobalInsidePolicy = MG12 ++  MG11 ++  MG10 ++  MG9 ++  MG8"
         
definition
 MandatoryGlobalOutsidePolicy :: "(networks,DummyContent) FWPolicy" where
"MandatoryGlobalOutsidePolicy =   MG13 ++ MG7 ++ MG6 ++ MG5 ++ MG4 ++ MG3 ++ MG2 ++ MG1"       

definition
 DefaultGlobalPolicy :: "(networks,DummyContent) FWPolicy" where
"DefaultGlobalPolicy = DG3 ++ DG2 ++ DG1"  

 lemmas PolicyLemmas =  
 FWLink_def
 any_def
 i26_def
 i27_def
 i32_def
 i4_def
 sip_def
 time_def
 dhcp_def
 i4_32_def
 i10_32_def
 eth_intern_def
 eth_private_def
 vpn_vpz_def
 MG1_def MG2_def MG3_def MG4_def MG5_def MG6_def MG7_def MG8_def MG9_def MG10_def MG11_def MG12_def MG13_def
 DG1_def DG2_def DG3_def
 MandatoryGlobalPolicy_def MandatoryGlobalOutsidePolicy_def MandatoryGlobalInsidePolicy_def DefaultGlobalPolicy_def
 src_def dest_def in_subnet_def dest_port_def

definition 
 not_in_same_net :: "(networks,DummyContent) packet \<Rightarrow> bool" where
"not_in_same_net x =  ((src x \<noteq> dest x))"                   

test_spec "not_in_same_net x \<longrightarrow> FUT x =  (deny_all ++ DefaultGlobalPolicy) x"
apply (prepare_fw_spec)
apply (simp add: PolicyLemmas PolicyCombinators PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def PolicyLemmas) 
store_test_thm "DG"

gen_test_data "DG"
thm DG.test_data


test_spec "not_in_same_net x \<longrightarrow> FUT x =  (deny_all ++ MandatoryGlobalInsidePolicy) x"
apply (prepare_fw_spec)
apply (simp add: PolicyLemmas PolicyCombinators PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def PolicyLemmas) 
store_test_thm "MGI"

gen_test_data "MGI"
thm MGI.test_data


test_spec "not_in_same_net x \<longrightarrow> FUT x =  (deny_all ++ MandatoryGlobalOutsidePolicy) x"
apply (prepare_fw_spec)
apply (simp add: PolicyLemmas PolicyCombinators PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def PolicyLemmas) 
store_test_thm "MGO"


test_spec "not_in_same_net x \<longrightarrow> FUT x =  (deny_all ++ MandatoryGlobalPolicy) x"
apply (prepare_fw_spec)
apply (simp add: PolicyLemmas PortCombinators PolicyCombinators PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def PolicyLemmas) 
store_test_thm "MG"



test_spec "not_in_same_net x \<longrightarrow> FUT x =  (deny_all ++ DefaultGlobalPolicy ++ MandatoryGLobalPolicy) x"
apply (prepare_fw_spec)
apply (simp add: PolicyLemmas  PolicyCombinators PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def PolicyLemmas) 
store_test_thm "all"


write_clocks "time_stats_ethz_datatype_base"



end
