theory FTPInterleaved
imports 
  FWTesting
begin
declare [[testgen_profiling=true]]


definition 
  intranet :: "adr\<^isub>i\<^isub>p net" where
  "intranet = {{(a,e) . a = 3}}"

definition 
  internet :: "adr\<^isub>i\<^isub>p net" where
  "internet = {{(a,c). a > 3}}"

definition 
  ftp_policy :: "(adr\<^isub>i\<^isub>p,FTP.msg) FWPolicy" where
  "ftp_policy = (allow_from_to_port (21::port) intranet internet \<Oplus> D\<^sub>U)"


definition 
  is_in_intranet :: "adr\<^isub>i\<^isub>p  \<Rightarrow> bool" where
  "is_in_intranet a = ((fst a) = 3)"

definition 
  is_in_internet :: "adr\<^isub>i\<^isub>p \<Rightarrow> bool" where
  "is_in_internet a =  ((fst a) > 3)"


definition 
  "\<sigma>_0_ftp" ::  "(adr\<^isub>i\<^isub>p, FTP.msg) history \<times>
                 (adr\<^isub>i\<^isub>p, FTP.msg) FWPolicy" 
where                                     
  "\<sigma>_0_ftp = ([],ftp_policy)"



declare [[testgen_depth=4]]


lemmas ST_simps = 
is_in_intranet_def ftp_policy_def 
is_in_internet_def intranet_def internet_def FTPLemmas



lemma "
((14::int) \<noteq> 21 \<and>
        (543::int) \<noteq> 21 \<and>
        (3::int) <24 \<and>
        3 < (34::int) \<and>
        (([], ftp_policy) \<Turnstile>
         ( os \<leftarrow> mbind [(1, (3,423), (34, 21), msg.ftp_init), (1, (3,423), (34, 21), msg.ftp_port_request 543), (1, (3,423), (34, 21), msg.ftp_close), (2, (3, 234), (2222, 21), msg.ftp_init),
                       (2, (3, 234), (2222, 21), msg.ftp_port_request 14), (2, (3, 234), (2222, 21), msg.ftp_close)]
                 (policy2MON TRPolicy); unit\<^isub>S\<^isub>E (os = [allow (),allow (),allow (),allow (),allow (),allow ()]))))"
apply (simp add: ST_simps)
done



definition 
  ftp_2_interleaved :: 
  "id \<Rightarrow> id \<Rightarrow> adr\<^isub>i\<^isub>p src \<Rightarrow> adr\<^isub>i\<^isub>p dest \<Rightarrow> port  \<Rightarrow>
               adr\<^isub>i\<^isub>p src \<Rightarrow> adr\<^isub>i\<^isub>p dest \<Rightarrow> port  \<Rightarrow>
            (adr\<^isub>i\<^isub>p,FTP.msg) history set" where
  "ftp_2_interleaved i1 i2 s1 d1 p1 s2 d2 p2 = 
      {x. (FTP.is_ftp ftp_states.S0 s1 d1 1 p1 (packet_with_id x i1)) \<and>
          (FTP.is_ftp ftp_states.S0 s2 d2 2 p2 (packet_with_id x i2))}"

definition "accept2_ftp" ::  
  "id \<Rightarrow> id \<Rightarrow> (adr\<^isub>i\<^isub>p, FTP.msg) history \<Rightarrow> bool" where
 "accept2_ftp i1 i2 t = 
  (\<exists> c1 s1 p1 c2 s2 p2. t \<in> ftp_2_interleaved i1 i2 c1 s1 p1 c2 s2 p2 \<and> 
                     is_in_intranet c1 \<and> is_in_internet s1 \<and> (snd s1) = 21 \<and>
                     is_in_intranet c2 \<and> is_in_internet s2 \<and> (snd s2) = 21 \<and>
                     p1 \<noteq> 21 \<and> p2 \<noteq> 21 \<and> i1 \<noteq> i2)"


declare[[testgen_depth=6]]



test_spec "\<lbrakk>length t = length X; ids {1,2} t;
  accept2_ftp 1 2 (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp \<Turnstile> (os \<leftarrow> mbind t (policy2MON TRPolicy); return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept2_ftp_def \<sigma>_0_ftp_def is_ftp_other_def)
apply (rule impI)+
apply (unfold is_single_ftp_run_def is_in_internet_def is_in_intranet_def ftp_2_interleaved_def)
apply (simp add: split_tupled_all)
apply (gen_test_cases "FUT"  simp:  NetworkCore.id_def)
apply (simp_all)
store_test_thm "interl6"


declare ST_simps[simp]
gen_test_data "interl6"
thm interl6.test_data
declare ST_simps[simp del]

declare[[testgen_depth=7]]
test_spec "\<lbrakk>length t = length X;  ids {1,2} t; length t <= 7;
  accept2_ftp 1 2 (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp  \<Turnstile> (os  \<leftarrow>  mbind t  FTP_TRPolicy; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept2_ftp_def \<sigma>_0_ftp_def is_ftp_other_def)
apply (rule impI)+
apply (unfold is_single_ftp_run_def is_in_internet_def is_in_intranet_def ftp_2_interleaved_def)
apply (simp add: split_tupled_all)
apply (gen_test_cases "FUT"  simp:  NetworkCore.id_def)
apply simp_all
store_test_thm "interl7"



declare[[testgen_depth=8]]
test_spec "\<lbrakk>length t = length X;  ids {1,2} t; length t <= 8;
  accept2_ftp 1 2 (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp \<Turnstile> (os  \<leftarrow>  mbind t  FTP_TRPolicy; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept2_ftp_def \<sigma>_0_ftp_def is_ftp_other_def)
apply (rule impI)+
apply (unfold is_single_ftp_run_def is_in_internet_def is_in_intranet_def ftp_2_interleaved_def)
apply (simp add: split_tupled_all)
apply (gen_test_cases "FUT"  simp:  NetworkCore.id_def)
store_test_thm "interl8"




declare[[testgen_depth=9]]
test_spec "\<lbrakk>length t = length X;  ids {1,2} t; length t <= 9;
  accept2_ftp 1 2 (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp  \<Turnstile> (os  \<leftarrow>  mbind t  FTP_TRPolicy; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept2_ftp_def \<sigma>_0_ftp_def is_ftp_other_def)
apply (rule impI)+
apply (unfold is_single_ftp_run_def is_in_internet_def is_in_intranet_def ftp_2_interleaved_def)
apply (simp add: split_tupled_all)
apply (gen_test_cases "FUT"  simp:  NetworkCore.id_def)
store_test_thm "interl9"




declare[[testgen_depth=10]]
test_spec "\<lbrakk>length t = length X;  ids {1,2} t; length t <= 10;
  accept2_ftp 1 2 (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp  \<Turnstile> (os  \<leftarrow>  mbind t  FTP_TRPolicy; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept2_ftp_def \<sigma>_0_ftp_def is_ftp_other_def)
apply (rule impI)+
apply (unfold is_single_ftp_run_def is_in_internet_def is_in_intranet_def ftp_2_interleaved_def)
apply (simp add: split_tupled_all)
apply (gen_test_cases "FUT"  simp:  NetworkCore.id_def)
store_test_thm "interl10"




write_clocks "time_stats_ftpInterleaved.tex"
declare [[testgen_profiling=false]]

end
