#!/bin/bash

 echo "

 
header{* File generated by script \"generate_nat_thy.sh\", on 
`date`.
*}
 
theory 
  fwnats
imports
  FWTesting
begin

declare [[testgen_profiling]]
declare [[testgen_iterations=50]]
declare [[testgen_SMT=true]]

definition
   filter_0 :: \"(adr\\<^isub>i\\<^isub>p, DummyContent) FWPolicy\" where
  \"filter_0 = deny_all\"

definition 
 nat_0  where 
 \"nat_0 = (A\<^sub>f(\<lambda>x. {x}))\"


definition subnet1 :: \"adr\\<^isub>i\\<^isub>p net\" where
 \"subnet1 = {{(d,e). d > 1 \<and> d < 256}}\"

definition subnet2 :: \"adr\\<^isub>i\\<^isub>p net\" where
 \"subnet2 = {{(d,e). d > 500 \<and> d < 1256}}\"


definition
  \"accross_subnets x  \<equiv> 
 ((src x \<sqsubset> subnet1 \<and> (dest x \<sqsubset> subnet2)) \<or>
  (src x \<sqsubset> subnet2 \<and> (dest x \<sqsubset> subnet1)))\"

lemmas UnfoldPolicy0 =filter_0_def nat_0_def 
                      NATLemmas 
                      ProtocolPortCombinators.ProtocolCombinators 
                      adr\\<^isub>i\\<^isub>pLemmas
                      packet_defs accross_subnets_def
         

lemmas subnets = subnet1_def subnet2_def 

" > fwnats.thy
test=10
for ((i=1; i<=5; i++))

do

  j=$i
  let "j = $j - 1"
  fm="filter_${i}_def"
  nm="nat_${i}_def"
  a1m="adr1${i}_def"
  a2m="adr2${i}_def"
  pm="policy_${i}_def"
  let "n1 = $i*2"
  let "n2 = $i*3" 
  let "m1 = $i*2+500"
  let "m2 = $i*3+500" 


echo "
definition
   filter_$i :: \"(adr\\<^isub>i\\<^isub>p, DummyContent) FWPolicy\" where
  \"filter_$i = filter_$j ++ allow_from_port_to ($i::port) subnet1 subnet2\"


definition adr1$i :: \"int set\"
where \"adr1$i = {d. d > $n1 \<and> d < $n2}\"

definition adr2$i :: \"int set\" where
 \"adr2$i = {d. d > $m1 \<and> d < $m2}\"


definition nat_$i where
 \"nat_$i = nat_$j ++ (srcPat2pool_IntPort adr1$i adr2$i)\"


definition policy_$i where
 \"policy_$i = ((\<lambda> (x,y). x) o_f 
 ((nat_$i \<Otimes>\<^sub>2 filter_$i) o (\<lambda> x. (x,x))))\"


lemmas UnfoldPolicy$i = UnfoldPolicy$j $fm $nm $a1m $a2m $pm
" >> fwnats.thy

done

for ((i=1; i<=10; i++))

do
#for ((j=1; j<=10; i++))

#do

echo "

test_spec 
\"accross_subnets x  \<longrightarrow> 
FUT x = policy_$i x\"
apply (prepare_fw_spec)
apply (gen_test_cases FUT simp: subnets UnfoldPolicy$i 
                                split_tupled_all)
apply (tactic 
 \"ALLGOALS(TestGen.ONLY_POS(
      (full_simp_tac (simpset_of @{context}addsimps @{thms  PO_def UnfoldPolicy$i}))
      THEN_ALL_NEW ( (rtac @{thm PO_add}))))\") 
store_test_thm "natT$i"

gen_test_data \"natT$i\"

" >> fwnats.thy

#fi

done

echo "

write_clocks \"time_stats_fwnats.tex\"

declare [[testgen_profiling = false]]

end" >> fwnats.thy
