(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * FWCompilationProof.thy --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: NormalisationGenericProofs.thy 9411 2012-10-07 15:38:35Z lukasbru $ *)

theory NormalisationGenericProofs
imports FWNormalisationCore
begin

text {* This theory contains the generic proofs of the normalisation
procedure, i.e. those which are independent from the concrete
semantical interpretation function.  *}


lemma domNMT: "dom X \<noteq> {} \<Longrightarrow> X \<noteq> \<emptyset>"
apply auto
done

lemma denyNMT: "deny_all \<noteq>  \<emptyset>"
apply (rule domNMT)
apply (simp add: deny_all_def dom_def)
done


lemma wellformed_policy1_charn[rule_format] : "wellformed_policy1 p \<longrightarrow>
DenyAll \<in> set p \<longrightarrow> (\<exists> p'. p = DenyAll # p' \<and> DenyAll \<notin> set p')"
by(induct p,simp_all)
 

lemma singleCombinatorsConc: "singleCombinators (x#xs) \<Longrightarrow> singleCombinators xs"
by (case_tac x,simp_all)

lemma aux0_0: "singleCombinators x \<Longrightarrow> \<not> (\<exists> a b. (a\<oplus>b) \<in> set x)"
apply (induct x, simp_all)
apply (rule allI)+
by (case_tac a,simp_all)


lemma aux0_4: "(a \<in> set x \<or> a \<in> set y) = (a \<in> set (x@y))"
by auto

lemma aux0_1: "\<lbrakk>singleCombinators xs; singleCombinators [x]\<rbrakk> \<Longrightarrow>
               singleCombinators (x#xs)"
by (case_tac x,simp_all) 


lemma aux0_6: "\<lbrakk>singleCombinators xs; \<not> (\<exists> a b. x = a \<oplus> b)\<rbrakk> \<Longrightarrow>
               singleCombinators(x#xs)"
apply (rule aux0_1,simp_all)
apply (case_tac x,simp_all)
apply auto
done

lemma aux0_5: " \<not> (\<exists> a b. (a\<oplus>b) \<in> set x) \<Longrightarrow> singleCombinators x"
apply (induct x)
apply simp_all
apply (rule aux0_6)
apply auto
done



lemma ANDConc[rule_format]: "allNetsDistinct (a#p) \<longrightarrow> allNetsDistinct (p)"
apply (simp add: allNetsDistinct_def)
apply (case_tac "a")
by simp_all


lemma aux6: "twoNetsDistinct a1 a2 a b \<Longrightarrow>
            dom (deny_all_from_to a1 a2) \<inter> dom (deny_all_from_to a b) = {}"
by (auto simp: twoNetsDistinct_def netsDistinct_def src_def dest_def 
           in_subnet_def PolicyCombinators.PolicyCombinators dom_def)

lemma aux5[rule_format]: "(DenyAllFromTo a b) \<in> set p \<longrightarrow> a \<in> set (net_list p)"
by (rule net_list_aux.induct,simp_all) 


lemma aux5a[rule_format]: "(DenyAllFromTo b a) \<in> set p \<longrightarrow> a \<in> set (net_list p)"
by (rule net_list_aux.induct,simp_all) 

lemma aux5c[rule_format]:
             "(AllowPortFromTo a b po) \<in> set p \<longrightarrow> a \<in> set (net_list p)"
by (rule net_list_aux.induct,simp_all) 

lemma aux5d[rule_format]:
            "(AllowPortFromTo b a po) \<in> set p \<longrightarrow> a \<in> set (net_list p)"
by (rule net_list_aux.induct,simp_all) 

lemma aux10[rule_format]: "a \<in> set (net_list p) \<longrightarrow> a \<in> set (net_list_aux p)"
by simp


lemma srcInNetListaux[simp]: "\<lbrakk>x \<in> set p; singleCombinators[x]; x \<noteq> DenyAll\<rbrakk> \<Longrightarrow>
                             srcNet x \<in> set (net_list_aux p)"
apply (induct p)
apply simp_all
apply (case_tac "x = a", simp_all)
apply (case_tac a, simp_all)+
done 


lemma destInNetListaux[simp]: "\<lbrakk>x \<in> set p; singleCombinators[x]; x \<noteq> DenyAll\<rbrakk> \<Longrightarrow>
                               destNet x \<in> set (net_list_aux p)"
apply (induct p)
apply simp_all
apply (case_tac "x = a", simp_all)
apply (case_tac a, simp_all)+
done 


lemma tND1: "\<lbrakk>allNetsDistinct p; x \<in> set p; y \<in> set p; a = srcNet x;
             b = destNet x; c = srcNet y; d = destNet y; a \<noteq> c;
             singleCombinators[x]; x \<noteq> DenyAll; singleCombinators[y];
             y \<noteq> DenyAll\<rbrakk> \<Longrightarrow> twoNetsDistinct a b c d"
apply (simp add: allNetsDistinct_def twoNetsDistinct_def)
done


lemma tND2: "\<lbrakk>allNetsDistinct p; x \<in> set p; y \<in> set p; a = srcNet x;
             b = destNet x; c = srcNet y; d = destNet y; b \<noteq> d;
             singleCombinators[x]; x \<noteq> DenyAll; singleCombinators[y];
             y \<noteq> DenyAll\<rbrakk> \<Longrightarrow> twoNetsDistinct a b c d"
apply (simp add: allNetsDistinct_def twoNetsDistinct_def)
done


lemma tND: "\<lbrakk>allNetsDistinct p; x \<in> set p; y \<in> set p; a = srcNet x;
            b = destNet x; c = srcNet y; d = destNet y; a \<noteq> c \<or> b \<noteq> d;
            singleCombinators[x]; x \<noteq> DenyAll; singleCombinators[y]; y \<noteq> DenyAll\<rbrakk>
            \<Longrightarrow> twoNetsDistinct a b c d"
apply (case_tac "a \<noteq> c", simp_all)
apply (erule_tac x = x and y =y in tND1, simp_all)
apply (erule_tac x = x and y =y in tND2, simp_all)
done


lemma aux7: "\<lbrakk>DenyAllFromTo a b \<in> set p; allNetsDistinct ((DenyAllFromTo c d)#p);
             a\<noteq> c\<or> b\<noteq> d\<rbrakk> \<Longrightarrow> twoNetsDistinct a b c d"
apply (erule_tac x = "DenyAllFromTo a b" and y = "DenyAllFromTo c d" in tND)
apply simp_all
done


lemma aux7a: "\<lbrakk>DenyAllFromTo a b  \<in> set p;
              allNetsDistinct ((AllowPortFromTo c d po)#p); a \<noteq> c\<or> b \<noteq> d\<rbrakk> \<Longrightarrow> 
              twoNetsDistinct a b c d"
apply (erule_tac x = "DenyAllFromTo a b" and
                 y = "AllowPortFromTo  c d po" in tND)
apply simp_all
done


lemma nDComm: assumes ab: "netsDistinct a b" shows ba: "netsDistinct b a"
apply (insert ab)
by (auto simp: netsDistinct_def  in_subnet_def)

lemma tNDComm: 
  assumes abcd: "twoNetsDistinct a b c d" shows "twoNetsDistinct c d a b"
apply (insert abcd)
apply (metis twoNetsDistinct_def nDComm)
done

lemma aux[rule_format]: "a \<in> set (removeShadowRules2 p) \<longrightarrow> a \<in> set p"
apply (case_tac a)
by (rule removeShadowRules2.induct, simp_all)+

lemma aux12: "\<lbrakk>a \<in> x; b \<notin> x\<rbrakk> \<Longrightarrow> a \<noteq> b"
by auto



lemma ND0aux1[rule_format]: "DenyAllFromTo x y \<in> set b \<Longrightarrow>  
                             x \<in> set (net_list_aux b)"
by (metis aux5 net_list.simps set_remdups)

lemma ND0aux2[rule_format]: "DenyAllFromTo x y \<in> set b \<Longrightarrow>  
                             y \<in> set (net_list_aux b)"
by (metis aux5a net_list.simps set_remdups)

lemma ND0aux3[rule_format]: "AllowPortFromTo x y p \<in> set b \<Longrightarrow>  
                             x \<in> set (net_list_aux b)"
by (metis aux5c net_list.simps set_remdups)

lemma ND0aux4[rule_format]: "AllowPortFromTo x y p \<in> set b \<Longrightarrow>  
                             y \<in> set (net_list_aux b)"
by (metis aux5d net_list.simps set_remdups)

lemma aNDSubsetaux[rule_format]: "singleCombinators a  \<longrightarrow> set a \<subseteq> set b \<longrightarrow> 
                                  set (net_list_aux a) \<subseteq> set (net_list_aux b)"
apply (induct a)
apply simp_all
apply clarify
apply (drule mp, erule singleCombinatorsConc)
apply (case_tac "a1")
apply (simp_all add: contra_subsetD)
apply (metis contra_subsetD)
apply (metis ND0aux1 ND0aux2 contra_subsetD)
apply (metis ND0aux3 ND0aux4 contra_subsetD)
done

lemma aNDSetsEqaux[rule_format]: "singleCombinators a \<longrightarrow> singleCombinators b \<longrightarrow> 
                 set a = set b \<longrightarrow> set (net_list_aux a) = set (net_list_aux b)"
apply (rule impI)+
apply (rule equalityI)
apply (rule aNDSubsetaux, simp_all)+
done

lemma aNDSubset: "\<lbrakk>singleCombinators a;set a \<subseteq> set b; allNetsDistinct b\<rbrakk> \<Longrightarrow> 
                  allNetsDistinct a"
apply (simp add: allNetsDistinct_def)
apply (rule allI)+
apply (rule impI)+
apply (drule_tac x = "aa" in spec, drule_tac x = "ba" in spec)
apply (metis subsetD aNDSubsetaux)
done

lemma aNDSetsEq: "\<lbrakk>singleCombinators a; singleCombinators b; set a = set b; 
                  allNetsDistinct b\<rbrakk> \<Longrightarrow> allNetsDistinct a"
apply (simp add: allNetsDistinct_def)
apply (rule allI)+
apply (rule impI)+
apply (drule_tac x = "aa" in spec, drule_tac x = "ba" in spec)
apply (metis aNDSetsEqaux )
done

lemma SCConca: "\<lbrakk>singleCombinators p; singleCombinators [a]\<rbrakk> \<Longrightarrow> 
                singleCombinators (a#p)"
by (case_tac "a",simp_all)

lemma aux3[simp]: "\<lbrakk>singleCombinators p; singleCombinators [a];  
                   allNetsDistinct (a#p)\<rbrakk> \<Longrightarrow> allNetsDistinct (a#a#p)"
apply (insert aNDSubset[of "(a#a#p)" "(a#p)"])
by (simp add: SCConca)





lemma wp1_aux1a[rule_format]: "xs \<noteq> [] \<longrightarrow> wellformed_policy1_strong (xs @ [x]) \<longrightarrow> 
                               wellformed_policy1_strong xs"
by (induct xs,simp_all)

lemma wp1alternative_RS1[rule_format]: "DenyAll \<in> set p \<longrightarrow> 
                              wellformed_policy1_strong (removeShadowRules1 p)"
by (induct p,simp_all)

lemma wellformed_eq: "DenyAll \<in> set p \<longrightarrow> 
                      ((wellformed_policy1 p) = (wellformed_policy1_strong p))"
by (induct p,simp_all)

lemma set_insort: "set(insort x xs l) = insert x (set xs)"
by (induct xs) auto

lemma set_sort[simp]: "set(sort xs l) = set xs"
by (induct xs) (simp_all add:set_insort)


lemma set_sortQ: "set(qsort xs l) = set xs"
by simp

lemma aux79[rule_format]: "y \<in> set (insort x a l) \<longrightarrow>  y \<noteq> x \<longrightarrow> y \<in> set a"
apply (induct a)
by auto

lemma aux80: "\<lbrakk>y \<notin> set p; y \<noteq> x\<rbrakk> \<Longrightarrow> y \<notin> set (insort x (sort p l) l)"
apply (metis aux79 set_sort)
done


lemma WP1Conca: "DenyAll \<notin> set p \<Longrightarrow> wellformed_policy1 (a#p)"
by (case_tac a,simp_all)


lemma saux[simp]: "(insort DenyAll p l) = DenyAll#p"
by (induct_tac p,simp_all)
 
lemma saux3[rule_format]: "DenyAllFromTo a b \<in> set list \<longrightarrow> 
                          DenyAllFromTo c d \<notin> set list \<longrightarrow> (a \<noteq> c) \<or> (b \<noteq> d)"
by blast

lemma waux2[rule_format]: " (DenyAll \<notin> set xs) \<longrightarrow> wellformed_policy1 xs"
by (induct_tac xs,simp_all)

lemma waux3[rule_format]: "\<lbrakk>x \<noteq> a;  x \<notin> set p\<rbrakk> \<Longrightarrow> x \<notin> set (insort a p l)"
by (metis aux79)

lemma wellformed1_sorted_aux[rule_format]: "wellformed_policy1 (x#p) \<Longrightarrow> 
                                            wellformed_policy1 (insort x p l)" 
apply (case_tac x,simp_all) 
by (rule waux2,rule waux3, simp_all)+

lemma wellformed1_sorted_auxQ[rule_format]: "wellformed_policy1 (p) \<Longrightarrow> 
                                            wellformed_policy1 (qsort p l)" 
apply (induct p)
apply simp_all
apply (case_tac a,simp_all)
apply (metis append_Cons append_Nil mem_def member_rec(1) member_set not_Cons_self qsort.simps(2) 
             set_qsort srcnets.simps(1) srcnets.simps(2) waux2)
apply (metis Combinators.simps(6) append_Cons append_Nil mem_def member_rec(1) member_set 
              qsort.simps(2) set_qsort waux2)
by (metis Combinators.simps(8) append_Cons append_Nil mem_def member_rec(1) member_set 
            qsort.simps(2) set_sortQ waux2) 


lemma SR1Subset: "set (removeShadowRules1 p) \<subseteq> set p"
apply (induct_tac p, simp_all)
apply (case_tac a, simp_all)
by auto

lemma SCSubset[rule_format]: " singleCombinators b \<longrightarrow> set a \<subseteq> set b \<longrightarrow>
                              singleCombinators a"
proof (induct a)
 case Nil thus ?case by simp
next
 case (Cons x xs) thus ?case 
   proof (cases x)
   case goal1 thus ?thesis by simp
   next
   case goal2 thus ?thesis by simp
   next
   case goal3 thus ?thesis by simp
   next 
   case (Conc c d)
   have f:  "c \<oplus> e \<in> set b \<longrightarrow> \<not> singleCombinators b"
       by (rule singleCombinators.induct,simp_all)
   from this show ?thesis 
    apply simp
    by (metis Conc aux0_0)
  qed
qed

lemma setInsert[simp]: "set list \<subseteq> insert a (set list)"
by auto

lemma SC_RS1[rule_format,simp]: "singleCombinators p \<longrightarrow> allNetsDistinct p \<longrightarrow>
    singleCombinators (removeShadowRules1 p)"
apply (induct_tac p)
apply simp_all
apply (rule impI)+
apply (drule mp)
apply (erule SCSubset,simp)
by (simp add: ANDConc)

lemma RS2Set[rule_format]: "set (removeShadowRules2 p) \<subseteq> set p"
apply (induct p, simp_all)
apply (case_tac a, simp_all)
apply auto
done

lemma WP1: "a \<notin> set list \<Longrightarrow> a \<notin> set (removeShadowRules2 list)"
apply (insert RS2Set [of list])
apply blast
done

lemma denyAllDom[simp]: "x \<in> dom (deny_all)"
by (simp add: PLemmas)

lemma lCdom2: "(list2FWpolicy (a @ (b @ c))) = (list2FWpolicy ((a@b)@c))"
by auto


lemma SCConcEnd: "singleCombinators (xs @ [xa]) \<Longrightarrow> singleCombinators xs"
by (induct "xs", simp_all, case_tac a, simp_all)

lemma list2FWpolicyconc[rule_format]: "a \<noteq> [] \<longrightarrow>
                               (list2FWpolicy (xa # a)) = (xa) \<oplus> (list2FWpolicy a)"
by (induct a,simp_all)


lemma wp1n_tl [rule_format]: "wellformed_policy1_strong p \<longrightarrow>
                              p = (DenyAll#(tl p))"
by (induct p, simp_all)

lemma foo2: 
"\<lbrakk>a \<notin> set ps; a \<notin> set ss; set p = set s;  p = (a#(ps)); s = (a#ss)\<rbrakk> \<Longrightarrow> 
 set (ps) = set (ss)"
by auto


lemma SCnotConc[rule_format,simp]: "a\<oplus>b \<in> set p \<longrightarrow> singleCombinators p \<longrightarrow>False"
by (induct p, simp_all, case_tac aa, simp_all)

lemma auxx8: "removeShadowRules1_alternative_rev [x] = [x]"
by (case_tac "x", simp_all)

lemma RS1End[rule_format]: "x \<noteq> DenyAll \<longrightarrow> removeShadowRules1 (xs @ [x]) = 
                            (removeShadowRules1 xs)@[x]"
by (induct_tac xs, simp_all)

lemma aux114: "x \<noteq> DenyAll \<Longrightarrow> removeShadowRules1_alternative_rev (x#xs) = 
               x#(removeShadowRules1_alternative_rev xs)"
apply (induct_tac xs)
apply (auto simp: auxx8)
by (case_tac "x", simp_all)

lemma aux115[rule_format]: "x \<noteq> DenyAll\<Longrightarrow>removeShadowRules1_alternative (xs@[x]) 
                            = (removeShadowRules1_alternative xs)@[x]"
apply (simp add: removeShadowRules1_alternative_def aux114)
done

lemma RS1_DA[simp]: "removeShadowRules1 (xs @ [DenyAll]) = [DenyAll]"
by (induct_tac xs, simp_all)

lemma rSR1_eq: "removeShadowRules1_alternative = removeShadowRules1"
apply (rule ext)
apply (simp add: removeShadowRules1_alternative_def)
apply (rule_tac xs = x in rev_induct)
apply simp_all
apply (case_tac "xa = DenyAll", simp_all)
apply (metis RS1End aux114 rev.simps(2))
done

lemma domInterMT[rule_format]: "\<lbrakk>dom a \<inter> dom b = {}; x \<in> dom a\<rbrakk> \<Longrightarrow> x \<notin> dom b"
by auto

lemma domComm: "dom a \<inter> dom b = dom b \<inter> dom a"
by auto


lemma r_not_DA_in_tl[rule_format]: "wellformed_policy1_strong p \<longrightarrow>  a \<in> set p\<longrightarrow>
                                    a \<noteq> DenyAll \<longrightarrow> a \<in> set (tl p)"
by (induct p,simp_all)

lemma wp1_aux1aa[rule_format]: "wellformed_policy1_strong p \<longrightarrow> DenyAll \<in> set p"
by (induct p,simp_all)

lemma mauxa: "(\<exists> r. a b = \<lfloor>r\<rfloor>) = (a b \<noteq> \<bottom>)"
by auto

lemma l2p_aux[rule_format]: "list \<noteq> [] \<longrightarrow>
                             list2FWpolicy (a # list) = a \<oplus>(list2FWpolicy list)"
by (induct "list", simp_all)

lemma l2p_aux2[rule_format]: "list = [] \<Longrightarrow> list2FWpolicy (a # list) = a"
by simp

lemma aux7aa: "\<lbrakk>AllowPortFromTo a b poo \<in> set p; 
              allNetsDistinct ((AllowPortFromTo c d po) # p); a \<noteq> c \<or> b \<noteq> d\<rbrakk> \<Longrightarrow>
               twoNetsDistinct a b c d"
apply (simp add: allNetsDistinct_def twoNetsDistinct_def)
apply (case_tac "a \<noteq> c")
apply (rule disjI1)
apply (drule_tac x = "a" in spec, drule_tac x = "c" in spec)
apply (simp split: if_splits)
apply (simp_all add: ND0aux3,metis)
apply (rule disjI2)
apply (drule_tac x = "b" in spec, drule_tac x = "d" in spec)
apply (simp split: if_splits)
apply (metis ND0aux4)+
done

lemma ANDConcEnd: "\<lbrakk> allNetsDistinct (xs @ [xa]); singleCombinators xs\<rbrakk> \<Longrightarrow> 
                   allNetsDistinct xs"
by (rule aNDSubset) auto

lemma WP1ConcEnd[rule_format]: 
  "wellformed_policy1 (xs@[xa]) \<longrightarrow> wellformed_policy1 xs"
by (induct xs, simp_all)

lemma NDComm: "netsDistinct a b = netsDistinct b a"
by (auto simp: netsDistinct_def in_subnet_def)


lemma wellformed1_sorted[simp]: 
  assumes wp1: "wellformed_policy1 p" 
  shows        "wellformed_policy1 (sort p l)"
 proof (cases p)
   case Nil thus ?thesis by simp
 next
   case (Cons x xs) thus ?thesis
    proof (cases "x = DenyAll") 
      case True thus ?thesis using assms Cons by simp 
      next
      case False thus ?thesis using assms  
       by (metis Cons set_sort False waux2 wellformed_eq 
                 wellformed_policy1_strong.simps(2)) 
    qed
 qed


lemma wellformed1_sortedQ[simp]: 
  assumes wp1: "wellformed_policy1 p" 
  shows        "wellformed_policy1 (qsort p l)"
 proof (cases p)
   case Nil thus ?thesis by simp
 next
   case (Cons x xs) thus ?thesis
    proof (cases "x = DenyAll") 
      case True thus ?thesis using assms Cons by simp 
      next
      case False thus ?thesis using assms  
       by (metis Cons set_qsort False waux2 wellformed_eq 
                 wellformed_policy1_strong.simps(2)) 
    qed
 qed




lemma SC1[simp]: "singleCombinators p \<Longrightarrow>singleCombinators (removeShadowRules1 p)"
by (erule SCSubset) (rule SR1Subset)

lemma SC2[simp]: "singleCombinators p \<Longrightarrow>singleCombinators (removeShadowRules2 p)"
by (erule SCSubset) (rule RS2Set)


lemma SC3[simp]: "singleCombinators p \<Longrightarrow> singleCombinators (sort p l)"
by (erule SCSubset) simp



lemma SC3Q[simp]: "singleCombinators p \<Longrightarrow> singleCombinators (qsort p l)"
by (erule SCSubset) simp

lemma aND_RS1[simp]: "\<lbrakk>singleCombinators p; allNetsDistinct p\<rbrakk> \<Longrightarrow> 
                      allNetsDistinct (removeShadowRules1 p)"
apply (rule aNDSubset)
apply (erule SC_RS1, simp_all)
apply (rule SR1Subset)
done

lemma aND_RS2[simp]: "\<lbrakk>singleCombinators p; allNetsDistinct p\<rbrakk> \<Longrightarrow> 
                      allNetsDistinct (removeShadowRules2 p)"
apply (rule aNDSubset)
apply (erule SC2, simp_all)
apply (rule RS2Set)
done
 
lemma aND_sort[simp]: "\<lbrakk>singleCombinators p; allNetsDistinct p\<rbrakk> \<Longrightarrow>  
                       allNetsDistinct (sort p l)"
apply (rule aNDSubset)
by (erule SC3, simp_all)



lemma aND_sortQ[simp]: "\<lbrakk>singleCombinators p; allNetsDistinct p\<rbrakk> \<Longrightarrow>  
                       allNetsDistinct (qsort p l)"
apply (rule aNDSubset)
by (erule SC3Q, simp_all)



lemma inRS2[rule_format,simp]: "x \<notin> set p \<longrightarrow> x \<notin> set (removeShadowRules2 p)"
apply (insert RS2Set [of p])
by blast

lemma distinct_RS2[rule_format,simp]: "distinct p \<longrightarrow> 
                                       distinct (removeShadowRules2 p)"
apply (induct p)
apply simp_all
apply clarify
apply (case_tac "a")
by auto

lemma setPaireq: " {x, y} = {a, b} \<Longrightarrow> x = a \<and> y = b \<or> x = b \<and> y = a"
by (metis  doubleton_eq_iff)

lemma position_positive[rule_format]: "a \<in> set l \<longrightarrow> position a l > 0"
by (induct l, simp_all)

lemma pos_noteq[rule_format]: 
  "a \<in> set l  \<longrightarrow> b  \<in> set l  \<longrightarrow>    c \<in> set l \<longrightarrow> a \<noteq> b \<longrightarrow> 
   (position a l) <= (position b l) \<longrightarrow>   
    (position b l) <= (position c l) \<longrightarrow> 
       a \<noteq> c"
apply (induct l)
apply simp_all
apply (rule conjI)
apply (rule impI)+
apply (simp add: position_positive)+
apply (metis gr_implies_not0 position_positive)
done

lemma setPair_noteq: "{a,b} \<noteq> {c,d} \<Longrightarrow> \<not> ((a = c) \<and> (b = d))" 
by auto

lemma setPair_noteq_allow: "{a,b} \<noteq> {c,d} \<Longrightarrow> \<not> ((a = c) \<and> (b = d) \<and> P)" 
by auto

lemma order_trans:  
 "\<lbrakk>in_list x l; in_list y l; in_list z l; singleCombinators [x]; 
  singleCombinators [y]; singleCombinators [z]; smaller x y l; smaller y z l\<rbrakk> \<Longrightarrow> 
  smaller x z l"
apply (case_tac x)
apply simp_all
apply (case_tac z)
apply simp_all
apply (case_tac y)
apply simp_all
apply (case_tac y)
apply simp_all
apply (rule conjI|rule impI)+
apply (rule setPaireq,simp)
apply (rule conjI|rule impI)+
apply (simp_all split: if_splits)
apply metis
apply metis
apply (simp add: setPair_noteq)
apply (rule impI, simp_all)
apply (erule setPaireq)
apply (rule impI)
apply (case_tac y, simp_all)
apply (simp_all split: if_splits)
apply metis
apply (simp_all add: setPair_noteq setPair_noteq_allow)
apply (case_tac z)
apply simp_all
apply (case_tac y)
apply simp_all
apply (case_tac y)
apply simp_all
apply (rule impI|rule conjI)+
apply (simp_all split: if_splits)
apply (simp add: setPair_noteq)
apply (erule pos_noteq)
apply simp_all
apply (rule impI)
apply (simp add: setPair_noteq)
apply (rule conjI)
apply (simp add: setPair_noteq_allow)
apply (erule pos_noteq, simp_all)
apply (rule impI)
apply (simp add: setPair_noteq_allow)
apply (rule impI)
apply (rule disjI2)
apply (case_tac y, simp_all)
apply (simp_all split: if_splits)
apply metis
apply (simp_all add: setPair_noteq_allow)
done

lemma sortedConcStart[rule_format]: 
 "sorted (a # aa # p) l \<longrightarrow> in_list a l \<longrightarrow> in_list aa l \<longrightarrow> all_in_list p l\<longrightarrow> 
  singleCombinators [a] \<longrightarrow> singleCombinators [aa] \<longrightarrow> singleCombinators p \<longrightarrow> 
  sorted (a#p) l"
apply (induct p)
apply simp_all
apply (rule impI)+
apply simp
apply (rule_tac y = "aa" in order_trans)
apply simp_all
apply (case_tac ab, simp_all)
done

lemma singleCombinatorsStart[simp]: "singleCombinators (x#xs) \<Longrightarrow> 
                                     singleCombinators [x]"
by (case_tac x, simp_all)


lemma sorted_is_smaller[rule_format]: 
 "sorted (a # p) l \<longrightarrow> in_list a l \<longrightarrow> in_list b l \<longrightarrow> all_in_list p l \<longrightarrow>  
  singleCombinators [a] \<longrightarrow> singleCombinators p \<longrightarrow> b \<in> set p \<longrightarrow> smaller a b l"
apply (induct p)
apply (auto intro: singleCombinatorsConc sortedConcStart) 
done

lemma sortedConcEnd[rule_format]: "sorted (a # p) l \<longrightarrow> in_list a l \<longrightarrow> 
                                   all_in_list p l \<longrightarrow> singleCombinators [a] \<longrightarrow> 
                                   singleCombinators p  \<longrightarrow> sorted p l"
apply (induct p)
apply (auto intro: singleCombinatorsConc sortedConcStart)
done

lemma in_set_in_list[rule_format]: "a \<in> set p \<longrightarrow> all_in_list p l\<longrightarrow> in_list a l"
by (induct p) auto
   
lemma sorted_Consb[rule_format]: 
   "all_in_list (x#xs) l \<longrightarrow> singleCombinators (x#xs) \<longrightarrow> 
    (sorted xs l & (ALL y:set xs. smaller x y l)) \<longrightarrow>  (sorted (x#xs) l) "
apply(induct xs arbitrary: x) 
apply simp
apply (auto simp: order_trans)
done

lemma sorted_Cons: "\<lbrakk>all_in_list (x#xs) l; singleCombinators (x#xs)\<rbrakk> \<Longrightarrow> 
              (sorted xs l & (ALL y:set xs. smaller x y l)) =  (sorted (x#xs) l)"
apply auto
apply (rule sorted_Consb, simp_all)
apply (metis singleCombinatorsConc singleCombinatorsStart sortedConcEnd)
apply (erule sorted_is_smaller)
apply (auto intro: singleCombinatorsConc singleCombinatorsStart in_set_in_list)
done

lemma smaller_antisym: "\<lbrakk>\<not> smaller a b l; in_list a l; in_list b l; 
                        singleCombinators[a]; singleCombinators [b]\<rbrakk> \<Longrightarrow>  
                        smaller b a l"
apply (case_tac a)
apply simp_all
apply (case_tac b)
apply simp_all
apply (simp_all split: if_splits)
apply (rule setPaireq)
apply simp
apply (case_tac b)
apply simp_all
apply (simp_all split: if_splits)
done

lemma set_insort_insert: "set (insort x xs l) \<subseteq> insert x (set xs)"
by (induct xs) auto

lemma all_in_listSubset[rule_format]: "all_in_list b l \<longrightarrow>singleCombinators a \<longrightarrow> 
                                      set a \<subseteq> set b \<longrightarrow> all_in_list a l"
by (induct_tac a) (auto intro: in_set_in_list singleCombinatorsConc)

lemma singleCombinators_insort: "\<lbrakk>singleCombinators [x]; singleCombinators xs\<rbrakk> \<Longrightarrow> 
                                 singleCombinators (insort x xs l)"
by (metis SCSubset SCConca set_insort set.simps(2) subset_refl)

lemma all_in_list_insort: "\<lbrakk>all_in_list xs l; singleCombinators (x#xs); 
                           in_list x l\<rbrakk> \<Longrightarrow>  all_in_list (insort x xs l) l"
apply (rule_tac b = "x#xs" in all_in_listSubset)
apply simp_all
apply (metis singleCombinatorsConc singleCombinatorsStart
             singleCombinators_insort)
apply (rule set_insort_insert)
done

lemma sorted_ConsA:"\<lbrakk>all_in_list (x#xs) l; singleCombinators (x#xs)\<rbrakk> \<Longrightarrow> 
              (sorted (x#xs) l)  = (sorted xs l & (ALL y:set xs. smaller x y l))"
by (metis sorted_Cons)

lemma is_in_insort: "y \<in> set xs \<Longrightarrow> y \<in> set (insort x xs l)"
by (metis ListMem_iff insert set_insort set.simps(2))

lemma sorted_insorta[rule_format]: 
  "sorted (insort x xs l) l \<longrightarrow> all_in_list (x#xs) l \<longrightarrow> distinct (x#xs) \<longrightarrow> 
   singleCombinators [x] \<longrightarrow> singleCombinators xs \<longrightarrow> sorted xs l"
apply (induct xs)
apply simp_all
apply (rule impI)+
apply simp
apply (auto intro: is_in_insort sorted_ConsA set_insort singleCombinators_insort 
  singleCombinatorsConc sortedConcEnd all_in_list_insort)
apply (metis sort.simps(2) set_sort SCSubset all_in_list_insort set_subset_Cons 
  singleCombinators.simps(3) singleCombinatorsConc singleCombinatorsStart 
  singleCombinators_insort sortedConcEnd)
apply (rule sorted_Consb)
apply simp_all
apply (rule ballI)
apply (rule_tac p = "insort x xs l" in sorted_is_smaller)
apply (auto intro: in_set_in_list all_in_listSubset singleCombinators_insort 
  singleCombinatorsConc set_insort_insert is_in_insort)
apply (rule_tac b = "x#xs" in all_in_listSubset)
apply simp_all
apply (erule singleCombinators_insort)
apply (erule singleCombinatorsConc)
apply (rule set_insort_insert)
done

lemma sorted_insortb[rule_format]: 
  "sorted xs l \<longrightarrow> all_in_list (x#xs) l \<longrightarrow> distinct (x#xs) \<longrightarrow> 
   singleCombinators [x] \<longrightarrow> singleCombinators xs \<longrightarrow> sorted (insort x xs l) l"
apply (induct xs)
apply simp_all
apply (rule impI)+
apply (subgoal_tac "sorted (insort x xs l) l")
apply simp_all
defer 1
apply (metis sorted_Cons all_in_list.simps(2) 
             singleCombinatorsConc)
apply (rule sorted_Consb)
apply simp_all
apply auto
apply (rule_tac b = "x#xs" in all_in_listSubset)
apply simp_all
apply (rule singleCombinators_insort, simp_all)
apply (erule singleCombinatorsConc)
apply (rule set_insort_insert)
apply (metis SCConca singleCombinatorsConc singleCombinatorsStart 
             singleCombinators_insort)
apply (case_tac "y = x")
apply simp_all
apply (rule smaller_antisym)
apply simp_all
apply (subgoal_tac "y \<in> set xs")
apply (auto intro: in_set_in_list all_in_list_insort aux0_1 singleCombinatorsConc
                   aux79 sorted_is_smaller smaller_antisym)
done

lemma sorted_insort: "\<lbrakk>all_in_list (x#xs) l; distinct(x#xs); singleCombinators [x];
                     singleCombinators xs\<rbrakk> \<Longrightarrow>
                  sorted (insort x xs l) l = sorted xs l"
by (auto intro: sorted_insorta sorted_insortb)

lemma distinct_insort: "distinct (insort x xs l) = (x \<notin> set xs \<and> distinct xs)"
by(induct xs)(auto simp:set_insort)

lemma distinct_sort[simp]: "distinct (sort xs l) = distinct xs"
by(induct xs)(simp_all add:distinct_insort)


lemma sort_is_sorted[rule_format]: "all_in_list p l \<longrightarrow> distinct p \<longrightarrow>
                                   singleCombinators p \<longrightarrow> sorted (sort p l) l"
apply (induct p)
apply (auto intro: SC3 all_in_listSubset singleCombinatorsConc sorted_insort)
apply (subst sorted_insort)
apply (auto intro: singleCombinatorsConc all_in_listSubset SC3)
apply (erule all_in_listSubset)
by (auto intro: SC3 singleCombinatorsConc sorted_insort)




lemma smaller_sym[rule_format]: "all_in_list [a] l \<longrightarrow> smaller a a l"
apply (case_tac a,simp_all)
done


lemma SC_sublist[rule_format]: "singleCombinators xs \<Longrightarrow>
    singleCombinators (qsort [y\<leftarrow>xs. P y] l)"
apply (auto intro: SCSubset)
done

lemma all_in_list_sublist[rule_format]: "singleCombinators xs \<longrightarrow> all_in_list xs l \<longrightarrow>
    all_in_list (qsort [y\<leftarrow>xs. P y] l) l"
  apply (auto intro: all_in_listSubset SC_sublist)
done


lemma SC_sublist2[rule_format]: "singleCombinators xs \<longrightarrow>
    singleCombinators ([y\<leftarrow>xs. P y])"
apply (auto intro: SCSubset)
done


lemma all_in_list_sublist2[rule_format]: "singleCombinators xs \<longrightarrow> all_in_list xs l \<longrightarrow>
    all_in_list ( [y\<leftarrow>xs. P y]) l"
apply (auto intro: all_in_listSubset SC_sublist2)
done


lemma all_in_listAppend[rule_format]: 
 "all_in_list (xs) l \<longrightarrow> all_in_list (ys) l \<longrightarrow> all_in_list (xs @ ys) l"
apply (induct xs)
apply simp_all
done


lemma distinct_sortQ[rule_format]: "singleCombinators xs \<longrightarrow>
  all_in_list xs l \<longrightarrow> distinct xs \<longrightarrow> distinct (qsort xs l)"
apply (induct xs l rule: qsort.induct) 
apply simp_all
apply (rule impI)+
apply (auto simp: SC_sublist2 singleCombinatorsConc all_in_list_sublist2)
done


lemma singleCombinatorsAppend[rule_format]: 
 "singleCombinators (xs) \<longrightarrow> singleCombinators (ys) \<longrightarrow> singleCombinators (xs @ ys)"
apply (induct xs)
apply simp_all
apply (rule impI)+
apply simp_all
apply (drule mp)+
apply (case_tac a,simp_all)+
done

lemma sorted_append1[rule_format]:
  "all_in_list xs l  \<longrightarrow> singleCombinators xs \<longrightarrow>
   all_in_list ys l  \<longrightarrow>
   singleCombinators ys \<longrightarrow> 
  (sorted (xs@ys) l \<longrightarrow> (sorted xs l & sorted ys l & (\<forall>x \<in> set xs. \<forall>y \<in> set ys. smaller x y l)))"
apply (induct xs)
apply simp_all
apply (rule impI)+
apply simp
apply (drule mp)+
apply (case_tac a,simp_all)
apply (drule mp)+
apply (erule sortedConcEnd)
apply simp_all
apply (rule all_in_listAppend,simp_all)
apply (rule singleCombinatorsAppend,simp_all)
apply (case_tac a,simp_all)
apply (rule conjI)
apply (smt all_in_list.simps(2) all_in_listAppend append_Cons aux0_4 sorted_Cons singleCombinatorsAppend)
by (metis all_in_list.simps(2) all_in_listAppend append_Cons aux0_4 sorted_Cons singleCombinatorsAppend) 


lemma sorted_append2[rule_format]:
  "all_in_list xs l\<longrightarrow> singleCombinators xs \<longrightarrow>
   all_in_list ys l \<longrightarrow>
   singleCombinators ys \<longrightarrow> 
   (sorted xs l & sorted ys l & (\<forall>x \<in> set xs. \<forall>y \<in> set ys. smaller x y l)) \<longrightarrow> (sorted (xs@ys) l)"
apply (induct xs)
apply simp_all
apply (rule impI)+
apply simp
apply (drule mp)+
apply (case_tac a,simp_all)
apply (drule mp)
apply (rule_tac a = a in sortedConcEnd)
apply simp_all
apply (case_tac a,simp_all)
by (smt all_in_list.simps(2) all_in_listAppend append_Cons aux0_4 sorted_Cons singleCombinatorsAppend sorted_Consb)

lemma sorted_append[rule_format]:
  "all_in_list xs l \<longrightarrow> singleCombinators xs \<longrightarrow>
   all_in_list ys l \<longrightarrow>
   singleCombinators ys \<longrightarrow> (sorted (xs@ys) l) =
   (sorted xs l & sorted ys l & (\<forall>x \<in> set xs. \<forall>y \<in> set ys. smaller x y l))"
apply (rule impI)+
apply (rule iffI)
apply (rule sorted_append1,simp_all)
apply (rule sorted_append2,simp_all)
done



lemma sort_is_sortedQ[rule_format]: "all_in_list p l  \<longrightarrow>
                                   singleCombinators p \<longrightarrow> sorted (qsort p l) l"
 apply (induct p l rule: qsort.induct)
 apply simp_all
 apply (rule impI)+
 apply (simp_all add: SC_sublist all_in_list_sublist all_in_list_sublist2 singleCombinatorsConc SC_sublist2)
apply (subst sorted_append)
apply (metis all_in_list_sublist singleCombinatorsConc)
apply (auto simp add: SC_sublist all_in_list_sublist SC_sublist2 all_in_list_sublist2 sorted_Cons sorted_append not_le less_imp_le )
apply (metis SC3Q SC_sublist2 singleCombinatorsConc)
apply (metis all_in_list_sublist singleCombinatorsConc)
apply (metis SC3Q SC_sublist2 aux0_1 singleCombinatorsConc singleCombinatorsStart)
apply (rule sorted_Consb)
apply (metis all_in_list.simps(2) all_in_list_sublist singleCombinatorsConc)
apply (metis SC3Q SC_sublist2 aux0_1 singleCombinatorsConc singleCombinatorsStart)
apply (rule conjI)
apply simp_all
apply (rule smaller_antisym)
apply simp_all
apply (metis in_set_in_list)
apply (erule SCSubset)
apply auto
apply (subgoal_tac "smaller xa x l")
apply (rule_tac y = x in order_trans)
apply simp_all
apply (metis in_set_in_list)
apply (metis in_set_in_list)
apply (erule SCSubset)
apply simp
apply (erule SCSubset)
apply simp
apply (rule smaller_antisym)
apply simp_all
apply (metis in_set_in_list)
apply (erule SCSubset)
apply simp
done


lemma inSet_not_MT: "a \<in> set p \<Longrightarrow> p \<noteq> []"
by auto 

lemma RS1n_assoc: "x \<noteq> DenyAll \<Longrightarrow>  removeShadowRules1_alternative xs @ [x] =
                                    removeShadowRules1_alternative (xs @ [x])"
by (simp add: removeShadowRules1_alternative_def aux114)

lemma RS1n_nMT[rule_format,simp]: "p \<noteq> []\<longrightarrow> removeShadowRules1_alternative p \<noteq> []"
apply (simp add: removeShadowRules1_alternative_def)
apply (rule_tac xs = p in rev_induct, simp_all) 
apply (case_tac "xs = []", simp_all)
apply (case_tac x, simp_all)
apply (rule_tac xs = "xs" in rev_induct, simp_all)
apply (case_tac x, simp_all)+
done

lemma RS1N_DA[simp]: "removeShadowRules1_alternative (a@[DenyAll]) = [DenyAll]"
by (simp add: removeShadowRules1_alternative_def)


lemma WP1n_DA_notinSet[rule_format]: "wellformed_policy1_strong p \<longrightarrow>
                                     DenyAll \<notin> set (tl p)"
by (induct p) (simp_all) 

lemma mt_sym: "dom a \<inter> dom b = {} \<Longrightarrow> dom b \<inter> dom a = {}"
by auto

lemma DAnotTL[rule_format]: 
    "xs \<noteq> [] \<longrightarrow> wellformed_policy1 (xs @ [DenyAll]) \<longrightarrow> False"
by (induct xs, simp_all)


lemma AND_tl[rule_format]: "allNetsDistinct ( p) \<longrightarrow> allNetsDistinct (tl p)"
apply (induct p, simp_all)
by (auto intro: ANDConc)


lemma distinct_tl[rule_format]: "distinct p \<longrightarrow> distinct (tl p)"
by (induct p, simp_all)


lemma SC_tl[rule_format]: "singleCombinators ( p) \<longrightarrow> singleCombinators (tl p)"
apply (induct p, simp_all)
by (auto intro: singleCombinatorsConc) 

lemma Conc_not_MT: "p = x#xs \<Longrightarrow> p \<noteq> []"
by auto

lemma wp1_tl[rule_format]: "p \<noteq> [] \<and> wellformed_policy1 p \<longrightarrow>
                            wellformed_policy1 (tl p)"
apply (induct p)
apply simp_all
apply (auto intro: waux2)
done

lemma wp1_eq[rule_format]: "wellformed_policy1_strong p \<Longrightarrow> wellformed_policy1 p"
apply (case_tac "DenyAll \<in> set p")
apply (subst wellformed_eq)
apply simp_all
apply (erule waux2)
done


lemma wellformed1_alternative_sorted: "wellformed_policy1_strong p \<Longrightarrow>
                                       wellformed_policy1_strong (sort p l)"
by (case_tac "p", simp_all)



lemma wp1n_RS2[rule_format]: "wellformed_policy1_strong p \<longrightarrow>
                              wellformed_policy1_strong (removeShadowRules2 p)"
by (induct p, simp_all)

lemma RS2_NMT[rule_format]: "p \<noteq> [] \<longrightarrow> removeShadowRules2 p \<noteq> []"
apply (induct p, simp_all)
apply (case_tac "p \<noteq> []", simp_all)
apply (case_tac "a", simp_all)+
done


lemma wp1_alternative_not_mt[simp]: "wellformed_policy1_strong p \<Longrightarrow> p \<noteq> []"
by auto

lemma AIL1[rule_format,simp]: "all_in_list p l \<longrightarrow>
                               all_in_list (removeShadowRules1 p) l"
by (induct_tac p, simp_all)



lemma wp1ID: "wellformed_policy1_strong (insertDeny (removeShadowRules1 p))"
by (induct p, simp_all, case_tac a, simp_all)

lemma dRD[simp]: "distinct (remdups p)"
by simp

lemma AILrd[rule_format,simp]: "all_in_list p l \<longrightarrow> all_in_list (remdups p) l"
by (induct p, simp_all)


lemma AILiD[rule_format,simp]: "all_in_list p l \<longrightarrow> all_in_list (insertDeny p) l"
apply (induct p, simp_all)
apply (rule impI, simp)
apply (case_tac "a", simp_all)
done

lemma SCrd[rule_format,simp]:"singleCombinators p\<longrightarrow> singleCombinators(remdups p)"
apply (induct p, simp_all)
apply (case_tac a)
apply simp_all
done

lemma SCRiD[rule_format,simp]: "singleCombinators p \<longrightarrow>
                                singleCombinators(insertDeny p)"
apply (induct p, simp_all)
apply (case_tac "a", simp_all)
done


lemma WP1rd[rule_format,simp]: "wellformed_policy1_strong p \<longrightarrow>
                                wellformed_policy1_strong (remdups p)"
apply (induct p, simp_all)
done

lemma ANDrd[rule_format,simp]: "singleCombinators p \<longrightarrow> allNetsDistinct p \<longrightarrow>
                                allNetsDistinct (remdups p)"
apply (rule impI)+
apply (rule_tac b = p in aNDSubset)
apply simp_all
done

lemma ANDiD[rule_format,simp]: "allNetsDistinct p \<longrightarrow>
                              allNetsDistinct (insertDeny p)"
apply (induct p, simp_all)
apply (simp add: allNetsDistinct_def)
apply (auto intro: ANDConc)
apply (case_tac "a")
apply (simp_all add: allNetsDistinct_def)
done


lemma mr_iD[rule_format]: "wellformed_policy1_strong p  \<longrightarrow>
                           matching_rule x p = matching_rule x (insertDeny p)"
by (induct p, simp_all)

lemma WP1iD[rule_format,simp]: "wellformed_policy1_strong p \<longrightarrow>
                                wellformed_policy1_strong (insertDeny p)"
by (induct p, simp_all)


lemma DAiniD: "DenyAll \<in> set (insertDeny p)"
by (induct p, simp_all, case_tac a, simp_all)

lemma p2lNmt: "policy2list p \<noteq> []"
by (rule policy2list.induct, simp_all)

lemma AIL2[rule_format,simp]: "all_in_list p l \<longrightarrow>
                               all_in_list (removeShadowRules2 p) l"
by (induct_tac p, simp_all, case_tac a, simp_all)
	
	
lemma SCConc: "\<lbrakk>singleCombinators x; singleCombinators y\<rbrakk> \<Longrightarrow>
	               singleCombinators (x@y)"
apply (rule aux0_5)
apply (metis aux0_0 aux0_4)
done 	

lemma SCp2l: "singleCombinators (policy2list p)"
apply (induct_tac p)
apply simp_all
apply (rule SCConc)
apply simp_all
done


lemma subnetAux: "Dd \<inter> A \<noteq> {} \<Longrightarrow> A \<subseteq> B \<Longrightarrow>  Dd \<inter> B \<noteq> {}"
apply auto
done

lemma soadisj: "\<lbrakk>x \<in> subnetsOfAdr a; y \<in> subnetsOfAdr a\<rbrakk> \<Longrightarrow> \<not> netsDistinct x y"
by (simp add: subnetsOfAdr_def netsDistinct_def, auto simp: PLemmas)

lemma not_member: "\<not> member a (x\<oplus>y) \<Longrightarrow> \<not> member a x"
apply auto
done

lemma soadisj2: "(\<forall> a x y. x \<in> subnetsOfAdr a \<and> y \<in> subnetsOfAdr a \<longrightarrow>
                 \<not> netsDistinct x y)"
by (simp add: subnetsOfAdr_def netsDistinct_def, auto simp: PLemmas)

lemma ndFalse1: "\<lbrakk>(\<forall>a b c d. (a,b)\<in>A  \<and> (c,d)\<in>B \<longrightarrow> netsDistinct a c);
        \<exists>(a, b)\<in>A. a \<in> subnetsOfAdr D; 
        \<exists>(a, b)\<in>B. a \<in> subnetsOfAdr D\<rbrakk>
    \<Longrightarrow> False"
apply (auto simp: soadisj)
apply (insert soadisj2)
apply (rotate_tac -1, drule_tac x = D in spec)
apply (rotate_tac -1, drule_tac x = a in spec)
apply (rotate_tac -1, drule_tac x = aa in spec)
by auto

lemma ndFalse2: "\<lbrakk>(\<forall>a b c d. (a,b)\<in>A  \<and> (c,d)\<in>B \<longrightarrow> netsDistinct b d);
        \<exists>(a, b)\<in>A. b \<in> subnetsOfAdr D; 
        \<exists>(a, b)\<in>B. b \<in> subnetsOfAdr D\<rbrakk>
    \<Longrightarrow> False"
apply (auto simp: soadisj)
apply (insert soadisj2)
apply (rotate_tac -1, drule_tac x = D in spec)
apply (rotate_tac -1, drule_tac x = b in spec)
apply (rotate_tac -1, drule_tac x = ba in spec)
apply simp
apply auto
done

lemma tndFalse: "\<lbrakk>(\<forall>a b c d. (a,b)\<in>A  \<and> (c,d)\<in>B \<longrightarrow> twoNetsDistinct a b c d);
        \<exists>(a, b)\<in>A. a \<in> subnetsOfAdr (D::('a::adr)) \<and> b \<in> subnetsOfAdr (F::'a); 
        \<exists>(a, b)\<in>B. a \<in> subnetsOfAdr D\<and> b\<in> subnetsOfAdr F\<rbrakk>
    \<Longrightarrow> False"
apply (simp add: twoNetsDistinct_def)
apply (auto simp: ndFalse1 ndFalse2)
apply (metis soadisj)
done

lemma sepnMT[rule_format]: "p \<noteq> [] \<longrightarrow> (separate p) \<noteq> []"
apply (rule separate.induct) back back back
by simp_all

lemma sepDA[rule_format]: "DenyAll \<notin> set p \<longrightarrow> DenyAll \<notin> set (separate p)"
apply (rule separate.induct) back
apply simp_all
done

lemma setnMT: "set a = set b \<Longrightarrow> a \<noteq> [] \<Longrightarrow> b \<noteq> []"
by auto

lemma sortnMT: "p \<noteq> [] \<Longrightarrow> sort p l \<noteq> []"
by (metis set_sort setnMT)

lemma idNMT[rule_format]: "p \<noteq> [] \<longrightarrow> insertDenies p \<noteq> []"
apply (induct p, simp_all)
apply (case_tac a, simp_all)
done

lemma OTNoTN[rule_format]: " OnlyTwoNets p \<longrightarrow> x \<noteq> DenyAll \<longrightarrow> x \<in> set p \<longrightarrow>
                            onlyTwoNets x"
apply (induct p, simp_all)
apply (rule impI)+
apply (rule conjI)
apply (rule impI)
apply simp
apply (case_tac a, simp_all)
apply (rule impI)
apply (drule mp, simp_all)
apply (case_tac a, simp_all)
done

lemma first_isIn[rule_format]:
  "\<not> member DenyAll x \<longrightarrow> (first_srcNet x,first_destNet x) \<in> sdnets x"
by (induct x,case_tac x, simp_all)

lemma sdnets2: "\<lbrakk>\<exists>a b. sdnets x = {(a, b), (b, a)}; \<not> member DenyAll x\<rbrakk> \<Longrightarrow>
                sdnets x = {(first_srcNet x, first_destNet x),
                            (first_destNet x, first_srcNet x)}"
apply (subgoal_tac "(first_srcNet x, first_destNet x) \<in> sdnets x")
apply (drule exE)
prefer 2
apply assumption
apply (drule exE)
prefer 2
apply assumption
apply simp
apply (case_tac "first_srcNet x = a \<and> first_destNet x = b")
apply simp_all
apply (metis insert_commute)
apply (erule first_isIn)
done

lemma alternativelistconc1[rule_format]: "a \<in> set (net_list_aux [x]) \<longrightarrow>
                                         a \<in> set (net_list_aux [x,y])"
by (induct x,simp_all)

lemma alternativelistconc2[rule_format]: "a \<in> set (net_list_aux [x]) \<longrightarrow>
                                         a \<in> set (net_list_aux [y,x])"
by (induct y, simp_all)


lemma noDA[rule_format]:"noDenyAll xs \<longrightarrow> s \<in> set xs \<longrightarrow> \<not> member DenyAll s"
by (induct xs, simp_all)

lemma isInAlternativeList:
  "(aa \<in> set (net_list_aux [a]) \<or> aa \<in> set (net_list_aux p))
       \<Longrightarrow> aa \<in> set (net_list_aux (a # p))"
apply (case_tac a,simp_all)
done

lemma netlistaux: "x \<in> set (net_list_aux (a # p))\<Longrightarrow>
                   x \<in> set (net_list_aux ([a])) \<or> x \<in> set (net_list_aux (p))"
apply (case_tac " x \<in> set (net_list_aux [a])")
apply simp_all
apply (case_tac a, simp_all)
done

lemma firstInNet[rule_format]: "\<not> member DenyAll a \<longrightarrow>
                                first_destNet a \<in> set (net_list_aux (a # p))"
apply (rule Combinators.induct) 
apply simp_all
apply (metis netlistaux)
done

lemma firstInNeta[rule_format]: "\<not> member DenyAll a \<longrightarrow>
                                 first_srcNet a \<in> set (net_list_aux (a # p))"
apply (rule Combinators.induct) 
apply simp_all
apply (metis netlistaux)
done  

lemma disjComm: "disjSD_2 a b \<Longrightarrow> disjSD_2 b a"
apply (simp add: disjSD_2_def)
apply (rule allI)+
apply (rule impI)
apply (rule conjI)
apply (drule_tac x = c in spec)
apply (drule_tac x = d in spec)
apply (drule_tac x = aa in spec)
apply (drule_tac x = ba in spec)
apply (metis tNDComm)
apply (drule_tac x = c in spec)
apply (drule_tac x = d in spec)
apply (drule_tac x = aa in spec)
apply (drule_tac x = ba in spec)
apply simp
apply (simp add: twoNetsDistinct_def)
apply (metis nDComm)+
done

lemma disjSD2aux: "\<lbrakk>disjSD_2 a b; \<not> member DenyAll a; \<not> member DenyAll b\<rbrakk>  \<Longrightarrow> 
 disjSD_2 (DenyAllFromTo (first_srcNet a) (first_destNet a) \<oplus>
 DenyAllFromTo (first_destNet a) (first_srcNet a) \<oplus> a)  b"
apply (drule disjComm)
apply (rule disjComm)
apply (simp add: disjSD_2_def)
apply (rule allI)+
apply (rule impI)+
apply safe
apply (drule_tac x = "aa" in spec, drule_tac x = ba in spec, 
      drule_tac x = "first_srcNet a" in spec,
      drule_tac x = "first_destNet a" in spec, auto intro: first_isIn)+
done


lemma noDA1eq[rule_format]: "noDenyAll p \<longrightarrow> noDenyAll1 p"
apply (induct p)
apply simp
apply (case_tac a, simp_all)
done

lemma noDA1C[rule_format]: "noDenyAll1 (a#p) \<longrightarrow> noDenyAll1 p"
apply (case_tac a, simp_all)
apply (rule impI, rule noDA1eq, simp)+
done

lemma disjSD_2IDa: "\<lbrakk>disjSD_2 x y; \<not> member DenyAll x; \<not> member DenyAll y;
                    a = (first_srcNet x); b = (first_destNet x)\<rbrakk>  \<Longrightarrow>
disjSD_2 ((DenyAllFromTo a b) \<oplus> (DenyAllFromTo b a) \<oplus> x) y"
apply simp
apply (rule disjSD2aux)
apply simp_all
done

lemma noDAID[rule_format]: "noDenyAll p \<longrightarrow> noDenyAll (insertDenies p)"
apply (induct p)
apply simp_all
apply (case_tac a, simp_all)
done

lemma isInIDo[rule_format]:  "noDenyAll p  \<longrightarrow> s \<in> set (insertDenies p) \<longrightarrow> 
 (\<exists>! a. s = (DenyAllFromTo (first_srcNet a) (first_destNet a)) \<oplus>
 (DenyAllFromTo (first_destNet a) (first_srcNet a)) \<oplus> a \<and> a \<in> set p)"
apply (induct p)
apply simp_all
apply (case_tac "a = DenyAll")
apply simp
apply (case_tac a, simp_all)
apply auto
done

lemma id_aux1[rule_format]: "DenyAllFromTo (first_srcNet s) (first_destNet s) \<oplus>
      DenyAllFromTo (first_destNet s) (first_srcNet s) \<oplus> s\<in> set (insertDenies p)
    \<longrightarrow> s \<in> set p"
apply (induct p)
apply simp_all
apply (case_tac a, simp_all)
done

lemma id_aux2:
 "\<lbrakk>noDenyAll p; (\<forall>s. s \<in> set p \<longrightarrow> disjSD_2 a s); \<not> member DenyAll a;
 ((DenyAllFromTo (first_srcNet s) (first_destNet s)) \<oplus> (DenyAllFromTo
   (first_destNet s) (first_srcNet s)) \<oplus> s) \<in> set (insertDenies p)\<rbrakk>  \<Longrightarrow>
  disjSD_2 a ((DenyAllFromTo (first_srcNet s) (first_destNet s)) \<oplus>
  (DenyAllFromTo (first_destNet s) (first_srcNet s)) \<oplus> s)"
apply (rule disjComm)
apply (rule disjSD_2IDa)
apply simp_all
apply (metis disjComm id_aux1)
apply (metis id_aux1 noDA)
done

lemma id_aux4[rule_format]: "\<lbrakk>noDenyAll p; (\<forall>s. s \<in> set p \<longrightarrow>
   disjSD_2 a s);s \<in> set (insertDenies p); \<not> member DenyAll a\<rbrakk> \<Longrightarrow>  disjSD_2 a s"
apply (subgoal_tac "   \<exists>a. s =
          DenyAllFromTo (first_srcNet a) (first_destNet a) \<oplus>
          DenyAllFromTo (first_destNet a) (first_srcNet a) \<oplus> a \<and>
          a \<in> set p")
apply (drule_tac Q = "disjSD_2 a s" in exE)
apply simp_all
apply (rule id_aux2, simp_all)
apply (rule ex1_implies_ex)
apply (rule isInIDo)
apply simp_all
done

lemma sepNetsID[rule_format]: "noDenyAll1 p \<longrightarrow> separated p \<longrightarrow>
                               separated (insertDenies p)"
apply (induct p)
apply simp_all
apply (rule impI)
apply (drule mp)
apply (erule noDA1C)
apply (rule impI)
apply (case_tac "a = DenyAll")
apply simp_all
apply (simp add: disjSD_2_def)
apply (case_tac a,simp_all)
apply auto
apply (rule disjSD_2IDa, simp_all, rule id_aux4, simp_all, metis noDA noDAID)+
done


lemma aNDDA[rule_format]: "allNetsDistinct p \<longrightarrow> allNetsDistinct(DenyAll#p)"
apply (case_tac p)
apply simp
apply (rule impI)
apply (simp add: allNetsDistinct_def)
apply (rule impI)
apply (auto)
apply (simp add: allNetsDistinct_def)
done

lemma OTNConc[rule_format]: "OnlyTwoNets (y # z) \<longrightarrow>  OnlyTwoNets z"
apply (case_tac y, simp_all)
done

lemma first_bothNetsd: "\<not> member DenyAll x \<Longrightarrow>
                        first_bothNet x = {first_srcNet x, first_destNet x}"
apply (induct x)
apply simp_all
done 

lemma bNaux:
"\<lbrakk>\<not>member DenyAll x; \<not> member DenyAll y;first_bothNet x = first_bothNet y\<rbrakk> 
 \<Longrightarrow> {first_srcNet x, first_destNet x} = {first_srcNet y, first_destNet y}"
apply (simp add: first_bothNetsd)
done

lemma setPair: "{a,b} = {a,d} \<Longrightarrow> b = d"
apply (metis setPaireq)
done

lemma setPair1: "{a,b} = {d,a} \<Longrightarrow> b = d"
apply (metis Un_empty_right Un_insert_right insert_absorb2 setPaireq)
done

lemma setPair4: "{a,b} = {c,d} \<Longrightarrow> a \<noteq> c \<Longrightarrow> a = d"
by auto

lemma otnaux1: " {x, y, x, y} = {x,y}"
by auto


lemma OTNIDaux4: "{x,y,x} = {y,x}"
by auto

lemma setPair5: "{a,b} = {c,d} \<Longrightarrow> a \<noteq> c \<Longrightarrow> a = d"
by auto

lemma otnaux: "   
  \<lbrakk>first_bothNet x = first_bothNet y; \<not> member DenyAll x; \<not> member DenyAll y; 
   onlyTwoNets y; onlyTwoNets x\<rbrakk> \<Longrightarrow> 
   onlyTwoNets (x \<oplus> y)"
apply (simp add: onlyTwoNets_def)
apply (subgoal_tac "{first_srcNet x, first_destNet x} =
                     {first_srcNet y, first_destNet y}")
apply (case_tac "(\<exists>a b. sdnets y = {(a, b)})")
apply simp_all
apply (case_tac "(\<exists>a b. sdnets x = {(a, b)})")
apply simp_all
apply (subgoal_tac "sdnets x = {(first_srcNet x, first_destNet x)}")
apply (subgoal_tac "sdnets y = {(first_srcNet y, first_destNet y)}")
apply simp
apply (case_tac "first_srcNet x = first_srcNet y")
apply simp_all
apply (rule disjI1)
apply (rule setPair)
apply simp
apply (subgoal_tac "first_srcNet x = first_destNet y")
apply simp
apply (subgoal_tac "first_destNet x = first_srcNet y")
apply simp
apply (rule_tac x ="first_srcNet y" in exI,
       rule_tac x = "first_destNet y" in exI,simp)
apply (rule setPair1)
apply simp
apply (rule setPair4)
apply simp_all
apply (metis first_isIn singletonE)
apply (metis first_isIn singletonE)
apply (subgoal_tac "sdnets x = {(first_srcNet x, first_destNet x),
                                (first_destNet x, first_srcNet x)}")
apply (subgoal_tac "sdnets y = {(first_srcNet y, first_destNet y)}")
apply simp
apply (case_tac "first_srcNet x = first_srcNet y")
apply simp_all
apply (subgoal_tac "first_destNet x = first_destNet y")
apply simp
apply (rule setPair)
apply simp
apply (subgoal_tac "first_srcNet x = first_destNet y")
apply simp
apply (subgoal_tac "first_destNet x = first_srcNet y")
apply simp
apply (rule_tac x ="first_srcNet y" in exI,
       rule_tac x = "first_destNet y" in exI)
apply (metis OTNIDaux4 insert_commute )
apply (rule setPair1)
apply simp
apply (rule setPair5)
apply assumption
apply simp
apply (metis first_isIn singletonE)
apply (rule sdnets2)
apply simp_all
apply (case_tac "(\<exists>a b. sdnets x = {(a, b)})")
apply simp_all
apply (subgoal_tac "sdnets x = {(first_srcNet x, first_destNet x)}")
apply (subgoal_tac "sdnets y = {(first_srcNet y, first_destNet y),
                                 (first_destNet y, first_srcNet y)}")
apply simp
apply (case_tac "first_srcNet x = first_srcNet y")
apply simp_all
apply (subgoal_tac "first_destNet x = first_destNet y")
apply simp
apply (rule_tac x ="first_srcNet y" in exI,
       rule_tac x = "first_destNet y" in exI)
apply (metis OTNIDaux4 insert_commute )
apply (rule setPair)
apply simp
apply (subgoal_tac "first_srcNet x = first_destNet y")
apply simp
apply (subgoal_tac "first_destNet x = first_srcNet y")
apply simp
apply (rule setPair1)
apply simp
apply (rule setPair4)
apply assumption
apply simp
apply (rule sdnets2)
apply simp
apply simp
apply (metis singletonE first_isIn)
apply (subgoal_tac "sdnets x = {(first_srcNet x, first_destNet x),
                                 (first_destNet x, first_srcNet x)}")
apply (subgoal_tac "sdnets y = {(first_srcNet y, first_destNet y),
                                (first_destNet y, first_srcNet y)}")
apply simp
apply (case_tac "first_srcNet x = first_srcNet y")
apply simp_all
apply (subgoal_tac "first_destNet x = first_destNet y")
apply simp
apply (rule_tac x ="first_srcNet y" in exI,
       rule_tac x = "first_destNet y" in exI)
apply (rule otnaux1)
apply (rule setPair)
apply simp
apply (subgoal_tac "first_srcNet x = first_destNet y")
apply simp
apply (subgoal_tac "first_destNet x = first_srcNet y")
apply simp
apply (rule_tac x ="first_srcNet y" in exI,
       rule_tac x = "first_destNet y" in exI)
apply (metis OTNIDaux4 insert_commute)
apply (rule setPair1)
apply simp
apply (rule setPair4)
apply assumption
apply simp
apply (rule sdnets2,simp_all)+
apply (rule bNaux, simp_all)
done

lemma OTNSepaux: "\<lbrakk>onlyTwoNets (a \<oplus> y) \<and> OnlyTwoNets z \<longrightarrow>
     OnlyTwoNets (separate (a \<oplus> y # z));
     \<not> member DenyAll a; 
      \<not> member DenyAll y; noDenyAll z;
     onlyTwoNets a; OnlyTwoNets (y # z);first_bothNet (a) = first_bothNet y\<rbrakk>
    \<Longrightarrow> OnlyTwoNets (separate (a \<oplus> y # z))"
apply (drule mp)
apply simp_all
apply (rule conjI)
apply (rule otnaux)
apply simp_all
apply (rule_tac p = "(y # z)" in OTNoTN)
apply simp_all
apply (metis member.simps(2))
apply (simp add: onlyTwoNets_def)
apply (rule_tac y = y in OTNConc,simp)
done

lemma OTNSEp[rule_format]: "noDenyAll1 p \<longrightarrow> OnlyTwoNets p  \<longrightarrow>
                            OnlyTwoNets (separate p)"
apply (rule separate.induct) back 
by (simp_all add: OTNSepaux noDA1eq)

lemma nda[rule_format]: "singleCombinators (a#p) \<longrightarrow> noDenyAll p \<longrightarrow>
                         noDenyAll1 (a # p)"
apply (induct p)
apply simp_all
apply (case_tac a, simp_all)
apply (case_tac a, simp_all)
done

lemma nDAcharn[rule_format]: "noDenyAll p = (\<forall> r \<in> set p. \<not> member DenyAll r)"
apply (induct p)
apply simp_all
done

lemma nDAeqSet: "set p = set s \<Longrightarrow> noDenyAll p = noDenyAll s"
apply (simp add: nDAcharn)
done

lemma nDASCaux[rule_format]: "DenyAll \<notin> set p \<longrightarrow> singleCombinators p \<longrightarrow>
                              r \<in> set p \<longrightarrow>  \<not> member DenyAll r"
apply (case_tac r) 
apply simp_all
apply (rule impI)
apply (rule impI)
apply (rule impI)
apply (rule FalseE)
apply (rule SCnotConc)
apply simp
apply simp
done

lemma nDASC[rule_format]: "wellformed_policy1 p \<longrightarrow> singleCombinators p \<longrightarrow>
                           noDenyAll1 p"
apply (induct p)
apply (rule impI)
apply simp_all
apply (rule impI)+
apply (drule mp)
apply (erule waux2)
apply (drule mp)
apply (erule singleCombinatorsConc)
apply (rule nda)
apply simp
apply (simp add: nDAcharn)
apply (rule ballI)
apply (rule nDASCaux)apply simp_all
apply (erule singleCombinatorsConc)
done

lemma noDAAll[rule_format]: "noDenyAll p = (\<not> memberP DenyAll p)"
apply (induct p)
apply simp_all
done

lemma memberPsep[symmetric]: "memberP x p = memberP x (separate p)"
apply (rule separate.induct) back 
apply simp_all
done

lemma noDAsep[rule_format]: "noDenyAll p \<Longrightarrow> noDenyAll (separate p)"
apply (simp add:noDAAll)
apply (subst memberPsep) 
apply simp
done

lemma noDA1sep[rule_format]: "noDenyAll1 p \<longrightarrow> noDenyAll1 (separate p)"
apply (rule separate.induct) back
apply simp_all
apply (rule impI)
apply (rule noDAsep)
apply simp
apply (rule impI)+
apply (rule noDAsep)
apply (case_tac y, simp_all)
apply (rule impI)+
apply (rule noDAsep)
apply (case_tac y, simp_all)
apply (rule impI)+
apply (rule noDAsep)
apply (case_tac y, simp_all)
done

lemma isInAlternativeLista: "(aa \<in> set (net_list_aux [a]))\<Longrightarrow>
                             aa \<in> set (net_list_aux (a # p))"
apply (case_tac a,simp_all)
apply safe
done

lemma isInAlternativeListb: "(aa \<in> set (net_list_aux p))\<Longrightarrow>
                             aa \<in> set (net_list_aux (a # p))"
apply (case_tac a,simp_all)
done

lemma ANDSepaux: "allNetsDistinct (x # y # z) \<Longrightarrow> allNetsDistinct (x \<oplus> y # z)"
apply (simp add: allNetsDistinct_def)
apply (rule allI)+
apply (rule impI)
apply (drule_tac x = a in spec, drule_tac x = b in spec)
apply simp
apply (drule mp)
apply (rule conjI, simp_all)
apply (metis isInAlternativeList)+
done

lemma netlistalternativeSeparateaux:
  "net_list_aux [y] @ net_list_aux z = net_list_aux (y # z)"
apply (case_tac y, simp_all)
done

lemma netlistalternativeSeparate: "net_list_aux p = net_list_aux (separate p)"
apply (rule separate.induct) back
apply simp_all
apply (simp_all add: netlistalternativeSeparateaux)
done

lemma  ANDSepaux2: "\<lbrakk>allNetsDistinct (x # y # z);
                    allNetsDistinct (separate (y # z))\<rbrakk>
       \<Longrightarrow> allNetsDistinct (x # separate (y # z))"
apply (simp add: allNetsDistinct_def)
apply (rule allI)+
apply (rule impI)
apply (drule_tac x = a in spec)
apply (rotate_tac -1)
apply (drule_tac x = b in spec)
apply (simp)
apply (drule mp)
apply (rule conjI)
apply (case_tac "a \<in> set (net_list_aux [x])")
apply simp_all
apply (rule isInAlternativeLista)
apply simp
apply (rule isInAlternativeListb)
apply (subgoal_tac "a \<in> set (net_list_aux (separate (y#z)))")
apply (metis netlistalternativeSeparate)
apply (metis netlistaux netlistalternativeSeparate)
apply (case_tac "b \<in> set (net_list_aux [x])")
apply (rule isInAlternativeLista)
apply simp
apply (rule isInAlternativeListb)
apply (subgoal_tac "b \<in> set (net_list_aux (separate (y#z)))")
apply (metis netlistalternativeSeparate)
apply (metis netlistaux netlistalternativeSeparate)
done



lemma ANDSep[rule_format]: "allNetsDistinct p \<longrightarrow> allNetsDistinct(separate p)"
apply (rule separate.induct) back
apply simp_all
apply (metis ANDConc aNDDA separate.simps(1))
apply (metis ANDConc ANDSepaux ANDSepaux2)
apply (metis ANDConc ANDSepaux ANDSepaux2)
apply (metis ANDConc ANDSepaux ANDSepaux2)
done

lemma wp1_alternativesep[rule_format]: "wellformed_policy1_strong p \<longrightarrow>
                                        wellformed_policy1_strong (separate p)"
apply (rule impI)
apply (subst wp1n_tl) back
apply simp
apply simp
apply (rule sepDA)
apply (erule WP1n_DA_notinSet)
done

lemma noDAsort[rule_format]: "noDenyAll1 p \<longrightarrow> noDenyAll1 (sort p l)"
apply (case_tac "p")
apply simp
apply simp
apply (case_tac "a = DenyAll")
apply simp_all 
apply (rule impI)
apply (subst nDAeqSet)
defer 1
apply simp
defer 1
apply (rule set_sort)
apply (rule impI)
apply (case_tac "insort a (sort list l) l")
apply simp_all
apply (rule noDA1eq)
apply (subgoal_tac "noDenyAll (a#list)")
defer 1
apply (case_tac a, simp,simp)
apply simp
apply simp
apply (subst nDAeqSet)
defer 1
apply assumption
apply (metis sort.simps(2) set_sort)
done

lemma OTNSC[rule_format]: "singleCombinators p \<longrightarrow> OnlyTwoNets p"
apply (induct p)
apply simp_all
apply (rule impI)
apply (drule mp)
apply (erule singleCombinatorsConc)
apply (case_tac a, simp_all)
apply (simp add: onlyTwoNets_def)+
done

lemma fMTaux: "\<not> member DenyAll x \<Longrightarrow> first_bothNet x \<noteq> {}"
apply (metis first_bothNetsd insert_commute insert_not_empty)
done

lemma fl2[rule_format]: "firstList (separate p) = firstList p"
apply (rule separate.induct)
apply simp_all
done

lemma fl3[rule_format]: "NetsCollected p \<longrightarrow> (first_bothNet x \<noteq> firstList p \<longrightarrow>
          (\<forall>a\<in>set p. first_bothNet x \<noteq> first_bothNet a))\<longrightarrow> NetsCollected (x#p)"
apply (induct p)
apply simp_all
done

lemma sortedConc[rule_format]: " sorted (a # p) l \<longrightarrow>  sorted p l"
apply (induct p)
apply simp_all
done

lemma smalleraux2: 
 "{a,b} \<in> set l \<Longrightarrow> {c,d} \<in> set l \<Longrightarrow> {a,b} \<noteq> {c,d} \<Longrightarrow> 
   smaller (DenyAllFromTo a b) (DenyAllFromTo c d) l \<Longrightarrow> 
  \<not> smaller (DenyAllFromTo c d) (DenyAllFromTo a b) l"
apply simp
apply (rule conjI)
apply (rule impI)
apply simp
apply (metis)
apply (metis pos_noteq)
done

lemma smalleraux2a: 
 "{a,b} \<in> set l \<Longrightarrow> {c,d} \<in> set l \<Longrightarrow> {a,b} \<noteq> {c,d} \<Longrightarrow> 
   smaller (DenyAllFromTo a b) (AllowPortFromTo c d p) l \<Longrightarrow> 
  \<not> smaller (AllowPortFromTo c d p) (DenyAllFromTo a b) l"
apply simp
apply (metis eq_imp_le pos_noteq)
done


lemma smalleraux2b: 
"{a,b} \<in> set l \<Longrightarrow> {c,d} \<in> set l \<Longrightarrow> {a,b} \<noteq> {c,d} \<Longrightarrow> y = DenyAllFromTo a b \<Longrightarrow>
   smaller (AllowPortFromTo  c d p) y l \<Longrightarrow> 
  \<not> smaller y (AllowPortFromTo  c d p) l"
apply simp
apply (metis eq_imp_le pos_noteq)
done

lemma smalleraux2c: 
"{a,b} \<in> set l\<Longrightarrow>{c,d}\<in>set l\<Longrightarrow>{a,b} \<noteq> {c,d} \<Longrightarrow> y = AllowPortFromTo a b q \<Longrightarrow> 
smaller (AllowPortFromTo  c d p) y l \<Longrightarrow> \<not> smaller y (AllowPortFromTo  c d p) l"
apply simp
apply (metis pos_noteq)
done

lemma smalleraux3: 
  assumes "x \<in> set l"
  assumes " y \<in> set l" 
  assumes "x \<noteq> y"
  assumes "x = bothNet a" 
  assumes "y = bothNet b"
  assumes "smaller a b l" 
  assumes "singleCombinators [a]" 
  assumes "singleCombinators [b]"  
  shows "\<not> smaller b a l"
proof (cases a)
 case DenyAll thus ?thesis using assms by (case_tac b,simp_all)  
 next
 case (DenyAllFromTo c d) thus ?thesis 
  proof (cases b)
   case DenyAll thus ?thesis using assms DenyAll DenyAllFromTo by simp
   next
   case (DenyAllFromTo e f) thus ?thesis using assms  DenyAllFromTo  apply simp
by (metis DenyAllFromTo `a = DenyAllFromTo c d`  bothNet.simps(2) smalleraux2)
   next
   case (AllowPortFromTo e f g) thus ?thesis using assms DenyAllFromTo AllowPortFromTo apply simp
    by (metis eq_imp_le pos_noteq)
   next
   case (Conc e f) thus ?thesis using assms by simp
  qed
 next
 case (AllowPortFromTo c d p) thus ?thesis
  proof (cases b)
   case DenyAll thus ?thesis using assms AllowPortFromTo DenyAll by simp
   next
   case (DenyAllFromTo e f) thus ?thesis using assms apply simp
by (metis AllowPortFromTo DenyAllFromTo bothNet.simps(3) smalleraux2a)
   next
   case (AllowPortFromTo e f g) thus ?thesis using assms apply simp
by (metis AllowPortFromTo `a = AllowPortFromTo c d p` bothNet.simps(3) smalleraux2c)
   next
   case (Conc e f) thus ?thesis using assms by simp
  qed
 next
 case (Conc c d) thus ?thesis using assms by simp
qed


lemma smalleraux3a: 
   "a \<noteq> DenyAll \<Longrightarrow> b \<noteq> DenyAll \<Longrightarrow> in_list b l \<Longrightarrow> in_list a l \<Longrightarrow>  
   bothNet a \<noteq> bothNet b \<Longrightarrow> smaller a b l \<Longrightarrow> singleCombinators [a] \<Longrightarrow>
    singleCombinators [b] \<Longrightarrow> \<not> smaller b a l"
apply (rule smalleraux3)
apply simp_all
apply (case_tac a, simp_all)
apply (case_tac b, simp_all)
done

lemma posaux[rule_format]: "position a l < position b l \<longrightarrow> a \<noteq> b"
apply (induct l)
apply simp_all
done

lemma posaux6[rule_format]: "a \<in> set l \<longrightarrow> b \<in> set l \<longrightarrow> a \<noteq> b \<longrightarrow>
   position a l \<noteq>  position b l"
apply (induct l)
apply simp_all
apply (rule conjI)
apply (rule impI)+
apply (rule conjI, rule impI,simp)
apply (erule position_positive)
apply (metis position_positive)
apply (metis position_positive)
done

lemma notSmallerTransaux[rule_format]: 
 "\<lbrakk>x \<noteq> DenyAll; r \<noteq> DenyAll; singleCombinators [x]; singleCombinators [y]; 
  singleCombinators [r]; \<not> smaller y x l; smaller x y l; smaller x r l;
  smaller y r l; in_list x l; in_list y l; in_list r l\<rbrakk> \<Longrightarrow>
     \<not> smaller r x l"
by (metis order_trans)


lemma notSmallerTrans[rule_format]: 
 "x \<noteq> DenyAll \<longrightarrow> r \<noteq> DenyAll \<longrightarrow> singleCombinators (x#y#z) \<longrightarrow> 
  \<not> smaller y x l \<longrightarrow> sorted (x#y#z) l \<longrightarrow> r \<in> set z \<longrightarrow> 
  all_in_list (x#y#z) l \<longrightarrow> \<not> smaller r x l"
apply (rule impI)+
apply (rule notSmallerTransaux)
apply simp_all
apply (metis singleCombinatorsConc singleCombinatorsStart)
apply (metis SCSubset equalityE remdups.simps(2) set_remdups
             singleCombinatorsConc singleCombinatorsStart)
apply metis
apply (metis sorted.simps(3) in_set_in_list singleCombinatorsConc
              singleCombinatorsStart sortedConcStart sorted_is_smaller)
apply (metis sorted_Cons all_in_list.simps(2)
             singleCombinatorsConc)
apply metis
apply (metis in_set_in_list)
done

lemma  NCSaux1[rule_format]:
 "noDenyAll p \<longrightarrow> {x, y} \<in> set l \<longrightarrow>  all_in_list p l\<longrightarrow> singleCombinators p \<longrightarrow> 
  sorted (DenyAllFromTo x y # p) l \<longrightarrow> {x, y} \<noteq> firstList p \<longrightarrow>
  DenyAllFromTo u v \<in> set p \<longrightarrow> {x, y} \<noteq> {u, v}"
proof (cases p)
 case Nil thus ?thesis by simp next
 case (Cons a p) thus ?thesis using assms apply simp
  apply (rule impI)+
  apply (rule conjI)
  apply (metis bothNet.simps(2) first_bothNet.simps(3))
  apply (rule impI) 
  apply (subgoal_tac "smaller (DenyAllFromTo x y) (DenyAllFromTo u v) l")
apply (subgoal_tac "\<not> smaller (DenyAllFromTo u v) (DenyAllFromTo x y) l")
apply (rule notI)
apply (case_tac "smaller (DenyAllFromTo u v) (DenyAllFromTo x y) l")
apply (simp del: smaller.simps)
apply simp
apply (case_tac "x = u")
apply simp
apply (case_tac "y = v")
apply simp
apply (subgoal_tac "u = v")
apply simp
apply simp
apply simp
apply metis
apply (rule_tac y = a and z = p in notSmallerTrans)
apply (simp_all del: smaller.simps)
apply (rule smalleraux3a)
apply (simp_all del: smaller.simps)
apply (case_tac a, simp_all del: smaller.simps)
apply (case_tac a, simp_all del: smaller.simps)
apply (rule_tac y = a in order_trans)
apply simp_all
apply (subgoal_tac "in_list (DenyAllFromTo u v) l")
apply simp
apply (rule_tac p = p in in_set_in_list)
apply simp
apply (case_tac a, simp_all del: smaller.simps)
apply (metis all_in_list.simps(2) sorted_Cons)
done
qed

lemma posaux3[rule_format]:
  "a \<in> set l \<longrightarrow> b \<in> set l \<longrightarrow> a \<noteq> b \<longrightarrow> position a l \<noteq> position b l"
apply (induct l)
apply simp_all
apply (rule conjI)
apply (rule impI)+
apply (rule conjI)
apply (rule impI)
apply simp_all
apply (metis position_positive)+
done

lemma posaux4[rule_format]: "singleCombinators [a] \<longrightarrow> a\<noteq> DenyAll \<longrightarrow>
                           b \<noteq> DenyAll \<longrightarrow> in_list a l \<longrightarrow>in_list b l \<longrightarrow>
                           smaller a b  l\<longrightarrow> x = (bothNet a) \<longrightarrow>
                           y = (bothNet b) \<longrightarrow> position x l <= position y l"
proof (cases a) 
 case DenyAll then show ?thesis by simp
 next
 case (DenyAllFromTo c d) thus ?thesis
  proof (cases b)
   case DenyAll thus ?thesis by simp next
   case (DenyAllFromTo e f) thus ?thesis using assms DenyAllFromTo
    by (auto simp: eq_imp_le  `a = DenyAllFromTo c d`)
   next
   case (AllowPortFromTo e f p) thus ?thesis using assms  `a = DenyAllFromTo c d` by simp next
   case (Conc e f) thus ?thesis using assms Conc `a = DenyAllFromTo c d` by simp
 qed
 next
 case (AllowPortFromTo c d p) thus ?thesis
  proof (cases b)
   case DenyAll thus ?thesis by simp next
   case (DenyAllFromTo e f) thus ?thesis using assms AllowPortFromTo by simp next
   case (AllowPortFromTo e f p2) thus ?thesis using assms   `a = AllowPortFromTo c d p` by simp next
   case (Conc e f) thus ?thesis using assms AllowPortFromTo by simp
 qed
 next
 case (Conc c d) thus ?thesis by simp
qed

lemma  NCSaux2[rule_format]:
  "noDenyAll p \<longrightarrow> {a, b} \<in> set l \<longrightarrow> all_in_list p l \<longrightarrow>singleCombinators p \<longrightarrow>
   sorted (DenyAllFromTo a b # p) l \<longrightarrow> {a, b} \<noteq> firstList p \<longrightarrow>
   AllowPortFromTo u v w \<in> set p \<longrightarrow>  {a, b} \<noteq> {u, v}"
apply (case_tac p)
apply simp_all
apply (rule impI)+
apply (rule conjI)
apply (rule impI)
apply (rotate_tac -1, drule sym)
apply simp
apply (rule impI)
apply (subgoal_tac "smaller (DenyAllFromTo a b) (AllowPortFromTo u v w) l")
apply (subgoal_tac "\<not> smaller (AllowPortFromTo u v w) (DenyAllFromTo a b) l")
defer 1
apply (rule_tac y = aa and z = list in notSmallerTrans)
apply (simp_all del: smaller.simps)
apply (rule smalleraux3a)
apply (simp_all del: smaller.simps)
apply (case_tac aa, simp_all del: smaller.simps)
apply (case_tac aa, simp_all del: smaller.simps)
apply (rule_tac y = aa in order_trans)
apply (simp_all del: smaller.simps)
apply (subgoal_tac "in_list (AllowPortFromTo u v w) l")
apply simp
apply (rule_tac p = list in in_set_in_list)
apply simp
apply simp
apply (metis all_in_list.simps(2) sorted_Cons)
apply (rule_tac l = l in posaux) 
apply (rule_tac y = "position (first_bothNet aa) l" in basic_trans_rules(22))
apply simp
apply (simp split: if_splits)
apply (case_tac aa, simp_all)
apply (case_tac "a = \<alpha>1 \<and> b = \<alpha>2")
apply simp_all
apply (case_tac "a = \<alpha>1")
apply simp_all
apply (rule basic_trans_rules(18))
apply simp
apply (rule posaux3)
apply simp
apply simp
apply simp
apply (rule basic_trans_rules(18))
apply simp
apply (rule posaux3)
apply simp
apply simp
apply simp
apply (rule basic_trans_rules(18))
apply simp
apply (rule posaux3)
apply simp
apply simp
apply simp
apply (rule basic_trans_rules(18))
apply (rule_tac a = "DenyAllFromTo a b" and b = aa in posaux4)
apply simp_all
apply (case_tac aa,simp_all)
apply (case_tac aa, simp_all)
apply (rule posaux3)
apply simp_all
apply (case_tac aa, simp_all)
apply (simp split: if_splits)
apply (rule_tac a = aa and b = "AllowPortFromTo u v w" in posaux4)
apply simp_all
apply (case_tac aa,simp_all)
apply (rule_tac p = list in sorted_is_smaller)
apply simp_all
apply (case_tac aa, simp_all)
apply (case_tac aa, simp_all)
apply (rule_tac a = aa and b = "AllowPortFromTo u v w" in posaux4)
apply simp_all
apply (case_tac aa,simp_all)
apply (subgoal_tac "in_list (AllowPortFromTo u v w) l")
apply simp
apply (rule_tac p = list in in_set_in_list)
apply simp
defer 1
apply simp_all
apply (metis all_in_list.simps(2) sorted_Cons)
apply (case_tac aa, simp_all)
done

lemma  NCSaux3[rule_format]:
 "noDenyAll p \<longrightarrow> {a, b} \<in> set l \<longrightarrow>  all_in_list p l \<longrightarrow>singleCombinators p \<longrightarrow> 
  sorted (AllowPortFromTo a b w # p) l \<longrightarrow> {a, b} \<noteq> firstList p \<longrightarrow>
  DenyAllFromTo u v \<in> set p \<longrightarrow> {a, b} \<noteq> {u, v}"
apply (case_tac p)
apply simp_all
apply (rule impI)+
apply (rule conjI)
apply (rule impI)
apply (rotate_tac -1, drule sym)
apply simp
apply (rule impI)
apply (subgoal_tac "smaller (AllowPortFromTo a b w) (DenyAllFromTo u v) l")
apply (subgoal_tac "\<not> smaller (DenyAllFromTo u v) (AllowPortFromTo a b w) l")
apply (simp split: if_splits)
apply (rule_tac y = aa and z = list in notSmallerTrans)
apply (simp_all del: smaller.simps)
apply (rule smalleraux3a)
apply (simp_all del: smaller.simps)
apply (case_tac aa, simp_all del: smaller.simps)
apply (case_tac aa, simp_all del: smaller.simps)
apply (rule_tac y = aa in order_trans)
apply (simp_all del: smaller.simps)
apply (subgoal_tac "in_list (DenyAllFromTo u v) l")
apply simp
apply (rule_tac p = list in in_set_in_list)
apply simp
apply simp
apply (rule_tac p = list in sorted_is_smaller)
apply (simp_all del: smaller.simps)
apply (subgoal_tac "in_list (DenyAllFromTo u v) l")
apply simp
apply (rule_tac p = list in in_set_in_list)
apply simp
apply simp
apply (erule singleCombinatorsConc)
done


lemma  NCSaux4[rule_format]:
"noDenyAll p \<longrightarrow> {a, b} \<in> set l \<longrightarrow>  all_in_list p l \<longrightarrow> singleCombinators p \<longrightarrow> 
 sorted (AllowPortFromTo a b c # p) l \<longrightarrow> {a, b} \<noteq> firstList p \<longrightarrow>
 AllowPortFromTo u v w \<in> set p \<longrightarrow> {a, b} \<noteq> {u, v}"
apply (case_tac p)
apply simp_all
apply (rule impI)+
apply (rule conjI)
apply (rule impI)
apply (rotate_tac -1, drule sym)
apply simp
apply (rule impI)
apply (subgoal_tac "smaller (AllowPortFromTo a b c) (AllowPortFromTo u v w) l")
apply (subgoal_tac "\<not> smaller (AllowPortFromTo u v w) (AllowPortFromTo a b c) l")
apply (simp split: if_splits)
apply (rule_tac y = aa and z = list in notSmallerTrans)
apply (simp_all del: smaller.simps)
apply (rule smalleraux3a)
apply (simp_all del: smaller.simps)
apply (case_tac aa, simp_all del: smaller.simps)
apply (case_tac aa, simp_all del: smaller.simps)
apply (case_tac aa, simp_all del: smaller.simps)
apply (rule_tac y = aa in order_trans)
apply (simp_all del: smaller.simps)
apply (subgoal_tac "in_list (AllowPortFromTo u v w) l")
apply simp
apply (rule_tac p = list in in_set_in_list)
apply simp
apply (case_tac aa, simp_all del: smaller.simps)
apply (rule_tac p = list in sorted_is_smaller)
apply (simp_all del: smaller.simps)
apply (subgoal_tac "in_list (AllowPortFromTo u v w) l")
apply simp
apply (rule_tac p = list in in_set_in_list)
apply simp
apply simp
apply (rule_tac y = aa in order_trans)
apply (simp_all del: smaller.simps)
apply (subgoal_tac "in_list (AllowPortFromTo u v w) l")
apply simp
apply (rule_tac p = list in in_set_in_list)
apply simp
apply simp
apply (rule_tac p = list in sorted_is_smaller)
apply (simp_all del: smaller.simps)
apply (subgoal_tac "in_list (AllowPortFromTo u v w) l")
apply simp
apply (rule_tac p = list in in_set_in_list)
apply simp_all
done


lemma NetsCollectedSorted[rule_format]: 
 "noDenyAll1 p \<longrightarrow> all_in_list p l \<longrightarrow> singleCombinators p \<longrightarrow> sorted p l \<longrightarrow>
  NetsCollected p"
apply (induct p)
apply simp
apply (rule impI)+
apply (drule mp)
apply (erule noDA1C)
apply (drule mp)
apply simp
apply (drule mp)
apply (erule singleCombinatorsConc)
apply (drule mp)
apply (erule sortedConc)

apply (rule fl3)
apply simp
apply simp
apply (case_tac a)
apply simp_all
apply (metis fMTaux noDA)
apply (case_tac aa)
apply simp_all
apply (rule NCSaux1, simp_all)
apply (rule NCSaux2, simp_all)
apply (metis aux0_0)
apply (case_tac "aa")
apply simp_all
apply (rule NCSaux3,simp_all)
apply (rule NCSaux4,simp_all)
apply (metis aux0_0)
done





lemma NetsCollectedSort: "distinct p \<Longrightarrow>noDenyAll1 p \<Longrightarrow> all_in_list p l \<Longrightarrow>
                          singleCombinators p \<Longrightarrow> NetsCollected (sort p l)"
apply (rule_tac l = l in NetsCollectedSorted)
apply (rule noDAsort)
apply simp_all
apply (rule_tac b=p in all_in_listSubset)
apply simp_all
apply (rule sort_is_sorted)
apply simp_all
done


lemma fBNsep[rule_format]: "(\<forall>a\<in>set z. {b,c} \<noteq> first_bothNet a) \<longrightarrow>
                           (\<forall>a\<in>set (separate z). {b,c} \<noteq> first_bothNet a)"
apply (rule separate.induct) back
apply simp
apply (rule impI, simp)+
done


lemma fBNsep1[rule_format]: " (\<forall>a\<in>set z. first_bothNet x \<noteq> first_bothNet a) \<longrightarrow>
                        (\<forall>a\<in>set (separate z). first_bothNet x \<noteq> first_bothNet a)"
apply (rule separate.induct) back
apply simp
apply (rule impI, simp)+
done


lemma NetsCollectedSepauxa:
  "\<lbrakk>{b,c} \<noteq> firstList z; noDenyAll1 z;
   (\<forall>a\<in>set z. {b,c} \<noteq> first_bothNet a); NetsCollected (z);
   NetsCollected (separate (z)); {b,c} \<noteq> firstList (separate (z));
   a \<in> set (separate (z))\<rbrakk> \<Longrightarrow>
   {b,c} \<noteq> first_bothNet a"
apply (rule fBNsep)
apply simp_all
done


lemma NetsCollectedSepaux:
"\<lbrakk>first_bothNet (x::('a,'b)Combinators) \<noteq> first_bothNet y; \<not> member DenyAll y \<and> 
 noDenyAll z;
 (\<forall>a\<in>set z. first_bothNet x \<noteq> first_bothNet a) \<and> NetsCollected (y # z);
NetsCollected (separate (y # z)); first_bothNet x \<noteq> firstList (separate (y # z));
 a \<in> set (separate (y # z))\<rbrakk> \<Longrightarrow>
 first_bothNet (x::('a,'b)Combinators) \<noteq> first_bothNet (a::('a,'b)Combinators)"
apply (rule fBNsep1)
apply simp_all
apply auto
done

lemma NetsCollectedSep[rule_format]: "noDenyAll1 p \<longrightarrow> NetsCollected p \<longrightarrow> 
                                      NetsCollected (separate p)"
apply (rule separate.induct) back
apply simp_all
apply (metis fMTaux noDA noDA1eq noDAsep)
apply (rule conjI|rule impI)+
apply simp
apply (metis fBNsep set_ConsD)
apply (metis noDA1eq noDenyAll.simps(1))
apply (rule conjI|rule impI)+
apply (metis fBNsep set_ConsD)
apply (metis noDA1eq noDenyAll.simps(1))
apply (rule conjI|rule impI)+
apply simp
apply (metis NetsCollected.simps(1) NetsCollectedSepaux firstList.simps(1) fl2 fl3 
  noDA1eq noDenyAll.simps(1))
apply (metis noDA1eq noDenyAll.simps(1))
done

lemma OTNaux: 
  "onlyTwoNets a \<Longrightarrow> \<not> member DenyAll a \<Longrightarrow> (x,y) \<in> sdnets a \<Longrightarrow> 
   (x = first_srcNet a \<and> y = first_destNet a) \<or> 
   (x = first_destNet a \<and> y = first_srcNet a)"
apply (case_tac "(x = first_srcNet a \<and> y = first_destNet a)")
apply simp_all
apply (simp add: onlyTwoNets_def)
apply (case_tac "(\<exists>aa b. sdnets a = {(aa, b)})")
apply simp_all
apply (subgoal_tac "sdnets a = {(first_srcNet a,first_destNet a)}")
apply simp_all
apply (metis singletonE first_isIn)
apply (subgoal_tac "sdnets a = {(first_srcNet a,first_destNet a),(first_destNet a,
                    first_srcNet a)}")
apply simp_all
apply (rule sdnets2)
apply simp_all
done

lemma sdnets_charn: "onlyTwoNets a \<Longrightarrow> \<not> member DenyAll a \<Longrightarrow>
sdnets a = {(first_srcNet a,first_destNet a)} \<or> 
sdnets a = {(first_srcNet a, first_destNet a),(first_destNet a, first_srcNet a)}"
apply (case_tac "sdnets a = {(first_srcNet a, first_destNet a)}")  
apply simp_all
apply (simp add: onlyTwoNets_def)
apply (case_tac "(\<exists>aa b. sdnets a = {(aa, b)})")
apply simp_all
apply (metis singletonE first_isIn)
apply (subgoal_tac "sdnets a = {(first_srcNet a,first_destNet a),
                   (first_destNet a,first_srcNet a)}")
apply simp_all
apply (rule sdnets2)
apply simp_all
done

lemma first_bothNet_charn[rule_format]: "\<not> member DenyAll a \<longrightarrow> 
                  first_bothNet a = {first_srcNet a, first_destNet a}"
apply (induct a)
apply simp_all
done

lemma sdnets_noteq:
  "\<lbrakk>onlyTwoNets a; onlyTwoNets aa; first_bothNet a \<noteq> first_bothNet aa; 
   \<not> member DenyAll a; \<not> member DenyAll aa\<rbrakk>
              \<Longrightarrow> sdnets a \<noteq> sdnets aa"
apply (insert sdnets_charn [of a])
apply (insert sdnets_charn [of aa])
apply (insert first_bothNet_charn [of a])
apply (insert first_bothNet_charn [of aa])
apply simp
apply (metis OTNaux first_bothNetsd first_isIn insert_absorb2 insert_commute)
done


lemma fbn_noteq: 
  "\<lbrakk>onlyTwoNets a; onlyTwoNets aa; first_bothNet a \<noteq> first_bothNet aa; 
   \<not> member DenyAll a; \<not> member DenyAll aa; allNetsDistinct [a, aa]\<rbrakk> \<Longrightarrow> 
   first_srcNet a \<noteq> first_srcNet aa \<or> first_srcNet a \<noteq> first_destNet aa \<or> 
   first_destNet a \<noteq> first_srcNet aa \<or> first_destNet a \<noteq> first_destNet aa" 
apply (insert sdnets_charn [of a])
apply (insert sdnets_charn [of aa])
apply simp
apply (insert sdnets_noteq [of a aa])
apply simp
apply (rule impI)+
apply simp
apply (case_tac " sdnets a = {(first_destNet aa, first_srcNet aa)}")
apply simp_all
apply (case_tac "sdnets aa = {(first_srcNet aa, first_destNet aa)}")
apply simp_all 
done

lemma NCisSD2aux: 
   "\<lbrakk>onlyTwoNets a; onlyTwoNets aa; first_bothNet a \<noteq> first_bothNet aa; 
   \<not> member DenyAll a; \<not> member DenyAll aa; allNetsDistinct [a, aa]\<rbrakk> \<Longrightarrow> 
     disjSD_2 a aa" 
apply (simp add: disjSD_2_def)
apply (rule allI)+
apply (rule impI)
apply (insert sdnets_charn [of a])
apply (insert sdnets_charn [of aa])
apply simp
apply (insert sdnets_noteq [of a aa])
apply (insert fbn_noteq [of a aa])
apply simp
apply (simp add: allNetsDistinct_def twoNetsDistinct_def)
apply (rule conjI)
apply (cases "sdnets a = {(first_srcNet a, first_destNet a)}")
apply (cases "sdnets aa = {(first_srcNet aa, first_destNet aa)}")
apply simp_all
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (case_tac "(c = first_srcNet aa \<and> d = first_destNet aa)")
apply simp_all
apply (case_tac "(first_srcNet a) \<noteq> (first_srcNet aa)")
apply (metis firstInNeta alternativelistconc2)
apply simp
apply (subgoal_tac "first_destNet a \<noteq> first_destNet aa") 
apply (metis  firstInNet alternativelistconc2)
apply (metis first_bothNetsd)
apply (case_tac "(first_destNet aa) \<noteq> (first_srcNet a)")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply simp
apply (case_tac "first_destNet aa \<noteq> first_destNet a")
apply simp
apply (subgoal_tac "first_srcNet aa \<noteq> first_destNet a")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (metis first_bothNetsd insert_commute)
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (case_tac "(c = first_srcNet aa \<and> d = first_destNet aa)")
apply simp_all
apply (case_tac "(ab = first_srcNet a \<and> b = first_destNet a)")
apply simp_all
apply (case_tac "(first_srcNet a) \<noteq> (first_srcNet aa)")
apply (metis firstInNeta alternativelistconc2)
apply simp
apply (subgoal_tac "first_destNet a \<noteq> first_destNet aa")
apply (metis  firstInNet alternativelistconc2)
apply (metis first_bothNetsd )
apply (case_tac "(first_destNet aa) \<noteq> (first_srcNet a)")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply simp
apply (case_tac "first_destNet aa \<noteq> first_destNet a")
apply simp
apply (subgoal_tac "first_srcNet aa \<noteq> first_destNet a")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (metis first_bothNetsd insert_commute)
apply (metis)
apply (case_tac "(ab = first_srcNet a \<and> b = first_destNet a)")
apply simp_all
apply (case_tac "c = first_srcNet aa")
apply simp_all
apply (metis OTNaux)
apply (subgoal_tac "c = first_destNet aa")
apply simp
apply (subgoal_tac "d = first_srcNet aa")
apply simp
apply (case_tac "(first_srcNet a) \<noteq> (first_destNet aa)")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply simp
apply (subgoal_tac "first_destNet a \<noteq> first_srcNet aa")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (metis first_bothNetsd insert_commute)
apply (metis OTNaux)
apply (metis OTNaux)
apply (case_tac "c = first_srcNet aa")
apply simp_all
apply (metis OTNaux)
apply (subgoal_tac "c = first_destNet aa")
apply simp
apply (subgoal_tac "d = first_srcNet aa")
apply simp
apply (case_tac "(first_destNet a) \<noteq> (first_destNet aa)")
apply (metis firstInNet alternativelistconc2)
apply simp
apply (subgoal_tac "first_srcNet a \<noteq> first_srcNet aa")
apply (metis firstInNeta alternativelistconc2)
apply (metis first_bothNetsd  insert_commute)
apply (metis OTNaux)
apply (metis OTNaux)
apply (cases "sdnets a = {(first_srcNet a, first_destNet a)}")
apply (cases "sdnets aa = {(first_srcNet aa, first_destNet aa)}")
apply simp_all
apply (case_tac "(c = first_srcNet aa \<and> d = first_destNet aa)")
apply simp_all
apply (case_tac "(first_srcNet a) \<noteq> (first_destNet aa)")
apply simp_all
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (subgoal_tac "first_destNet a \<noteq> first_srcNet aa")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (metis first_bothNetsd insert_commute)
apply (case_tac "(c = first_srcNet aa \<and> d = first_destNet aa)")
apply simp_all
apply (case_tac "(ab = first_srcNet a \<and> b = first_destNet a)")
apply simp_all
apply (case_tac "(first_destNet a) \<noteq> (first_srcNet aa)")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply simp
apply (subgoal_tac "first_srcNet a \<noteq> first_destNet aa")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (metis first_bothNetsd insert_commute)
apply (case_tac "(first_srcNet aa) \<noteq> (first_srcNet a)")
apply (metis firstInNeta alternativelistconc2)
apply simp
apply (case_tac "first_destNet aa \<noteq> first_destNet a")
apply (metis firstInNet alternativelistconc2)
apply simp
apply (metis first_bothNetsd)
apply (cases "sdnets aa = {(first_srcNet aa, first_destNet aa)}")
apply simp_all
apply (case_tac "(c = first_srcNet aa \<and> d = first_destNet aa)")
apply simp_all
apply (case_tac "(ab = first_srcNet a \<and> b = first_destNet a)")
apply simp_all
apply (case_tac "(first_destNet a) \<noteq> (first_srcNet aa)")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply simp
apply (subgoal_tac "first_srcNet a \<noteq> first_destNet aa")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (metis first_bothNetsd insert_commute)
apply (case_tac "(first_srcNet aa) \<noteq> (first_srcNet a)")
apply (metis firstInNeta alternativelistconc2)
apply simp
apply (case_tac "first_destNet aa \<noteq> first_destNet a")
apply (metis firstInNet alternativelistconc2)
apply simp
apply (metis first_bothNetsd)
apply (case_tac "(c = first_srcNet aa \<and> d = first_destNet aa)")
apply simp_all
apply (case_tac "(ab = first_srcNet a \<and> b = first_destNet a)")
apply simp_all
apply (case_tac "(first_destNet a) \<noteq> (first_srcNet aa)")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply simp
apply (subgoal_tac "first_srcNet a \<noteq> first_destNet aa")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply (metis insert_commute)
apply (case_tac "(first_srcNet aa) \<noteq> (first_srcNet a)")
apply (metis firstInNeta alternativelistconc2)
apply simp
apply (case_tac "first_destNet aa \<noteq> first_destNet a")
apply (metis firstInNet alternativelistconc2)
apply simp
apply (case_tac "(ab = first_srcNet a \<and> b = first_destNet a)")
apply simp_all
apply (case_tac "(first_destNet a) \<noteq> (first_srcNet aa)")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply simp
apply (subgoal_tac "first_srcNet a \<noteq> first_srcNet aa")
apply (metis firstInNeta firstInNet)
apply (metis)
apply (case_tac "(first_srcNet aa) \<noteq> (first_destNet a)")
apply (metis firstInNeta firstInNet alternativelistconc2)
apply simp
apply (case_tac "first_destNet aa \<noteq> first_srcNet a")
apply (metis firstInNeta firstInNet alternativelistconc2 )
apply simp
apply (metis insert_commute)
done

lemma ANDaux3[rule_format]: "y \<in> set xs \<longrightarrow> a \<in> set (net_list_aux [y]) \<longrightarrow> 
                           a \<in> set (net_list_aux xs)"
apply (induct xs)
apply simp_all
apply (rule conjI)
apply (rule impI)+
apply simp
apply (metis isInAlternativeList)
apply (rule impI)+
apply simp
apply (erule isInAlternativeListb)
done

lemma ANDaux2: "allNetsDistinct (x # xs) \<Longrightarrow> y \<in> set xs  
              \<Longrightarrow> allNetsDistinct [x,y]"
apply (simp add: allNetsDistinct_def)
apply (rule allI)
apply (rule allI)
apply (rule impI)+
apply (drule_tac x = a in spec)
apply (drule_tac x = b in spec)
apply simp
apply (drule mp)
apply simp_all
apply (rule conjI)
apply (case_tac "a \<in> set (net_list_aux [x])")
apply (erule isInAlternativeLista)
apply (rule isInAlternativeListb)
apply (rule ANDaux3)
apply simp_all
apply (metis netlistaux)
apply (case_tac "b \<in> set (net_list_aux [x])")
apply (erule isInAlternativeLista)
apply (rule isInAlternativeListb)
apply (rule ANDaux3)
apply simp_all
apply (metis netlistaux)
done
 
lemma NCisSD2[rule_format]: "
\<lbrakk>\<not> member DenyAll a; OnlyTwoNets (a#p); NetsCollected2 (a # p); 
 NetsCollected (a#p);noDenyAll ( p); allNetsDistinct (a # p); s \<in> set p\<rbrakk> \<Longrightarrow>
 disjSD_2 a s"
apply (case_tac p)
apply simp_all
apply (rule NCisSD2aux)
apply simp_all 
apply (rule OTNoTN)
apply simp 
apply (case_tac a, simp_all)
apply (rule OTNoTN)
apply simp
apply (metis member.simps(2) noDA)
apply simp
apply metis
apply (metis noDA)
apply (rule ANDaux2)
apply simp_all
apply simp
done

lemma separatedNC[rule_format]: 
  "OnlyTwoNets p \<longrightarrow> NetsCollected2 p \<longrightarrow> NetsCollected p \<longrightarrow> noDenyAll1 p \<longrightarrow>  
   allNetsDistinct p  \<longrightarrow> separated p"
apply (induct p)
apply simp_all
apply (case_tac "a = DenyAll")
apply simp_all
defer 1
apply (rule impI)+
apply (drule mp)
apply (erule OTNConc)
apply (drule mp)
apply (case_tac p, simp_all)
apply (drule mp)
apply (erule noDA1C)
apply (rule conjI)
apply (rule allI)
apply (rule impI) 
apply (rule NCisSD2)
apply simp_all
apply (case_tac a, simp_all)
apply (case_tac a, simp_all)
apply (drule mp)
apply (erule ANDConc)
apply simp
apply (rule impI)+
apply (simp)
apply (drule mp)
apply (erule noDA1eq)
apply (drule mp)
apply (erule ANDConc)
apply simp
apply (simp add: disjSD_2_def)
done

lemma NC2Sep[rule_format]: "noDenyAll1 p \<longrightarrow> NetsCollected2 (separate p)"
apply (rule separate.induct) back
apply simp_all
apply (rule impI, drule mp)
apply (erule noDA1eq)
apply (case_tac "separate x = []")
apply simp_all
apply (case_tac x, simp_all)
apply (metis fMTaux firstList.simps(1) fl2 )
apply (rule impI)+
apply simp
apply (drule mp)
apply (rule noDA1eq)
apply (case_tac y, simp_all)
apply (metis firstList.simps(1) fl2)
apply (rule impI)+
apply simp
apply (drule mp)
apply (rule noDA1eq)
apply (case_tac y, simp_all)
apply (metis firstList.simps(1) fl2)
apply (rule impI)+
apply simp
apply (drule mp)
apply (rule noDA1eq)
apply (case_tac y, simp_all)
apply (metis firstList.simps(1) fl2)
done

lemma separatedSep[rule_format]: 
 "OnlyTwoNets p \<longrightarrow> NetsCollected2 p \<longrightarrow> NetsCollected p \<longrightarrow> noDenyAll1 p \<longrightarrow> 
  allNetsDistinct p  \<longrightarrow> separated (separate p)"
apply (rule impI)+
apply (rule separatedNC)
apply (rule OTNSEp)
apply simp_all
apply (erule NC2Sep)
apply (erule NetsCollectedSep)
apply simp
apply (erule noDA1sep)
apply (erule ANDSep)
done

lemma rADnMT[rule_format]: "p \<noteq> []  \<longrightarrow> removeAllDuplicates p \<noteq> []"
apply (induct p)
apply simp_all
done

(*
 TODO: Prove this lemma:
lemma all2: "all_in_list (policy2list p) (Nets_List p)"
apply (induct "policy2list p" "Nets_List p" rule: all_in_list.induct)
apply simp_all

*)


lemma remDupsNMT[rule_format]: "p \<noteq> [] \<longrightarrow>
    remdups p \<noteq> []"
by (metis remdups_eq_nil_iff) 


lemma sets_distinct1: "(n::int) \<noteq> m \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
by auto

lemma sets_distinct2: "(m::int) \<noteq> n \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
by auto

lemma sets_distinct5: "(n::int) < m \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
by auto

lemma sets_distinct6: "(m::int) < n \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
by auto



end
