
theory NormalisationIPPProofs
imports NormalisationIntegerPortProof
begin

text{* Normalisation proofs which are specific to the IntegerProtocol address representation. *} 

lemma ConcAssoc: "Cp((A \<oplus> B) \<oplus> D) = Cp(A \<oplus> (B \<oplus> D))"
apply (simp add: Cp.simps)
  done


lemma aux26[simp]: "twoNetsDistinct a b c d \<Longrightarrow> 
             dom (Cp (AllowPortFromTo a b p)) \<inter> dom (Cp (DenyAllFromTo c d)) = {}"
by (auto simp: PLemmas twoNetsDistinct_def netsDistinct_def) auto


lemma wp2_aux[rule_format]: "wellformed_policy2Pr (xs @ [x]) \<longrightarrow> 
                             wellformed_policy2Pr xs"
apply (induct xs, simp_all)
apply (case_tac "a", simp_all)
done 

lemma Cdom2: "x \<in> dom(Cp b) \<Longrightarrow> Cp (a \<oplus> b) x = (Cp b) x"
by (auto simp: Cp.simps)

lemma wp2Conc[rule_format]: "wellformed_policy2Pr (x#xs) \<Longrightarrow> wellformed_policy2Pr xs"
by (case_tac "x",simp_all)

lemma DAimpliesMR_E[rule_format]: "DenyAll \<in> set p \<longrightarrow>
                                   (\<exists> r. applied_rule_rev Cp x p = Some r)"
apply (simp add: applied_rule_rev_def)
apply (rule_tac xs = p in rev_induct)
apply simp_all
by (metis Cp.simps(1) denyAllDom)



lemma DAimplieMR[rule_format]: "DenyAll \<in> set p \<Longrightarrow> applied_rule_rev Cp x p \<noteq> None"
by (auto intro: DAimpliesMR_E)

lemma MRList1[rule_format]: "x \<in> dom (Cp a) \<Longrightarrow> applied_rule_rev Cp x (b@[a]) = Some a"
by (simp add: applied_rule_rev_def)

lemma MRList2: "x \<in> dom (Cp a) \<Longrightarrow> applied_rule_rev Cp x (c@b@[a]) = Some a"
by (simp add: applied_rule_rev_def)

lemma MRList3: "x \<notin> dom (Cp xa) \<Longrightarrow>
            applied_rule_rev Cp x (a @ b # xs @ [xa]) = applied_rule_rev Cp x (a @ b # xs)"
by (simp add: applied_rule_rev_def)

lemma CConcEnd[rule_format]: "Cp a x = Some y \<longrightarrow>
                              Cp (list2FWpolicy (xs @ [a])) x = Some y"
 (is "?P xs")
apply (rule_tac P = ?P in list2FWpolicy.induct)
by (simp_all add:Cp.simps)

lemma CConcStartaux: " \<lbrakk>Cp a x = None\<rbrakk> \<Longrightarrow> (Cp aa ++ Cp a) x = Cp aa x"
by (simp add: PLemmas)

lemma CConcStart[rule_format]: "xs \<noteq> [] \<longrightarrow> Cp a x = None \<longrightarrow>
                             Cp (list2FWpolicy (xs @ [a])) x = Cp (list2FWpolicy xs) x"
apply (rule list2FWpolicy.induct)
by (simp_all add: PLemmas)

lemma mrNnt[simp]: "applied_rule_rev Cp x p = Some a \<Longrightarrow> p \<noteq> []"
apply (simp add: applied_rule_rev_def)
by auto

lemma mr_is_C[rule_format]: "applied_rule_rev Cp x p = Some a \<longrightarrow>
                             Cp (list2FWpolicy (p)) x = Cp a x"
apply (simp add: applied_rule_rev_def)
apply (rule rev_induct)
apply simp_all
apply safe
apply (metis CConcEnd rotate_simps)
apply (metis CConcEnd)
by (metis CConcStart applied_rule_rev_def mrNnt option.exhaust)


lemma CConcStart2: "\<lbrakk>p \<noteq> []; x \<notin> dom (Cp a)\<rbrakk> \<Longrightarrow>
                    Cp (list2FWpolicy (p@[a])) x = Cp (list2FWpolicy p)x"
by (erule CConcStart,simp add: PLemmas)

lemma CConcEnd1: "\<lbrakk>q@p \<noteq> []; x \<notin> dom (Cp a)\<rbrakk> \<Longrightarrow>
                  Cp (list2FWpolicy (q@p@[a])) x = Cp (list2FWpolicy (q@p))x"
apply (subst lCdom2)
by (rule CConcStart2, simp_all)

lemma CConcEnd2[rule_format]: "x \<in> dom (Cp a) \<longrightarrow>
                               Cp (list2FWpolicy (xs @ [a])) x = Cp a x"
 (is "?P xs")
apply (rule_tac P = ?P in list2FWpolicy.induct)
by (auto simp:Cp.simps)


lemma bar3: "x \<in> dom (Cp (list2FWpolicy (xs @ [xa]))) \<Longrightarrow>
             x \<in> dom (Cp (list2FWpolicy xs)) \<or> x \<in> dom (Cp xa)"
apply auto
by (metis CConcStart eq_Nil_appendI l2p_aux2 option.simps(3))


lemma CeqEnd[rule_format,simp]: "a \<noteq> [] \<longrightarrow> x \<in> dom (Cp (list2FWpolicy a)) \<longrightarrow> 
                                Cp (list2FWpolicy (b@a)) x = (Cp (list2FWpolicy a)) x"
apply (rule rev_induct,simp_all) 
apply (case_tac "xs \<noteq> []", simp_all)
apply (case_tac "x \<in> dom (Cp xa)")
apply (metis CConcEnd2 MRList2 mr_is_C rotate_simps)
apply (metis CConcEnd1 CConcStart2 Nil_is_append_conv bar3 rotate_simps)
apply (metis MRList2 eq_Nil_appendI mr_is_C rotate_simps)
done

lemma CConcStartA[rule_format,simp]: " x \<in> dom (Cp a) \<longrightarrow>
                                      x \<in> dom (Cp (list2FWpolicy (a # b)))"
(is "?P b")
apply (rule_tac P = ?P in list2FWpolicy.induct)  
apply (simp_all add: Cp.simps)
done


lemma domConc: "\<lbrakk>x \<in> dom (Cp (list2FWpolicy b)); b \<noteq> []\<rbrakk>  \<Longrightarrow>
                x \<in> dom (Cp (list2FWpolicy (a@b)))"
by (auto simp: PLemmas)

lemma CeqStart[rule_format,simp]:
                          "x \<notin> dom (Cp (list2FWpolicy a)) \<longrightarrow> a \<noteq> [] \<longrightarrow> b \<noteq> [] \<longrightarrow>
                           Cp (list2FWpolicy (b@a)) x = (Cp (list2FWpolicy b)) x"
apply (rule list2FWpolicy.induct,simp_all)
apply (auto simp: list2FWpolicyconc PLemmas)
done 

lemma C_eq_if_mr_eq2: "\<lbrakk>applied_rule_rev Cp x a = Some r; applied_rule_rev Cp x b = Some r;
                       a \<noteq> []; b \<noteq> []\<rbrakk>  \<Longrightarrow>
                       (Cp (list2FWpolicy a)) x = (Cp (list2FWpolicy b)) x"
by (metis mr_is_C)

lemma nMRtoNone[rule_format]: "p \<noteq> [] \<longrightarrow> applied_rule_rev Cp x p = None \<longrightarrow>
                               Cp (list2FWpolicy p) x = None"
apply (rule rev_induct, simp_all)
apply (case_tac "xs = []", simp_all)
by (simp_all add: applied_rule_rev_def dom_def)

lemma C_eq_if_mr_eq: 
 "\<lbrakk>applied_rule_rev Cp x b = applied_rule_rev Cp x a; a \<noteq> []; b \<noteq> []\<rbrakk> \<Longrightarrow>  
  (Cp (list2FWpolicy a)) x = (Cp (list2FWpolicy b)) x"
apply (cases "applied_rule_rev Cp x a = None") 
apply simp_all
apply (subst nMRtoNone)
apply (simp_all)
apply (subst nMRtoNone)
apply simp_all
by (auto intro: C_eq_if_mr_eq2)



lemma notmatching_notdom: "applied_rule_rev Cp x (p@[a]) \<noteq> Some a \<Longrightarrow> x \<notin> dom (Cp a)"
by (simp add: applied_rule_rev_def split: if_splits)

lemma foo3a[rule_format]: "applied_rule_rev Cp x (a@[b]@c) = Some b \<longrightarrow>  r \<in> set c \<longrightarrow> 
                           b \<notin> set c \<longrightarrow> x \<notin> dom (Cp r)"
apply (rule rev_induct) 
apply simp_all
apply (rule impI|rule conjI|simp)+
apply (rule_tac p = "a @ b # xs" in notmatching_notdom,simp_all)
apply (rule impI,simp)+
apply (drule sym,drule mp, simp_all)
apply (rule MRList3[symmetric],drule sym)
apply (rule_tac p = "a @ b # xs" in notmatching_notdom,simp_all)
done

lemma foo3D: "\<lbrakk>wellformed_policy1 p; p = (DenyAll#ps); 
              applied_rule_rev Cp x p = Some DenyAll; r \<in> set ps\<rbrakk> \<Longrightarrow> x \<notin> dom (Cp r)"
by (rule_tac a = "[]" and b = "DenyAll" and c = "ps"  in foo3a, simp_all)

lemma foo4[rule_format]: "set p = set s \<and> (\<forall> r. r \<in> set p \<longrightarrow> x \<notin> dom (Cp r)) \<longrightarrow> 
                          (\<forall> r .r \<in> set s \<longrightarrow> x \<notin> dom (Cp r))"
by simp

lemma foo5b[rule_format]: "x \<in> dom (Cp b) \<longrightarrow> (\<forall> r. r \<in> set c \<longrightarrow> x \<notin> dom (Cp r))\<longrightarrow> 
                           applied_rule_rev Cp x (b#c) = Some b"
apply (simp add: applied_rule_rev_def)
apply (rule_tac xs = c in rev_induct, simp_all)  
done

lemma mr_first: "\<lbrakk>x \<in> dom (Cp b); (\<forall> r. r \<in> set c \<longrightarrow> x \<notin> dom (Cp r)); s = b#c\<rbrakk> \<Longrightarrow> 
                           applied_rule_rev Cp x s = Some b"
by (simp add: foo5b)

lemma mr_charn[rule_format]: "a \<in> set p \<longrightarrow> (x \<in> dom (Cp a)) \<longrightarrow> 
                             (\<forall> r. r \<in> set p \<and> x \<in> dom (Cp r) \<longrightarrow> r = a) \<longrightarrow>  
                              applied_rule_rev Cp x p = Some a"
apply (rule_tac xs = p in rev_induct) 
by (simp_all add: applied_rule_rev_def)

lemma foo8: "\<lbrakk>(\<forall> r. r \<in> set p \<and> x \<in> dom (Cp r) \<longrightarrow> r = a); set p = set s\<rbrakk> \<Longrightarrow> 
             (\<forall> r. r \<in> set s \<and> x \<in> dom (Cp r) \<longrightarrow> r = a)"
by auto

lemma mrConcEnd[rule_format]: "applied_rule_rev Cp x (b # p) = Some a \<longrightarrow> a \<noteq> b \<longrightarrow> 
                               applied_rule_rev Cp x p = Some a"
apply (simp add: applied_rule_rev_def)
apply (rule_tac xs = p in rev_induct,simp_all) 
by auto


lemma wp3tl[rule_format]: "wellformed_policy3Pr p \<longrightarrow> wellformed_policy3Pr (tl p)"
by (induct p, simp_all, case_tac a, simp_all)

lemma wp3Conc[rule_format]: "wellformed_policy3Pr (a#p) \<longrightarrow> wellformed_policy3Pr p"
by (induct p, simp_all, case_tac a, simp_all)


lemma foo98[rule_format]:"applied_rule_rev Cp x (aa # p) = Some a \<longrightarrow> x \<in> dom (Cp r) \<longrightarrow> 
            r \<in> set p
           \<longrightarrow> a \<in> set p"
apply (simp add: applied_rule_rev_def)
apply (rule rev_induct)
apply simp_all
apply (case_tac "r = xa", simp_all)
done


lemma mrMTNone[simp]: "applied_rule_rev Cp x [] = None"
by (simp add: applied_rule_rev_def)

lemma DAAux[simp]: "x \<in> dom (Cp DenyAll)"
by (simp add: dom_def PolicyCombinators.PolicyCombinators Cp.simps)

lemma mrSet[rule_format]: "applied_rule_rev Cp x p = Some r \<longrightarrow> r \<in> set p"
apply (simp add: applied_rule_rev_def)
apply (rule_tac xs=p in rev_induct)
apply simp_all
done

lemma mr_not_Conc: "singleCombinators p \<Longrightarrow> applied_rule_rev Cp x p \<noteq> Some (a\<oplus>b)"
apply (auto simp:  mrSet)
apply (drule mrSet)
apply (erule SCnotConc,simp)
done


lemma foo25[rule_format]: "wellformed_policy3Pr (p@[x]) \<longrightarrow> wellformed_policy3Pr p"
by (induct p, simp_all, case_tac a, simp_all)

lemma mr_in_dom[rule_format]: "applied_rule_rev Cp x p = Some a \<longrightarrow> x \<in> dom (Cp a)"
apply (rule_tac xs = p in rev_induct)
by (auto simp: applied_rule_rev_def)


lemma wp3EndMT[rule_format]: "wellformed_policy3Pr (p@[xs]) \<longrightarrow> 
                             AllowPortFromTo a b po \<in> set p \<longrightarrow>
                             dom (Cp (AllowPortFromTo a b po)) \<inter> dom (Cp xs) = {}"
apply (induct p,simp_all)
apply (rule impI)+
apply (drule mp)
apply (erule wp3Conc)
by clarify auto

lemma foo29: "\<lbrakk>dom (Cp a) \<noteq> {}; dom (Cp a) \<inter> dom (Cp b) = {}\<rbrakk> \<Longrightarrow> a \<noteq> b"
by auto

lemma foo28:  "\<lbrakk>AllowPortFromTo a b po \<in> set p; 
               dom (Cp (AllowPortFromTo a b po)) \<noteq> {}; (wellformed_policy3Pr (p@[x]))\<rbrakk> 
               \<Longrightarrow> x \<noteq> AllowPortFromTo a b po"
by (metis foo29 Cp.simps(3) wp3EndMT)

lemma foo28a[rule_format]: "x \<in> dom (Cp a) \<Longrightarrow> dom (Cp a) \<noteq> {}"
by auto

lemma allow_deny_dom[simp]: "dom (Cp (AllowPortFromTo a b po)) \<subseteq> 
                            dom (Cp (DenyAllFromTo a b))"
by (simp_all add: twoNetsDistinct_def netsDistinct_def PLemmas) auto 

lemma DenyAllowDisj: "dom (Cp (AllowPortFromTo a b p)) \<noteq> {} \<Longrightarrow> 
             dom (Cp (DenyAllFromTo a b)) \<inter> dom (Cp (AllowPortFromTo a b p))  \<noteq> {}"
by (metis Int_absorb1 allow_deny_dom)

lemma foo31: "\<lbrakk>(\<forall> r. r \<in> set p \<and> x \<in> dom (Cp r) \<longrightarrow> 
           (r = AllowPortFromTo a b po \<or> r = DenyAllFromTo a b \<or> r = DenyAll)); 
            set p = set s\<rbrakk> \<Longrightarrow> 
            (\<forall> r. r \<in> set s \<and> x \<in> dom (Cp r) \<longrightarrow> 
           (r = AllowPortFromTo a b po \<or> r = DenyAllFromTo a b \<or> r = DenyAll))"
by auto



lemma wp1_auxa: "wellformed_policy1_strong p\<Longrightarrow>(\<exists> r. applied_rule_rev Cp x p = Some r)"
apply (rule DAimpliesMR_E)
by (erule wp1_aux1aa)



lemma deny_dom[simp]:  "twoNetsDistinct a b c d \<Longrightarrow> dom (Cp (DenyAllFromTo a b)) \<inter>
                       dom (Cp (DenyAllFromTo c d)) = {}"
apply (simp add: Cp.simps)
by (erule aux6)

lemma domTrans: "\<lbrakk>dom a \<subseteq> dom b; dom(b) \<inter> dom (c) = {}\<rbrakk> \<Longrightarrow> dom(a) \<inter> dom(c) = {}" 
by auto

lemma DomInterAllowsMT: "\<lbrakk> twoNetsDistinct a b c d\<rbrakk>  \<Longrightarrow>
        dom (Cp (AllowPortFromTo a b p)) \<inter> dom (Cp (AllowPortFromTo c d po)) = {}"
apply (case_tac "p = po", simp_all)
apply (rule_tac b = "Cp (DenyAllFromTo a b)" in domTrans, simp_all)
apply (metis domComm aux26 tNDComm)
apply (simp add: twoNetsDistinct_def netsDistinct_def PLemmas) 
by (auto simp: prod_eqI)


lemma DomInterAllowsMT_Ports: "\<lbrakk>p \<noteq> po\<rbrakk> \<Longrightarrow>
        dom (Cp (AllowPortFromTo a b p)) \<inter> dom (Cp (AllowPortFromTo c d po)) = {}"
apply (simp add: twoNetsDistinct_def netsDistinct_def PLemmas)
by (auto simp: prod_eqI)


lemma wellformed_policy3_charn[rule_format]: 
  "singleCombinators p \<longrightarrow> distinct p \<longrightarrow> allNetsDistinct p \<longrightarrow> 
   wellformed_policy1 p \<longrightarrow> wellformed_policy2Pr p  \<longrightarrow> wellformed_policy3Pr p"
apply (induct_tac p)   
apply simp_all
apply clarify
apply simp_all
apply (auto intro: singleCombinatorsConc ANDConc waux2 wp2Conc)
apply (case_tac a) 
apply simp_all
apply clarify
apply (case_tac r)
apply simp_all
apply (metis Int_commute) 
apply (metis DomInterAllowsMT aux7aa DomInterAllowsMT_Ports)
apply (metis aux0_0 mem_def)
done



lemma DistinctNetsDenyAllow: 
 "\<lbrakk>DenyAllFromTo b c \<in> set p; AllowPortFromTo a d po \<in> set p; allNetsDistinct p;
  dom (Cp (DenyAllFromTo b c)) \<inter> dom (Cp (AllowPortFromTo a d po)) \<noteq> {}\<rbrakk>
       \<Longrightarrow> b = a \<and> c = d" 
apply (simp add: allNetsDistinct_def)
apply (frule_tac x = "b" in spec) 
apply (drule_tac x = "d" in spec)
apply (drule_tac x = "a" in spec)
apply (drule_tac x = "c" in spec)
apply (metis Int_commute ND0aux1 ND0aux3 NDComm aux26 twoNetsDistinct_def 
             ND0aux2 ND0aux4)
done

lemma DistinctNetsAllowAllow: 
 "\<lbrakk>AllowPortFromTo b c poo \<in> set p; AllowPortFromTo a d po \<in> set p; 
  allNetsDistinct p; dom (Cp (AllowPortFromTo b c poo)) \<inter> 
  dom (Cp (AllowPortFromTo a d po)) \<noteq> {}\<rbrakk>
       \<Longrightarrow> b = a \<and> c = d \<and> poo = po" 
apply (simp add: allNetsDistinct_def)
apply (frule_tac x = "b" in spec) 
apply (drule_tac x = "d" in spec)
apply (drule_tac x = "a" in spec)
apply (drule_tac x = "c" in spec)
apply (metis DomInterAllowsMT DomInterAllowsMT_Ports ND0aux3 ND0aux4 NDComm 
             twoNetsDistinct_def)
done



lemma WP2RS2[simp]: 
  "\<lbrakk>singleCombinators p;
 distinct p;
 allNetsDistinct p\<rbrakk> \<Longrightarrow> 
   wellformed_policy2Pr (removeShadowRules2 p)"
 proof (induct p)
   case Nil thus ?case by simp
 next   
   case (Cons x xs) 
      have wp_xs: "wellformed_policy2Pr (removeShadowRules2 xs)" using assms
by (metis Cons ANDConc distinct.simps(2) singleCombinatorsConc)
      show ?case
      proof (cases x) 
         case DenyAll thus ?thesis using wp_xs by simp
      next  
         case (DenyAllFromTo a b) thus ?thesis
           using wp_xs Cons  
          by (simp,metis DenyAllFromTo aux aux7 tNDComm deny_dom)
      next
         case (AllowPortFromTo a b p) thus ?thesis
             using assms wp_xs 
         by (simp, metis aux26 AllowPortFromTo Cons(4) aux aux7a tNDComm)
      next
         case (Conc a b) thus ?thesis
             using assms by (metis Conc Cons(2) singleCombinators.simps(2)) 
      qed
 qed



lemma AD_aux: "\<lbrakk>AllowPortFromTo a b po \<in> set p ;DenyAllFromTo c d \<in> set p; 
                allNetsDistinct  p ; singleCombinators p; 
        a \<noteq> c \<or> b \<noteq> d\<rbrakk>
    \<Longrightarrow> dom (Cp (AllowPortFromTo a b po)) \<inter> dom (Cp (DenyAllFromTo c d)) = {}"
apply (rule aux26)
apply (rule_tac x ="AllowPortFromTo a b po" and y = "DenyAllFromTo c d" in tND)
apply auto 
done



lemma sorted_WP2[rule_format]: "sorted p l \<longrightarrow> all_in_list p l \<longrightarrow> distinct p \<longrightarrow> 
            allNetsDistinct p \<longrightarrow> singleCombinators p \<longrightarrow> wellformed_policy2Pr p"
proof (induct p)
 case Nil thus ?case by simp 
next
 case (Cons a p) thus ?case
  proof (cases a)
   case DenyAll thus ?thesis using assms Cons
               by (auto intro: ANDConc singleCombinatorsConc sortedConcEnd)
  next
   case (DenyAllFromTo c d) thus ?thesis using assms Cons
     apply simp
     apply (rule impI)+
     apply (rule conjI)
     apply (rule allI)+
     apply (rule impI)+
     apply (rule deny_dom)
     apply (auto intro: aux7 tNDComm ANDConc singleCombinatorsConc sortedConcEnd) 
     done
   next
   case (AllowPortFromTo c d e) thus ?thesis using Cons assms
      apply simp
      apply (rule impI|rule conjI|rule allI)+
      apply (rule aux26)
      apply (rule_tac x = "AllowPortFromTo c d e" and
                      y = "DenyAllFromTo aa b" in tND)
      apply (assumption,simp_all)
      apply (subgoal_tac "smaller (AllowPortFromTo c d e) (DenyAllFromTo aa b) l") 	
      apply (simp split: if_splits)
      apply metis
      apply (erule sorted_is_smaller)
      apply simp_all
      apply (metis bothNet.simps(2) in_list.simps(2)
                   in_set_in_list)
      by (auto intro: aux7 tNDComm ANDConc singleCombinatorsConc sortedConcEnd)
   next
   case (Conc a b) thus ?thesis using Cons by simp 
   qed
qed




lemma wellformed2_sorted[simp]: "\<lbrakk>all_in_list p l; distinct p; allNetsDistinct p;
                         singleCombinators p\<rbrakk> \<Longrightarrow> wellformed_policy2Pr (sort p l)"
apply (rule sorted_WP2)
apply (erule sort_is_sorted, simp_all)
apply (erule all_in_listSubset)
apply (auto intro: SC3 singleCombinatorsConc sorted_insort)
done


lemma wellformed2_sortedQ[simp]: "\<lbrakk>all_in_list p l; distinct p; allNetsDistinct p;
                         singleCombinators p\<rbrakk> \<Longrightarrow> wellformed_policy2Pr (qsort p l)"
apply (rule sorted_WP2)
apply (erule sort_is_sortedQ, simp_all)
apply (erule all_in_listSubset)
apply (auto intro: SC3Q singleCombinatorsConc distinct_sortQ)
done


lemma C_DenyAll[simp]: "Cp (list2FWpolicy (xs @ [DenyAll])) x = Some (deny ())"
by (auto simp: PLemmas)


lemma C_eq_RS1n:
   "Cp(list2FWpolicy (removeShadowRules1_alternative p)) = Cp(list2FWpolicy p)"
apply (case_tac "p = []")
apply simp_all 
apply (metis list2FWpolicy.simps(1) rSR1_eq removeShadowRules1.simps(2))
apply (rule rev_induct)
apply (metis rSR1_eq removeShadowRules1.simps(2))
apply (case_tac "xs = []", simp_all)
apply (simp add: removeShadowRules1_alternative_def)
apply (case_tac x, simp_all)
apply (rule ext) 
apply (case_tac "x = DenyAll")
apply (simp_all add: PLemmas)
apply (rule_tac t = "removeShadowRules1_alternative (xs @ [x])" and
               s = "(removeShadowRules1_alternative xs)@[x]" in subst)
apply (erule RS1n_assoc)
apply (case_tac "xa \<in> dom (Cp x)")
apply simp_all
done


lemma C_eq_RS1[simp]: "p \<noteq> [] \<Longrightarrow>
                       Cp(list2FWpolicy (removeShadowRules1 p)) = Cp(list2FWpolicy p)"
by (metis rSR1_eq C_eq_RS1n)


lemma EX_MR_aux[rule_format]: "applied_rule_rev Cp x (DenyAll # p) \<noteq> Some DenyAll \<longrightarrow>
                               (\<exists>y. applied_rule_rev Cp x p = Some y)"
apply (simp add: applied_rule_rev_def)
apply (rule_tac xs = p in rev_induct, simp_all)  
done

lemma EX_MR : "\<lbrakk>applied_rule_rev Cp x p \<noteq> (Some DenyAll); p = DenyAll#ps\<rbrakk> \<Longrightarrow> 
                           (applied_rule_rev Cp x p = applied_rule_rev Cp x ps)"
apply auto

apply (subgoal_tac "applied_rule_rev Cp x (DenyAll#ps) \<noteq> None")
apply auto
apply (metis mrConcEnd)
by (metis DAimpliesMR_E hd.simps hd_in_set list.simps(3) not_Some_eq)



lemma mr_not_DA:
  "\<lbrakk>wellformed_policy1_strong s; applied_rule_rev Cp x p = Some (DenyAllFromTo a ab);
    set p = set s\<rbrakk> \<Longrightarrow> applied_rule_rev Cp x s \<noteq> Some DenyAll"
apply (subst wp1n_tl, simp_all)
apply (subgoal_tac "x \<in> dom (Cp (DenyAllFromTo a ab))")
apply (subgoal_tac "DenyAllFromTo a ab \<in> set (tl s)")
apply (metis wp1n_tl foo98 wellformed_policy1_strong.simps(2))
apply (erule r_not_DA_in_tl, simp_all)
apply (subgoal_tac "DenyAllFromTo a ab \<in> set p", simp)
apply (erule mrSet)
apply (erule mr_in_dom)
done

lemma domsMT_notND_DD: 
   "\<lbrakk>dom (Cp (DenyAllFromTo a b)) \<inter> dom (Cp (DenyAllFromTo c d)) \<noteq> {}\<rbrakk> \<Longrightarrow>
     \<not> netsDistinct a c"
apply (erule contrapos_nn)
apply (simp add: Cp.simps)
apply (rule aux6)
apply (simp add: twoNetsDistinct_def)
done


lemma domsMT_notND_DD2: 
  "\<lbrakk>dom (Cp (DenyAllFromTo a b)) \<inter> dom (Cp (DenyAllFromTo c d)) \<noteq> {}\<rbrakk> \<Longrightarrow>
   \<not> netsDistinct b d"
apply (erule contrapos_nn)
apply (simp add: Cp.simps)
apply (rule aux6)
apply (simp add: twoNetsDistinct_def)
done

lemma domsMT_notND_DD3: 
  "\<lbrakk>x \<in> dom (Cp (DenyAllFromTo a b)); x \<in> dom (Cp (DenyAllFromTo c d))\<rbrakk> \<Longrightarrow>
   \<not> netsDistinct a c"
apply (rule domsMT_notND_DD)
apply auto
done

lemma domsMT_notND_DD4: 
  "\<lbrakk>x \<in> dom (Cp (DenyAllFromTo a b)); x \<in> dom (Cp (DenyAllFromTo c d))\<rbrakk> \<Longrightarrow>
   \<not> netsDistinct b d"
apply (rule domsMT_notND_DD2)
apply auto
done

lemma NetsEq_if_sameP_DD: 
  "\<lbrakk>allNetsDistinct p; u\<in> set p; v\<in> set p; u = (DenyAllFromTo a b); 
   v = (DenyAllFromTo c d); x \<in> dom (Cp (u)); x \<in> dom (Cp (v))\<rbrakk> \<Longrightarrow> 
    a = c \<and> b = d"
apply (simp add: allNetsDistinct_def)
apply (metis ND0aux1 ND0aux2 domsMT_notND_DD3 domsMT_notND_DD4 mem_def)
done


lemma rule_charn1: 
 assumes aND: "allNetsDistinct p"
 and mr_is_allow: "applied_rule_rev Cp x p = Some (AllowPortFromTo a b po)"
 and SC: "singleCombinators p"
 and inp: "r \<in> set p" 
 and inDom: "x \<in> dom (Cp r)"
 shows "(r = AllowPortFromTo a b po \<or> r = DenyAllFromTo a b \<or> r = DenyAll)"
proof (cases r) 
  case DenyAll show ?thesis  by (metis DenyAll)
next
  case (DenyAllFromTo x y) show ?thesis using Cons assms DenyAllFromTo
     apply (simp,rule_tac  p = p and po =po in DistinctNetsDenyAllow, simp_all)
     apply (metis mrSet)
     by (metis Int_iff mr_in_dom inSet_not_MT set_empty2)
next
  case (AllowPortFromTo x y b) show ?thesis using assms AllowPortFromTo Cons
     apply (simp) 
     apply( rule DistinctNetsAllowAllow)
     apply  (simp_all)
     apply (metis mrSet)
     by (metis Int_iff mr_in_dom inSet_not_MT set_empty2)
next
  case (Conc x y) thus ?thesis using assms by (metis aux0_0)
qed




lemma noneMTsubset[rule_format]: "noneMT Cp a \<longrightarrow> set b \<subseteq> set a \<longrightarrow> noneMT Cp b"
apply (induct b,simp_all)
by (metis notMTnMT)


lemma nMTSort: "noneMT Cp p \<Longrightarrow> noneMT Cp (sort p l)"
by (metis set_sort nMTeqSet)



lemma nMTSortQ: "noneMT Cp p \<Longrightarrow> noneMT Cp (qsort p l)"
by (metis set_sortQ nMTeqSet)

lemma wp3char[rule_format]: "noneMT Cp xs \<and>  Cp (AllowPortFromTo a b po) = empty \<and> 
                            wellformed_policy3Pr (xs @ [DenyAllFromTo a b]) \<longrightarrow> 
                             AllowPortFromTo a b po \<notin> set xs"
apply (induct xs)
apply simp_all
by (metis domNMT wp3Conc)



lemma wp3charn[rule_format]: 
assumes domAllow: "dom (Cp (AllowPortFromTo a b po)) \<noteq> {}" 
and wp3: "wellformed_policy3Pr (xs @ [DenyAllFromTo a b])"
shows allowNotInList: "AllowPortFromTo a b po \<notin> set xs"
apply (insert assms)
proof (induct xs)
 case Nil show ?case by simp
next
 case (Cons x xs) show ?case using Cons
 by (simp,auto intro: wp3Conc) (auto simp: DenyAllowDisj domAllow)
qed


lemma rule_charn2: 
 assumes aND: "allNetsDistinct p"
 and wp1: "wellformed_policy1 p"
 and SC: "singleCombinators p"
 and wp3: "wellformed_policy3Pr p"
 and allow_in_list: "AllowPortFromTo c d po \<in> set p"
 and x_in_dom_allow: "x \<in> dom (Cp (AllowPortFromTo c d po))"
 shows  "applied_rule_rev Cp x p = Some (AllowPortFromTo c d po)"
 proof  (insert assms, induct p rule: rev_induct)    
   case Nil show ?case using Nil by simp
 next 
   case (snoc y ys) show ?case using snoc
     apply simp
     apply (case_tac "y = (AllowPortFromTo c d po)")
     apply (simp add: applied_rule_rev_def)
     apply simp_all
     apply (subgoal_tac "ys \<noteq> []")
     apply (subgoal_tac "applied_rule_rev Cp x ys = Some (AllowPortFromTo c d po)")
     defer 1
     apply (metis ANDConcEnd SCConcEnd WP1ConcEnd foo25) 
     apply (metis inSet_not_MT)
     proof (cases y)
      case DenyAll thus ?thesis using DenyAll snoc    
            apply simp
            by (metis DAnotTL DenyAll inSet_not_MT policy2list.simps(2))
       next
      case (DenyAllFromTo a b) thus ?thesis using snoc apply simp
     apply (simp_all add: applied_rule_rev_def)
     apply (rule conjI)
     apply (metis domInterMT  wp3EndMT)
     apply (rule impI)
     by (metis ANDConcEnd DenyAllFromTo SCConcEnd WP1ConcEnd foo25)
       next
     case (AllowPortFromTo a1 a2 b) thus ?thesis using AllowPortFromTo snoc apply simp
     apply (simp_all add: applied_rule_rev_def)
     apply (rule conjI)
     apply (metis domInterMT  wp3EndMT)
 by (metis ANDConcEnd AllowPortFromTo SCConcEnd WP1ConcEnd foo25 x_in_dom_allow)
     next
     case (Conc a b) thus ?thesis using Conc snoc apply simp
        by (metis Conc aux0_0 in_set_conv_decomp) 
     qed
qed

lemma rule_charn3: 
 "\<lbrakk>wellformed_policy1 p; allNetsDistinct p; singleCombinators p; 
  wellformed_policy3Pr p; applied_rule_rev Cp x p = Some (DenyAllFromTo c d); 
  AllowPortFromTo a b po \<in> set p\<rbrakk> \<Longrightarrow> x \<notin> dom (Cp (AllowPortFromTo a b po))"
by (clarify, auto simp: rule_charn2 dom_def) 

lemma rule_charn4: 
assumes wp1: "wellformed_policy1 p" 
and aND: "allNetsDistinct p" 
and SC: "singleCombinators p"
and wp3: "wellformed_policy3Pr p"  
and DA: "DenyAll \<notin> set p" 
and mr: "applied_rule_rev Cp x p = Some (DenyAllFromTo a b)"
and rinp: "r \<in> set p"
and xindom: "x \<in> dom (Cp r)"
shows "r = DenyAllFromTo a b"
proof (cases r)
  case DenyAll thus ?thesis using DenyAll assms by simp 
next
  case (DenyAllFromTo c d) thus ?thesis using assms apply simp
    apply (erule_tac x = x and p = p and v = "(DenyAllFromTo a b)" and
                     u = "(DenyAllFromTo c d)"  in NetsEq_if_sameP_DD)
    apply simp_all
    apply (erule mrSet)
    by (erule mr_in_dom) 
next
  case (AllowPortFromTo c d e) thus ?thesis using assms apply simp
    apply (subgoal_tac "x \<notin> dom (Cp  (AllowPortFromTo c d e))")
    apply simp 
    apply (rule_tac p = p in rule_charn3)
    by (auto intro: SCnotConc) 
next
  case (Conc a b) thus ?thesis using assms apply simp 
      by (metis Conc aux0_0) 
qed



lemma foo31a: "\<lbrakk>(\<forall> r. r \<in> set p \<and> x \<in> dom (Cp r) \<longrightarrow>
           (r = AllowPortFromTo a b po \<or> r = DenyAllFromTo a b \<or> r = DenyAll));
           set p = set s; r \<in> set s ; x \<in> dom (Cp r)\<rbrakk> \<Longrightarrow>
           (r = AllowPortFromTo a b po \<or> r = DenyAllFromTo a b \<or> r = DenyAll)"
by auto


lemma aux4[rule_format]: 
  "applied_rule_rev Cp x (a#p) = Some a \<longrightarrow> a \<notin> set (p) \<longrightarrow> applied_rule_rev Cp x p = None"
apply (rule rev_induct)
apply simp_all
apply (rule impI)+
apply simp
apply (simp add: applied_rule_rev_def)
apply (simp split: if_splits)
done

lemma mrDA_tl: 
 assumes mr_DA: "applied_rule_rev Cp x p = Some DenyAll"
 and wp1n: "wellformed_policy1_strong p"
 shows "applied_rule_rev Cp x (tl p) = None"
 apply (rule aux4 [where a = DenyAll])
 apply (metis wp1n_tl mr_DA wp1n)
 by (metis WP1n_DA_notinSet wp1n)

lemma rule_charnDAFT: 
"\<lbrakk>wellformed_policy1_strong p; allNetsDistinct p; singleCombinators p; 
  wellformed_policy3Pr p; applied_rule_rev Cp x p = Some (DenyAllFromTo a b);
  r \<in> set (tl p); x \<in> dom (Cp r)\<rbrakk> \<Longrightarrow> r = DenyAllFromTo a b"
apply (subgoal_tac "p = DenyAll#(tl p)")
apply (rule_tac p = "tl p" in rule_charn4)
apply simp_all
apply (metis wellformed_policy1_strong.simps(1) wp1_eq wp1_tl)
apply (erule AND_tl)
apply (erule SC_tl)
apply (erule wp3tl)
apply (erule WP1n_DA_notinSet)
apply (metis Combinators.simps(4) EX_MR option.inject)
apply (metis  wp1n_tl) 
done

lemma mrDenyAll_is_unique: 
  "\<lbrakk>wellformed_policy1_strong p; applied_rule_rev Cp x p = Some DenyAll;
   r \<in> set (tl p)\<rbrakk> \<Longrightarrow> x \<notin> dom (Cp r)"
apply (rule_tac a = "[]" and b = "DenyAll" and c = "tl p"  in foo3a, simp_all)  
apply (metis wp1n_tl)
by (metis WP1n_DA_notinSet)


theorem  C_eq_Sets_mr: 
 assumes sets_eq: "set p = set s"
 and SC: "singleCombinators p"
 and wp1_p: "wellformed_policy1_strong p"
 and wp1_s: "wellformed_policy1_strong s"
 and wp3_p: "wellformed_policy3Pr p"       
 and wp3_s: "wellformed_policy3Pr s"  
 and aND: "allNetsDistinct p"        
   
 shows "applied_rule_rev Cp x p = applied_rule_rev Cp x s"
proof (cases "applied_rule_rev Cp x p")

 case None 
     have DA: "DenyAll \<in> set p" using wp1_p by (auto simp: wp1_aux1aa)
     have notDA: "DenyAll \<notin> set p" using None by (auto simp: DAimplieMR) 
     thus ?thesis using DA by (contradiction)
 next
 
 case (Some y) thus ?thesis
   proof (cases y) 
      have tl_p: "p = DenyAll#(tl p)" by (metis wp1_p wp1n_tl) 
      have tl_s: "s = DenyAll#(tl s)" by (metis wp1_s wp1n_tl)
      have tl_eq: "set (tl p) = set (tl s)"
       by (metis tl.simps(2) WP1n_DA_notinSet sets_eq foo2
                 wellformed_policy1_charn wp1_aux1aa wp1_eq wp1_p wp1_s)
    {
   case DenyAll
     have mr_p_is_DenyAll: "applied_rule_rev Cp x p = Some DenyAll"
        by (simp add: DenyAll Some)
     hence x_notin_tl_p: "\<forall> r. r \<in> set (tl p) \<longrightarrow>  x \<notin> dom (Cp r)" using wp1_p
        by (auto simp: mrDenyAll_is_unique)
     hence x_notin_tl_s: "\<forall> r. r \<in> set (tl s) \<longrightarrow>  x \<notin> dom (Cp r)" using tl_eq
        by auto
     hence mr_s_is_DenyAll: "applied_rule_rev Cp x s = Some DenyAll" using tl_s
        by (auto simp: mr_first)
     thus ?thesis using mr_p_is_DenyAll by simp
}
{   case (DenyAllFromTo a b) 
     have mr_p_is_DAFT: "applied_rule_rev Cp x p = Some (DenyAllFromTo a b)"
        by (simp add: DenyAllFromTo Some)
     have DA_notin_tl: "DenyAll \<notin> set (tl p)"
        by (metis WP1n_DA_notinSet wp1_p)
     have mr_tl_p: "applied_rule_rev Cp x p = applied_rule_rev Cp x (tl p)"
        by (metis Combinators.simps(4) DenyAllFromTo Some mrConcEnd tl_p)
     have dom_tl_p: "\<And> r. r \<in> set (tl p) \<and> x \<in> dom (Cp r) \<Longrightarrow> 
                     r = (DenyAllFromTo a b)"
       using wp1_p aND SC wp3_p mr_p_is_DAFT
       by (auto simp: rule_charnDAFT)
     hence dom_tl_s: "\<And> r. r \<in> set (tl s) \<and> x \<in> dom (Cp r) \<Longrightarrow> 
                      r = (DenyAllFromTo a b)"
       using tl_eq  by auto
     have DAFT_in_tl_s: "DenyAllFromTo a b \<in> set (tl s)" using mr_tl_p
       by (metis DenyAllFromTo mrSet mr_p_is_DAFT tl_eq)
     have x_in_dom_DAFT: "x \<in> dom (Cp (DenyAllFromTo a b))"
       by (metis mr_p_is_DAFT DenyAllFromTo mr_in_dom)
     hence mr_tl_s_is_DAFT: "applied_rule_rev Cp x (tl s) = Some (DenyAllFromTo a b)"
       using DAFT_in_tl_s dom_tl_s  by (auto simp: mr_charn)
     hence mr_s_is_DAFT: "applied_rule_rev Cp x s = Some (DenyAllFromTo a b)"
       using tl_s
          by (metis  DA_notin_tl DenyAllFromTo EX_MR mrDA_tl 
                     not_Some_eq tl_eq wellformed_policy1_strong.simps(2))
     thus ?thesis using mr_p_is_DAFT by simp
}
{  case (AllowPortFromTo a b c)
     have wp1s: "wellformed_policy1 s" by (metis wp1_eq wp1_s)
     have mr_p_is_A: "applied_rule_rev Cp x p = Some (AllowPortFromTo a b c)"
       by (simp add: AllowPortFromTo Some)
     hence A_in_s: "AllowPortFromTo a b c \<in> set s" using sets_eq
       by (auto intro: mrSet)
     have x_in_dom_A: "x \<in> dom (Cp (AllowPortFromTo a b c))"
       by (metis mr_p_is_A AllowPortFromTo mr_in_dom)
     have SCs: "singleCombinators s" using SC sets_eq
       by (auto intro: SCSubset)
     hence ANDs: "allNetsDistinct s" using aND sets_eq SC
       by (auto intro: aNDSetsEq)
     hence mr_s_is_A: "applied_rule_rev Cp x s = Some (AllowPortFromTo a b c)"
       using A_in_s wp1s mr_p_is_A aND SCs wp3_s x_in_dom_A
          by (simp add: rule_charn2)
     thus ?thesis using mr_p_is_A by simp
}
    case (Conc a b) thus ?thesis by (metis Some mr_not_Conc SC) 
 qed
qed

lemma C_eq_Sets: 
"\<lbrakk>singleCombinators p; wellformed_policy1_strong p; wellformed_policy1_strong s;
wellformed_policy3Pr p; wellformed_policy3Pr s; allNetsDistinct p; set p = set s\<rbrakk> \<Longrightarrow>
  Cp (list2FWpolicy p) x  = Cp (list2FWpolicy s) x"
  apply (rule C_eq_if_mr_eq)
  apply (rule C_eq_Sets_mr [symmetric])
  apply simp_all
done


lemma C_eq_sorted: "\<lbrakk>distinct p; all_in_list p l; singleCombinators p;
        wellformed_policy1_strong p; wellformed_policy3Pr p; allNetsDistinct p\<rbrakk> \<Longrightarrow>
        Cp (list2FWpolicy (sort p l))= Cp (list2FWpolicy p)"
 apply (rule ext)
 apply (rule C_eq_Sets)
 apply (auto simp: nMTSort wellformed1_alternative_sorted 
                  wellformed_policy3_charn  wp1_eq)
done


lemma C_eq_sortedQ: "\<lbrakk>distinct p; all_in_list p l; singleCombinators p;
        wellformed_policy1_strong p; wellformed_policy3Pr p; allNetsDistinct p\<rbrakk> \<Longrightarrow>
        Cp (list2FWpolicy (qsort p l))= Cp (list2FWpolicy p)"
 apply (rule ext)
 apply (rule C_eq_Sets)
 apply (auto simp: nMTSortQ wellformed1_alternative_sorted distinct_sortQ
                   wellformed_policy3_charn wp1_eq)
by (metis mem_def member_rec(1) member_set wellformed1_sortedQ wellformed_eq wp1_eq
         set_qsort wp1n_tl)





lemma C_eq_RS2_mr: "applied_rule_rev Cp x (removeShadowRules2 p)= applied_rule_rev Cp x p"
proof (induct p)
 case Nil thus ?case by simp next
 case (Cons y ys) thus ?case
 proof (cases "ys = []")
  case True thus ?thesis by (cases y, simp_all) next
  case False thus ?thesis
   proof (cases y)
    case DenyAll thus ?thesis by (simp, metis Cons DenyAll mreq_end2) next
    case (DenyAllFromTo a b) thus ?thesis
        by (simp, metis Cons DenyAllFromTo mreq_end2)
     next
    case (AllowPortFromTo a b p) thus ?thesis
      proof (cases "DenyAllFromTo a b \<in> set ys")
        case True thus ?thesis using AllowPortFromTo Cons
          apply (cases "applied_rule_rev Cp x ys = None", simp_all)
          apply (subgoal_tac "x \<notin> dom (Cp (AllowPortFromTo a b p))")
          apply (subst mrconcNone, simp_all)
          apply (simp add: applied_rule_rev_def )
          apply (rule contra_subsetD [OF allow_deny_dom])
          apply (erule mrNoneMT,simp)
          apply (metis AllowPortFromTo mrconc)
          done
        next
        case False thus ?thesis using False Cons AllowPortFromTo
                     by (simp, metis AllowPortFromTo Cons mreq_end2) qed
    next
    case (Conc a b) thus ?thesis
                      by (metis Cons mreq_end2 removeShadowRules2.simps(4))
   qed
  qed
qed



lemma C_eq_None[rule_format]: "p \<noteq> []  --> applied_rule_rev Cp x p = None \<longrightarrow>
                               Cp (list2FWpolicy p) x = None"
apply (simp add: applied_rule_rev_def)
apply (rule rev_induct, simp_all)
apply (rule impI)+
apply simp
apply (case_tac "xs \<noteq> []")
apply (simp_all add: dom_def)
done

lemma C_eq_None2:
  "\<lbrakk>a \<noteq> []; b \<noteq> []; applied_rule_rev Cp x a = None; applied_rule_rev Cp x b = None\<rbrakk> \<Longrightarrow>
  (Cp (list2FWpolicy a)) x = (Cp (list2FWpolicy b)) x"
by (auto simp: C_eq_None)

lemma C_eq_RS2: "wellformed_policy1_strong p \<Longrightarrow> 
 Cp (list2FWpolicy (removeShadowRules2 p))= Cp (list2FWpolicy p)"
apply (rule ext)
apply (rule C_eq_if_mr_eq)
apply (rule C_eq_RS2_mr [symmetric], simp_all)
apply (metis wp1_alternative_not_mt wp1n_RS2)
done



lemma noneMTRS2: "noneMT Cp p \<Longrightarrow> noneMT Cp (removeShadowRules2 p)"
by (auto simp: RS2Set noneMTsubset)

lemma CconcNone: "\<lbrakk>dom (Cp a) = {}; p \<noteq> []\<rbrakk> \<Longrightarrow>
                  Cp (list2FWpolicy (a # p)) x = Cp (list2FWpolicy p) x"
apply (case_tac "p = []", simp_all)
apply (case_tac "x\<in> dom (Cp (list2FWpolicy(p)))")
apply (metis Cdom2 list2FWpolicyconc)
apply (metis Cp.simps(4) map_add_dom_app_simps(2) inSet_not_MT list2FWpolicyconc set_empty2)
done


lemma noneMTrd[rule_format]: "noneMT Cp p \<longrightarrow> noneMT Cp (remdups p)"
by (induct p, simp_all)

lemma DARS3[rule_format]:"DenyAll \<notin> set p\<longrightarrow>DenyAll \<notin> set (removeShadowRules3 Cp p)"
by (induct p, simp_all)

lemma DAnMT: "dom (Cp DenyAll) \<noteq> {}"
by (simp add: dom_def Cp.simps PolicyCombinators.PolicyCombinators)

lemma DAnMT2: "Cp DenyAll \<noteq> empty"
by (metis DAAux dom_eq_empty_conv empty_iff) 




lemma wp1n_RS3[rule_format,simp]: "wellformed_policy1_strong p \<longrightarrow>
                               wellformed_policy1_strong (removeShadowRules3 Cp p)"
apply (induct p, simp_all)
apply (rule conjI| rule impI|simp)+
apply (metis DAnMT)
apply (metis DARS3) 
done

lemma AILRS3[rule_format,simp]: "all_in_list p l \<longrightarrow>
                                 all_in_list (removeShadowRules3 Cp p) l"
by (induct p, simp_all)

lemma SCRS3[rule_format,simp]: "singleCombinators p \<longrightarrow>
                                singleCombinators(removeShadowRules3 Cp p)"
apply (induct p, simp_all)
apply (case_tac "a", simp_all)
done


lemma RS3subset: "set (removeShadowRules3 Cp p)  \<subseteq> set p "
by (induct p, auto)


lemma ANDRS3[simp]: "\<lbrakk>singleCombinators p; allNetsDistinct p\<rbrakk> \<Longrightarrow>
                     allNetsDistinct (removeShadowRules3 Cp p)"
apply (rule_tac b = p in aNDSubset)
apply simp_all
apply (rule RS3subset)
done

lemma nlpaux: "x \<notin> dom (Cp b) \<Longrightarrow> Cp (a \<oplus> b) x = Cp a x"
by (metis Cp.simps(4) map_add_dom_app_simps(3))

lemma notindom[rule_format]: "a \<in> set p \<longrightarrow>  x \<notin> dom (Cp (list2FWpolicy p)) \<longrightarrow>
                             x \<notin> dom (Cp a)"
apply (induct p)
apply simp_all
apply (rule conjI| rule impI)+
apply (metis CConcStartA)
apply (rule impI)+
apply simp
apply (metis Cdom2 List.set.simps(2) domIff insert_absorb list.simps(2) list2FWpolicyconc set_empty)
done

lemma C_eq_rd[rule_format]: "p \<noteq> [] \<Longrightarrow>
                            Cp (list2FWpolicy (remdups p)) = Cp (list2FWpolicy p)"
apply (rule ext)
proof (induct p) 
  case Nil thus ?case by simp next
  case (Cons y ys) thus ?case
    proof (cases "ys = []")
      case True thus ?thesis by simp next
      case False thus ?thesis using Cons apply simp
       apply (rule conjI, rule impI)
       apply (cases "x \<in> dom (Cp (list2FWpolicy ys))")
       apply (metis Cdom2 False list2FWpolicyconc)
       apply (metis False domIff list2FWpolicyconc nlpaux notindom)
       apply (rule impI)
       apply (cases "x \<in> dom (Cp (list2FWpolicy ys))")
       apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy (remdups ys)))")
       apply (metis Cdom2 False list2FWpolicyconc remdups_eq_nil_iff)
       apply (metis domIff)
       apply (subgoal_tac "x \<notin> dom (Cp (list2FWpolicy (remdups ys)))")
       apply (metis False list2FWpolicyconc nlpaux remdups_eq_nil_iff)
       apply (metis domIff)
       done
  qed
qed
   


lemma nMT_domMT: "\<lbrakk>\<not> notMTpolicy Cp  p; p \<noteq> []\<rbrakk> \<Longrightarrow> r \<notin> dom (Cp (list2FWpolicy p))"
 proof (induct p)
   case Nil thus ?case by simp next
   case (Cons x xs) thus ?case apply simp
      apply (simp split: if_splits)
      apply (cases "xs = []")
      apply (simp_all )
by (metis CconcNone domIff)
qed
            
lemma C_eq_RS3_aux[rule_format]: "notMTpolicy Cp  p \<Longrightarrow>
                  Cp (list2FWpolicy p) x = Cp (list2FWpolicy (removeShadowRules3 Cp p)) x"
proof (induct p)
 case Nil thus ?case by simp next
 case (Cons y ys) thus ?case
   proof (cases "notMTpolicy Cp  ys")
     case True thus ?thesis using Cons apply simp
     apply (rule conjI, rule impI, simp)
     apply (metis CconcNone True notMTpolicyimpnotMT)
     apply (rule impI, simp)
     apply (cases "x \<in> dom (Cp (list2FWpolicy ys))")
     apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy (removeShadowRules3 Cp ys)))")
     apply (metis Cdom2  NMPRS3 True l2p_aux notMTpolicyimpnotMT)
     apply (simp add: domIff)
     apply (subgoal_tac "x \<notin>  dom (Cp (list2FWpolicy (removeShadowRules3 Cp ys)))") 
     apply (metis l2p_aux l2p_aux2 nlpaux)
     apply (metis domIff)
     done
     next 
     case False thus ?thesis using Cons False
        proof (cases "ys = []") 
          case True thus ?thesis using Cons by (simp) (rule impI, simp) next
          case False thus ?thesis using Cons False `\<not> notMTpolicy Cp ys` 
             apply (simp)
            apply (rule conjI| rule impI| simp)+
            apply (subgoal_tac "removeShadowRules3 Cp ys = []")
            apply (subgoal_tac "x \<notin> dom (Cp (list2FWpolicy ys))") 
            apply simp_all 
            apply (metis l2p_aux nlpaux)    
            apply (erule nMT_domMT, simp_all)
            by (metis SR3nMT)
        qed
     qed
qed       



lemma C_eq_id: "wellformed_policy1_strong p \<Longrightarrow>
                Cp(list2FWpolicy (insertDeny p)) = Cp (list2FWpolicy p)"
apply (rule ext)
apply (rule C_eq_if_mr_eq)
apply simp_all
apply (erule mr_iD)
done

lemma C_eq_RS3: "notMTpolicy Cp  p \<Longrightarrow>
                 Cp(list2FWpolicy (removeShadowRules3 Cp p)) = Cp (list2FWpolicy p)"
apply (rule ext)
by (erule C_eq_RS3_aux[symmetric])



lemma NMPrd[rule_format]: "notMTpolicy Cp  p \<longrightarrow> notMTpolicy Cp  (remdups p)"
apply (induct p, simp_all)
by (auto simp: NMPcharn)



lemma NMPDA[rule_format]: "DenyAll \<in> set p \<longrightarrow> notMTpolicy Cp  p"
by (induct p, simp_all add: DAnMT)


lemma NMPiD[rule_format]: "notMTpolicy Cp  (insertDeny p)"
apply (insert DAiniD [of p])
apply (erule NMPDA)
done



lemma list2FWpolicy2list[rule_format]: "Cp (list2FWpolicy(policy2list p)) = (Cp p)"
apply (rule ext)
apply (induct_tac p, simp_all)
apply (case_tac "x \<in> dom (Cp (Combinators2))") 
apply (metis Cdom2 CeqEnd domIff p2lNmt)
apply (metis CeqStart domIff p2lNmt nlpaux)
done



lemmas C_eq_Lemmas = noneMTRS2 noneMTrd  SCp2l
              wp1n_RS2  wp1ID NMPiD wp1alternative_RS1
              p2lNmt list2FWpolicy2list wellformed_policy3_charn waux2 wp1_eq

lemmas C_eq_subst_Lemmas = C_eq_sorted C_eq_sortedQ C_eq_RS2 C_eq_rd C_eq_RS3 C_eq_id 

lemma C_eq_All_untilSorted: 
"\<lbrakk>DenyAll \<in> set (policy2list p);  all_in_list (policy2list p) l;
  allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow>
  Cp(list2FWpolicy (sort (removeShadowRules2 (remdups (removeShadowRules3 Cp
    (insertDeny (removeShadowRules1  (policy2list p)))))) l)) = Cp p"
apply (subst C_eq_sorted)
apply (simp_all add: C_eq_Lemmas)
apply (subst C_eq_RS2)
apply (simp_all add: C_eq_Lemmas) 
apply (subst C_eq_rd)
apply (simp_all add: C_eq_Lemmas)
apply (subst C_eq_RS3)
apply (simp_all add: C_eq_Lemmas)
apply (subst C_eq_id)
apply (simp_all add: C_eq_Lemmas)
done


lemma C_eq_All_untilSortedQ: 
"\<lbrakk>DenyAll \<in> set (policy2list p);  all_in_list (policy2list p) l;
  allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow>
  Cp(list2FWpolicy (qsort (removeShadowRules2 (remdups (removeShadowRules3 Cp
    (insertDeny (removeShadowRules1  (policy2list p)))))) l)) = Cp p"
apply (subst C_eq_sortedQ)
apply (simp_all add: C_eq_Lemmas)
apply (subst C_eq_RS2)
apply (simp_all add: C_eq_Lemmas) 
apply (subst C_eq_rd)
apply (simp_all add: C_eq_Lemmas)
apply (subst C_eq_RS3)
apply (simp_all add: C_eq_Lemmas)
apply (subst C_eq_id)
apply (simp_all add: C_eq_Lemmas)
done


(* or, even shorter *)

lemma C_eq_All_untilSorted_withSimps: 
"\<lbrakk>DenyAll \<in> set (policy2list p);  all_in_list (policy2list p) l;
 allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow>
 Cp(list2FWpolicy (sort (removeShadowRules2 (remdups (removeShadowRules3 Cp (insertDeny
   (removeShadowRules1  (policy2list p)))))) l)) = Cp p"
by (simp_all add: C_eq_Lemmas C_eq_subst_Lemmas)


lemma C_eq_All_untilSorted_withSimpsQ: 
"\<lbrakk>DenyAll \<in> set (policy2list p);  all_in_list (policy2list p) l;
 allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow>
 Cp(list2FWpolicy (qsort (removeShadowRules2 (remdups (removeShadowRules3 Cp (insertDeny
   (removeShadowRules1  (policy2list p)))))) l)) = Cp p"
by (simp_all add: C_eq_Lemmas C_eq_subst_Lemmas)




lemma InDomConc[rule_format]: "p \<noteq> [] \<longrightarrow> x \<in> dom (Cp (list2FWpolicy (p))) \<longrightarrow>
                               x \<in>  dom (Cp (list2FWpolicy (a#p)))"
apply (induct p)
apply simp_all
apply (case_tac "p = []")
apply (simp_all add: dom_def Cp.simps)
done

lemma not_in_member[rule_format]: "member a b \<longrightarrow> x \<notin> dom (Cp b) \<longrightarrow> x \<notin> dom (Cp a)"
apply (induct b)
apply (simp_all add: dom_def Cp.simps)
done



lemma src_in_sdnets[rule_format]: "\<not> member DenyAll x \<longrightarrow> p \<in> dom (Cp x) \<longrightarrow>
                             subnetsOfAdr (src p) \<inter> (fst_set (sdnets x)) \<noteq> {}"
apply (induct rule: Combinators.induct)
apply simp
apply (simp add: fst_set_def subnetsOfAdr_def PLemmas)
apply (simp add: fst_set_def subnetsOfAdr_def PLemmas)
apply (rule impI)+
apply (simp add: fst_set_def)
apply (case_tac "p \<in> dom (Cp Combinators2)")
apply simp_all
apply (rule subnetAux)
apply assumption
apply (auto simp: PLemmas)
done

lemma dest_in_sdnets[rule_format]: "\<not> member DenyAll x \<longrightarrow> p \<in> dom (Cp x) \<longrightarrow>
                              subnetsOfAdr (dest p) \<inter> (snd_set (sdnets x)) \<noteq> {}"
apply (induct rule: Combinators.induct)
apply simp
apply (simp add: snd_set_def subnetsOfAdr_def PLemmas)
apply (simp add: snd_set_def subnetsOfAdr_def PLemmas)
apply (rule impI)+
apply (simp add: snd_set_def)
apply (case_tac "p \<in> dom (Cp Combinators2)")
apply simp_all
apply (rule subnetAux)
apply assumption
apply (auto simp: PLemmas)
done



lemma sdnets_in_subnets[rule_format]: "p\<in> dom (Cp x) \<longrightarrow> \<not> member DenyAll x \<longrightarrow>
        (\<exists> (a,b)\<in>sdnets x. a \<in> subnetsOfAdr (src p) \<and> b \<in> subnetsOfAdr (dest p))"
apply (rule Combinators.induct)
apply simp_all
apply (simp add: PLemmas subnetsOfAdr_def)
apply (simp add: PLemmas subnetsOfAdr_def)
apply (rule impI)+
apply simp
apply (case_tac "p \<in> dom (Cp (Combinators2))")
apply simp_all
apply (auto simp: PLemmas subnetsOfAdr_def)
done

lemma disjSD_no_p_in_both[rule_format]:   
 "\<lbrakk>disjSD_2 x y; \<not> member DenyAll x; \<not> member DenyAll y;  
   p \<in> dom (Cp x); p \<in> dom (Cp y)\<rbrakk> \<Longrightarrow> False"
apply (rule_tac A = "sdnets x" and B = "sdnets y" and D = "src p"
             and F = "dest p" in tndFalse)
by (auto simp: dest_in_sdnets src_in_sdnets sdnets_in_subnets disjSD_2_def)

lemma list2FWpolicy_eq: "zs \<noteq> [] \<Longrightarrow>
             Cp (list2FWpolicy (x \<oplus> y # z)) p = Cp (x \<oplus> list2FWpolicy (y # z)) p"
by (metis ConcAssoc l2p_aux list2FWpolicy.simps(2))



lemma dom_sep[rule_format]: "x \<in> dom (Cp (list2FWpolicy p)) \<longrightarrow>
                             x \<in> dom (Cp (list2FWpolicy(separate p)))"
apply (rule separate.induct) back
apply simp_all
apply (rule conjI)
apply (rule impI)+
apply simp
apply (thin_tac "False \<Longrightarrow> ?S")
apply (drule mp)
apply (case_tac "x \<in> dom (Cp (DenyAllFromTo v va))")
apply (metis CConcStartA domIff l2p_aux2
             list2FWpolicyconc not_Cons_self )
apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy (y #z)))")
apply (metis CConcStartA Cdom2 InDomConc domIff l2p_aux2 list2FWpolicyconc nlpaux)
apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy ((DenyAllFromTo v va)#y#z)))")
apply (simp add: dom_def Cp.simps)
apply simp
apply simp
apply (rule impI)+
apply simp
apply (thin_tac "False \<Longrightarrow> ?S")
apply (case_tac "x \<in> dom (Cp (DenyAllFromTo v va))")
apply simp_all
apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy (y #z)))")
apply (metis InDomConc sepnMT list.simps(2))
apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy ((DenyAllFromTo v va)#y#z)))")
apply (simp add: dom_def Cp.simps)
apply simp
apply (rule impI | rule conjI)+
apply simp
apply (case_tac "x \<in> dom (Cp (AllowPortFromTo v va vb))")
apply (metis CConcStartA domIff  l2p_aux2
            list2FWpolicyconc not_Cons_self )
apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy (y #z)))")
apply simp
apply (metis CConcStartA Cdom2 InDomConc domIff l2p_aux2 list2FWpolicyconc nlpaux)
apply (simp add: dom_def Cp.simps)
apply (rule impI)+
apply simp_all
apply (thin_tac "False \<Longrightarrow> ?S")
apply (case_tac "x \<in> dom (Cp (AllowPortFromTo v va vb))")
apply simp_all 
apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy (y #z)))")
apply simp
apply (metis Conc_not_MT InDomConc sepnMT)
apply (metis domIff nlpaux)
apply (rule conjI | rule impI)+
apply simp
apply (thin_tac "False \<Longrightarrow> ?S")
apply (drule mp)
apply (case_tac "x \<in> dom (Cp ((v \<oplus> va)))")
apply (metis Cp.simps(4) CConcStartA ConcAssoc domIff 
    list2FWpolicy2list list2FWpolicyconc  p2lNmt)
defer 1
apply simp_all
apply (rule impI)+
apply simp
apply (thin_tac "False \<Longrightarrow> ?S")
apply (case_tac "x \<in> dom (Cp ((v \<oplus> va)))")
apply (metis CConcStartA)
apply (drule mp)
apply (simp add: Cp.simps dom_def)
apply (metis InDomConc list.simps(2)sepnMT)
apply (subgoal_tac "x \<in> dom (Cp (list2FWpolicy (y#z)))")
apply (case_tac "x \<in> dom (Cp y)")
apply simp_all
apply (metis CConcStartA Cdom2 ConcAssoc domIff)
apply (metis InDomConc domIff l2p_aux2 list2FWpolicyconc nlpaux)
apply (case_tac "x \<in> dom (Cp y)")
apply simp_all
apply (metis domIff nlpaux)
done

lemma domdConcStart[rule_format]: "   x \<in> dom (Cp (list2FWpolicy (a#b))) \<longrightarrow>
        x \<notin> dom (Cp (list2FWpolicy b))
       \<longrightarrow> x \<in> dom (Cp (a))"
apply (induct b, simp_all)
apply (auto simp: PLemmas)
done

lemma sep_dom2_aux: "\<lbrakk>x \<in> dom (Cp (list2FWpolicy (a \<oplus> y # z)))\<rbrakk>
       \<Longrightarrow> x \<in> dom (Cp (a \<oplus> list2FWpolicy (y # z)))"
apply auto
by (metis list2FWpolicy_eq p2lNmt)


lemma sep_dom2_aux2: 
"\<lbrakk>(x \<in> dom (Cp (list2FWpolicy (separate (y # z)))) \<longrightarrow>
    x \<in> dom (Cp (list2FWpolicy (y # z))));
   x \<in> dom (Cp (list2FWpolicy (a # separate (y # z))))\<rbrakk>
\<Longrightarrow> x \<in> dom (Cp (list2FWpolicy (a \<oplus> y # z)))"
by (metis CConcStartA InDomConc domdConcStart list.simps(2) list2FWpolicy.simps(2) list2FWpolicyconc)


lemma sep_dom2[rule_format]: 
 "x \<in> dom (Cp (list2FWpolicy (separate p))) \<longrightarrow> x \<in> dom (Cp (list2FWpolicy( p)))"
apply (rule separate.induct)
by (simp_all add: sep_dom2_aux sep_dom2_aux2)

lemma sepDom: "dom (Cp (list2FWpolicy p)) = dom (Cp (list2FWpolicy (separate p)))"
apply (rule equalityI)
by (rule subsetI, (erule dom_sep|erule sep_dom2))+

lemma C_eq_s_ext[rule_format]: "p \<noteq> [] \<longrightarrow>
                         Cp (list2FWpolicy (separate p)) a  = Cp (list2FWpolicy p) a "
 proof (induct rule: separate.induct) 
  case goal1 thus ?case 
    apply simp
     apply (cases "x = []")
     apply (metis l2p_aux2 separate.simps(5))
     apply simp
     apply (cases "a \<in> dom (Cp (list2FWpolicy x))")
     apply (subgoal_tac "a \<in> dom (Cp (list2FWpolicy (separate x)))")
     apply (metis Cdom2 list2FWpolicyconc sepDom sepnMT)
     apply (metis sepDom)
     apply (subgoal_tac "a \<notin> dom (Cp (list2FWpolicy (separate x)))")
     apply (subst list2FWpolicyconc) 
     apply (simp add: sepnMT)
     apply (subst list2FWpolicyconc) 
     apply (simp add: sepnMT)
     apply (metis nlpaux sepDom)
     apply (metis sepDom)
     done
  next
  case goal2 thus ?case 
     apply simp 
     apply (cases "z = []")
     apply simp_all
     apply (rule conjI|rule impI|simp)+
     apply (subst list2FWpolicyconc)
      apply (metis not_Cons_self sepnMT)
     apply (metis Cp.simps(4) CConcStartaux Cdom2 domIff)
     apply (rule conjI|rule impI|simp)+
     apply (erule list2FWpolicy_eq)
     apply (rule impI, simp)
     apply (subst list2FWpolicyconc) 
     apply (metis list.simps(2) sepnMT) 
     by (metis Cp.simps(4) CConcStartaux Cdom2 domIff)
  next
  case goal3 thus ?case
  apply simp 
     apply (cases "z = []")
     apply simp_all
     apply (rule conjI|rule impI|simp)+
     apply (subst list2FWpolicyconc)
      apply (metis not_Cons_self sepnMT)
     apply (metis Cp.simps(4) CConcStartaux Cdom2 domIff)
     apply (rule conjI|rule impI|simp)+
     apply (erule list2FWpolicy_eq)
     apply (rule impI, simp)
     apply (subst list2FWpolicyconc) 
     apply (metis list.simps(2) sepnMT) 
     by (metis Cp.simps(4) CConcStartaux Cdom2 domIff)
  next
  case goal4 thus ?case
  apply simp 
     apply (cases "z = []")
     apply simp_all
     apply (rule conjI|rule impI|simp)+
     apply (subst list2FWpolicyconc)
      apply (metis not_Cons_self sepnMT)
     apply (metis Cp.simps(4) CConcStartaux Cdom2 domIff)
     apply (rule conjI|rule impI|simp)+
     apply (erule list2FWpolicy_eq)
     apply (rule impI, simp)
     apply (subst list2FWpolicyconc) 
     apply (metis list.simps(2) sepnMT) 
     by (metis Cp.simps(4) CConcStartaux Cdom2 domIff)
   next
  case goal5 thus ?case by simp next
  case goal6 thus ?case by simp next
  case goal7 thus ?case by simp next
  case goal8 thus ?case by simp next
 qed

lemma C_eq_s: "p \<noteq> [] \<Longrightarrow> Cp (list2FWpolicy (separate p)) = Cp (list2FWpolicy p)  "
apply (rule ext)
apply (rule C_eq_s_ext)
apply simp
done

(*MOVE FORWARD*)
lemma sortnMTQ: "p \<noteq> [] \<Longrightarrow> qsort p l \<noteq> []"
by (metis set_sortQ setnMT)


lemmas C_eq_Lemmas_sep =
       C_eq_Lemmas sortnMT sortnMTQ RS2_NMT NMPrd NMPRS3 


lemma C_eq_until_separated:
"\<lbrakk>DenyAll \<in> set (policy2list p); all_in_list (policy2list p) l;
  allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow>
 Cp (list2FWpolicy (separate (sort (removeShadowRules2 (remdups (removeShadowRules3 Cp
  (insertDeny (removeShadowRules1 (policy2list p)))))) l))) = Cp p"
apply (subst  C_eq_s)
apply (simp_all add: C_eq_Lemmas_sep)
apply (rule C_eq_All_untilSorted)
apply simp_all
done

lemma C_eq_until_separatedQ:
"\<lbrakk>DenyAll \<in> set (policy2list p); all_in_list (policy2list p) l;
  allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow>
 Cp (list2FWpolicy (separate (qsort (removeShadowRules2 (remdups (removeShadowRules3 Cp
  (insertDeny (removeShadowRules1 (policy2list p)))))) l))) = Cp p"
apply (subst  C_eq_s)
apply (simp_all add: C_eq_Lemmas_sep)
apply (rule C_eq_All_untilSortedQ)
apply simp_all
done



lemma domID[rule_format]: "p \<noteq> [] \<and> x \<in> dom(Cp(list2FWpolicy p)) \<longrightarrow>
                           x \<in> dom (Cp(list2FWpolicy(insertDenies p)))"
proof(induct p)
  case Nil then show ?case by simp
next 
  case (Cons a p) then show ?case
    proof(cases "p=[]")
       case goal1 then show ?case
          apply(simp) apply(rule impI)
          apply (cases a, simp_all)
          apply (simp_all add: Cp.simps dom_def)+ 
          by auto
    next
       case goal2 then show ?case
         proof(cases "x \<in> dom(Cp(list2FWpolicy p))")
            case goal1 then show ?case
               apply simp apply (rule impI)
               apply (cases a, simp_all)
               apply (metis InDomConc goal1(2) idNMT)
               apply (rule InDomConc, simp_all add: idNMT)+
               done
         next 
            case goal2 then show ?case  
               apply simp apply (rule impI)
               proof(cases "x \<in> dom (Cp (list2FWpolicy (insertDenies p)))")
                  case goal1 then show ?case
                     proof(induct a)
                        case DenyAll then show ?case by simp
                     next               
                        case (DenyAllFromTo src dest) then show ?case
                           apply simp by( rule InDomConc, simp add: idNMT)
                     next  
                       case (AllowPortFromTo src dest port) then show ?case
                           apply simp by(rule InDomConc, simp  add: idNMT)
                     next 
                       case (Conc _ _) then show ?case
                           apply simp by(rule InDomConc, simp add: idNMT)
                     qed
               next
                  case goal2 then show ?case
                     proof (induct a)
                       case DenyAll then show ?case by simp
                     next
                       case (DenyAllFromTo src dest) then show ?case
                   by(simp,metis domIff CConcStartA list2FWpolicyconc nlpaux Cdom2)
                     next
                       case (AllowPortFromTo src dest port) then show ?case
                   by(simp,metis domIff CConcStartA list2FWpolicyconc nlpaux Cdom2)
                     next
                       case (Conc _ _) then show ?case
                    apply simp
by (metis CConcStartA Cdom2 Conc(5) ConcAssoc domIff domdConcStart goal2(2)) 
                     qed
               qed
         qed
    qed
 qed
   

lemma DA_is_deny: 
  "x \<in> dom (Cp (DenyAllFromTo a b \<oplus> DenyAllFromTo b a \<oplus> DenyAllFromTo a b)) \<Longrightarrow>
Cp (DenyAllFromTo a b\<oplus>DenyAllFromTo b a \<oplus> DenyAllFromTo a b) x = Some (deny ())"
apply (case_tac "x \<in> dom (Cp (DenyAllFromTo a b))")
apply (simp_all add: PLemmas)
apply (simp_all split: if_splits)
done

lemma iDdomAux[rule_format]: 
  "p \<noteq> [] \<longrightarrow> x \<notin> dom (Cp (list2FWpolicy p)) \<longrightarrow>
   x \<in> dom (Cp (list2FWpolicy (insertDenies p))) \<longrightarrow>
   Cp (list2FWpolicy (insertDenies p)) x = Some (deny ())"
 proof (induct p)
  case Nil thus ?case by simp
  next
  case (Cons y ys) thus ?case
    proof (cases y) 
      case DenyAll then show ?thesis by simp next
      case (DenyAllFromTo a b) then show ?thesis using DenyAllFromTo Cons
        apply simp
        apply (rule impI)+
        proof (cases "ys = []")
          case goal1 then show ?case by (simp add: DA_is_deny) next
          case goal2 then show ?case
            apply simp
            apply (drule mp)
            apply (metis DenyAllFromTo InDomConc goal2(3) goal2(5))
            apply (cases "x \<in> dom (Cp (list2FWpolicy (insertDenies ys)))")
            apply simp_all
            apply (metis Cdom2 DenyAllFromTo goal2(5) idNMT list2FWpolicyconc)
            apply (subgoal_tac "Cp (list2FWpolicy (DenyAllFromTo a b \<oplus>
             DenyAllFromTo b a \<oplus> DenyAllFromTo a b#insertDenies ys)) x =
            Cp ((DenyAllFromTo a b \<oplus> DenyAllFromTo b a \<oplus> DenyAllFromTo a b)) x ")
            apply simp
            apply (rule DA_is_deny)
            apply (metis DenyAllFromTo domdConcStart goal2(4))
            apply (metis DenyAllFromTo l2p_aux2 list2FWpolicyconc nlpaux)
            done 
        qed
      next
      case (AllowPortFromTo a b c) then show ?thesis using Cons AllowPortFromTo
        proof (cases "ys = []")
          case goal1 then show ?case 
          apply simp
          apply (rule impI)+
       apply (subgoal_tac "x \<in> dom (Cp (DenyAllFromTo a b \<oplus> DenyAllFromTo b a))")
          apply (simp_all add: PLemmas)
          apply (simp split: if_splits) apply auto 
          done next
          case goal2 then show ?case
            apply simp
            apply (rule impI)+
            apply (drule mp)
            apply (metis AllowPortFromTo InDomConc goal2(4))
            apply (cases "x \<in> dom (Cp (list2FWpolicy (insertDenies ys)))")
            apply simp_all
            apply (metis AllowPortFromTo Cdom2 goal2(4) idNMT list2FWpolicyconc)
            apply (subgoal_tac "Cp (list2FWpolicy (DenyAllFromTo a b \<oplus>
                DenyAllFromTo b a \<oplus> AllowPortFromTo a b c#insertDenies ys)) x =
               Cp ((DenyAllFromTo a b \<oplus> DenyAllFromTo b a)) x ")
            apply simp
            defer 1 
            apply (metis AllowPortFromTo CConcStartA ConcAssoc goal2(4) idNMT
                         list2FWpolicyconc nlpaux)
            apply (simp add: PLemmas, simp split: if_splits) apply auto 
          done
         qed
      next
      case (Conc a b) then show ?thesis
 proof (cases "ys = []")
          case goal1 then show ?case 
            apply simp
            apply (rule impI)+
            apply (subgoal_tac "x \<in> dom (Cp (DenyAllFromTo (first_srcNet a)
         (first_destNet a) \<oplus> DenyAllFromTo (first_destNet a) (first_srcNet a)))")
            apply (simp_all add: PLemmas)
            apply (simp split: if_splits) apply auto 
            done next
          case goal2 then show ?case
            apply simp
            apply (rule impI)+
            apply (cases "x \<in> dom (Cp (list2FWpolicy (insertDenies ys)))")
           apply (metis Cdom2 Conc Cons InDomConc goal2(2) idNMT list2FWpolicyconc)
            apply (subgoal_tac "Cp (list2FWpolicy (DenyAllFromTo (first_srcNet a) 
            (first_destNet a) \<oplus> DenyAllFromTo (first_destNet a) (first_srcNet a)
            \<oplus> a\<oplus> b#insertDenies ys)) x =
            Cp ((DenyAllFromTo (first_srcNet a) (first_destNet a)  \<oplus>
            DenyAllFromTo (first_destNet a)  (first_srcNet a) \<oplus> a\<oplus> b)) x ")
            apply simp
            defer 1 
            apply (metis Conc l2p_aux2 list2FWpolicyconc nlpaux)
            apply (subgoal_tac "Cp ((DenyAllFromTo (first_srcNet a)
            (first_destNet a) \<oplus> DenyAllFromTo (first_destNet a)
            (first_srcNet a) \<oplus> a\<oplus> b)) x = Cp ((DenyAllFromTo (first_srcNet a)
     (first_destNet a)  \<oplus> DenyAllFromTo (first_destNet a)  (first_srcNet a))) x ")
            apply simp
            defer 1 
            apply (metis CConcStartA Conc ConcAssoc nlpaux)
            apply (simp add: PLemmas, simp split: if_splits) apply auto 
          done
        qed
    qed
qed         

lemma iD_isD[rule_format]: "p \<noteq> [] \<longrightarrow> x \<notin> dom (Cp (list2FWpolicy p))
          \<longrightarrow> Cp (DenyAll \<oplus> list2FWpolicy (insertDenies p)) x = Cp DenyAll x"
apply (case_tac "x \<in> dom (Cp (list2FWpolicy (insertDenies p)))")
apply (rule impI)+
apply (metis Cp.simps(1) deny_all_def iDdomAux Cdom2)
apply (rule impI)+
apply (subst nlpaux)
apply simp_all
done


lemma inDomConc:"\<lbrakk> x\<notin>dom (Cp a); x\<notin>dom (Cp (list2FWpolicy p))\<rbrakk> \<Longrightarrow>
                 x \<notin> dom (Cp (list2FWpolicy(a#p)))"
by (metis domdConcStart)

lemma domsdisj[rule_format]: "p \<noteq> [] \<longrightarrow> (\<forall> x s. s \<in> set p \<and> x \<in> dom (Cp A) \<longrightarrow> 
                             x \<notin> dom (Cp s)) \<longrightarrow> y \<in> dom (Cp A) \<longrightarrow>
                             y \<notin> dom (Cp (list2FWpolicy p))"
apply (induct p)
apply simp
apply (case_tac "p = []")
apply simp
apply (rule_tac x = y in spec)
apply (simp add: split_tupled_all)
apply (rule impI)+
apply (rule inDomConc)
apply (drule_tac x = y in spec, drule_tac x = a in spec)
apply auto
done

lemma isSepaux:
  "\<lbrakk>p \<noteq> []; noDenyAll (a#p); separated (a # p);
   x \<in> dom (Cp (DenyAllFromTo (first_srcNet a) (first_destNet a) \<oplus>
   DenyAllFromTo (first_destNet a) (first_srcNet a) \<oplus> a))\<rbrakk> \<Longrightarrow>
   x \<notin> dom (Cp (list2FWpolicy p))"
apply (rule_tac A = "(DenyAllFromTo (first_srcNet a) (first_destNet a) \<oplus>
             DenyAllFromTo (first_destNet a) (first_srcNet a) \<oplus> a)" in domsdisj)
apply simp_all
apply (rule notI)
apply (rule_tac p = xa and x ="(DenyAllFromTo (first_srcNet a) (first_destNet a) 
                \<oplus> DenyAllFromTo (first_destNet a) (first_srcNet a) \<oplus> a)" and
                y = s in disjSD_no_p_in_both)
apply simp_all
apply (simp add: disjSD_2_def)
apply (rule allI)+
apply (metis first_isIn tNDComm twoNetsDistinct_def)
apply (metis noDA)
done



lemma noneMTsep[rule_format]: "noneMT Cp p \<longrightarrow> noneMT Cp (separate p)"
apply (rule separate.induct) back
apply (simp_all add: Cp.simps map_add_le_mapE map_le_antisym)
done



lemma dom_id: 
  "\<lbrakk>noDenyAll (a#p); separated (a#p); p \<noteq> []; x \<notin> dom (Cp (list2FWpolicy p));
   x \<in> dom (Cp (a))\<rbrakk>
       \<Longrightarrow> x \<notin> dom (Cp (list2FWpolicy (insertDenies p)))"
apply (rule_tac a = a in isSepaux)
apply simp_all
apply (rule idNMT)
apply simp
apply (rule noDAID)
apply simp
apply (rule conjI)
apply (rule allI)
apply (rule impI)
apply (rule id_aux4)
apply simp_all
apply (rule sepNetsID)
apply simp_all
apply (metis noDA1eq)
apply (simp add: Cp.simps)
done

lemma C_eq_iD_aux2[rule_format]: 
 "noDenyAll1 p \<longrightarrow>
separated p\<longrightarrow>
p \<noteq> []\<longrightarrow>
x \<in> dom (Cp (list2FWpolicy p))\<longrightarrow>
Cp(list2FWpolicy (insertDenies p)) x = Cp(list2FWpolicy p) x"
proof (induct p)
case Nil thus ?case by simp 
next
case (Cons y ys)  thus ?case using Cons
 proof (cases y)
 case DenyAll thus ?thesis using Cons DenyAll apply simp
   apply (case_tac "ys = []")
   apply simp_all
   apply (case_tac "x \<in> dom (Cp (list2FWpolicy ys))") 
   apply simp_all
apply (metis Cdom2 domID idNMT list2FWpolicyconc noDA1eq)
apply (metis DenyAll iD_isD idNMT list2FWpolicyconc nlpaux)
   done
 next
 case (DenyAllFromTo a b) thus ?thesis using Cons apply simp
   apply (rule impI|rule allI|rule conjI|simp)+
   apply (case_tac "ys = []")
   apply simp_all
   apply (metis Cdom2 ConcAssoc DenyAllFromTo)
   apply (case_tac "x \<in> dom (Cp (list2FWpolicy ys))")
   apply simp_all
   apply (drule mp)
   apply (metis noDA1eq)
   apply (case_tac "x \<in> dom (Cp (list2FWpolicy (insertDenies ys)))")
   apply (metis Cdom2 DenyAllFromTo idNMT list2FWpolicyconc)
   apply (metis domID)
   apply (case_tac "x \<in> dom (Cp (list2FWpolicy (insertDenies ys)))")
   apply (subgoal_tac "Cp (list2FWpolicy (DenyAllFromTo a b \<oplus> DenyAllFromTo b a \<oplus>
                       DenyAllFromTo a b # insertDenies ys)) x = Some (deny ())")
   apply simp_all
   apply (subgoal_tac "Cp (list2FWpolicy (DenyAllFromTo a b # ys)) x =
                       Cp ((DenyAllFromTo a b)) x")
   apply (simp add: PLemmas, simp split: if_splits)
   apply (metis list2FWpolicyconc nlpaux)
   apply (metis Cdom2 DenyAllFromTo iD_isD iDdomAux idNMT list2FWpolicyconc)
apply (metis Cdom2 DenyAllFromTo domIff idNMT list2FWpolicyconc nlpaux)
   done
 next
 case (AllowPortFromTo a b c) thus ?thesis using AllowPortFromTo Cons apply simp
   apply (rule impI|rule allI|rule conjI|simp)+
   apply (case_tac "ys = []")
   apply simp_all
   apply (metis Cdom2 ConcAssoc AllowPortFromTo)
   apply (case_tac "x \<in> dom (Cp (list2FWpolicy ys))")
   apply simp_all
   apply (drule mp)
   apply (metis noDA1eq)
   apply (case_tac "x \<in> dom (Cp (list2FWpolicy (insertDenies ys)))")
   apply (metis Cdom2 AllowPortFromTo idNMT list2FWpolicyconc)
   apply (metis domID)
   apply (subgoal_tac "x \<in> dom (Cp (AllowPortFromTo a b c))")
   apply (case_tac "x \<notin> dom (Cp (list2FWpolicy (insertDenies ys)))")
   apply simp_all
   apply (metis AllowPortFromTo Cdom2 ConcAssoc l2p_aux2 list2FWpolicyconc nlpaux)
   apply (metis AllowPortFromTo Combinators.simps(6) member.simps(4) 
                dom_id noDenyAll.simps(1) separated.simps(1))
   apply (metis AllowPortFromTo domdConcStart)
   done
 next
 case (Conc a b) thus ?thesis using Cons Conc 
   apply simp
   apply (rule impI|rule allI|rule conjI|simp)+
   apply (case_tac "ys = []")
   apply simp_all
   apply (metis Cdom2 ConcAssoc Conc)
   apply (case_tac "x \<in> dom (Cp (list2FWpolicy ys))")
   apply simp_all
   apply (drule mp)
   apply (metis noDA1eq)
   apply (case_tac "x \<in> dom (Cp (a \<oplus> b))")
   apply (case_tac "x \<notin> dom (Cp (list2FWpolicy (insertDenies ys)))")
   apply simp_all
   apply (subst list2FWpolicyconc)
   apply (rule idNMT, simp)
   apply (metis domID)
   apply (metis Cdom2 Conc idNMT list2FWpolicyconc)
   apply (metis Cdom2 Conc domIff idNMT list2FWpolicyconc )
   apply (case_tac "x \<in> dom (Cp (a \<oplus> b))")
   apply (case_tac "x \<notin> dom (Cp (list2FWpolicy (insertDenies ys)))")
   apply simp_all
   apply (subst list2FWpolicyconc)
   apply (rule idNMT, simp)
   apply (metis Cdom2 Conc ConcAssoc list2FWpolicyconc nlpaux)
   apply (metis Conc member.simps(1) dom_id 
                noDenyAll.simps(1) separated.simps(1))
   apply (metis Conc domdConcStart)
   done
 qed
qed

lemma C_eq_iD: "\<lbrakk>separated p; noDenyAll1 p; wellformed_policy1_strong p\<rbrakk>  \<Longrightarrow> 
 Cp (list2FWpolicy (insertDenies p)) = Cp (list2FWpolicy p)"
apply (rule ext)
apply (rule C_eq_iD_aux2)
apply simp_all 
apply (subgoal_tac "DenyAll \<in> set p")
apply (metis CConcStartA DAAux wp1n_tl)
apply (erule wp1_aux1aa)
done

lemma noDAsortQ[rule_format]: "noDenyAll1 p \<longrightarrow> noDenyAll1 (qsort p l)"
apply (case_tac "p")
apply simp
apply simp
apply (case_tac "a = DenyAll")
apply simp_all 
apply (rule impI)
apply (subst nDAeqSet)
defer 1
apply simp
defer 1
apply (rule set_sortQ)
apply (rule impI)
apply (rule noDA1eq)
apply (subgoal_tac "noDenyAll (a#list)")
defer 1
apply (case_tac a, simp,simp)
apply simp
apply simp
apply (subst nDAeqSet)
defer 1
apply assumption
by (metis append_Cons append_Nil qsort.simps(2) set_qsort)

(*MOVE FORWARD*)

lemma NetsCollectedSortQ: "distinct p \<Longrightarrow>noDenyAll1 p \<Longrightarrow> all_in_list p l \<Longrightarrow>
                          singleCombinators p \<Longrightarrow> NetsCollected (qsort p l)"
apply (rule_tac l = l in NetsCollectedSorted)
apply (rule noDAsortQ)
apply simp_all
apply (rule_tac b=p in all_in_listSubset)
apply simp_all
apply (rule sort_is_sortedQ)
apply simp_all
done



lemmas CLemmas =  nMTSort nMTSortQ noneMTRS2 noneMTrd
  noDAsort noDAsortQ nDASC wp1_eq  wp1ID  
  SCp2l ANDSep   wp1n_RS2 
  OTNSEp OTNSC noDA1sep wp1_alternativesep wellformed_eq 
  wellformed1_alternative_sorted  
    

lemmas C_eqLemmas_id = CLemmas  NC2Sep NetsCollectedSep 
                       NetsCollectedSort NetsCollectedSortQ separatedNC



lemma C_eq_Until_InsertDenies: "\<lbrakk>DenyAll \<in> set (policy2list p); all_in_list 
                          (policy2list p) l; allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow> 
Cp (list2FWpolicy ((insertDenies (separate (sort (removeShadowRules2 (remdups
(removeShadowRules3 Cp (insertDeny (removeShadowRules1 (policy2list p)))))) l))))) =
                                Cp p"
apply (subst C_eq_iD)
apply (simp_all add: C_eqLemmas_id)
apply (rule C_eq_until_separated)
apply simp_all
done

lemma C_eq_Until_InsertDeniesQ: "\<lbrakk>DenyAll \<in> set (policy2list p); all_in_list 
                          (policy2list p) l; allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow> 
Cp (list2FWpolicy ((insertDenies (separate (qsort (removeShadowRules2 (remdups
(removeShadowRules3 Cp (insertDeny (removeShadowRules1 (policy2list p)))))) l))))) =
                                Cp p"
apply (subst C_eq_iD)
apply (simp_all add: C_eqLemmas_id)
apply (metis WP1rd set_qsort wellformed1_sortedQ wellformed_eq wp1ID wp1_alternativesep wp1_aux1aa wp1n_RS2 wp1n_RS3)
apply (rule C_eq_until_separatedQ)
apply simp_all
done


lemma C_eq_RD_aux[rule_format]: "Cp (p) x = Cp (removeDuplicates p) x" 
apply (induct p)
apply simp_all
apply (rule conjI, rule impI)
apply (metis Cdom2 domIff nlpaux not_in_member)
apply (metis Cp.simps(4) CConcStartaux Cdom2 domIff)
done

lemma C_eq_RAD_aux[rule_format]: 
  "p \<noteq> []  \<longrightarrow> Cp (list2FWpolicy p) x = Cp (list2FWpolicy (removeAllDuplicates p)) x"
apply (induct p)
apply simp_all
apply (case_tac "p = []")
apply simp_all
apply (metis C_eq_RD_aux)
apply (subst list2FWpolicyconc)
apply simp
apply (case_tac "x \<in> dom (Cp (list2FWpolicy p))")
apply (subst list2FWpolicyconc)
apply (rule rADnMT)
apply simp
apply (subst Cdom2)
apply simp
apply (drule sym)
apply (subst Cdom2)
apply (simp add: dom_def)
apply simp
apply (drule sym)
apply (subst nlpaux)
apply simp
apply (subst list2FWpolicyconc)
apply (rule rADnMT)
apply simp
apply (subst nlpaux)
apply (simp add: dom_def)
apply (rule C_eq_RD_aux)
done

lemma C_eq_RAD: 
  "p \<noteq> []  \<Longrightarrow> Cp (list2FWpolicy p) = Cp (list2FWpolicy (removeAllDuplicates p)) "
apply (rule ext)
apply (erule C_eq_RAD_aux)
done



lemma C_eq_compile: 
"\<lbrakk>DenyAll \<in> set (policy2list p); all_in_list (policy2list p) l; 
 allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow> 
Cp (list2FWpolicy (removeAllDuplicates (insertDenies (separate (sort 
(removeShadowRules2 (remdups (removeShadowRules3 Cp (insertDeny 
 (removeShadowRules1 (policy2list p)))))) l))))) = Cp p"
apply (subst C_eq_RAD[symmetric])
apply (rule idNMT)
apply (simp add: C_eqLemmas_id)
apply (rule C_eq_Until_InsertDenies) 
apply simp_all
done


lemma C_eq_compileQ: 
"\<lbrakk>DenyAll \<in> set (policy2list p); all_in_list (policy2list p) l; 
 allNetsDistinct (policy2list p)\<rbrakk> \<Longrightarrow> 
Cp (list2FWpolicy (removeAllDuplicates (insertDenies (separate (qsort 
(removeShadowRules2 (remdups (removeShadowRules3 Cp (insertDeny 
 (removeShadowRules1 (policy2list p)))))) l))))) = Cp p"
apply (subst C_eq_RAD[symmetric])
apply (rule idNMT)
apply (metis WP1rd sepnMT sortnMTQ wellformed_policy1_strong.simps(1) wp1ID wp1n_RS2 wp1n_RS3)
apply (rule C_eq_Until_InsertDeniesQ) 
apply simp_all
done



lemma C_eq_normalizePr: 
"\<lbrakk>DenyAll \<in> set (policy2list p); 
 allNetsDistinct (policy2list p);
all_in_list (policy2list p) (Nets_List p)\<rbrakk> \<Longrightarrow> 
Cp (list2FWpolicy (normalizePr p)) = Cp p"
apply (simp add: normalizePr_def)
apply (rule C_eq_compile)
apply simp_all
done




lemma C_eq_normalizePrQ: 
"\<lbrakk>DenyAll \<in> set (policy2list p); 
 allNetsDistinct (policy2list p);
all_in_list (policy2list p) (Nets_List p)\<rbrakk> \<Longrightarrow> 
Cp (list2FWpolicy (normalizePrQ p)) = Cp p"
apply (simp add: normalizePrQ_def)
apply (rule C_eq_compileQ)
apply simp_all
done






lemma domSubset3: "dom (Cp (DenyAll \<oplus> x)) = dom (Cp (DenyAll))"
apply (simp add: PLemmas split_tupled_all split: option.splits)
done



lemma domSubset4: "dom (Cp (DenyAllFromTo x y \<oplus> DenyAllFromTo y x \<oplus> AllowPortFromTo x y dn)) = 
                  dom (Cp (DenyAllFromTo x y \<oplus> DenyAllFromTo y x))"
apply (simp add: PLemmas )
apply (simp split: option.splits decision.splits)
apply auto
done

lemma domSubset5: 
 "dom (Cp (DenyAllFromTo x y \<oplus> DenyAllFromTo y x \<oplus> AllowPortFromTo y x dn)) = 
                  dom (Cp (DenyAllFromTo x y \<oplus> DenyAllFromTo y x))"
apply (simp add: PLemmas)
apply (simp split: option.splits decision.splits)
apply auto
done



lemma domSubset1: "dom (Cp (DenyAllFromTo one two \<oplus> DenyAllFromTo two one \<oplus> AllowPortFromTo one two dn \<oplus> x)) = 
                  dom (Cp (DenyAllFromTo one two \<oplus> DenyAllFromTo two one \<oplus> x))"
apply (simp add: PLemmas)
apply (simp split: option.splits decision.splits)
apply (simp add: allow_all_def deny_all_def)
apply auto
done

lemma domSubset2: 
 "dom (Cp (DenyAllFromTo one two \<oplus> DenyAllFromTo two one \<oplus> AllowPortFromTo two one dn \<oplus> x)) = 
                  dom (Cp (DenyAllFromTo one two \<oplus> DenyAllFromTo two one \<oplus> x))"
apply (simp add: PLemmas)
apply (simp split: option.splits decision.splits)
apply (simp add: allow_all_def deny_all_def)
apply auto
done


lemma ConcAssoc2: "Cp (X \<oplus> Y \<oplus> ((A \<oplus> B) \<oplus> D)) = Cp (X \<oplus> Y \<oplus> A \<oplus> B \<oplus> D)"
apply (simp add: Cp.simps)
done


lemma ConcAssoc3: "Cp (X \<oplus> ((Y \<oplus> A) \<oplus> D)) = Cp (X \<oplus> Y \<oplus> A \<oplus> D)"
apply (simp add: Cp.simps)
done

lemma RS3_NMT[rule_format]: "DenyAll \<in> set p \<longrightarrow>
    removeShadowRules3 Cp p \<noteq> []"
apply (induct_tac p) 
apply (simp_all add: PLemmas)
done

lemma norm_notMT: "DenyAll \<in> set (policy2list p) \<Longrightarrow> normalizePr p \<noteq> []"
apply (simp add: normalizePr_def)
apply (rule rADnMT)
apply (rule idNMT)
apply (rule sepnMT)
apply (rule sortnMT)
apply (rule RS2_NMT)
apply (rule remDupsNMT) 
apply (rule RS3_NMT)
apply (rule DAiniD)
done


lemma norm_notMTQ: "DenyAll \<in> set (policy2list p) \<Longrightarrow> normalizePrQ p \<noteq> []"
apply (simp add: normalizePrQ_def)
apply (rule rADnMT)
apply (rule idNMT)
apply (rule sepnMT)
apply (rule sortnMTQ)
apply (rule RS2_NMT)
apply (rule remDupsNMT) 
apply (rule RS3_NMT)
apply (rule DAiniD)
done



lemma domDA: "dom (Cp (DenyAll \<oplus> A)) = dom (Cp (DenyAll))"
apply (rule domSubset3)
done


lemmas domain_reasoningPr = domDA ConcAssoc2 domSubset1 domSubset2 
 domSubset3 domSubset4  domSubset5 domSubsetDistr1
 domSubsetDistr2 domSubsetDistrA domSubsetDistrD coerc_assoc ConcAssoc 
 ConcAssoc3




text {* The following lemmas help with the normalisation *}
lemma list2policyR_Start[rule_format]: "p \<in> dom (Cp a) \<longrightarrow>
                 Cp (list2policyR (a # list)) p = Cp a p"
apply (rule list2policyR.induct) back
apply (auto simp: Cp.simps dom_def map_add_def) 
done


lemma list2policyR_End: "p \<notin> dom (Cp a) \<Longrightarrow>
        Cp (list2policyR (a # list)) p = (Cp a \<Oplus> list2policy (map Cp list)) p"
apply (rule list2policyR.induct)
apply (simp_all add: Cp.simps dom_def map_add_def list2policy_def split: option.splits)
done


lemma l2polR_eq_el[rule_format]: "N \<noteq> [] \<longrightarrow>
 Cp( list2policyR N) p =  (list2policy (map Cp N)) p"
apply (induct_tac N)
apply (simp_all add: list2policy_def)
apply (case_tac "p \<in> dom (Cp a)")
apply (simp add: domStart) 
apply (rule list2policyR_Start)
apply simp_all
apply (rule list2policyR.induct)
apply simp_all
apply (simp_all add: Cp.simps dom_def map_add_def)
apply (simp split: option.splits)
done


lemma l2polR_eq: "N \<noteq> [] \<Longrightarrow>
 Cp( list2policyR N) =  (list2policy (map Cp N))"
by (auto simp: list2policy_def l2polR_eq_el )




lemma list2FWpolicys_eq_el[rule_format]: 
 "Filter \<noteq> []  \<longrightarrow>  Cp (list2policyR Filter) p =  Cp (list2FWpolicy (rev Filter)) p"
apply (induct_tac Filter)
apply simp_all
apply (case_tac "list = []")
apply simp_all
apply (case_tac "p \<in> dom (Cp a)")
apply simp_all
apply (rule list2policyR_Start)
apply simp_all
apply (subgoal_tac "Cp (list2policyR (a # list)) p = Cp (list2policyR list) p")
apply (subgoal_tac "Cp (list2FWpolicy (rev list @ [a])) p = Cp (list2FWpolicy (rev list)) p")
apply simp
apply (rule CConcStart2)
apply simp
apply simp
apply (thin_tac "?S")
apply (case_tac list,simp_all)
apply (simp_all add: Cp.simps dom_def map_add_def)
done

lemma list2FWpolicys_eq: 
 "Filter \<noteq> []  \<Longrightarrow>
  Cp (list2policyR Filter) =  Cp (list2FWpolicy (rev Filter))"
by (rule ext, erule list2FWpolicys_eq_el)


lemma list2FWpolicys_eq_sym: 
 "Filter \<noteq> []  \<Longrightarrow>
  Cp (list2policyR (rev Filter)) =  Cp (list2FWpolicy Filter)"
by (metis list2FWpolicys_eq rev_is_Nil_conv rev_rev_ident)


lemma p_eq[rule_format]: "p \<noteq> [] \<longrightarrow> 
 list2policy (map Cp (rev p)) = Cp (list2FWpolicy p)"
by (metis l2polR_eq list2FWpolicys_eq_sym rev.simps(1) rev_rev_ident)


lemma p_eq2[rule_format]: "normalizePr x \<noteq> [] \<longrightarrow> 
  Cp (list2FWpolicy (normalizePr x)) = Cp x \<longrightarrow>
 list2policy (map Cp (rev (normalizePr x))) = Cp x"
apply (simp add: p_eq)
done



lemma p_eq2Q[rule_format]: "normalizePrQ x \<noteq> [] \<longrightarrow> 
  Cp (list2FWpolicy (normalizePrQ x)) = Cp x \<longrightarrow>
 list2policy (map Cp (rev (normalizePrQ x))) = Cp x"
apply (simp add: p_eq)
done


lemma list2listNMT[rule_format]: "x \<noteq> [] \<longrightarrow>map sem x \<noteq> []"
apply (case_tac x)
apply simp_all
done


lemma Norm_Distr2: 
 "r o_f ((P \<Otimes>\<^sub>2 (list2policy Q)) o d) = 
  (list2policy ((P \<Otimes>\<^sub>L Q) (op \<Otimes>\<^sub>2) r d))"
apply (rule ext, rule Norm_Distr_2)
done


lemma NATDistr: "\<lbrakk>N \<noteq> []; F = Cp (list2policyR N)\<rbrakk> \<Longrightarrow> 
  ((\<lambda> (x,y). x) o_f ((NAT \<Otimes>\<^sub>2 F) o (\<lambda> x. (x,x)))  = 
   (list2policy (  ((NAT \<Otimes>\<^sub>L (map Cp N)) (op \<Otimes>\<^sub>2) 
    (\<lambda> (x,y). x) (\<lambda> x. (x,x))))))"
apply simp
apply (simp add: l2polR_eq) 
apply (rule ext)
apply (rule Norm_Distr_2)
done


lemma C_eq_normalize_manual: 
"\<lbrakk>DenyAll \<in> set (policy2list p); 
 allNetsDistinct (policy2list p);
all_in_list (policy2list p) l\<rbrakk> \<Longrightarrow> 
Cp (list2FWpolicy (normalize_manual_orderPr p l)) = Cp p"
apply (simp add: normalize_manual_orderPr_def)
apply (rule C_eq_compile)
apply simp_all
done




lemma p_eq2_manualQ[rule_format]: "normalize_manual_orderPrQ x l \<noteq> [] \<longrightarrow> 
  Cp (list2FWpolicy (normalize_manual_orderPrQ x l)) = Cp x \<longrightarrow>
 list2policy (map Cp (rev (normalize_manual_orderPrQ x l))) = Cp x"
apply (simp add: p_eq)
done


lemma norm_notMT_manualQ: "DenyAll \<in> set (policy2list p) \<Longrightarrow> normalize_manual_orderPrQ p l \<noteq> []"
apply (simp add: normalize_manual_orderPrQ_def)
apply (rule rADnMT)
apply (rule idNMT)
apply (rule sepnMT)
apply (rule sortnMTQ)
apply (rule RS2_NMT)
apply (rule remDupsNMT) 
apply (rule RS3_NMT)
apply (rule DAiniD)
done


lemma C_eq_normalizePr_manualQ: 
"\<lbrakk>DenyAll \<in> set (policy2list p); 
 allNetsDistinct (policy2list p);
all_in_list (policy2list p) l\<rbrakk> \<Longrightarrow> 
Cp (list2FWpolicy (normalize_manual_orderPrQ p l)) = Cp p"
apply (simp add: normalize_manual_orderPrQ_def)
apply (rule C_eq_compileQ)
apply simp_all
done




lemma p_eq2_manual[rule_format]: "normalize_manual_orderPr x l \<noteq> [] \<longrightarrow> 
  Cp (list2FWpolicy (normalize_manual_orderPr x l)) = Cp x \<longrightarrow>
 list2policy (map Cp (rev (normalize_manual_orderPr x l))) = Cp x"
apply (simp add: p_eq)
done


lemma norm_notMT_manual: "DenyAll \<in> set (policy2list p) \<Longrightarrow> normalize_manual_orderPr p l \<noteq> []"
apply (simp add: normalize_manual_orderPr_def)
apply (rule rADnMT)
apply (rule idNMT)
apply (rule sepnMT)
apply (rule sortnMT)
apply (rule RS2_NMT)
apply (rule remDupsNMT) 
apply (rule RS3_NMT)
apply (rule DAiniD)
done





text{* As an example, how this theorems can be used for a concrete
normalisation instantiation. *}

lemma normalizePrNAT: "\<lbrakk>DenyAll \<in> set (policy2list Filter);
 allNetsDistinct (policy2list Filter); 
 all_in_list (policy2list Filter) (Nets_List Filter)\<rbrakk> \<Longrightarrow> 
 ((\<lambda> (x,y). x) o_f (((NAT \<Otimes>\<^sub>2 Cp Filter) o (\<lambda>x. (x,x)))))  = 
  list2policy ((NAT \<Otimes>\<^sub>L (map Cp (rev (normalizePr Filter)))) (op \<Otimes>\<^sub>2) (\<lambda> (x,y). x) (\<lambda> x. (x,x)))"
apply (rule NATDistr)
apply simp_all
apply (metis norm_notMT)
by (metis C_eq_normalizePr list2FWpolicys_eq_sym norm_notMT)





lemma domSimpl[simp]: "dom (Cp (A \<oplus> DenyAll)) = dom (Cp (DenyAll))"
apply (simp add: PLemmas)
done




end


