(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * IntegerPort.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: IntegerPort_TCPUDP.thy 8994 2010-07-09 13:59:10Z lukasbru $ *)
 
theory IntegerPort_TCPUDP
imports NetworkCore
begin

text{* A theory describing addresses which are modelled as a pair of
Integers - the first being the host address, the second the port
number.*}


 
type_synonym 
  address = int

type_synonym 
  port = int 

type_synonym 
  adr\<^isub>i\<^isub>p\<^isub>p = "address \<times> port \<times> protocol" 

instance protocol :: adr ..

defs (overloaded)
src_port_def: "src_port (x::(adr\<^isub>i\<^isub>p\<^isub>p,'\<beta>) packet)  \<equiv> (fst o snd o fst  o snd) x"
dest_port_def: "dest_port  (x::(adr\<^isub>i\<^isub>p\<^isub>p,'\<beta>) packet) \<equiv> (fst o snd o fst o  snd o snd) x"
subnet_of_def: "subnet_of (x::(adr\<^isub>i\<^isub>p\<^isub>p)) \<equiv> {{(a,b,c). a = fst x}}" 
src_protocol_def: "src_protocol (x::(adr\<^isub>i\<^isub>p\<^isub>p,'\<beta>) packet) \<equiv> (snd o snd o fst o snd) x"
dest_protocol_def: "dest_protocol (x::(adr\<^isub>i\<^isub>p\<^isub>p,'\<beta>) packet) \<equiv> (snd o snd o fst o snd o snd) x"


lemma src_port: "src_port (a,x::adr\<^isub>i\<^isub>p\<^isub>p,d,e) = fst (snd x)"
  by (simp add: src_port_def in_subnet)

lemma dest_port: "dest_port (a,d,x::adr\<^isub>i\<^isub>p\<^isub>p,e) = fst (snd x)"
  by (simp add: dest_port_def in_subnet)

text {* Common test constraints: *}


definition port_positive :: "(adr\<^isub>i\<^isub>p\<^isub>p,'b) packet \<Rightarrow> bool" where
 "port_positive x = (dest_port x > (0::port))"

definition fix_values :: "(adr\<^isub>i\<^isub>p\<^isub>p,DummyContent) packet \<Rightarrow> bool" where
 "fix_values x =  (src_port x = (1::port) \<and> src_protocol x = udp \<and> content x = data \<and> id x = 1)"


lemmas adr\<^isub>i\<^isub>p\<^isub>pLemmas = src_port dest_port src_port_def dest_port_def src_protocol_def dest_protocol_def subnet_of_def 

lemmas adr\<^isub>i\<^isub>p\<^isub>pTestConstraints = port_positive_def fix_values_def



end
