(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * TestGen.thy --- the core of the HOL-TestGen System.
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2012 Universit� Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id:$ *)

header {* The TestGen Core*}

theory TestGen imports Complex_Main
  uses  "config.sml"
        "kernel_ext/isabelle2009_kernel_patch.ML" 
        ("log_thy.sml")
        ("clocks.ML")
      	("TestEnv.ML")
        ("TestGen.ML")
        ("isar_setup.ML")
begin 

ML {*
val testgen_version = Int.toString(testgen_ver_major)^"."^Int.toString(testgen_ver_minor)^"."
                        ^Int.toString(testgen_ver_micro)^testgen_ver_tag

val version = "HOL-TestGen "^testgen_version^" , based on "^(Distribution.version); 
*}

use "log_thy.sml"
use "clocks.ML"


section{*General Framework for Unit-Test*}

definition THYP :: "bool \<Rightarrow> bool"
  where "THYP x = x"

definition PO :: "bool \<Rightarrow> bool"
  where "PO x = x"

definition RSF :: "bool"
  where "RSF = True" -- {* just to denote random solver failures *}

lemma  THYP_triv [simp]: "THYP True = True" by(simp add: THYP_def)

lemma  THYP_I    : "A \<Longrightarrow> THYP A" by(simp add: THYP_def)

lemma  THYP_D    : "THYP A \<Longrightarrow> A" by(simp add: THYP_def) 

lemma  THYP_E    : "\<lbrakk> THYP A; (A \<Longrightarrow> R) \<rbrakk> \<Longrightarrow> R" by(simp add: THYP_def)

lemma  THYP_spec : "THYP(\<forall>x. P x) \<Longrightarrow> THYP(P x)" by(simp add: THYP_def)

lemma  THYP_app1 : "\<lbrakk> THYP(A \<longrightarrow> B); A \<rbrakk> \<Longrightarrow> B" by(simp add: THYP_def) 
       (* to be used with etac in order to apply uniformity hyp *)

lemma  THYP_app1_rev : "\<lbrakk> A; THYP(A \<longrightarrow> B) \<rbrakk> \<Longrightarrow> B" by(simp add: THYP_def) 
       (* to be used with etac in order to apply uniformity hyp *)

lemma  THYP_app2 : "\<lbrakk> THYP(A); A \<Longrightarrow> B \<rbrakk> \<Longrightarrow> B" by(simp add: THYP_def) 
       (* to be used with RS in order to apply regularity hyp *)


lemma  THYP_E_reg :  "\<lbrakk> size x \<le> k \<Longrightarrow> B; THYP((k < size x) \<longrightarrow> B)\<rbrakk> \<Longrightarrow> B" 
       by(simp add: THYP_def, auto)
       (* to be used with rtac  in order to introduce regularity hyp *)

lemma  PO_I : "(PO P) \<and> Q \<Longrightarrow> P \<and> Q" by(simp add: PO_def) 

lemma  PO_grow : "PO(P \<and> Q) \<and> R \<Longrightarrow> (PO P) \<and> (Q \<and> R)" by(simp add: PO_def)

lemma  RSF_E : "\<lbrakk> RSF \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
       by(auto simp: RSF_def)
consts_code  "RSF" ("TestHarness.rsf")

(* We use Z3 as the default SMT solver. *)
declare [[ smt_solver = z3 ]]
declare [[ z3_options = "AUTO_CONFIG=false MBQI=false" ]]
declare [[smt_datatypes=true]]
declare [[smt_oracle = true]]

(* The following lemmas are for SMT solving. *)

lemma add_bound: "\<lbrakk> P \<and> (size x) <= k \<rbrakk> \<Longrightarrow> P" by(auto)

abbreviation "trigger == SMT.trigger"
abbreviation "pat == SMT.pat"

lemma trigger_intr: "(x = y) \<Longrightarrow> trigger [[pat (x)]] (x = y)" by(simp add: trigger_def)

section {* Tool Setup *}

use "TestEnv.ML"
use "TestGen.ML"
use "isar_setup.ML"

setup {* TestEnv.setup #> TestGen.setup *}

attribute_setup test = {*
  Scan.lift Args.name >> (fn name =>
      Thm.declaration_attribute (fn thm =>
        TestEnv.map_data (TestEnv.add_test_data (name, thm))))
*} "declare theorems for test case generation"

attribute_setup testgen_type_candidates = {*
  Scan.repeat1 Args.typ >> (fn Ts =>
      Thm.declaration_attribute (fn _ =>
        TestEnv.map_data (TestEnv.type_candidates_update Ts)))
*} "declare testgen type candidates"

method_setup gen_test_cases = {*
  Scan.lift (Scan.option (Parse.nat -- Parse.nat) -- Scan.repeat1 (Scan.unless (Scan.first clasimp_modifiers) Args.name)) --|
      Method.sections clasimp_modifiers >>
    (fn (params, prog) => fn ctxt =>
      let 
        val _ = LogThy.start ();
        val ctxt' =
          (case params of
            NONE => ctxt
          | SOME (depth, breadth) =>
              ctxt
              |> Config.put TestEnv.depth depth
              |> Config.put TestEnv.breadth breadth);
      in
        Method.SIMPLE_METHOD (CHANGED_PROP (TestGen.gen_test_case_tac ctxt' prog))
      end)
  *} "generate symbolic test cases"

method_setup gen_test_cases_old = {*
  Scan.lift (Scan.option (Parse.nat -- Parse.nat) -- Scan.repeat1 (Scan.unless (Scan.first clasimp_modifiers) Args.name)) --|
      Method.sections clasimp_modifiers >>
    (fn (params, prog) => fn ctxt =>
      let 
        val _ = LogThy.start ();
        val ctxt' =
          (case params of
            NONE => ctxt
          | SOME (depth, breadth) =>
              ctxt
              |> Config.put TestEnv.depth depth
              |> Config.put TestEnv.breadth breadth);
      in
        Method.SIMPLE_METHOD (CHANGED_PROP (TestGen.gen_test_case_old_tac ctxt' prog))
      end)
  *} "old variant of gen_test_cases"

end
