(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * TestGen_test.ML  --- testbed for test data/case generation routines 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: TestGen_test.ML 8982 2010-06-28 20:17:31Z makarius $ *)

open TestGen;

val thy = Main.thy;





(* map fst (Symtab.dest(Datatype.get_all thy)); 

   val it =
   ["*", "+", "Datatype.option", "List.char", "List.list", "List.nibble",
      "Numeral.bin", "Product_Type.unit", "bool", "nat"] 
 *)

get_exhaust_of "List.list" Main.thy;
get_nchotomy_of "List.list" Main.thy;



fun get_tycons thy =
    Symtab.dest(#tycons(Type.rep_tsig(#tsig(Sign.rep_sg(sign_of thy)))));

(*
get_tycons Main.thy;
val it =
   [("#prop", 0), ("*", 2), ("+", 2), ("Datatype.option", 1),
      ("Datatype_Universe.node", 2), ("Extraction.sumbool", 0),
      ("Fun.updbind", 0), ("Fun.updbinds", 0), ("HOL.case_syn", 0),
      ("HOL.cases_syn", 0), ("HOL.letbind", 0), ("HOL.letbinds", 0),
      ("IntDef.int", 0), ("List.char", 0), ("List.list", 1),
      ("List.lupdbind", 0), ("List.lupdbinds", 0), ("List.nibble", 0),
      ("Nat.ind", 0), ("Numeral.bin", 0), ("Product_Type.patterns", 0),
      ("Product_Type.tuple_args", 0), ("Product_Type.unit", 0),
      ("Record.field", 0), ("Record.field_type", 0),
      ("Record.field_types", 0), ("Record.fields", 0), ("Record.ident", 0),
      ("Record.update", 0), ("Record.updates", 0), ("any", 0), ("aprop", 0),
      ("args", 0), ("asms", 0), ("bool", 0), ("cargs", 0), ("classes", 0),
      ("dummy", 0), ("fun", 2), ("id", 0), ("idt", 0), ("idts", 0),
      ("index", 0), ("itself", 1), ("logic", 0), ("longid", 0), ("nat", 0),
      ("num", 0), ("num_const", 0), ("prop", 0), ("pttrn", 0), ("pttrns", 0),
      ("set", 1), ("sort", 0), ("struct", 0), ("tid", 0), ("tvar", 0),
      ("type", 0), ("types", 0), ("var", 0), ("xnum", 0), ("xstr", 0)]

  d.h. da sind auch syntaktische Typen drin.
*)



(* test ... >>>> *)
val [] = OldGoals.goal Main.thy 
              "!!P. (!!a list. [| y = [a]; list = [] |] ==> P) ==> \
              \     (!!a. [| y = [a] |] ==> P)";
val Const("all",_) $ (Abs(s,t,H)) = Logic.dest_goal(concl_of (OldGoals.topthm()));
val Const("==>",_) $ H1 $ H2 = subst_bounds([Free(s,t)],H);
(* <<<<<<<<<<<<< *)

prems_hyp_subst_tac Main.thy "list" H1;
val m = get_exhaust_of "List.list" Main.thy;
fun number S      = (1 upto length S) ~~ S;
        
val cands         = distinct(filter (fn(_,ty)=> isDataType thy ty) 
                                    (maps Logic.strip_params (prems_of m)))

val res = refine_data_separation_lemma false 2 cands m;
val res = nrefine_data_separation_lemma 1 m;
val res = nrefine_data_separation_lemma 2 m;
val res = nrefine_data_separation_lemma 3 m;
val res = nrefine_data_separation_lemma 4 m;
val res = nrefine_data_separation_lemma 10 m;


(* under AVL:
cd "examples/AVL";
use_thy "AVL";

val m = tree.exhaust;

nrefine_data_separation_lemma 3 tree.exhaust
nrefine_data_separation_lemma 5 tree.exhaust

cd "../..";
 *)

(* under RBT:

cd "examples/RBT";
use_thy "RBT";

val m = tree.exhaust;

nrefine_data_separation_lemma 3 tree.exhaust

nrefine_data_separation_lemma 5 tree.exhaust

nrefine_data_separation_lemma 8 tree.exhaust

cd "../..";



OldGoals.goal (theory"Testing")
"!!x. a * x = x * a ==> x = y";
by(uniformityI_tac 1);
val firstT = uresult();

OldGoals.goal (theory"Testing")
"!!x. a * x = x * a ==> THYP(x = y)";
by(uniformityI_tac 1);
(* oops *)

OldGoals.goal (theory"Testing")
"!!x. THYP(a * x = x * a) ==> x = y";
by(uniformityI_tac 1);
(* oops *)


 *)

