(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * TestGen_test.ML  --- testbed for test data/case generation routines 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: TestGen_test.ML 6644 2007-06-29 20:49:59Z brucker $ *)

theory TestGen_test
imports Testing
begin

section{* Setup for Testing *}

section{* Structural Splitting *}

datatype  Xcolor = R | B

find_theorems name:"Xcolor"

lemmas Xcolor_split'pos = 
       TestGen_test.Xcolor.split[THEN iffD2,OF conjI, OF impI, OF _ impI]

lemmas Xcolor_split'neg_0 =
       TestGen_test.Xcolor.split[THEN iffD1, THEN conjE]

text{* The form it should have ... *}
lemma Xcolor_split'neg:
"\<lbrakk>P (case x of R \<Rightarrow> f1 | B \<Rightarrow> f2);
     \<lbrakk>x = R; P f1\<rbrakk> \<Longrightarrow> Q; \<lbrakk>x = B; P f2\<rbrakk> \<Longrightarrow> Q\<rbrakk>
    \<Longrightarrow> Q"
  apply(insert TestGen_test.Xcolor.nchotomy)
  apply(erule_tac x=x in allE, auto)
done

lemma test: "(case x of R \<Rightarrow> f1 | B \<Rightarrow> f2) = X \<Longrightarrow> Q"
apply(erule Xcolor_split'neg) 
oops (* *)


find_theorems "~ _" name: "HOL"
find_theorems "If" name:"HOL"
ML{* @{term "If"}*}

lemmas split_if'pos = split_if[THEN iffD2, OF conjI, OF impI, OF _ impI]

lemma split_if'neg : 
  "\<lbrakk>P (if Q then x else y);   (\<not> (?Q \<and> \<not> ?P ?x \<or> \<not> ?Q \<and> \<not> ?P ?y)) "

ML{*  @{const_name If} *}



find_theorems name:"cases" name:"List"

ML{* 
val OldGoals. = Unsynchronized.ref(@{thm add_commute});
val ctxt = Unsynchronized.ref(@{context});
fun spy_tac pctxt thm = (OldGoals.topthm:=thm; ctxt:=pctxt; all_tac thm)
*}

ML{* print_depth 100; *}


section{* Queries to the Datatype Package *}

text{* We begin with some auxilliary test-cases ... *}


datatype  color = R | B

datatype 'a tree = E | T color "'a tree" "'a" "'a tree"

lemma "size (T R (T x E 0 E) 0 (T x E 0 (T x E 0 E))) = ?X"
apply simp
done

record 'a monoid_sig =
  times :: "'a => 'a => 'a"
  one :: 'a


datatype 'a forest = E1 | mk1 "'a forest list" | mk2 "'a forest2"
     and 'a forest2 = E2 | mk3 "'a forest"



ML{* map fst (Symtab.dest(Datatype.get_all @{theory}));  *}


text{*
This leads to the following list of registered types :
\begin{verbatim}
val it =
   ["*", "+", "nat", "bool", "List.char", "List.list", "List.nibble",
      "Option.option", "Predicate.seq", "Code_Eval.term", "Predicate.pred",
      "Typerep.typerep", "Product_Type.unit", "TestGen_test.tree",
      "Extraction.sumbool", "TestGen_test.color",
      "Code_Message.message_string"] : Table().key list
\end{verbatim}

Note that \verb+monoid+ is not registered here ...

*}


text{* Access to data-type package functionality: *}
ML{* 

TestGen.get_exhaust_of "List.list" @{theory};
TestGen.get_nchotomy_of "List.list" @{theory};

TestGen.get_exhaust_of "TestGen_test.tree" @{theory};
TestGen.get_nchotomy_of "TestGen_test.tree" @{theory};

TestGen.get_exhaust_of "*" @{theory};

*}
text{*
\begin{verbatim}
val it =
   "ALL v.
         v = [] |
         (EX a list.
             v = a # list)" : Thm.thm
val it =
   "[| ?y = E ==> ?P;
         !!color tree1 a tree2.
            ?y =
            T color tree1 a tree2
            ==> ?P |] ==> ?P" : Thm.thm
val it =
   "ALL v.
         v = E |
         (EX color tree1 a tree2.
             v =
             T color tree1 a
              tree2)" : Thm.thm
val it =
   "(!!a b.
          ?y = (a, b) ==> ?P)
      ==> ?P" : Thm.thm
\end{verbatim}
*}

text{* What I found out about the record package ... *}

ML{*
Record.print_records @{theory};
Record.get_extinjects @{theory};
Record.get_extension @{theory} "TestGen_test.monoid_sig"
*}

lemma recordtest : 
"reverse M = M (| times := \<lambda>x y. times M y x |)"
ML_prf{* val ctxt = @{context} 
         val ctxt_alt = Isar.context()
         val thm_state = Isar.goal() *}
ML_prf{*Logic.unprotect (concl_of thm_state) ;*}
apply(tactic "spy_tac (@{context})")
sorry 
ML{*Logic.unprotect (concl_of (!OldGoals.topthm)) ;*}

text{* Internally, this leads to a term representation as 
follows:
\verb+Free("M","('a,'b) TestGen_test.monoid_sig.monoid_sig_ext_type").
However, Datatype and Record Package are not very well integrated:
\verb+TestGen.get_exhaust_of "TestGen_test.monoid_sig.monoid_sig_ext" @ {theory}+
leads to an Option exception, since this type is unknown in the data-
type package's internal Repsig-Table 
\verb+#exhaust(the(Symtab.lookup(Datatype.get_all thy)+; 
however, it is stored inside the global signature representation
map (of course). *}

ML{*
fun get_tycons thy =
     NameSpace.dest_table(#types(Type.rep_tsig(#tsig(Sign.rep_sg thy))));
get_tycons @{theory};
*}


(*

val get_tycons = fn
:
   theory ->
   (string * ((Type.decl * Properties.T) * Library.serial)) list
val it =
   [("#prop", ((Nonterminal, []), 48)),
      ("*",
         ((LogicalType 2,
               [("line", "129"), ("column", "1"), ("offset", "2671"),
                  ("end_line", "129"), ("end_column", "8"),
                  ("end_offset", "2678"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Product_Type.thy"),
                  ("theory_name", "Product_Type")]),
            36356)),
      ("+",
         ((LogicalType 2,
               [("line", "25"), ("column", "1"), ("offset", "497"),
                  ("end_line", "25"), ("end_column", "8"),
                  ("end_offset", "504"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Sum_Type.thy"),
                  ("theory_name", "Sum_Type")]),
            33098)),
      ("Datatype.item",
         ((Abbreviation
                  (["'a"],
                     "('a, Product_Type.unit) Datatype.node
                      => bool",
                     false),
               [("line", "21"), ("column", "1"), ("offset", "603"),
                  ("end_line", "21"), ("end_column", "6"),
                  ("end_offset", "608"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Datatype.thy")]),
            117731)),
      ("Datatype.node",
         ((LogicalType 2,
               [("line", "14"), ("column", "1"), ("offset", "365"),
                  ("end_line", "14"), ("end_column", "8"),
                  ("end_offset", "372"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Datatype.thy"),
                  ("theory_name", "Datatype")]),
            117623)),
      ("Predicate.pred",
         ((LogicalType 1,
               [("line", "351"), ("column", "1"), ("offset", "9300"),
                  ("end_line", "351"), ("end_column", "9"),
                  ("end_offset", "9308"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Predicate.thy"),
                  ("theory_name", "Predicate")]),
            137552)),
      ("Predicate.seq",
         ((LogicalType 1,
               [("line", "499"), ("column", "1"), ("offset", "13005"),
                  ("end_line", "499"), ("end_column", "9"),
                  ("end_offset", "13013"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Predicate.thy"),
                  ("theory_name", "Predicate")]),
            139184)),
      ("any", ((Nonterminal, []), 47)), ("aprop", ((Nonterminal, []), 45)),
      ("args", ((Nonterminal, []), 39)), ("asms", ((Nonterminal, []), 46)),
      ("bool",
         ((LogicalType 0,
               [("line", "60"), ("column", "1"), ("offset", "1488"),
                  ("end_line", "60"), ("end_column", "9"),
                  ("end_offset", "1496"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy"),
                  ("theory_name", "HOL")]),
            972)),
      ("cargs", ((Nonterminal, []), 40)),
      ("HOL.case_syn",
         ((Nonterminal,
               [("line", "120"), ("column", "1"), ("offset", "3059"),
                  ("end_line", "120"), ("end_column", "13"),
                  ("end_offset", "3071"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy")]),
            1075)),
      ("HOL.cases_syn",
         ((Nonterminal,
               [("line", "120"), ("column", "1"), ("offset", "3059"),
                  ("end_line", "120"), ("end_column", "13"),
                  ("end_offset", "3071"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy")]),
            1076)),
      ("List.char",
         ((LogicalType 0,
               [("line", "3445"), ("column", "1"), ("offset", "101860"),
                  ("end_line", "3445"), ("end_column", "9"),
                  ("end_offset", "101868"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy"),
                  ("theory_name", "List")]),
            262415)),
      ("classes", ((Nonterminal, []), 38)),
      ("TestGen_test.color",
         ((LogicalType 0, [("theory_name", "TestGen_test")]), 394026)),
      ("Datatype.dtree",
         ((Abbreviation
                  (["'a", "'b"],
                     "('a, 'b) Datatype.node
                      => bool",
                     false),
               [("line", "21"), ("column", "1"), ("offset", "603"),
                  ("end_line", "21"), ("end_column", "6"),
                  ("end_offset", "608"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Datatype.thy")]),
            117733)),
      ("dummy", ((LogicalType 0, [("theory_name", "Pure")]), 23)),
      ("Record.field",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39406)),
      ("Record.field_type",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39404)),
      ("Record.field_types",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39405)),
      ("Record.fields",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39407)), ("float_const", ((Nonterminal, []), 50)),
      ("float_token", ((Nonterminal, []), 31)),
      ("fun", ((LogicalType 2, [("theory_name", "Pure")]), 20)),
      ("id", ((Nonterminal, []), 25)),
      ("Record.ident",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39403)), ("idt", ((Nonterminal, []), 43)),
      ("idts", ((Nonterminal, []), 44)),
      ("Nat.ind",
         ((LogicalType 0,
               [("line", "23"), ("column", "1"), ("offset", "523"),
                  ("end_line", "23"), ("end_column", "9"),
                  ("end_offset", "531"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Nat.thy"),
                  ("theory_name", "Nat")]),
            107840)),
      ("index", ((Nonterminal, []), 51)),
      ("Int.int",
         ((LogicalType 0,
               [("line", "26"), ("column", "1"), ("offset", "762"),
                  ("end_line", "26"), ("end_column", "8"),
                  ("end_offset", "769"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Int.thy"),
                  ("theory_name", "Int")]),
            171383)),
      ("itself", ((LogicalType 1, [("theory_name", "Pure")]), 22)),
      ("List.lc_qual",
         ((Nonterminal,
               [("line", "320"), ("column", "1"), ("offset", "9983"),
                  ("end_line", "320"), ("end_column", "13"),
                  ("end_offset", "9995"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            248543)),
      ("List.lc_quals",
         ((Nonterminal,
               [("line", "320"), ("column", "1"), ("offset", "9983"),
                  ("end_line", "320"), ("end_column", "13"),
                  ("end_offset", "9995"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            248544)),
      ("HOL.letbind",
         ((Nonterminal,
               [("line", "120"), ("column", "1"), ("offset", "3059"),
                  ("end_line", "120"), ("end_column", "13"),
                  ("end_offset", "3071"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy")]),
            1074)),
      ("HOL.letbinds",
         ((Nonterminal,
               [("line", "120"), ("column", "1"), ("offset", "3059"),
                  ("end_line", "120"), ("end_column", "13"),
                  ("end_offset", "3071"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy")]),
            1073)),
      ("List.list",
         ((LogicalType 1,
               [("line", "12"), ("column", "1"), ("offset", "214"),
                  ("end_line", "12"), ("end_column", "9"),
                  ("end_offset", "222"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy"),
                  ("theory_name", "List")]),
            245550)),
      ("logic", ((Nonterminal, []), 34)),
      ("longid", ((Nonterminal, []), 26)),
      ("List.lupdbind",
         ((Nonterminal,
               [("line", "46"), ("column", "1"), ("offset", "1353"),
                  ("end_line", "46"), ("end_column", "13"),
                  ("end_offset", "1365"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            246179)),
      ("List.lupdbinds",
         ((Nonterminal,
               [("line", "46"), ("column", "1"), ("offset", "1353"),
                  ("end_line", "46"), ("end_column", "13"),
                  ("end_offset", "1365"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            246178)),
      ("Map.maplet",
         ((Nonterminal,
               [("line", "58"), ("column", "1"), ("offset", "1493"),
                  ("end_line", "58"), ("end_column", "13"),
                  ("end_offset", "1505"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Map.thy")]),
            290740)),
      ("Map.maplets",
         ((Nonterminal,
               [("line", "58"), ("column", "1"), ("offset", "1493"),
                  ("end_line", "58"), ("end_column", "13"),
                  ("end_offset", "1505"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Map.thy")]),
            290739)),
      ("Code_Message.message_string",
         ((LogicalType 0,
               [("line", "13"), ("column", "1"), ("offset", "241"),
                  ("end_line", "13"), ("end_column", "9"),
                  ("end_offset", "249"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Code_Message.thy"),
                  ("theory_name", "Code_Message")]),
            284614)),
      ("TestGen_test.monoid_sig",
         ((Abbreviation
                  (["'a"],
                     "('a,
                       Product_Type.unit) TestGen_test.monoid_sig.monoid_sig_ext_type",
                     false),
               []),
            396231)),
      ("TestGen_test.monoid_sig.monoid_sig_ext_type",
         ((LogicalType 2, [("theory_name", "TestGen_test")]), 395829)),
      ("TestGen_test.monoid_sig_scheme",
         ((Abbreviation
                  (["'a", "'z"],
                     "('a,
                       'z) TestGen_test.monoid_sig.monoid_sig_ext_type",
                     false),
               []),
            396229)),
      ("nat",
         ((LogicalType 0,
               [("line", "45"), ("column", "1"), ("offset", "911"),
                  ("end_line", "45"), ("end_column", "8"),
                  ("end_offset", "918"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Nat.thy"),
                  ("theory_name", "Nat")]),
            108090)),
      ("List.nibble",
         ((LogicalType 0,
               [("line", "3431"), ("column", "1"), ("offset", "101348"),
                  ("end_line", "3431"), ("end_column", "9"),
                  ("end_offset", "101356"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy"),
                  ("theory_name", "List")]),
            260880)),
      ("num", ((Nonterminal, []), 30)),
      ("num_const", ((Nonterminal, []), 49)),
      ("Option.option",
         ((LogicalType 1,
               [("line", "11"), ("column", "1"), ("offset", "140"),
                  ("end_line", "11"), ("end_column", "9"),
                  ("end_offset", "148"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Option.thy"),
                  ("theory_name", "Option")]),
            143136)),
      ("Product_Type.patterns",
         ((Nonterminal,
               [("line", "163"), ("column", "1"), ("offset", "3564"),
                  ("end_line", "163"), ("end_column", "13"),
                  ("end_offset", "3576"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Product_Type.thy")]),
            36536)),
      ("prop", ((LogicalType 0, [("theory_name", "Pure")]), 21)),
      ("pttrn", ((Nonterminal, []), 41)),
      ("pttrns", ((Nonterminal, []), 42)),
      ("set",
         ((Abbreviation (["'a"], "'a => bool", false),
               [("line", "18"), ("column", "1"), ("offset", "272"),
                  ("end_line", "18"), ("end_column", "6"),
                  ("end_offset", "277"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Set.thy")]),
            20546)), ("sort", ((Nonterminal, []), 37)),
      ("List.string",
         ((Abbreviation ([], "List.char List.list", false),
               [("line", "3489"), ("column", "1"), ("offset", "103237"),
                  ("end_line", "3489"), ("end_column", "6"),
                  ("end_offset", "103242"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            264108)), ("struct", ((Nonterminal, []), 52)),
      ("Extraction.sumbool",
         ((LogicalType 0,
               [("line", "50"), ("column", "1"), ("offset", "1735"),
                  ("end_line", "50"), ("end_column", "9"),
                  ("end_offset", "1743"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Extraction.thy"),
                  ("theory_name", "Extraction")]),
            150691)),
      ("Code_Eval.term",
         ((LogicalType 0,
               [("line", "15"), ("column", "1"), ("offset", "285"),
                  ("end_line", "15"), ("end_column", "9"),
                  ("end_offset", "293"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Code_Eval.thy"),
                  ("theory_name", "Code_Eval")]),
            288124)),
      ("tid", ((Nonterminal, []), 28)),
      ("TestGen_test.tree",
         ((LogicalType 1, [("theory_name", "TestGen_test")]), 394962)),
      ("Product_Type.tuple_args",
         ((Nonterminal,
               [("line", "163"), ("column", "1"), ("offset", "3564"),
                  ("end_line", "163"), ("end_column", "13"),
                  ("end_offset", "3576"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Product_Type.thy")]),
            36535)), ("tvar", ((Nonterminal, []), 29)),
      ("type", ((Nonterminal, []), 35)),
      ("Typerep.typerep",
         ((LogicalType 0,
               [("line", "11"), ("column", "1"), ("offset", "182"),
                  ("end_line", "11"), ("end_column", "9"),
                  ("end_offset", "190"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Typerep.thy"),
                  ("theory_name", "Typerep")]),
            285665)),
      ("types", ((Nonterminal, []), 36)),
      ("Product_Type.unit",
         ((LogicalType 0,
               [("line", "42"), ("column", "1"), ("offset", "963"),
                  ("end_line", "42"), ("end_column", "8"),
                  ("end_offset", "970"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Product_Type.thy"),
                  ("theory_name", "Product_Type")]),
            35631)),
      ("Record.update",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39408)),
      ("Record.updates",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39409)),
      ("Fun.updbind",
         ((Nonterminal,
               [("line", "358"), ("column", "1"), ("offset", "9421"),
                  ("end_line", "358"), ("end_column", "13"),
                  ("end_offset", "9433"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Fun.thy")]),
            32481)),
      ("Fun.updbinds",
         ((Nonterminal,
               [("line", "358"), ("column", "1"), ("offset", "9421"),
                  ("end_line", "358"), ("end_column", "13"),
                  ("end_offset", "9433"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Fun.thy")]),
            32480)), ("var", ((Nonterminal, []), 27)),
      ("xnum", ((Nonterminal, []), 32)), ("xstr", ((Nonterminal, []), 33)),
      ("Map.~=>",
         ((Abbreviation
                  (["'a", "'b"],
                     "'a => 'b Option.option",
                     false),
               [("line", "14"), ("column", "1"), ("offset", "261"),
                  ("end_line", "14"), ("end_column", "6"),
                  ("end_offset", "266"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Map.thy")]),
            290482))]
: (string * ((Type.decl * Properties.T) * Library.serial)) list

val it =
   [("#prop", ((Nonterminal, []), 48)),
      ("*",
         ((LogicalType 2,
               [("line", "129"), ("column", "1"), ("offset", "2671"),
                  ("end_line", "129"), ("end_column", "8"),
                  ("end_offset", "2678"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Product_Type.thy"),
                  ("theory_name", "Product_Type")]),
            36356)),
      ("+",
         ((LogicalType 2,
               [("line", "25"), ("column", "1"), ("offset", "497"),
                  ("end_line", "25"), ("end_column", "8"),
                  ("end_offset", "504"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Sum_Type.thy"),
                  ("theory_name", "Sum_Type")]),
            33098)),
      ("Datatype.item",
         ((Abbreviation
                  (["'a"],
                     "('a, Product_Type.unit) Datatype.node
                      => bool",
                     false),
               [("line", "21"), ("column", "1"), ("offset", "603"),
                  ("end_line", "21"), ("end_column", "6"),
                  ("end_offset", "608"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Datatype.thy")]),
            117731)),
      ("Datatype.node",
         ((LogicalType 2,
               [("line", "14"), ("column", "1"), ("offset", "365"),
                  ("end_line", "14"), ("end_column", "8"),
                  ("end_offset", "372"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Datatype.thy"),
                  ("theory_name", "Datatype")]),
            117623)),
      ("Predicate.pred",
         ((LogicalType 1,
               [("line", "351"), ("column", "1"), ("offset", "9300"),
                  ("end_line", "351"), ("end_column", "9"),
                  ("end_offset", "9308"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Predicate.thy"),
                  ("theory_name", "Predicate")]),
            137552)),
      ("Predicate.seq",
         ((LogicalType 1,
               [("line", "499"), ("column", "1"), ("offset", "13005"),
                  ("end_line", "499"), ("end_column", "9"),
                  ("end_offset", "13013"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Predicate.thy"),
                  ("theory_name", "Predicate")]),
            139184)),
      ("any", ((Nonterminal, []), 47)), ("aprop", ((Nonterminal, []), 45)),
      ("args", ((Nonterminal, []), 39)), ("asms", ((Nonterminal, []), 46)),
      ("bool",
         ((LogicalType 0,
               [("line", "60"), ("column", "1"), ("offset", "1488"),
                  ("end_line", "60"), ("end_column", "9"),
                  ("end_offset", "1496"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy"),
                  ("theory_name", "HOL")]),
            972)),
      ("cargs", ((Nonterminal, []), 40)),
      ("HOL.case_syn",
         ((Nonterminal,
               [("line", "120"), ("column", "1"), ("offset", "3059"),
                  ("end_line", "120"), ("end_column", "13"),
                  ("end_offset", "3071"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy")]),
            1075)),
      ("HOL.cases_syn",
         ((Nonterminal,
               [("line", "120"), ("column", "1"), ("offset", "3059"),
                  ("end_line", "120"), ("end_column", "13"),
                  ("end_offset", "3071"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy")]),
            1076)),
      ("List.char",
         ((LogicalType 0,
               [("line", "3445"), ("column", "1"), ("offset", "101860"),
                  ("end_line", "3445"), ("end_column", "9"),
                  ("end_offset", "101868"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy"),
                  ("theory_name", "List")]),
            262415)),
      ("classes", ((Nonterminal, []), 38)),
      ("Datatype.dtree",
         ((Abbreviation
                  (["'a", "'b"],
                     "('a, 'b) Datatype.node
                      => bool",
                     false),
               [("line", "21"), ("column", "1"), ("offset", "603"),
                  ("end_line", "21"), ("end_column", "6"),
                  ("end_offset", "608"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Datatype.thy")]),
            117733)),
      ("dummy", ((LogicalType 0, [("theory_name", "Pure")]), 23)),
      ("Record.field",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39406)),
      ("Record.field_type",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39404)),
      ("Record.field_types",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39405)),
      ("Record.fields",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39407)), ("float_const", ((Nonterminal, []), 50)),
      ("float_token", ((Nonterminal, []), 31)),
      ("fun", ((LogicalType 2, [("theory_name", "Pure")]), 20)),
      ("id", ((Nonterminal, []), 25)),
      ("Record.ident",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39403)), ("idt", ((Nonterminal, []), 43)),
      ("idts", ((Nonterminal, []), 44)),
      ("Nat.ind",
         ((LogicalType 0,
               [("line", "23"), ("column", "1"), ("offset", "523"),
                  ("end_line", "23"), ("end_column", "9"),
                  ("end_offset", "531"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Nat.thy"),
                  ("theory_name", "Nat")]),
            107840)),
      ("index", ((Nonterminal, []), 51)),
      ("Int.int",
         ((LogicalType 0,
               [("line", "26"), ("column", "1"), ("offset", "762"),
                  ("end_line", "26"), ("end_column", "8"),
                  ("end_offset", "769"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Int.thy"),
                  ("theory_name", "Int")]),
            171383)),
      ("itself", ((LogicalType 1, [("theory_name", "Pure")]), 22)),
      ("List.lc_qual",
         ((Nonterminal,
               [("line", "320"), ("column", "1"), ("offset", "9983"),
                  ("end_line", "320"), ("end_column", "13"),
                  ("end_offset", "9995"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            248543)),
      ("List.lc_quals",
         ((Nonterminal,
               [("line", "320"), ("column", "1"), ("offset", "9983"),
                  ("end_line", "320"), ("end_column", "13"),
                  ("end_offset", "9995"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            248544)),
      ("HOL.letbind",
         ((Nonterminal,
               [("line", "120"), ("column", "1"), ("offset", "3059"),
                  ("end_line", "120"), ("end_column", "13"),
                  ("end_offset", "3071"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy")]),
            1074)),
      ("HOL.letbinds",
         ((Nonterminal,
               [("line", "120"), ("column", "1"), ("offset", "3059"),
                  ("end_line", "120"), ("end_column", "13"),
                  ("end_offset", "3071"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/HOL.thy")]),
            1073)),
      ("List.list",
         ((LogicalType 1,
               [("line", "12"), ("column", "1"), ("offset", "214"),
                  ("end_line", "12"), ("end_column", "9"),
                  ("end_offset", "222"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy"),
                  ("theory_name", "List")]),
            245550)),
      ("logic", ((Nonterminal, []), 34)),
      ("longid", ((Nonterminal, []), 26)),
      ("List.lupdbind",
         ((Nonterminal,
               [("line", "46"), ("column", "1"), ("offset", "1353"),
                  ("end_line", "46"), ("end_column", "13"),
                  ("end_offset", "1365"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            246179)),
      ("List.lupdbinds",
         ((Nonterminal,
               [("line", "46"), ("column", "1"), ("offset", "1353"),
                  ("end_line", "46"), ("end_column", "13"),
                  ("end_offset", "1365"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            246178)),
      ("Map.maplet",
         ((Nonterminal,
               [("line", "58"), ("column", "1"), ("offset", "1493"),
                  ("end_line", "58"), ("end_column", "13"),
                  ("end_offset", "1505"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Map.thy")]),
            290740)),
      ("Map.maplets",
         ((Nonterminal,
               [("line", "58"), ("column", "1"), ("offset", "1493"),
                  ("end_line", "58"), ("end_column", "13"),
                  ("end_offset", "1505"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Map.thy")]),
            290739)),
      ("Code_Message.message_string",
         ((LogicalType 0,
               [("line", "13"), ("column", "1"), ("offset", "241"),
                  ("end_line", "13"), ("end_column", "9"),
                  ("end_offset", "249"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Code_Message.thy"),
                  ("theory_name", "Code_Message")]),
            284614)),
      ("nat",
         ((LogicalType 0,
               [("line", "45"), ("column", "1"), ("offset", "911"),
                  ("end_line", "45"), ("end_column", "8"),
                  ("end_offset", "918"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Nat.thy"),
                  ("theory_name", "Nat")]),
            108090)),
      ("List.nibble",
         ((LogicalType 0,
               [("line", "3431"), ("column", "1"), ("offset", "101348"),
                  ("end_line", "3431"), ("end_column", "9"),
                  ("end_offset", "101356"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy"),
                  ("theory_name", "List")]),
            260880)),
      ("num", ((Nonterminal, []), 30)),
      ("num_const", ((Nonterminal, []), 49)),
      ("Option.option",
         ((LogicalType 1,
               [("line", "11"), ("column", "1"), ("offset", "140"),
                  ("end_line", "11"), ("end_column", "9"),
                  ("end_offset", "148"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Option.thy"),
                  ("theory_name", "Option")]),
            143136)),
      ("Product_Type.patterns",
         ((Nonterminal,
               [("line", "163"), ("column", "1"), ("offset", "3564"),
                  ("end_line", "163"), ("end_column", "13"),
                  ("end_offset", "3576"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Product_Type.thy")]),
            36536)),
      ("prop", ((LogicalType 0, [("theory_name", "Pure")]), 21)),
      ("pttrn", ((Nonterminal, []), 41)),
      ("pttrns", ((Nonterminal, []), 42)),
      ("set",
         ((Abbreviation (["'a"], "'a => bool", false),
               [("line", "18"), ("column", "1"), ("offset", "272"),
                  ("end_line", "18"), ("end_column", "6"),
                  ("end_offset", "277"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Set.thy")]),
            20546)), ("sort", ((Nonterminal, []), 37)),
      ("List.string",
         ((Abbreviation ([], "List.char List.list", false),
               [("line", "3489"), ("column", "1"), ("offset", "103237"),
                  ("end_line", "3489"), ("end_column", "6"),
                  ("end_offset", "103242"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/List.thy")]),
            264108)), ("struct", ((Nonterminal, []), 52)),
      ("Extraction.sumbool",
         ((LogicalType 0,
               [("line", "50"), ("column", "1"), ("offset", "1735"),
                  ("end_line", "50"), ("end_column", "9"),
                  ("end_offset", "1743"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Extraction.thy"),
                  ("theory_name", "Extraction")]),
            150691)),
      ("Code_Eval.term",
         ((LogicalType 0,
               [("line", "15"), ("column", "1"), ("offset", "285"),
                  ("end_line", "15"), ("end_column", "9"),
                  ("end_offset", "293"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Code_Eval.thy"),
                  ("theory_name", "Code_Eval")]),
            288124)),
      ("tid", ((Nonterminal, []), 28)),
      ("Product_Type.tuple_args",
         ((Nonterminal,
               [("line", "163"), ("column", "1"), ("offset", "3564"),
                  ("end_line", "163"), ("end_column", "13"),
                  ("end_offset", "3576"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Product_Type.thy")]),
            36535)), ("tvar", ((Nonterminal, []), 29)),
      ("type", ((Nonterminal, []), 35)),
      ("Typerep.typerep",
         ((LogicalType 0,
               [("line", "11"), ("column", "1"), ("offset", "182"),
                  ("end_line", "11"), ("end_column", "9"),
                  ("end_offset", "190"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Typerep.thy"),
                  ("theory_name", "Typerep")]),
            285665)),
      ("types", ((Nonterminal, []), 36)),
      ("Product_Type.unit",
         ((LogicalType 0,
               [("line", "42"), ("column", "1"), ("offset", "963"),
                  ("end_line", "42"), ("end_column", "8"),
                  ("end_offset", "970"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Product_Type.thy"),
                  ("theory_name", "Product_Type")]),
            35631)),
      ("Record.update",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39408)),
      ("Record.updates",
         ((Nonterminal,
               [("line", "28"), ("column", "1"), ("offset", "560"),
                  ("end_line", "28"), ("end_column", "13"),
                  ("end_offset", "572"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Record.thy")]),
            39409)),
      ("Fun.updbind",
         ((Nonterminal,
               [("line", "358"), ("column", "1"), ("offset", "9421"),
                  ("end_line", "358"), ("end_column", "13"),
                  ("end_offset", "9433"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Fun.thy")]),
            32481)),
      ("Fun.updbinds",
         ((Nonterminal,
               [("line", "358"), ("column", "1"), ("offset", "9421"),
                  ("end_line", "358"), ("end_column", "13"),
                  ("end_offset", "9433"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Fun.thy")]),
            32480)), ("var", ((Nonterminal, []), 27)),
      ("xnum", ((Nonterminal, []), 32)), ("xstr", ((Nonterminal, []), 33)),
      ("Map.~=>",
         ((Abbreviation
                  (["'a", "'b"],
                     "'a => 'b Option.option",
                     false),
               [("line", "14"), ("column", "1"), ("offset", "261"),
                  ("end_line", "14"), ("end_column", "6"),
                  ("end_offset", "266"),
                  ("file",
                     "/usr/local/isabelle/Isabelle2009/src/HOL/Map.thy")]),
            290482))]
: (string * ((Type.decl * Properties.T) * Library.serial)) list
*)

text{* This infrastructure was formerly 
       ( has been refined dramatically !!! ): *}

(* 
val it =
   [("#prop", 0), ("*", 2), ("+", 2), ("Datatype.option", 1),
      ("Datatype_Universe.node", 2), ("Extraction.sumbool", 0),
      ("Fun.updbind", 0), ("Fun.updbinds", 0), ("HOL.case_syn", 0),
      ("HOL.cases_syn", 0), ("HOL.letbind", 0), ("HOL.letbinds", 0),
      ("IntDef.int", 0), ("List.char", 0), ("List.list", 1),
      ("List.lupdbind", 0), ("List.lupdbinds", 0), ("List.nibble", 0),
      ("Nat.ind", 0), ("Numeral.bin", 0), ("Product_Type.patterns", 0),
      ("Product_Type.tuple_args", 0), ("Product_Type.unit", 0),
      ("Record.field", 0), ("Record.field_type", 0),
      ("Record.field_types", 0), ("Record.fields", 0), ("Record.ident", 0),
      ("Record.update", 0), ("Record.updates", 0), ("any", 0), ("aprop", 0),
      ("args", 0), ("asms", 0), ("bool", 0), ("cargs", 0), ("classes", 0),
      ("dummy", 0), ("fun", 2), ("id", 0), ("idt", 0), ("idts", 0),
      ("index", 0), ("itself", 1), ("logic", 0), ("longid", 0), ("nat", 0),
      ("num", 0), ("num_const", 0), ("prop", 0), ("pttrn", 0), ("pttrns", 0),
      ("set", 1), ("sort", 0), ("struct", 0), ("tid", 0), ("tvar", 0),
      ("type", 0), ("types", 0), ("var", 0), ("xnum", 0), ("xstr", 0)]

  d.h. da sind auch syntaktische Typen drin.
*)

section{* Simulation of a Data-Separation theorem.*}

lemma T1:
 "!!P. (!!a list. [| y = [a]; list = [] |] ==> P) 
   ==> (!!a. [| y = [a] |] ==> P)"
apply(tactic "spy_tac (@{context})")
sorry


ML{*
val Const("all",_) $ (Abs(s,t,H)) = Logic.unprotect (concl_of (!OldGoals.topthm)) ;
val Const("==>",_) $ H1 $ H2 = subst_bounds([Free(s,t)],H);
(* val thy = theory_of_thm (!OldGoals.topthm) *)
val thy = ProofContext.theory_of (!ctxt) 

*}

ML{* (* open TestGen --- equiv:  Just redeclare: *)

fun thm |$> tac =
    case Seq.pull (tac thm) of
         SOME (m,_) => m
        |_ => error"|$> : TACTIC FAILED"

*}

ML{*
val m = TestGen.get_exhaust_of "List.list" @{theory};
fun number S      = (1 upto length S) ~~ S;
        
val cands         = distinct (op =)
                             (filter (fn(_,ty)=> TestGen.isDataType thy ty) 
                                    (maps Logic.strip_params (prems_of m)))

*}
ML{*
val css = ( Clasimp.clasimpset_of(!ctxt))
val res = TestGen.refine_data_separation_lemma (!ctxt)css  false 2 cands m;
*}
text{*New Target: low-level simulation of \verb+refine_data_separation_lemma+ *}
ML{*
(* New Target: *)
TestGen.refine_data_separation_lemma (!ctxt) css  false 2 cands m;

val pctxt = (!ctxt)
val clasimp = css
val bal = false
val prem_no = 2
val cands = cands
val thm = TestGen.get_exhaust_of "List.list" @{theory};

*}

ML{*
    val thy    = ProofContext.theory_of pctxt
    val prem   = Logic.nth_prem(prem_no, prop_of thm)
    val params = Logic.strip_params(prem)
    val cands' = if bal then
                    let val term   = HOLogic.dest_Trueprop(hd
                                            (Logic.strip_assums_hyp prem))
                        val (k,mp) = TestGen.max_depth(Term.subst_bounds
                                            ((rev(map Free params)),term))
                    in filter(fn p => forall(fn p'=> not(p=p')) (mp)) cands end
                 else cands;

    val var_no = find_index (fn(s,t)=> exists(fn (name,ty)=>
                                             s = name andalso t = ty) cands')
                             params
    val (name,ty)  = if var_no > ~1 then List.nth(params,var_no)
                     else raise (ERROR("no param found"))
    val Type(tn,_) = ty;
    val m          = TestGen.get_exhaust_of tn thy;
    val n          = m RSN (prem_no,thm);
    (* pre: all occurrences of ?indname t1 ... tn have the same
            arguments and types  and occur in a critical pattern : 
             ?indname t1 ... tn = C x1 ... xn 
        ... *)
    fun find_mvar_pat indname goal_no thm =
        let val goal = Logic.get_goal thm goal_no
            val params = Logic.strip_params goal
            fun  cc(Const("==>", _) 
                    $ (Const("Trueprop", _) 
                       $ (Const(@{const_name HOL.eq},t) $ A $ B)) 
                    $ C) = (case strip_comb(subst_bounds(
                                     rev(map Free params), A)) of
                             (Var(s,_),S) =>(if s = indname 
                                             then (map dest_Free S, head_of B)
                                             else cc(C))
                            | (_, _)  => cc C)
               | cc _ = error("find_mvar_pat: No pattern found")        
        in cc (strip_all_body goal) end
    fun lifter name S =
        let fun ll k [] = if 0 <=k then Bound(k) 
                          else error"find_mvar:impossible lifter"
               |ll k ((s,t)::R) = 
                      if s = name then Abs(s,t, ll 0 R)
                      else Abs(s,t, ll (k+1) R)
        in ll (~(length S)) S end
    fun typ_of_lifter (name:string) S = 
        let fun tll S [] = error"find_mvar:impossible type for lifter"
               |tll S ((s,t)::R) = 
                       if name = s 
                       then ((rev S)@(t::(map snd R)), t)
                       else tll (t::S) R
        in  (op --->)(tll [] S) end   
    val (lifter_params, constr) = find_mvar_pat ("y",1) prem_no (prop_of n)
    val subst0 = lifter "list" lifter_params 
    val ty_lift = fastype_of subst0
    val ty_constr = fastype_of constr
    val subst = cterm_of thy subst0 
    val ty_lift_body = body_type ty_lift
    val ty_lift_constr = body_type ty_constr
    val tyenv = Sign.typ_match thy (ty_lift_constr,ty_lift_body) Vartab.empty
    fun tysubs2ctyps(x,(s,t)) = (ctyp_of thy (TVar(x,s)),ctyp_of thy t)
    val tsubst = map tysubs2ctyps (Vartab.dest tyenv)
    val ()    = (TestGen.changed:=true)
    val var = cterm_of thy (Var(("y",1),ty_lift))
    val result = (instantiate (tsubst, [(var,subst)]) n)
                 |$> (ALLGOALS(fn n => TRY(hyp_subst_tac n))) 
                 |$> prune_params_tac
*}

ML{*standard result*}



ML{*
m;
val res = TestGen.nrefine_data_separation_lemma (!ctxt) css 1 m;
val res = TestGen.nrefine_data_separation_lemma (!ctxt) css 3 m;
val res = TestGen.nrefine_data_separation_lemma (!ctxt) css 5 m;
val res = TestGen.nrefine_data_separation_lemma (!ctxt) css 10 m;

*}



lemma T2:
 "sort (x::'a::linorder list) = y"
apply(tactic "TestGen.data_separation_tac @{context} @{clasimpset} 4  1")
sorry

text{* Test for correct context adaption*}


test_spec "sort l = PUT  l"
(*
  apply(tactic "TestGen.gen_test_case_tac @{context} \"PUT\" ")
*)
  apply(gen_test_cases 4 1 "PUT")

(* The analysis of prior versions revealed that the
unfolding of sort, ins, and the resulting case-splits 
did happen in version 2005, did not happen in 2008 and
where the behaviour is pretty much the same under 2009
now. Probably did the use of the old goalpackages in 
new context management alrady result in a few uncatched 
ugs \<dots> *)

(*
  apply(tactic "TestGen.ALLCASES 
                       (TestGen.DEPTH_SPLIT 
                           (TestGen.unfold ((thms\"sort.simps\") @ 
                                            (thms\"ins.simps\")) 
                            THEN' 
                            TestGen.if_splitter @{claset}))")

   apply(tactic "TestGen.ALLCASES(asm_full_simp_tac @{simpset})")

txt {* 
      which leads to the following test partitioning (excerpt):
      @{subgoals [display, goals_limit=10]}
     *}
*)
  store_test_thm "test_insertion_sort'"


primrec  ins :: "('a::linorder) \<Rightarrow> 'a list \<Rightarrow> 'a list"
where   "ins x [] = [x]" |
        "ins x (y#ys) = (if (x < y) then x#y#ys else (y#(ins x ys)))"
primrec  sort:: "('a::linorder) list \<Rightarrow> 'a list"
where   "sort [] = [] " |
        "sort (x#xs) = ins x (sort xs)"


test_spec "sort l = PUT  l"
(*
  apply(tactic "TestGen.gen_test_case_tac @{context} \"PUT\" ")
*)
  apply(gen_test_cases 5 1 "PUT")

(* The analysis of prior versions revealed that the
unfolding of sort, ins, and the resulting case-splits 
did happen in version 2005, did not happen in 2008 and
where the behaviour is pretty much the same under 2009
now. Probably did the use of the old goalpackages in 
new context management alrady result in a few uncatched 
ugs \<dots> *)

  apply(tactic "TestGen.ALLCASES 
                       (TestGen.DEPTH_SPLIT 
                           (TestGen.unfold ((thms\"sort.simps\") @ 
                                            (thms\"ins.simps\")) 
                            THEN' 
                            TestGen.if_splitter @{claset}))")

   apply(tactic "TestGen.ALLCASES(asm_full_simp_tac @{simpset})")

txt {* 
      which leads to the following test partitioning (excerpt):
      @{subgoals [display, goals_limit=10]}
     *}
  store_test_thm "test_insertion_sort"


lemma hh :
"\<not>(X2X18 < X1X16 \<and> X4X22 < X1X16 \<and> X4X22 < X3X20 \<and>
 X4X22 < X3X20 \<and> X4X22 < X1X16 \<and> X3X20 < X1X16 \<and> 
 X4X22 < X2X18 \<and> X4X22 < X3X20 \<and> X4X22 < X2X18 \<and>
 X3X20 < X2X18)"
quickcheck
oops


ML{*

*}

(* under RBT:

cd "examples/RBT";
use_thy "RBT";

val m = tree.exhaust;

nrefine_data_separation_lemma 3 tree.exhaust

nrefine_data_separation_lemma 5 tree.exhaust

nrefine_data_separation_lemma 8 tree.exhaust

cd "../..";



OldGoals.goal (theory"Testing")
"!!x. a * x = x * a ==> x = y";
by(uniformityI_tac 1);
val firstT = uresult();

OldGoals.goal (theory"Testing")
"!!x. a * x = x * a ==> THYP(x = y)";
by(uniformityI_tac 1);


OldGoals.goal (theory"Testing")
"!!x. THYP(a * x = x * a) ==> x = y";
by(uniformityI_tac 1);



 *)

end
