(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * harness_demo.sml --- example using the HOL-TestGen test harness
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007 ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: harness_demo.sml 8022 2008-05-22 14:36:16Z brucker $ *)

(* load test harness *)
use "harness.sml";

(* define function under test *)
fun fut x y = x div y;

(**************************************************************)
(* run test harness without observing the return value of fut *)
(**************************************************************)
(* define stub for geting return value and converting the 
   return value to string *)
fun retval () = NONE;
fun toString a = "";


(* pre evaluates to true and post evaluates to true *)
val pre_0   = [fn () => true, fn () => 5 = 5]; 
val post_0  = fn () => (fut  10 2)  = 5;
val ret_0   = (TestHarness.check retval pre_0 post_0);
val _       = TestHarness.printResult toString 0 ret_0;

(* pre evaluates to true and post evaluates to false *)
val pre_1   = [fn () => true, fn () => 5 = 5]; 
val post_1  = fn () => (fut  10 2)  = 3;
val ret_1   = (TestHarness.check retval pre_1 post_1);
val _       = TestHarness.printResult toString 1 ret_1;

(* pre evaluates to true and post throws exception*)
val pre_2   = [fn () => true, fn () => 5 = 5]; 
val post_2  = fn () => (fut  10 0)  = 3;
val ret_2   = (TestHarness.check retval pre_2 post_2);
val _       = TestHarness.printResult toString 2 ret_2;

(* pre evaluates to false and post evaluates to true *)
val pre_3   = [fn () => false, fn () => 5 = 5];
val post_3  = fn () => (fut 12 2) = 6;
val ret_3   = (TestHarness.check retval pre_3 post_3);
val _       = TestHarness.printResult toString 3 ret_3;

(* pre evaluates to false and post evaluates to false *)
val pre_4   = [fn () => false, fn () => 5 = 5]; 
val post_4  = fn () => (fut  10 2)  = 3;
val ret_4   = (TestHarness.check retval pre_4 post_4);
val _       = TestHarness.printResult toString 4 ret_4;

(* pre evaluates to false and post throws exception*)
val pre_5   = [fn () => false, fn () => 5 = 5]; 
val post_5  = fn () => (fut  10 0)  = 3;
val ret_5   = (TestHarness.check retval pre_5 post_5);
val _       = TestHarness.printResult toString 5 ret_5;

(* pre throws exception and post evaluates to true *)
val pre_6   = [fn () => (fut 10 0)=1, fn () => 5 = 5];
val post_6  = fn () => (fut 12 2) = 6;
val ret_6   = (TestHarness.check retval pre_6 post_6);
val _       = TestHarness.printResult toString 6 ret_6;

(* pre throws exception and post evaluates to false *)
val pre_7   = [fn () => (fut 10 0) = 1, fn () => 5 = 5]; 
val post_7  = fn () => (fut  10 2)  = 3;
val ret_7   = (TestHarness.check retval pre_7 post_7);
val _       = TestHarness.printResult toString 7 ret_7;

(* pre throws exception and post throws exception*)
val pre_8   = [fn () => (fut 10 0) =1, fn () => 5 = 5]; 
val post_8  = fn () => (fut  10 0)  = 3;
val ret_8   = (TestHarness.check retval pre_8 post_8);
val _       = TestHarness.printResult toString 8 ret_8;
    
val results = [ret_0,ret_1,ret_2,ret_3,ret_4,ret_5,ret_6,ret_7,ret_8];
val _       = TestHarness.printList toString results;



(**************************************************************)
(* run test harness with observing the return value of fut    *)
(**************************************************************)
(* define functions for getting return value and converting the 
   return value to string *)
val return = ref 0;
fun eval (f:int) =  ((return := f);f);
fun retval () = SOME(!return);
fun toString a = Int.toString a;




(* pre evaluates to true and post evaluates to true *)
val pre_0   = [fn () => true, fn () => 5 = 5]; 
val post_0  = fn () => eval (fut  10 2)  = 5;
val ret_0   = (TestHarness.check retval pre_0 post_0);
val _       = TestHarness.printResult toString 0 ret_0;

(* pre evaluates to true and post evaluates to false *)
val pre_1   = [fn () => true, fn () => 5 = 5]; 
val post_1  = fn () => eval (fut  10 2)  = 3;
val ret_1   = (TestHarness.check retval pre_1 post_1);
val _       = TestHarness.printResult toString 1 ret_1;

(* pre evaluates to true and post throws exception*)
val pre_2   = [fn () => true, fn () => 5 = 5]; 
val post_2  = fn () => eval (fut  10 0)  = 3;
val ret_2   = (TestHarness.check retval pre_2 post_2);
val _       = TestHarness.printResult toString 2 ret_2;

(* pre evaluates to false and post evaluates to true *)
val pre_3   = [fn () => false, fn () => 5 = 5];
val post_3  = fn () => eval (fut 12 2) = 6;
val ret_3   = (TestHarness.check retval pre_3 post_3);
val _       = TestHarness.printResult toString 3 ret_3;

(* pre evaluates to false and post evaluates to false *)
val pre_4   = [fn () => false, fn () => 5 = 5]; 
val post_4  = fn () => eval (fut  10 2)  = 3;
val ret_4   = (TestHarness.check retval pre_4 post_4);
val _       = TestHarness.printResult toString 4 ret_4;

(* pre evaluates to false and post throws exception*)
val pre_5   = [fn () => false, fn () => 5 = 5]; 
val post_5  = fn () => eval (fut  10 0)  = 3;
val ret_5   = (TestHarness.check retval pre_5 post_5);
val _       = TestHarness.printResult toString 5 ret_5;

(* pre throws exception and post evaluates to true *)
val pre_6   = [fn () => (fut 10 0)=1, fn () => 5 = 5];
val post_6  = fn () => eval (fut 12 2) = 6;
val ret_6   = (TestHarness.check retval pre_6 post_6);
val _       = TestHarness.printResult toString 6 ret_6;

(* pre throws exception and post evaluates to false *)
val pre_7   = [fn () => (fut 10 0) = 1, fn () => 5 = 5]; 
val post_7  = fn () => eval (fut  10 2)  = 3;
val ret_7   = (TestHarness.check retval pre_7 post_7);
val _       = TestHarness.printResult toString 7 ret_7;

(* pre throws exception and post throws exception*)
val pre_8   = [fn () => (fut 10 0) =1, fn () => 5 = 5]; 
val post_8  = fn () => eval (fut  10 0)  = 3;
val ret_8   = (TestHarness.check retval pre_8 post_8);
val _       = TestHarness.printResult toString 8 ret_8;
    
val results = [ret_0,ret_1,ret_2,ret_3,ret_4,ret_5,ret_6,ret_7,ret_8];
val _       = TestHarness.printList toString results;


